/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#define OWL_TRACE OWL_TRACE_INFO
#define OWL_FILE "main"

#include <owl/core/owl_defs.h>
#include <owl/core/owl_err.h>
#include <owl/core/owl_debug.h>
#include <owl/core/owl_shell.h>
#include <owl/core/owl_timer.h>
#include <owl/drivers/owl_board.h>
#include <owl/drivers/owl_uart.h>
#include <owl/drivers/owl_spi.h>
#include <owl/drivers/owl_gpio.h>
#include <owl/components/hdg205.h>
# include <owl/util/cbuf.h>

#if WLP_CONFIG_STACK_OFFLOAD
# include <pico/wlp_ttcp.h>
# include <pico/wlp_ntp.h>
# include <pico/wlp_ping.h>
#include <pico/httpd/httpd.h>
#else /* WLP_CONFIG_STACK_OFFLOAD */
# include <owl/core/owl_lwip.h>
# include <owl/util/ttcp.h>
#endif /* WLP_CONFIG_STACK_OFFLOAD */

#include <wlp_api.h>
#include <wlp_rftest.h>
#include <wlp_inet.h>
#include <hdg205/wlp_hdg205.h>

#include <pico/wlp_shell.h>
#include <pico/wlp_apitest.h>

#include "upgrade.h"
#include "http.h"

static int fd;
static int quit = 0;
#define OWL_BOARD_HDG205_UART ((void *) fd)

int 
uart_exchange(void *ctx, const void *tx, void *rx, int len)
{
        if (tx)
                return owluart_write(ctx, tx, len);
        if (rx)
                return owluart_read(ctx, rx, len);
}

static void
error_cb(void *ctx, int err)
{
        struct owlsh_ctx *sh = ctx;
        owlsh_printf(sh, "err_cb err:%d\n\r", err);
}

static void
link_cb(void *ctx, int link)
{
        struct owlsh_ctx *sh = ctx;
        owlsh_printf(sh, "link_cb link:%d\n\r", link);

#if !WLP_CONFIG_STACK_OFFLOAD
        if (link > 0)
                owlip_linkup(0);
        else 
                owlip_linkdown(0);
#endif /* !WLP_CONFIG_STACK_OFFLOAD */
}

static void
addr_cb(void *ctx, const struct ip_addr *ip)
{
        struct owlsh_ctx *sh = ctx;
        if (ip == NULL) {
                owlsh_printf(sh, "addr_cb ip:none\n\r");
        } else {
                owlsh_printf(sh, "addr_cb ip:%s\n\r", ip_ntoa(ip));
        }
}


void
systick_irq(void)
{
        owltmr_tick(OWL_BOARD_SYSTICK_PERIOD);
        debug_time += OWL_BOARD_SYSTICK_PERIOD;
}

static void
select_cb(void *bus, int select)
{
        return;
}


static void quit_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void init_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void shutdown_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void reset_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void get_version_f(struct owlsh_ctx *sh, int argc, const char *argv[]);

static const struct owlsh_cmd cmd_quit = {
        "quit", quit_f, "quit"
};

/**
 * Intialize a Wi-Fi device by downloadings its firmware and making it ready
 * for use. When the Wi-Fi device is successfully initialized, "Ready"
 * will be printed.
 */
static const struct owlsh_cmd cmd_init = {
        "init", init_f, "wlp_init()"
};

/**
 * Shutdown the Wi-Fi device by unloading its firmware and freeing all
 * resources.
 */
static const struct owlsh_cmd cmd_shutdown = {
        "shutdown", shutdown_f, "wlp_shutdown()"
};

/**
 *
 */
static const struct owlsh_cmd cmd_get_version = {
        "get_version", get_version_f, "wlp_get_version()"
};

/**
 *
 */
static const struct owlsh_cmd cmd_reset = {
        "reset", reset_f, "wlp_reset()"
};

static void
quit_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        quit = 1;
}

static void
init_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int err;
        int flags = 0;
                
        /* Setup the owl pico library. We pass the functions
         * that the library will use to read from and write to
         * the WiFi module.
         */
        if ((err = wlp_init(
                     uart_exchange, NULL, NULL, owluart_poll,
                     OWL_BOARD_HDG205_UART, WLP_HPI_POLL)) < 0) {
                owl_err("wlp_init failed err:%d(%s)", err, owl_error(err));
                return;
        }

#if !WLP_CONFIG_STACK_OFFLOAD
        owlip_add(0);
#endif /* !WLP_CONFIG_STACK_OFFLOAD */
        
        /* We register callbacks so that we get notified when the
         * WiFi link is established (link up) or torn down (link down)
         * and when and IP address has been assigned to the interface
         * (since this can happen at some point after the WiFi link
         * comes up if DHCP is used).
         */
        wlp_set_error_cb(error_cb, sh);
        wlp_set_link_cb(link_cb, sh);
        wlp_set_ipaddr_cb(addr_cb, sh);
        
        owlsh_printf(sh, "Ready\n\r");
}

static void
shutdown_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        owl_unused(sh);
        owl_unused(argc);
        owl_unused(argv);

        wlp_shutdown();
}

static void
reset_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        owl_unused(sh);
        owl_unused(argc);
        owl_unused(argv);

        wlp_reset();
}

static void
get_version_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        char driver_v[16];
        char firmware_v[16];

        owl_unused(sh);
        owl_unused(argc);
        owl_unused(argv);

        wlp_get_version(
                driver_v, sizeof(driver_v), firmware_v, sizeof(firmware_v));
        
        owlsh_printf(sh, "driver:%s firmware:%s\r\n", driver_v, firmware_v);
}

int
main(int argc, char *argv[])
{
        int err;
        int shflags = OWLSH_ESCAPECHARS;
        struct owlsh_ctx *sh;
        uint8_t buf[1024];
        
        char *devname;

        owl_os_init();
        
	shflags |= OWLSH_NOECHO;
        
        if (argc > 1 && !owl_os_strcmp(argv[1], "-h")) {
                printf("%s [port]\n", argv[0]);
                return 0;
        }

        if(argc > 1)
                devname = argv[1];
        else
                devname = "/dev/ttyUSB0";


        if ((fd = open(devname, O_RDWR | O_NOCTTY)) < 0) {
                owl_err("Failed to open device '%s'", devname);
                return -1;
        }

        owlboard_init();
	
        if ((err = owltmr_init()) < 0) {
                owl_err("owltmr_init failed err:%d(%s)", err, owl_error(err));
                return -1;
        }

        /* Initialize the console UART */
        if (owluart_init(OWL_BOARD_SHELL_UART, UART_BAUDRATE_115200,
                         UART_DATABITS_8, UART_PARITY_NONE,
                         UART_STOPBITS_1, 0) < 0)
                return -1;
        
        /* Initialize the host console input/output system. We will
         * only use the console for status output in the example
         * (using owlsh_print()) so the console command handling will
         * be unused.
         */
        if ((sh = owlsh_create(NULL, owluart_write, OWL_BOARD_SHELL_UART,
                               shflags)) == NULL) {
                owl_err("owlsh_create failed");
                return -1;
        }
        
        /* Initialize the host-WiFi UART */
        if ((err = owluart_init(OWL_BOARD_HDG205_UART, UART_BAUDRATE_115200,
                                UART_DATABITS_8, UART_PARITY_NONE,
                                UART_STOPBITS_1, 0)) < 0) {
                owl_err("uart_init pico err:%d(%s)", err, owl_error(err));
                return -1;
        }

#if !WLP_CONFIG_STACK_OFFLOAD
        if (owlip_init()) {
                owlsh_printf(sh, "net init failed\r\n");
                return;
        }
#endif /* !WLP_CONFIG_STACK_OFFLOAD */

        /* Add commands which can be used to control the WiFi link.
         * link_cb() will be called when the connection is up.
         */
        owlsh_addcmd(sh, &cmd_init, NULL);
#if WLP_CONFIG_STACK_OFFLOAD
        owlsh_addcmd(sh, &cmd_program, NULL);
#endif /* WLP_CONFIG_STACK_OFFLOAD */
        owlsh_addcmd(sh, &cmd_get_version, NULL);
        owlsh_addcmd(sh, &cmd_shutdown, NULL);
        owlsh_addcmd(sh, &cmd_reset, NULL);
        owlsh_addcmd(sh, &cmd_heartbeat, NULL);
        owlsh_addcmd(sh, &cmd_loopback, NULL);
        owlsh_addcmd(sh, &cmd_scan, NULL);
        owlsh_addcmd(sh, &cmd_set_wps, NULL);
        owlsh_addcmd(sh, &cmd_linkup, NULL);
        owlsh_addcmd(sh, &cmd_linkdown, NULL);
        owlsh_addcmd(sh, &cmd_get_hwaddr, NULL);
        owlsh_addcmd(sh, &cmd_get_network, NULL);
        owlsh_addcmd(sh, &cmd_get_rate, NULL);
        owlsh_addcmd(sh, &cmd_get_linkctx, buf);
        owlsh_addcmd(sh, &cmd_set_linkctx, buf);
        owlsh_addcmd(sh, &cmd_set_mode, NULL);
        owlsh_addcmd(sh, &cmd_get_mode, NULL);
        owlsh_addcmd(sh, &cmd_set_ps, NULL);
        owlsh_addcmd(sh, &cmd_conf_ps, NULL);
        owlsh_addcmd(sh, &cmd_set_channel, NULL);
        owlsh_addcmd(sh, &cmd_add_mcast, NULL);
        owlsh_addcmd(sh, &cmd_del_mcast, NULL);

        /* Add commands which can be used to control the ip address settings */
        owlsh_addcmd(sh, &cmd_set_ipaddr, NULL);
        owlsh_addcmd(sh, &cmd_get_ipaddr, NULL);
        owlsh_addcmd(sh, &cmd_set_dhcp, NULL);
        owlsh_addcmd(sh, &cmd_get_dhcp, NULL);

        owlsh_addcmd(sh, &cmd_ttcp, NULL);
        owlsh_addcmd(sh, &cmd_start_tx, NULL);
        owlsh_addcmd(sh, &cmd_stop_tx, NULL);
        owlsh_addcmd(sh, &cmd_clr_rxstats, NULL);
        owlsh_addcmd(sh, &cmd_get_rxstats, NULL);

        owlsh_addcmd(sh, &cmd_apitest, NULL);
        owlsh_addcmd(sh, &cmd_help, NULL);

#if WLP_CONFIG_STACK_OFFLOAD
        owlsh_addcmd(sh, &cmd_set_dhcpd, NULL);
        owlsh_addcmd(sh, &cmd_get_hostbyname, NULL);

        /* Add commands which can be used to play around with the socket
         * interface of owl pico
         */
        owlsh_addcmd(sh, &cmd_socket, NULL);
        owlsh_addcmd(sh, &cmd_bind, NULL);
        owlsh_addcmd(sh, &cmd_listen, NULL);
        owlsh_addcmd(sh, &cmd_connect, NULL);
        owlsh_addcmd(sh, &cmd_close, NULL);
        owlsh_addcmd(sh, &cmd_send, NULL);
        owlsh_addcmd(sh, &cmd_sendto, NULL);

        owlsh_addcmd(sh, &cmd_ntp, NULL);
        owlsh_addcmd(sh, &cmd_ping, NULL);

#endif /* WLP_CONFIG_STACK_OFFLOAD */
        
        owlsh_addcmd(sh, &cmd_quit, NULL);
        
#if WLP_CONFIG_STACK_OFFLOAD
        struct httpd_ctx *httpd;

        /* Create the HTTP server */
        if ((httpd = httpd_create()) == NULL) {
                owl_err("failed to create httpd");
                return -1;
        }
        
        http_server_config(httpd, sh);
        owlsh_addcmd(sh, &cmd_httpd, httpd);
#endif /* WLP_CONFIG_STACK_OFFLOAD */

        /* Start the main poll loop to check the console input and making sure
         * that the owl pico API makes forward progress.
         */
        while (!quit) {
                
                if (owluart_poll(OWL_BOARD_SHELL_UART))
                        owlsh_key(sh, owluart_inb(OWL_BOARD_SHELL_UART, 0));

                if (owluart_poll(OWL_BOARD_HDG205_UART))
                        wlp_irq();
                
                wlp_poll();
                owltmr_poll();
        }                      
        return 0;
}

/*!  @} */
