/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#define OWL_TRACE OWL_TRACE_DBG
#define OWL_FILE "upgrade"

#include <owl/core/owl_defs.h>
#include <owl/core/owl_err.h>
#include <owl/core/owl_debug.h>
#include <owl/core/owl_shell.h>
#include <owl/core/owl_timer.h>
#include <owl/util/cbuf.h>
#include <wlp_inet.h>
#include <wlp_socket.h>
#include <wlp_api.h>
#include <hdg205/wlp_hdg205.h>


struct upgrade_ctx {
        int sockid;
        int lsockid;
        /* Needs to be at least 8Kb */
#define UPGRADE_BUF_SIZE 1024 * 8
        struct cbuf *cbuf;
        int dots;
        struct owlsh_ctx *sh; 
};

static int started;

static void program_f(struct owlsh_ctx *sh, int argc, const char *argv[]);

static int
program_server_cb(void *ctx, void *buf, int offset, int len)
{
        struct upgrade_ctx *upgrade = ctx;
        int err = 0;
        
        if (buf == NULL) {
                owlsh_printf(upgrade->sh, "Done\r\n");
                wlp_close(upgrade->sockid);
                cbuf_destroy(upgrade->cbuf);
                owl_os_free(upgrade);
                started = 0;
                return 0;
        }
        
        /* Enough data in our recv buffer? */
        if ((err = cbuf_rdlen_tot(upgrade->cbuf) >= len)) {
                cbuf_get_data(upgrade->cbuf, buf, len);
                wlp_recved(upgrade->sockid, len);
                
                if (!(upgrade->dots % 32))
                        owlsh_printf(upgrade->sh, "\r\n");
                else
                        owlsh_printf(upgrade->sh, ".");
                
                upgrade->dots++;
                return len;
                
        }
        /* Make forward progress */
        wlp_poll();
        owltmr_poll();

        return 0;
}

static int
program_host_cb(void *ctx, void *buf, int offset, int len)
{
        struct owlsh_ctx *sh = ctx;

        const void *p = &firmware;
        
        if (len > 0) {
                owl_os_memcpy(buf, p + offset, len);
                return len;
        } else {
                owlsh_perr(sh, len);
                return 0;
        }
}

static void
recv_cb(void *ctx, int sockid, const struct ip_addr *ip, uint16_t port,
        void *data, int len)
{
        struct upgrade_ctx *upgrade = ctx;
        int err = 0;

        err = cbuf_put_data(upgrade->cbuf, data, len);
}

static void
listen_cb(void *ctx, int lsockid, int sockid,
          const struct ip_addr *ip, uint16_t port)
{
        struct upgrade_ctx *upgrade = ctx;
        
        owl_dbg("listen_cb lsockid:%d sockid:%d ip:%s port:%d",
                lsockid, sockid, ip_ntoa(ip), port);

        wlp_set_recv_cb(sockid, recv_cb, upgrade);
        wlp_accept(lsockid);
        wlp_close(upgrade->lsockid);
        upgrade->sockid = sockid;
        owlsh_perr(upgrade->sh, sockid);
}

static int
parse_int(const char *str, int *out)
{
        char *end;
        *out = owl_os_strtol(str, &end, 0);
        if (end == str)
                return 0;

        return 1;
}

static void
program_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int err;
        int port;
 
        if (argc < 2)
                goto err_parse;

        if (!owl_os_strcmp(argv[1], "server")) {
                if (started) {
                        owlsh_printf(sh, "upgrade already in progress \r\n");
                        return;
                }
                        
                if(!parse_int(argv[2], &port))
                        goto err_parse;

                struct upgrade_ctx *upgrade = NULL;

                if ((upgrade = owl_os_zalloc(sizeof(*upgrade))) == NULL) {
                        err = OWL_ERR_MEM;
                        owlsh_perr(sh, err);
                        return;
                }

                upgrade->dots = 0;
                upgrade->sh = sh;

                if((upgrade->cbuf = cbuf_create(UPGRADE_BUF_SIZE)) == NULL)
                        goto err;
                
                if((upgrade->lsockid = wlp_socket(WLP_SOCK_STREAM, 0)) < 0) 
                        goto err;
 
                if((err = wlp_bind(upgrade->lsockid, IP_ADDR_ANY, port)) < 0) {
                        wlp_close(upgrade->lsockid);
                        goto err;
                }
                if((err = wlp_listen(upgrade->lsockid, 1, 
                                     listen_cb, upgrade)) < 0) {
                        wlp_close(upgrade->lsockid);
                        goto err;
                }
                if ((err = wlp_program(program_server_cb, upgrade)) < 0) {
                        wlp_close(upgrade->sockid);
                        goto err;
                } else {
                        started = 1;
                        return;
                }
        err:
                if (upgrade->cbuf)
                        cbuf_destroy(upgrade->cbuf);
                
                if (upgrade)
                        owl_os_free(upgrade);

                owlsh_perr(sh, err);
                return;
                
        } else if (!owl_os_strcmp(argv[1], "host")) {
                if ((err = wlp_program(program_host_cb, sh))) 
                        owlsh_perr(sh, err);
        } else {
                goto err_parse;
        }
        return;

        /* if wlp_program() is successful, program_cb() will print the
           status upon completion */
err_parse:
        owlsh_printf(sh, "usage: program <server [port]/host> \r\n"); 
}

const struct owlsh_cmd cmd_program = {
        "program", program_f, "wlp_program()"
};
