/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#define OWL_TRACE OWL_TRACE_ERR
#define OWL_FILE "owl_uart"

#include <owl/core/owl_defs.h>
#include <owl/core/owl_err.h>
#include <owl/core/owl_debug.h>
#include <owl/drivers/owl_uart.h>

/*
 * NOTE: We treat stdin, stdout and stderr as UARTs in the linux implemenation
 * to somewhay unify the application code.
 *
 */
 

int owluart_init(void *uart,
                 uint32_t baudrate, enum uart_databits databits,
                 enum uart_parity parity, enum uart_stopbits stopbits,
                 int rtscts)
{
        int fd = (int) uart;
        int err;
        int spd;

        /* No need to initialize standard files */
        if (fd == STDIN_FILENO || fd == STDOUT_FILENO || fd == STDERR_FILENO)
                return fd;

        if (databits != 8 || stopbits != 1 || parity != UART_PARITY_NONE)
                return OWL_ERR_NOTSUPP;
        
        struct termios ti;
        if ((err = tcgetattr(fd, &ti)) < 0) {
                owl_err("Failed to tcgetattr:%d", err);
                return OWL_FAILURE;
        }
        
        cfmakeraw(&ti);
        ti.c_oflag = 0;
        if (rtscts)
                ti.c_cflag |= CRTSCTS;
        else
                ti.c_cflag &= ~(CRTSCTS);

        switch (baudrate) {
        case 300: spd = B300; break;
        case 1200: spd = B1200; break;
        case 2400: spd = B2400; break;
        case 4800: spd = B4800; break;
        case 9600: spd = B9600; break;
        case 19200: spd = B19200; break;
        case 38400: spd = B38400; break;
        case 57600: spd = B57600; break;
        case 115200: spd = B115200; break;
        case 230400: spd = B230400; break;
        case 460800: spd = B460800; break;
        case 500000: spd = B500000; break;
        case 576000: spd = B576000; break;
        case 921600: spd = B921600; break;
        case 1000000: spd = B1000000; break;
        case 1500000: spd = B1500000; break;
        case 2000000: spd = B2000000; break;
        case 2500000: spd = B2500000; break;
        case 3000000: spd = B3000000; break;
        case 3500000: spd = B3500000; break;
        case 4000000: spd = B4000000; break;
        default: spd = B57600; break;
        }

        
        if ((err = cfsetispeed(&ti, spd)) < 0) {
                owl_err("cfsetispeed:%d", err);
                return OWL_FAILURE;
        }
        if ((err = cfsetospeed(&ti, spd)) < 0) {
                owl_err("cfsetospeed:%d", err);
                return OWL_FAILURE;
        }
	
        if ((err = tcsetattr(fd, TCSANOW, &ti)) < 0) {
                owl_err("Failed to tcsetattr:%d", err);
                return OWL_FAILURE;
        }

        owl_info("uart configured with %d (%d) 8-N-1 rtscts:%d\n",
                 spd, baudrate, rtscts);
                 
        return fd;
}

int owluart_read(void *ctx, void *buf, int len)
{
        int fd = (int) ctx;
        int err;
        int pos = 0;
        void *nullbuf = NULL;
        
        if (buf == NULL) {
                /* support NULL buffer in e.g. wlp_recv() */
                if ((nullbuf = owl_os_alloc(len)) == NULL)
                        return OWL_ERR_MEM;
                buf = nullbuf;
        }
        
        owl_dbg("enter len:%d", len);

        /* Reading from stdout not possible, assume that stdin is desired */
        if (fd == STDOUT_FILENO) {
                fd = STDIN_FILENO;
        } else if (fd == STDERR_FILENO) {
                owl_os_free(nullbuf);
                return OWL_FAILURE;
        }
        
        while (pos < len) {
                if ((err = read(fd, buf + pos, len - pos)) < 0)
                        owl_err("read failed, fd:%d len:%d err:%d",
                                fd, len, err);
                else
                        pos += err;
        }

        owl_os_free(nullbuf);
        owl_hexdump("uart rx", buf, len);

        return len;
}

int owluart_write(void *ctx, const void *buf, int len)
{
        int fd = (int) ctx;
        int err;
        int pos = 0;
        
        /* see wlp_api.c */
        if (buf == NULL)
                return 0;
        
        owl_hexdump("uart tx", buf, len);

        while (pos < len) {
                if ((err = write(fd, buf + pos, len - pos)) < 0) {
                        owl_err("write failed, fd:%d len:%d err:%d",
                                fd, len, err);
                        return OWL_FAILURE;
                }
            
            pos += err;
        }
        
        return err;
}

void
owluart_outb(void *ctx, int c)
{
        int fd = (int) ctx;
        int err;
        int len = 1;
        char data = c;

        if ((err = write(fd, &data, len)) != len)
                owl_err("write failed fd:%d len:%d er:%d", fd, len, err);
}


int
owluart_inb(void *ctx, unsigned int timeout)
{
        int fd = (int) ctx;
        int err;
        int len = 1;
        char data;

        /* Reading from stdout not possible, assume that stdin is desired */
        if (fd == STDOUT_FILENO)
                fd = STDIN_FILENO;
        else if (fd == STDERR_FILENO)
                return OWL_FAILURE;
        
        if (timeout > 0) {
                owl_err("timeout:%d not supported");
                return OWL_ERR_NOTSUPP;
        }
        
        if ((err = read(fd, &data, len)) < 0) {
                owl_err("read failed fd:%d len:%d err:%d", fd, len, err);
                return OWL_FAILURE;
        }

        return data;
}


int
owluart_poll(void *ctx)
{
        int fd = (int) ctx;

        /* Reading from stdout not possible, assume that stdin is desired */
        if (fd == STDOUT_FILENO)
                fd = STDIN_FILENO;
        else if (fd == STDERR_FILENO)
                return OWL_FAILURE;
        
        struct timeval tv;
        fd_set fds;
        tv.tv_sec = 0;
        tv.tv_usec = 0;
        FD_ZERO(&fds);
        FD_SET(fd, &fds);
        if (select(fd + 1, &fds, NULL, NULL, &tv) < 0) {
                return 0;
        }
        return FD_ISSET(fd, &fds);
}
