/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#define OWL_TRACE OWL_TRACE_DBG
#define OWL_FILE "main"

#include <owl/core/owl_defs.h>
#include <owl/core/owl_err.h>
#include <owl/core/owl_debug.h>
#include <owl/core/owl_shell.h>
#include <owl/core/owl_timer.h>
#include <owl/drivers/owl_board.h>
#include <owl/drivers/owl_uart.h>
#include <owl/drivers/owl_rtc.h>
#include <owl/drivers/owl_led.h>
#include <owl/drivers/owl_spi.h>
#include <owl/drivers/owl_gpio.h>
#include <owl/drivers/owl_sdio.h>
#include <owl/components/hdg205.h>

#if WLP_CONFIG_STACK_OFFLOAD
# include <pico/wlp_ttcp.h>
# include <pico/wlp_ntp.h>
# include <pico/wlp_ping.h>
# include <pico/httpd/httpd.h>
#else
# include <owl/core/owl_lwip.h>
# include <owl/util/ttcp.h>
# include <owl/util/iperf.h>
#endif /* WLP_CONFIG_STACK_OFFLOAD */

#include <wlp_api.h>
#include <wlp_inet.h>
#include <hdg205/wlp_hdg205.h>

#include <pico/wlp_shell.h>
#include <pico/wlp_apitest.h>

#include "board.h"

#if WLP_CONFIG_STACK_OFFLOAD
static int httpd_led_cb(void *ctx, int seq, const char *reqbuf, uint16_t reqlen,
                        char *rspbuf, uint16_t rsplen);
static int httpd_index_cb(void *ctx, int seq, const char *reqbuf,
                          uint16_t reqlen,
                          char *rspbuf, uint16_t rsplen);
static int httpd_logo_cb(void *ctx, int seq, const char *reqbuf, uint16_t reqlen,
                         char *rspbuf, uint16_t rsplen);
#endif /* WLP_CONFIG_STACK_OFFLOAD */

static void
error_cb(void *ctx, int err)
{
        struct owlsh_ctx *sh = ctx;
        owlsh_printf(sh, "err_cb err:%d\n\r", err);
}

static void
link_cb(void *ctx, int link)
{
        struct owlsh_ctx *sh = ctx;
        owlsh_printf(sh, "link_cb link:%d\n\r", link);

#if !WLP_CONFIG_STACK_OFFLOAD
        if (link > 0)
                owlip_linkup(0);
        else 
                owlip_linkdown(0);
#endif /* !WLP_CONFIG_STACK_OFFLOAD */
}

static void
addr_cb(void *ctx, const struct ip_addr *ip)
{
        struct owlsh_ctx *sh = ctx;
        if (ip == NULL) {
                owlsh_printf(sh, "addr_cb ip:none\n\r");
        } else {
                owlsh_printf(sh, "addr_cb ip:%s\n\r", ip_ntoa(ip));
        }
}

#if WLP_CONFIG_HPI == WLP_CONFIG_HPI_SPI
static void
select_cb(void *ctx, int select)
{ 

        owl_unused(ctx);
        if (select)
                owlspi_claim(OWL_BOARD_HDG205_SPI, OWL_BOARD_HDG205_SPI_CS);
        else
                owlspi_release(OWL_BOARD_HDG205_SPI, OWL_BOARD_HDG205_SPI_CS);
}
#endif /* WLP_CONFIG_HPI_SPI */

static int
program_cb(void *ctx, void *buf, int offset, int len)
{
        struct owlsh_ctx *sh = ctx;
        const void *p = &firmware;
        
        if (len > 0) {
                owl_os_memcpy(buf, p + offset, len);
                return len;
        } else {
                owlsh_perr(sh, len);
                return 0;
        } 
}

static void init_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void shutdown_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void reset_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void program_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void get_version_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
#if WLP_CONFIG_HPI == WLP_CONFIG_HPI_UART 
static void baudrate_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
#endif /* WLP_CONFIG_HPI == WLP_CONFIG_HPI_UART */

/**
 * Intialize a Wi-Fi device by downloadings its firmware and making it ready
 * for use. When the Wi-Fi device is successfully initialized, "Ready"
 * will be printed.
 */
static const struct owlsh_cmd cmd_init = {
        "init", init_f, "wlp_init()"
};

/**
 * Shutdown the Wi-Fi device by unloading its firmware and freeing all
 * resources.
 */
static const struct owlsh_cmd cmd_shutdown = {
        "shutdown", shutdown_f, "wlp_shutdown()"
};

/**
 *
 */
static const struct owlsh_cmd cmd_program = {
        "program", program_f, "wlp_program()"
};

/**
 *
 */
static const struct owlsh_cmd cmd_get_version = {
        "get_version", get_version_f, "wlp_get_version()"
};

/**
 *
 */
static const struct owlsh_cmd cmd_reset = {
        "reset", reset_f, "wlp_reset()"
};

#if WLP_CONFIG_HPI == WLP_CONFIG_HPI_UART
/**
 *
 */
static const struct owlsh_cmd cmd_set_baudrate = {
        "set_baudrate", baudrate_f, "wlp_set_baudrate()"
};

static void
baudrate_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        uint32_t baudrate;
        int err;
        int ms;

        if (argc < 2)
                goto err_parse;

        baudrate = owl_os_strtol(argv[1]);

        if ((err = wlp_set_baudrate(baudrate))) {
                owlsh_perr(sh, err);
                return;
        }

        ms = owltmr_get_ms();
        while (!owltmr_intvl_passed(ms, owltmr_get_ms(), 100))
                wlp_poll();               
                
        err = owluart_init(
                OWL_BOARD_HDG205_UART, baudrate, UART_DATABITS_8, 
                UART_PARITY_NONE, UART_STOPBITS_1, 0);

        owlsh_perr(sh, err);
        return;
        
err_parse:
        owlsh_printf(sh, "usage: set_baudrate <baudrate>\r\n");
}
#endif /* WLP_CONFIG_HPI_UART */

static void
init_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int err;

#if WLP_CONFIG_HPI == WLP_CONFIG_HPI_UART
        if ((err = wlp_init(
                     owluart_exchange, NULL, NULL, owluart_poll,
                     OWL_BOARD_HDG205_UART, WLP_HPI_POLL)) < 0) {
                owl_err("wlp_init failed err:%d(%s)", err, owl_error(err));
                return;
        }
#endif /* WLP_CONFIG_HPI_UART */

#if WLP_CONFIG_HPI == WLP_CONFIG_HPI_SPI
        if ((err = wlp_init(
                     owlspi_exchange, select_cb, NULL,
                     OWL_BOARD_HDG205_SPI, 0)) < 0) {
                owl_err("wlp_init failed err:%d(%s)", err, owl_error(err));
                return;
        }
#endif /* WLP_CONFIG_HPI_SPI */

#if WLP_CONFIG_HPI == WLP_CONFIG_HPI_SDIO
        if ((err = wlp_init(
                     owlsdio_exchange, NULL, owlsdio_cmd, NULL,
                     OWL_BOARD_HDG205_SDIO, WLP_SDIO_4BIT) < 0)) {
                owl_err("wlp_init failed err:%d(%s)", err, owl_error(err));
                return;
        }
#endif /* WLP_CONFIG_HPI_SDIO */


#if !WLP_CONFIG_STACK_OFFLOAD
        if (owlip_add(0)) {
                owlsh_printf(sh, "fail to add iface:0\r\n");
                return;
        }
#endif /* !WLP_CONFIG_STACK_OFFLOAD */
   	
        /* We register callbacks so that we get notified when the
         * WiFi link is established (link up) or torn down (link down)
         * and when and IP address has been assigned to the interface
         * (since this can happen at some point after the WiFi link
         * comes up if DHCP is used).
         */
        wlp_set_error_cb(error_cb, sh);
        wlp_set_link_cb(link_cb, sh);
        wlp_set_ipaddr_cb(addr_cb, sh);
        
        owlsh_printf(sh, "Ready\n\r");
}

static void
shutdown_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        owl_unused(sh);
        owl_unused(argc);
        owl_unused(argv);

        wlp_shutdown();
}

static void
program_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int err;
        owl_unused(argc);
        owl_unused(argv);

        if ((err = wlp_program(program_cb, sh))) {
                owlsh_perr(sh, err);
        }

        /* if wlp_program() is successful, program_cb() will print the
           status upon completion */
}

static void
reset_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        owl_unused(sh);
        owl_unused(argc);
        owl_unused(argv);

        wlp_reset();
}

static void
get_version_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        char driver_v[16];
        char firmware_v[16];

        owl_unused(sh);
        owl_unused(argc);
        owl_unused(argv);

        wlp_get_version(
                driver_v, sizeof(driver_v), firmware_v, sizeof(firmware_v));
        
        owlsh_printf(sh, "driver:%s firmware:%s\r\n", driver_v, firmware_v);
}

int main(int argc, char *argv[])
{
        int err;
	struct owlsh_ctx *sh;
	uint8_t buf[512]; /* linkctx */

        owl_unused(argc);
        owl_unused(argv);
      	
	owl_os_init();
        owlboard_init();
  
#if defined(OWL_BOARD_SHELL_UART)
        /* Initialize the console UART */
        if (owluart_init(OWL_BOARD_SHELL_UART, UART_BAUDRATE_115200,
                         UART_DATABITS_8, UART_PARITY_NONE,
                         UART_STOPBITS_1, 0) < 0)
                return -1;

	/* Initialize the host console input/output system. We will
         * only use the console for status output in the example
         * (using owlsh_print()) so the console command handling will
         * be unused.
         */
        if ((sh = owlsh_create(
                     NULL, owluart_write, OWL_BOARD_SHELL_UART, 0)) == NULL) {
                owl_err("owlsh_create failed");
                return -1;
        }
#endif
        if ((err = owltmr_init()) < 0) {
                owl_err("owltmr_init failed err:%d(%s)", err, owl_error(err));
                return -1;
        }

#if !WLP_CONFIG_STACK_OFFLOAD
        owlip_init();
#endif

#if WLP_CONFIG_HPI == WLP_CONFIG_HPI_UART
        if ((err = owluart_init(OWL_BOARD_HDG205_UART, UART_BAUDRATE_115200,
                                UART_DATABITS_8, UART_PARITY_NONE,
                                UART_STOPBITS_1, 0)) < 0) {
                return -1;
        }
        
        if ((err = owluart_set_bufsize(OWL_BOARD_HDG205_UART, 
                                       WLP_MAX_PKT_LEN)) < 0) {
		return -1;
        }

#endif /* WLP_CONFIG_HPI_UART */

#if WLP_CONFIG_HPI == WLP_CONFIG_HPI_SPI
        /* Initialize the host-WiFi SPI bus */
        if ((err = owlspi_init(OWL_BOARD_HDG205_SPI,
                               OWL_SPI_MODE_MASTER)) < 0) {
                owl_err("spi err:%d(%s)", err, owl_error(err));
                return -1;
        }

        if ((err = owlspi_setup_cs(OWL_BOARD_HDG205_SPI,
                                   OWL_BOARD_HDG205_SPI_CS,
                                   OWL_BOARD_HDG205_SPI_FREQ,
                                   OWL_SPI_TRANSFER_MODE_3, 8)) < 0) {
                owl_err("spi err:%d(%s)", err, owl_error(err));
                return -1;
        }
#endif /* WLP_CONFIG_HPI_SPI */

#if WLP_CONFIG_HPI == WLP_CONFIG_HPI_SDIO
        if ((err = owlsdio_init((void *) OWL_BOARD_HDG205_SDIO, 
                OWL_SDIO_MODE_MASTER, OWL_SDIO_FLAG_DMA | 
                OWL_SDIO_FLAG_4BIT)) < 0) {
                owl_err("sdio err:%d(%s)", err, owl_error(err));
                return -1;
        }
#endif /* WLP_CONFIG_HPI_SDIO */
        
        /* Powercycle WiFi */
        hdg205_suspend();
        hdg205_resume();

        /* Add commands which can be used to control the WiFi link.
         * link_cb() will be called when the connection is up.
         */
        owlsh_addcmd(sh, &cmd_init, NULL);
        owlsh_addcmd(sh, &cmd_program, NULL);
        owlsh_addcmd(sh, &cmd_get_version, NULL);
        owlsh_addcmd(sh, &cmd_shutdown, NULL);
        owlsh_addcmd(sh, &cmd_reset, NULL);
        owlsh_addcmd(sh, &cmd_loopback, NULL);
        owlsh_addcmd(sh, &cmd_scan, NULL);
        owlsh_addcmd(sh, &cmd_set_wps, NULL);
        owlsh_addcmd(sh, &cmd_linkup, NULL);
        owlsh_addcmd(sh, &cmd_linkdown, NULL);
        owlsh_addcmd(sh, &cmd_get_hwaddr, NULL);
        owlsh_addcmd(sh, &cmd_get_network, NULL);
        owlsh_addcmd(sh, &cmd_get_rate, NULL);
        owlsh_addcmd(sh, &cmd_get_linkctx, buf);
        owlsh_addcmd(sh, &cmd_set_linkctx, buf);
        owlsh_addcmd(sh, &cmd_set_mode, NULL);
        owlsh_addcmd(sh, &cmd_get_mode, NULL);
        owlsh_addcmd(sh, &cmd_set_ps, NULL);
        owlsh_addcmd(sh, &cmd_conf_ps, NULL);
        owlsh_addcmd(sh, &cmd_set_channel, NULL);
        owlsh_addcmd(sh, &cmd_add_mcast, NULL);
        owlsh_addcmd(sh, &cmd_del_mcast, NULL);
        
        owlsh_addcmd(sh, &cmd_start_tx, NULL);
        owlsh_addcmd(sh, &cmd_stop_tx, NULL);
        owlsh_addcmd(sh, &cmd_get_rxstats, NULL);
        owlsh_addcmd(sh, &cmd_clr_rxstats, NULL);

        /* Add commands which can be used to control the ip address settings */
        owlsh_addcmd(sh, &cmd_set_ipaddr, NULL);
        owlsh_addcmd(sh, &cmd_get_ipaddr, NULL);
        owlsh_addcmd(sh, &cmd_set_dhcp, NULL);
        owlsh_addcmd(sh, &cmd_get_dhcp, NULL);

        owlsh_addcmd(sh, &cmd_ttcp, NULL);
#if WLP_CONFIG_HPI == WLP_CONFIG_HPI_UART
        owlsh_addcmd(sh, &cmd_set_baudrate, NULL);
#endif /* WLP_CONFIG_HPI == WLP_CONFIG_HPI_UART */

#if !WLP_CONFIG_STACK_OFFLOAD
        owlsh_addcmd(sh, &cmd_iperf, NULL);
#endif /* !WLP_CONFIG_STACK_OFFLOAD */

        owlsh_addcmd(sh, &cmd_apitest, NULL);
        owlsh_addcmd(sh, &cmd_help, NULL);


#if WLP_CONFIG_STACK_OFFLOAD
        owlsh_addcmd(sh, &cmd_set_dhcpd, NULL);
        owlsh_addcmd(sh, &cmd_get_hostbyname, NULL);

        /* Add commands which can be used to play around with the socket
         * interface of owl pico
         */
        owlsh_addcmd(sh, &cmd_socket, NULL);
        owlsh_addcmd(sh, &cmd_bind, NULL);
        owlsh_addcmd(sh, &cmd_listen, NULL);
        owlsh_addcmd(sh, &cmd_connect, NULL);
        owlsh_addcmd(sh, &cmd_close, NULL);
        owlsh_addcmd(sh, &cmd_send, NULL);
        owlsh_addcmd(sh, &cmd_sendto, NULL);

        owlsh_addcmd(sh, &cmd_ntp, NULL);
        owlsh_addcmd(sh, &cmd_ping, NULL);

        struct httpd_ctx *httpd;

        /* Create the HTTP server */
        if ((httpd = httpd_create()) == NULL) {
                owl_err("failed to create httpd");
                return -1;
        }

        /* Add documents page to HTTP server */
        if ((httpd_add_document(
                     httpd, HTTP_METHOD_GET, "/index.html?p=LED1",
                     httpd_led_cb, (void *) OWL_BOARD_LED1)) == NULL) {
                owl_err("failed to add http document");
                return -1;
        }
        if ((httpd_add_document(
                     httpd, HTTP_METHOD_GET, "/index.html?p=LED2",
                     httpd_led_cb, (void *) OWL_BOARD_LED2)) == NULL) {
                owl_err("failed to add http document");
                return -1;
        }
        if ((httpd_add_document(
                     httpd, HTTP_METHOD_GET, "/index.html?p=LED3",
                     httpd_led_cb, (void *) OWL_BOARD_LED3)) == NULL) {
                owl_err("failed to add http document");
                return -1;
        }
        if ((httpd_add_document(
                     httpd, HTTP_METHOD_GET, "/index.html?p=LED4",
                     httpd_led_cb, (void *) OWL_BOARD_LED4)) == NULL) {
                owl_err("failed to add http document");
                return -1;
        }
        if ((httpd_add_document(
                     httpd, HTTP_METHOD_GET, "/index.html",
                     httpd_index_cb, NULL)) == NULL) {
                owl_err("failed to add http document");
                return -1;
        }
        if ((httpd_add_document(
                     httpd, HTTP_METHOD_GET, "/image.gif",
                     httpd_logo_cb, NULL)) == NULL) {
                owl_err("failed to add http document");
                return -1;
        }
        owlsh_addcmd(sh, &cmd_httpd, httpd);

#endif /* WLP_CONFIG_STACK_OFFLOAD */

        /* Start the main poll loop to check the console input and making sure
         * that the owl pico API makes forward progress.
         */

        for (;;) {               
                if (owluart_poll(OWL_BOARD_SHELL_UART))
                        owlsh_key(sh, owluart_inb(OWL_BOARD_SHELL_UART, 0));

                wlp_poll();
                owltmr_poll();
        }                      

        return 0;
}

#if WLP_CONFIG_STACK_OFFLOAD
static int
httpd_index_cb(void *ctx, int seq, const char *reqbuf, uint16_t reqlen,
               char *rspbuf, uint16_t rsplen)
{
        int pos = 0;
        int remain = rsplen;
        int err;

        switch (seq) {
                
        case 0: {

                if ((err = owl_os_snprintf(
                             rspbuf + pos, remain,
                             "HTTP/1.0 200 OK\r\n"
                             "Server: lwIP\r\n"
                             "Content-type: text/html\r\n\r\n"
                             "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 "
                             "Frameset//EN\"\"http://www.w3.org/TR/html4/"
                             "frameset.dtd\">")) > remain)
                        return OWL_ERR_MEM;
                pos += err; remain -= err;
                   
                if ((err = owl_os_snprintf(
                             rspbuf + pos, remain,
                             "<html>"
                             "<head><title>HDG2XX 802.11bgn on Linux</title>"
                             "</head><body bgcolor=white text=black>"
                             "<table width=100%%>"
                             "<tr valign=top><td width=80>"
                             "<a href=index.html>"
                             "<img src=image.gif border=0 alt=logo"
                             "title=logo></a>" 
                             "</td><td width=500>"
                             "<h2>802.11bg Wi-Fi on STM32</h2>"
                             "<p>The web page is served by an STM32 "
                             "application, using the "
                             "H&D Wireless&rsquo; unique Wi-Fi solution, "
                             "consisting of the HDG8XX WLAN SiP and the "
                             "oWL Software Toolkit. The web server is "
                             "running on top of the oWL Pico API."
                             "</p><hr><table>")) > remain)
                        return OWL_ERR_MEM;
                pos += err; remain -= err;
                break;

        }
                
        case 1: {
                struct wlp_mac_addr hwaddr;
                struct ip_addr ipaddr;
                wlp_get_hwaddr(&hwaddr);
                wlp_get_ipaddr(&ipaddr, NULL, NULL, NULL);

                if ((err = owl_os_snprintf(
                             rspbuf + pos, remain,
                             "<tr><td>IP Address: </td><td><b>%s</b></td></tr>"
                             "<tr><td>MAC Address: </td><td><b>"
                             "%02x:%02x:%02x:%02x:%02x:%02x</b></td></tr>",
                             ip_ntoa(&ipaddr),
                             hwaddr.octet[0], hwaddr.octet[1],
                             hwaddr.octet[2], hwaddr.octet[3],
                             hwaddr.octet[4], hwaddr.octet[5])) > remain)
                        return OWL_ERR_MEM;
                pos += err; remain -= err;

                
                if ((err = owl_os_snprintf(
                             rspbuf + pos, remain,
                             "<tr><td><form method=get action=index.html>"
                             "<fieldset style='width: 300;'>"
                             "<legend>LED Control</legend>"
                             "<input type=submit style='height: 50px' "
                             "name=p value=LED1>&nbsp&nbsp"
                             "<input type=submit style='height: 50px' "
                             "name=p value=LED2>&nbsp&nbsp"
                             "<input type=submit style='height: 50px' "
                             "name=p value=LED3>&nbsp&nbsp"
                             "<input type=submit style='height: 50px' "
                             "name=p value=LED4>&nbsp&nbsp"
                             "</fieldset></form>"
                             "</td><td>&nbsp;</td></tr></table>"
                             "</body></html>")) > remain)
                        return OWL_ERR_MEM;
                pos += err; remain -= err;

                break;
        }

        default: /* done */
                return 0;
                
        }
        
        return pos;
}


static int
httpd_led_cb(void *ctx, int seq, const char *reqbuf, uint16_t reqlen,
             char *rspbuf, uint16_t rsplen)
{
        int led = (int) ctx;
     
        if (!seq)
                owlled_toggle(led);
        
        return httpd_index_cb(ctx, seq, reqbuf, reqlen, rspbuf, rsplen);
}

static int
httpd_logo_cb(void *ctx, int seq, const char *reqbuf, uint16_t reqlen,
              char *rspbuf, uint16_t rsplen)
{
        static const char hdwireless_gif[] = {
                /* 
                 * "HTTP/1.0 200 OK\n"
                 * "Server: lwIP\n"
                 * "Content-type: image/gif\n\n",
                 * 
                 */
                0x48, 0x54, 0x54, 0x50, 0x2f, 0x31, 0x2e, 0x30, 
                0x20, 0x32, 0x30, 0x30, 0x20, 0x4f, 0x4b, 0x0d, 
                0x0a, 0x53, 0x65, 0x72, 0x76, 0x65, 0x72, 0x3a, 
                0x20, 0x6c, 0x77, 0x49, 0x50, 0x0d, 0x0a, 0x43, 
                0x6f, 0x6e, 0x74, 0x65, 0x6e, 0x74, 0x2d, 0x74, 
                0x79, 0x70, 0x65, 0x3a, 0x20, 0x69, 0x6d, 0x61, 
                0x67, 0x65, 0x2f, 0x67, 0x69, 0x66, 0x0d, 0x0a, 
                0x0d, 0x0a,

                /* gif data */
                0x47, 0x49, 0x46, 0x38, 0x39, 0x61, 0x45, 0x01, 
                0x7b, 0x00, 0xa1, 0x00, 0x00, 0xff, 0xff, 0xff, 
                0x00, 0x66, 0x99, 0x00, 0x00, 0x00, 0xff, 0xff, 
                0xff, 0x21, 0xfe, 0x11, 0x43, 0x72, 0x65, 0x61, 
                0x74, 0x65, 0x64, 0x20, 0x77, 0x69, 0x74, 0x68, 
                0x20, 0x47, 0x49, 0x4d, 0x50, 0x00, 0x2c, 0x00, 
                0x00, 0x00, 0x00, 0x45, 0x01, 0x7b, 0x00, 0x00, 
                0x02, 0xfe, 0x84, 0x8f, 0xa9, 0xcb, 0xed, 0x0f, 
                0xa3, 0x9c, 0xb4, 0xda, 0x8b, 0xb3, 0xde, 0xbc, 
                0xfb, 0x0f, 0x86, 0xe2, 0x48, 0x96, 0xe6, 0x89, 
                0xa6, 0xea, 0xca, 0xb6, 0xee, 0x0b, 0xc7, 0xf2, 
                0x4c, 0xd7, 0xf6, 0x8d, 0xe7, 0xfa, 0xce, 0xf7, 
                0xfe, 0x0f, 0x0c, 0x0a, 0x87, 0xc4, 0xa2, 0xf1, 
                0x88, 0x4c, 0x2a, 0x97, 0xcc, 0xa6, 0xf3, 0x09, 
                0x8d, 0x4a, 0xa7, 0xd4, 0xaa, 0xf5, 0x8a, 0x65, 
                0x05, 0xb6, 0xdc, 0xae, 0xf7, 0x0b, 0x0e, 0x8b, 
                0xc7, 0x64, 0x71, 0xf6, 0x0c, 0x2d, 0xab, 0xd7, 
                0xec, 0x36, 0x19, 0x0d, 0x3f, 0xba, 0xe7, 0xf4, 
                0xfa, 0x3b, 0x8e, 0xff, 0xd9, 0xf7, 0x7c, 0x7b, 
                0xfe, 0x9f, 0xd3, 0x27, 0x38, 0xc8, 0x06, 0x68, 
                0x38, 0x43, 0x98, 0xa8, 0x18, 0x76, 0xd8, 0xd8, 
                0xb2, 0x08, 0x19, 0x19, 0xe0, 0x48, 0x79, 0x22, 
                0x79, 0xa9, 0x58, 0xa9, 0x29, 0x82, 0xd9, 0x29, 
                0xb8, 0x09, 0xda, 0xe1, 0x39, 0xca, 0x17, 0x6a, 
                0x8a, 0x41, 0x9a, 0x4a, 0x77, 0xca, 0x3a, 0xa1, 
                0xfa, 0xea, 0xd6, 0x2a, 0xeb, 0x00, 0x5b, 0xab, 
                0x36, 0x8b, 0xab, 0x60, 0xbb, 0x6b, 0x96, 0xeb, 
                0x0b, 0xc0, 0x1b, 0xec, 0xf5, 0xeb, 0x2b, 0x6c, 
                0xbc, 0x45, 0x8c, 0x7b, 0x6c, 0x9c, 0x3c, 0xbb, 
                0x2c, 0xdc, 0x2c, 0xfb, 0x1c, 0x1c, 0xdd, 0x3a, 
                0xcd, 0x5b, 0xcd, 0x7a, 0xbd, 0x9b, 0x7d, 0xba, 
                0x6d, 0xdb, 0x6d, 0xfa, 0x5d, 0x1b, 0x1e, 0x3a, 
                0x0e, 0x5b, 0x0e, 0x7a, 0xfe, 0x9a, 0xbe, 0xb9, 
                0xae, 0xda, 0xae, 0xf9, 0x9e, 0x1a, 0x5f, 0x39, 
                0x4f, 0x5a, 0x4f, 0x79, 0x3f, 0x9a, 0xef, 0xb8, 
                0xef, 0xd9, 0xaf, 0xd1, 0xbf, 0x4e, 0x01, 0x0f, 
                0x0d, 0xc4, 0x54, 0xd0, 0xd0, 0xc1, 0x4b, 0x09, 
                0x01, 0x2d, 0x94, 0xd4, 0xf0, 0xcf, 0xc3, 0x48, 
                0x11, 0xf3, 0x4c, 0x84, 0x54, 0x11, 0xcf, 0x1d, 
                0xfe, 0x5d, 0x60, 0x1a, 0x8c, 0xf1, 0x58, 0xa6, 
                0x42, 0x1b, 0x04, 0x02, 0x4a, 0x9a, 0x14, 0xb0, 
                0xe0, 0xe4, 0xc9, 0x94, 0x2a, 0x51, 0x1e, 0x68, 
                0x29, 0x03, 0x66, 0xb8, 0x8d, 0x09, 0x18, 0x31, 
                0xb0, 0xf9, 0x80, 0xa6, 0xab, 0x5b, 0x24, 0x65, 
                0xf6, 0xf4, 0xf9, 0x12, 0x28, 0x00, 0xa1, 0x2d, 
                0x88, 0x36, 0xd3, 0x79, 0x00, 0x27, 0xc7, 0x61, 
                0x12, 0x3e, 0x5e, 0xe0, 0x19, 0x54, 0xa5, 0x82, 
                0x96, 0x25, 0xa7, 0x0a, 0x35, 0xba, 0x02, 0x2b, 
                0x31, 0xa4, 0x06, 0x94, 0xd6, 0x64, 0xda, 0xb4, 
                0xd7, 0x53, 0x9d, 0x54, 0x13, 0x50, 0x35, 0x69, 
                0xf6, 0xaa, 0xd4, 0x18, 0x5a, 0x7f, 0x71, 0x05, 
                0xd6, 0xf1, 0x26, 0xd8, 0xb0, 0x71, 0x51, 0x39, 
                0x8d, 0xba, 0xf2, 0x67, 0x59, 0xbd, 0x68, 0xf9, 
                0xba, 0x84, 0xd1, 0xb6, 0xd8, 0xdd, 0xaf, 0x5f, 
                0x40, 0x72, 0x91, 0x3b, 0x69, 0x69, 0xe1, 0x0c, 
                0x64, 0xd5, 0x9e, 0x4d, 0xbb, 0x76, 0x47, 0xe0, 
                0x5c, 0x6f, 0xbd, 0x22, 0x98, 0x7b, 0x19, 0x99, 
                0x62, 0xcc, 0x49, 0x0f, 0xd3, 0x12, 0x8b, 0xb7, 
                0xaf, 0x81, 0xb3, 0x55, 0xfd, 0xfa, 0x98, 0xac, 
                0x6c, 0x70, 0xe6, 0xc5, 0x0b, 0x58, 0xaf, 0xde, 
                0xcc, 0xb9, 0x6b, 0x6c, 0xd9, 0x96, 0xf7, 0x0e, 
                0x25, 0x5d, 0xfa, 0x76, 0xe4, 0xd0, 0x7f, 0x47, 
                0xfb, 0xcc, 0xcb, 0x12, 0x78, 0x83, 0xc9, 0xa8, 
                0x79, 0xe7, 0x7e, 0xf0, 0xb8, 0x47, 0xe5, 0xba, 
                0xb0, 0x5b, 0x27, 0x26, 0xcc, 0x7c, 0x75, 0x17, 
                0xc4, 0xd1, 0x75, 0x03, 0xc7, 0x9d, 0xdb, 0x36, 
                0x6f, 0xc8, 0x2b, 0x89, 0x62, 0xef, 0xcd, 0xfd, 
                0x78, 0x78, 0xed, 0x56, 0xb1, 0x0f, 0xff, 0xce, 
                0xe3, 0xad, 0xc8, 0xd9, 0xce, 0x41, 0x4b, 0xf7, 
                0xdc, 0xde, 0xb5, 0xef, 0xc8, 0xe8, 0xe7, 0x0b, 
                0xb7, 0x2f, 0xbe, 0x2c, 0xd0, 0xef, 0xf7, 0xfe, 
                0xf1, 0x83, 0xf7, 0x8f, 0x5b, 0x70, 0xfc, 0x95, 
                0x37, 0x60, 0x20, 0xaa, 0xd9, 0xc5, 0xde, 0x7b, 
                0xd5, 0xd1, 0x36, 0x1d, 0x75, 0xb1, 0xfd, 0x66, 
                0x1e, 0x80, 0x04, 0xe6, 0x57, 0x20, 0x80, 0x01, 
                0x8e, 0x27, 0x60, 0x85, 0x16, 0x92, 0x36, 0x5e, 
                0x84, 0x38, 0xa8, 0x47, 0xc1, 0x82, 0xd0, 0xb9, 
                0xd7, 0x19, 0x7b, 0x24, 0x5a, 0x87, 0x96, 0x4c, 
                0x2a, 0xee, 0xb6, 0x9d, 0x5f, 0xda, 0xf1, 0xc7, 
                0x22, 0x8a, 0xff, 0xc1, 0x78, 0x9f, 0x63, 0xc2, 
                0xd9, 0x28, 0x9a, 0x0d, 0x20, 0xee, 0x04, 0x5f, 
                0x4e, 0x21, 0xc5, 0xd7, 0xe3, 0x88, 0x0d, 0x76, 
                0x28, 0xa3, 0x7f, 0x32, 0xfe, 0x57, 0xa4, 0x8b, 
                0x10, 0xc6, 0x48, 0x1e, 0x8e, 0xe2, 0x1d, 0x69, 
                0xda, 0x04, 0xc5, 0xdd, 0xb0, 0x23, 0x5d, 0xcf, 
                0x41, 0x00, 0x9f, 0x88, 0x22, 0xc2, 0xb5, 0xa0, 
                0x7e, 0x6b, 0xc1, 0x44, 0x9e, 0x91, 0x4a, 0x22, 
                0x89, 0x95, 0x93, 0x18, 0x66, 0xd8, 0x5b, 0x98, 
                0xe7, 0xc5, 0xe8, 0x43, 0x95, 0x11, 0x04, 0x89, 
                0xe0, 0x95, 0x0a, 0xca, 0x39, 0xa7, 0x66, 0x1d, 
                0xda, 0xa6, 0x61, 0x8b, 0xc6, 0x21, 0x79, 0x24, 
                0x8c, 0x51, 0xee, 0x59, 0x61, 0x9e, 0x63, 0xf6, 
                0xf7, 0xe1, 0x81, 0x0c, 0x0e, 0xb9, 0x14, 0x07, 
                0xb3, 0x6d, 0x79, 0x62, 0x9f, 0x37, 0x7a, 0x78, 
                0xe6, 0x9e, 0x0c, 0xd0, 0xb8, 0x1f, 0x8b, 0x94, 
                0xee, 0x46, 0x23, 0x9a, 0x1c, 0xea, 0xb0, 0x9c, 
                0x7c, 0x99, 0x21, 0xc6, 0x23, 0x9d, 0x87, 0xda, 
                0x09, 0x1b, 0x9c, 0x1b, 0xe6, 0x58, 0xdf, 0x84, 
                0xe0, 0xa9, 0x79, 0x2a, 0xa4, 0x4e, 0x66, 0x3a, 
                0x28, 0x9b, 0xb0, 0x16, 0xda, 0xa8, 0x65, 0x9d, 
                0x01, 0xf9, 0x26, 0xa2, 0x73, 0x3a, 0x38, 0x5b, 
                0xaa, 0x8e, 0xe6, 0xf8, 0xa7, 0x84, 0x93, 0x5e, 
                0x8a, 0xaa, 0xa5, 0xc4, 0x3e, 0xf9, 0xfe, 0x2b, 
                0xa1, 0xc9, 0x02, 0x4b, 0x43, 0xa7, 0x26, 0x8a, 
                0xea, 0xa9, 0x90, 0xbc, 0x1a, 0x26, 0x22, 0xa4, 
                0xad, 0x0e, 0x5b, 0xa3, 0x56, 0xac, 0x3a, 0x60, 
                0xa6, 0x07, 0x9b, 0x22, 0xf7, 0x6d, 0x0d, 0xce, 
                0xea, 0xfa, 0x29, 0x90, 0xa4, 0x6e, 0x46, 0x9d, 
                0x8f, 0xb6, 0x9e, 0xaa, 0x2a, 0x9b, 0x49, 0x1a, 
                0xb7, 0xa6, 0xb2, 0xdc, 0x1a, 0xeb, 0x6e, 0x06, 
                0x53, 0xbe, 0xab, 0xa3, 0xa1, 0x5c, 0x3e, 0x6b, 
                0x18, 0x8f, 0x1b, 0xe8, 0xfb, 0x6b, 0xbb, 0xcc, 
                0xea, 0x29, 0xac, 0xa6, 0xab, 0x76, 0x2b, 0xa9, 
                0x8b, 0x14, 0x6c, 0xab, 0xaa, 0x81, 0xb5, 0x32, 
                0x2a, 0x2a, 0x63, 0x11, 0x5b, 0xb9, 0xe5, 0x85, 
                0x83, 0xc6, 0x4b, 0x61, 0xbd, 0xcb, 0x7a, 0xc7, 
                0x24, 0xb1, 0x09, 0xeb, 0x99, 0x1c, 0xc1, 0xe2, 
                0xea, 0xbb, 0x2e, 0x97, 0x1e, 0x98, 0x4a, 0x71, 
                0xc5, 0x21, 0xdf, 0xc9, 0x27, 0xbe, 0x50, 0x62, 
                0x2c, 0x28, 0xc7, 0x94, 0xb2, 0x1c, 0xae, 0xc7, 
                0xb4, 0xae, 0x5b, 0x32, 0xb9, 0x16, 0xe8, 0x9c, 
                0xab, 0x7a, 0xe1, 0x8e, 0x09, 0x33, 0x91, 0xf3, 
                0xfa, 0xd9, 0xb0, 0xc0, 0x2b, 0xb7, 0xfa, 0xaa, 
                0xc5, 0x54, 0x92, 0x0c, 0x31, 0xca, 0xed, 0xbd, 
                0x37, 0xd6, 0x8f, 0xd8, 0xba, 0xfb, 0x73, 0xb0, 
                0x2f, 0x83, 0xab, 0x74, 0xa4, 0x46, 0x0f, 0x0c, 
                0x28, 0xd5, 0x82, 0x8e, 0xfc, 0x70, 0xb4, 0xa3, 
                0x4e, 0xbc, 0x6f, 0x83, 0x09, 0x96, 0x0a, 0xf0, 
                0xd5, 0x9a, 0x0e, 0x0d, 0x2c, 0xc3, 0x45, 0xb3, 
                0x8d, 0xac, 0xd0, 0x11, 0x1c, 0x0d, 0x73, 0xcb, 
                0xf9, 0x5e, 0x94, 0x48, 0x46, 0x71, 0xe0, 0x9d, 
                0x89, 0xde, 0x68, 0xf0, 0x9d, 0xb7, 0xdf, 0x67, 
                0x00, 0x4e, 0x88, 0xe0, 0x83, 0x13, 0xfe, 0x89, 
                0xe1, 0x58, 0x20, 0x9e, 0xb8, 0xe2, 0x56, 0x30, 
                0xde, 0x87, 0xe3, 0x57, 0x40, 0xad, 0x5e, 0x8a, 
                0xe4, 0x55, 0x50, 0xbe, 0x87, 0xe5, 0x97, 0x63, 
                0x5e, 0x87, 0xe6, 0x54, 0x70, 0xde, 0xb9, 0xe7, 
                0x52, 0x80, 0xbe, 0x8a, 0xe8, 0x51, 0x90, 0x3e, 
                0x87, 0xe9, 0xa7, 0xa3, 0x3e, 0x92, 0xea, 0x4f, 
                0xb0, 0xde, 0xba, 0xeb, 0x4d, 0xc0, 0x5e, 0x88, 
                0xec, 0xb3, 0xd3, 0x0e, 0x95, 0xed, 0x4a, 0xe0, 
                0x9e, 0xbb, 0xee, 0x48, 0xf0, 0x2e, 0xb5, 0xef, 
                0xbf, 0x03, 0x9f, 0xb6, 0xf0, 0x40, 0x10, 0x5f, 
                0xbc, 0xf1, 0x6d, 0x22, 0x5f, 0xb2, 0xf2, 0x7a, 
                0x30, 0xbf, 0xa5, 0xf3, 0xcb, 0x43, 0x7f, 0xb6, 
                0xf4, 0x3b, 0x50, 0x5f, 0xbd, 0xf5, 0x9c, 0x62, 
                0xef, 0xb4, 0xf6, 0xca, 0x71, 0x4f, 0xb6, 0xf7, 
                0xd7, 0x83, 0x2f, 0x3e, 0x11, 0xe4, 0x97, 0x6f, 
                0x3e, 0xf6, 0xe8, 0xa7, 0x4f, 0xfd, 0xfa, 0xec, 
                0x23, 0xef, 0x7e, 0x11, 0xcc, 0xc7, 0x2f, 0x3f, 
                0xf1, 0xf4, 0x1b, 0x01, 0xfc, 0xfd, 0xc3, 0xb3, 
                0xae, 0x7f, 0x12, 0xa4, 0xf7, 0x7f, 0x3b, 0xc6, 
                0x01, 0xf0, 0x73, 0x07, 0x19, 0xa0, 0x01, 0x0f, 
                0x88, 0xc0, 0x04, 0x2a, 0x70, 0x81, 0x0c, 0x6c, 
                0xa0, 0x03, 0x1f, 0x08, 0xc1, 0x08, 0x4a, 0x70, 
                0x82, 0x14, 0xac, 0xa0, 0x05, 0x2f, 0x88, 0x41, 
                0xcd, 0x15, 0x00, 0x00, 0x3b
        };

        static int pos;
        static int remain;
        int len;

        /* initialize state upon first call */
        if (!seq) {
                pos = 0;
                remain = sizeof(hdwireless_gif);
        }

        /* no more data? */
        if (!remain)
                return 0;

        /* send as much data as we can */
        if (remain > rsplen)
                len = rsplen;
        else
                len = remain;
        
        owl_os_memcpy(rspbuf, hdwireless_gif + pos, len);
        pos += len;
        remain -= len;
        return len;
}
#endif /* WLP_CONFIG_STACK_OFFLOAD */

#include <sys/types.h>
#include <errno.h>

/**
 * sbrk
 *
 * Increase program data space. Malloc and related functions depend on this.
 */
caddr_t _sbrk(int incr)
{

        extern char _ebss; /* Defined by the linker */
        static char *heap_end;
        char *prev_heap_end;

        if (heap_end == 0)
                heap_end = &_ebss;

        prev_heap_end = heap_end;

        char *stack = (char *) __get_MSP();
        if (heap_end + incr >  stack) {
                errno = ENOMEM;
                return  (caddr_t) -1;
        }

        heap_end += incr;
        return (caddr_t) prev_heap_end;
}

void EXTI9_5_IRQHandler(void)
{       
#if defined(OWL_BOARD_HDG205_SPI_IRQ_PIN)
        if (EXTI_GetITStatus(EXTI_Line9) != RESET) {
                wlp_irq();
                
                EXTI_ClearITPendingBit(EXTI_Line9);
        }
#endif
}

void
SysTick_Handler(void)
{
        owltmr_tick(OWL_BOARD_SYSTICK_PERIOD);
        debug_time += OWL_BOARD_SYSTICK_PERIOD;
}

void SDIO_IRQHandler(void)
{
        SDIO_ClearITPendingBit(SDIO_IT_SDIOIT);
        wlp_irq();
}


/*!  @} */
