/*
 * Note that this driver currently has these limitations:
 *
 * - Only size alignment 4 works
 * - Min transfer size seems to be > 8.
 *
 */
#define OWL_TRACE OWL_TRACE_ERR

#include <stdint.h>
#include <stdlib.h> /* for NULL */

#include <owl/core/owl_defs.h>
#include <owl/core/owl_err.h>
#include <owl/core/owl_debug.h>
#include <owl/drivers/owl_sdio.h>

#if defined(STM32F40XX)
#include "stm324xg_eval.h"
#include "stm32f4xx_sdio.h"
#include "stm32f4xx_dma.h"
#else
#include "stm32_eval.h"
#include "stm32f10x_sdio.h"
#endif

static struct sdio_dev {
        int flags;
} sdio_ctx;

/* Used to determine the transfer direction from sdio cmd52/cmd53 */
#define CMD_RW_BIT(arg) (arg & (1 << 31))

static struct sdio_dev *priv = NULL;

int owlsdio_init(void *bus, int mode, int flags)
{
        SDIO_InitTypeDef SDIO_InitStructure;

        owl_unused(bus);

        if (mode != OWL_SDIO_MODE_MASTER) {
                owl_err("mode slave not supported");
                return OWL_ERR_NOTSUPP;
        }
        
        owl_dbg("flags:%p", flags);
 
        priv = &sdio_ctx;
        priv->flags = flags;

/* Limitations using SD-microSD adapter */
        if (priv->flags & OWL_SDIO_FLAG_DMA) {
#if defined (STM32F40XX)
                SDIO_InitStructure.SDIO_ClockDiv = 2;
#else
                SDIO_InitStructure.SDIO_ClockDiv = 4;
#endif	
        } else {
		
                SDIO_InitStructure.SDIO_ClockDiv = 2;
        }
        
        SDIO_InitStructure.SDIO_ClockEdge = SDIO_ClockEdge_Rising;
        SDIO_InitStructure.SDIO_ClockBypass = SDIO_ClockBypass_Disable;

        /* NOTE: SDIO IRQ's are not properly detected if SDIO_ClockPowerSave
           is enabled. Seems like the DAT1 is only checked for pending 
           irq when the clock is running */
        SDIO_InitStructure.SDIO_ClockPowerSave = SDIO_ClockPowerSave_Disable;

        if (priv->flags & OWL_SDIO_FLAG_4BIT)
                SDIO_InitStructure.SDIO_BusWide = SDIO_BusWide_4b;
        else
                SDIO_InitStructure.SDIO_BusWide = SDIO_BusWide_1b;

        SDIO_InitStructure.SDIO_HardwareFlowControl =
                SDIO_HardwareFlowControl_Enable;

        owl_info("clk_div:%d", SDIO_InitStructure.SDIO_ClockDiv);

        SDIO_Init(&SDIO_InitStructure);
        
        /* Set Power State to ON */
        SDIO_SetPowerState(SDIO_PowerState_ON);

        /* Enable SDIO Clock */
        SDIO_ClockCmd(ENABLE);
        
        /* Configure SDIO interrupts */
        SDIO_SetSDIOOperation(ENABLE);
        NVIC_EnableIRQ(SDIO_IRQn);

        SDIO_ITConfig(SDIO_IT_SDIOIT, ENABLE);
        return 0;
}

/**
 * This function will be invoked when an sdio cmd should be sent to the
 * device.
 *
 * @param idx is the sdio command number
 * @param arg is the sdio command argument
 * @param rsp should hold the command response upon return. If null, the
 *        response can be ignored.
 * @param data holds a pointer to any data that might follow the command. This
 *        allows the sdio driver to setup dma transfers while waiting for the
 *        command response. NULL if no data transfer will follow. Note that
 *        the same data pointer will be passed to owlsdio_write(), which should
 *        start the actual transfer.
 * @param len is the length of the data buffer.
 *
 */

/* SD_LowLevel_DMA_RxConfig does not allow const data, so disable cast-qual
   here */
#pragma GCC diagnostic ignored "-Wcast-qual" 
void owlsdio_cmd(void *bus, uint8_t idx, uint32_t arg, uint32_t *rsp, 
                 const void *data, int len)
{
        uint32_t status;
        uint32_t respcmd; /* SDIO->RESPCMD */
        uint32_t resp1;   /* SDIO->RESP1   */

        SDIO_CmdInitTypeDef SDIO_CmdInitStructure;
        SDIO_DataInitTypeDef SDIO_DataInitStructure;   

        owl_unused(bus);

        owl_dbg("idx:0x%x arg:0x%lx", idx, arg);
        owl_dbg("rsp:%p data:%p len:%d", rsp, data, len);
        
        /* setup any data transfer */
        if (data != NULL) {
                SDIO_DataInitStructure.SDIO_DataTimeOut = 0xFFFFF;
                SDIO_DataInitStructure.SDIO_DataLength = len;
                SDIO_DataInitStructure.SDIO_DataBlockSize =
                        SDIO_DataBlockSize_1b;
                SDIO_DataInitStructure.SDIO_TransferMode =
                        SDIO_TransferMode_Stream;
                SDIO_DataInitStructure.SDIO_DPSM = SDIO_DPSM_Enable;
                
                if (!CMD_RW_BIT(arg))
                        SDIO_DataInitStructure.SDIO_TransferDir =
                                SDIO_TransferDir_ToSDIO;
                else
                        SDIO_DataInitStructure.SDIO_TransferDir =
                                SDIO_TransferDir_ToCard;

                SDIO_DataConfig(&SDIO_DataInitStructure);

                if (priv->flags & OWL_SDIO_FLAG_DMA) {
                        if (!CMD_RW_BIT(arg)) {
                                /* Prepare RX-DMA here since if we 
                                   take too long after sending the 
                                   command but before enabling DMA we might
                                   fill up the fifo (128b) and lose data. 
                                   This would usually happen with 
                                   traces enabled */
				
#if defined(STM32F40XX) /* STM32F40XX DMA RX */
                                uint32_t * MemoryDST = (uint32_t*) data;
                                
                                DMA_InitTypeDef SDDMA_InitStructure;	 

                                DMA_ClearFlag(SD_SDIO_DMA_STREAM, 
                                              SD_SDIO_DMA_FLAG_FEIF | 
                                              SD_SDIO_DMA_FLAG_DMEIF | 
                                              SD_SDIO_DMA_FLAG_TEIF | 	
                                              SD_SDIO_DMA_FLAG_HTIF | 
                                              SD_SDIO_DMA_FLAG_TCIF);
                                
                                /* DMA2 Stream3 disable */
                                DMA_Cmd(SD_SDIO_DMA_STREAM, DISABLE);
                                
                                /* DMA2 Stream3 Deinit */
                                DMA_DeInit(SD_SDIO_DMA_STREAM);
                                
                                /* Setup DMA control registers.
                                   Peripheral base adress is SDIO fifo, 
                                   Direction is from SDIO to Memory,
                                   Memory pointer is incremented, 
                                   fifo size is 32 bits (one word),
                                   each burst is 32 bits (one word). 
                                   BufferSize is not used when using 
                                   SDIO flowcontrol. */
                                
                                SDDMA_InitStructure.DMA_Channel = 
                                        SD_SDIO_DMA_CHANNEL;
                                SDDMA_InitStructure.DMA_PeripheralBaseAddr = 
                                        (uint32_t)SDIO_FIFO_ADDRESS;
                                SDDMA_InitStructure.DMA_Memory0BaseAddr = 
                                        (uint32_t) MemoryDST;
                                SDDMA_InitStructure.DMA_DIR = 
                                        DMA_DIR_PeripheralToMemory;
                                SDDMA_InitStructure.DMA_BufferSize = 
                                        len / 4;	 
                                SDDMA_InitStructure.DMA_PeripheralInc = 
                                        DMA_PeripheralInc_Disable;
                                SDDMA_InitStructure.DMA_MemoryInc = 
                                        DMA_MemoryInc_Enable;
                                SDDMA_InitStructure.DMA_PeripheralDataSize = 
                                        DMA_PeripheralDataSize_Word;
                                SDDMA_InitStructure.DMA_MemoryDataSize = 
                                        DMA_MemoryDataSize_Word;
                                SDDMA_InitStructure.DMA_Mode = 
                                        DMA_Mode_Normal;
                                SDDMA_InitStructure.DMA_Priority = 
                                        DMA_Priority_High;
                                SDDMA_InitStructure.DMA_FIFOMode = 
                                        DMA_FIFOMode_Enable;
                                SDDMA_InitStructure.DMA_FIFOThreshold = 
                                        DMA_FIFOThreshold_Full;
                                SDDMA_InitStructure.DMA_MemoryBurst = 
                                        DMA_MemoryBurst_INC4;
                                SDDMA_InitStructure.DMA_PeripheralBurst = 
                                        DMA_PeripheralBurst_INC4;  	  
                               
                                DMA_Init(
                                        SD_SDIO_DMA_STREAM, 
                                        &SDDMA_InitStructure);
  
                                DMA_FlowControllerConfig(
                                        SD_SDIO_DMA_STREAM, 
                                        DMA_FlowCtrl_Peripheral);    
                                
                                /* DMA2 Stream3 enable */
                                DMA_Cmd(SD_SDIO_DMA_STREAM, ENABLE);
                                
                                /* Start DMA transfer */
                                SDIO_DMACmd(ENABLE);
                                owl_dbg("SDIO_DMA ENABLE");
                                
#else /* STM32F1 DMA RX */
                                SDIO_DMACmd(ENABLE);
                                SD_LowLevel_DMA_RxConfig(
                                        (uint32_t *) data, len);
#endif		
                        }
                } 
        }
        /* send the command */
        SDIO_CmdInitStructure.SDIO_Argument = arg;
        SDIO_CmdInitStructure.SDIO_CmdIndex = idx;
        SDIO_CmdInitStructure.SDIO_Response = SDIO_Response_Short;
        SDIO_CmdInitStructure.SDIO_Wait = SDIO_Wait_No;
        SDIO_CmdInitStructure.SDIO_CPSM = SDIO_CPSM_Enable;
        SDIO_SendCommand(&SDIO_CmdInitStructure);
        
        /* wait until response is obtained or an error occurs */
        status = SDIO->STA;
        while (!(status & (SDIO_FLAG_CCRCFAIL | SDIO_FLAG_CTIMEOUT |
                           SDIO_FLAG_CMDREND))) {
                status = SDIO->STA;
        }
   
        /* check for transfer errors */
        if (status & SDIO_FLAG_CTIMEOUT) {
                SDIO_ClearFlag(SDIO_FLAG_CTIMEOUT);
                owl_err("ctimeout");
                owl_err("idx:0x%x arg:0x%lx", idx, arg);
                owl_err("rsp:%p data:%p len:%d", rsp, data, len);
                return;
                
        } else if (status & SDIO_FLAG_CCRCFAIL) {
                SDIO_ClearFlag(SDIO_FLAG_CCRCFAIL);
                owl_err("ccrcfail");
                owl_err("idx:0x%x arg:0x%lx", idx, arg);
                owl_err("rsp:%p data:%p len:%d", rsp, data, len);
                return;
        }

        /* Get the command index from the response.
         *
         * NOTE: we will get response timeout for cmd52 if we don't read
         * this register first!
         */
        respcmd = SDIO_GetCommandResponse();
        if (respcmd != idx) {
                owl_err("got response from wrong command:%lu expected:%d",
                        respcmd, idx);
                return;
        }

        /* read the response */
        resp1 = SDIO_GetResponse(SDIO_RESP1);
        SDIO_ClearFlag(SDIO_FLAG_CMDREND);

        owl_dbg("respcmd:0x%02lx resp1:0x%08lx", respcmd, resp1);
        
        if (rsp != NULL)
                *rsp = resp1;
}
#pragma GCC diagnostic warning "-Wcast-qual"

int 
owlsdio_exchange(void *bus, const void *tx, void *rx, int len)
{
        int err = 0;

        if (tx) 
                err = owlsdio_write(bus, tx, len);        
        if (rx) 
                err = owlsdio_read(bus, rx, len);
        
        return err;
}

/**
 * This function is called when data should be transmitted to the device.
 *
 * @param data holds a pointer to the data to transmit, the pointer is the
 *            same as the one passed to owlsdio_cmd().
 * @param len is the number of bytes that should be transmitted.
 *
 */

/* SD_LowLevel_DMA_TxConfig does not allow const data, so disable cast-qual
   here */
#pragma GCC diagnostic ignored "-Wcast-qual" 
int 
owlsdio_write(void *bus, const void *data, int len)
{
        int i;
        uint8_t remain; /* bytes remaining in data array */

        uint8_t tail[16]; /* put remaining bytes in 
                             this buffer */

        owl_unused(bus);

        /* the asserts below will hold if wl_conf_alignment() 
           was set to 16 */
        owl_assert(len % 4 == 0); /* SDIO host can not handle non-size 
                                     aligned data */
        owl_assert(len > 8); /* SDIO host can not handle buffers that are 8
                                bytes or smaller */
        
        owl_dbg("data:%p len:%d", data, len);

        if (!(priv->flags & OWL_SDIO_FLAG_DMA)) {
                const uint32_t *data32 = (const uint32_t *) data;
        
                for (i = 0; i < len / 4;) {
                        uint32_t status;
                        
                        status = SDIO->STA;
                        if (status & (SDIO_FLAG_DTIMEOUT | 
                                      SDIO_FLAG_DCRCFAIL |
                                      SDIO_FLAG_TXUNDERR | 
                                      SDIO_FLAG_STBITERR |
                                      SDIO_FLAG_RXOVERR)) {
                                owl_err("status:0x%08lx", status);
                                return OWL_ERR_CONN;
                        }

                        /* don't write if fifo is full */
                        if (status & SDIO_FLAG_TXFIFOF)
                                continue;
                        
                        /* write 4 bytes at a time */
                        SDIO_WriteData(data32[i++]);
                }
                
        } else {
                if (len / 4 > 0) {
               	
#if defined(STM32F40XX)
                        uint32_t * MemorySRC = (uint32_t*) data;       
                        uint32_t BufferSize = 4 * (len / 4);
                        
                        DMA_InitTypeDef SDDMA_InitStructure;
                        
                        DMA_ClearFlag(SD_SDIO_DMA_STREAM, 
                                      SD_SDIO_DMA_FLAG_FEIF | 
                                      SD_SDIO_DMA_FLAG_DMEIF | 
                                      SD_SDIO_DMA_FLAG_TEIF | 
                                      SD_SDIO_DMA_FLAG_HTIF | 
                                      SD_SDIO_DMA_FLAG_TCIF);
                        
                        /* DMA2 Stream3 disable */
                        DMA_Cmd(SD_SDIO_DMA_STREAM, DISABLE);
                        
                        /* DMA2 Stream3 Deinit */
                        DMA_DeInit(SD_SDIO_DMA_STREAM);
                        
                        /* Setup DMA control registers.
                           Peripheral base adress is SDIO fifo, 
                           Direction is from memory to SDIO,
                           Memory pointer is incremented, 
                           fifo size is 32 bits (one word),
                           each burst is 32 bits (one word). 
                           BufferSize is not used when using SDIO 
                           flowcontrol.*/
                        
                        SDDMA_InitStructure.DMA_Channel = 
                                SD_SDIO_DMA_CHANNEL;
                        SDDMA_InitStructure.DMA_PeripheralBaseAddr = 
                                (uint32_t)SDIO_FIFO_ADDRESS;   
                        SDDMA_InitStructure.DMA_Memory0BaseAddr = 
                                (uint32_t)MemorySRC ;  	    
                        SDDMA_InitStructure.DMA_DIR = 
                                DMA_DIR_MemoryToPeripheral;	 	    
                        SDDMA_InitStructure.DMA_BufferSize = 
                                BufferSize / 4;          	            
                        SDDMA_InitStructure.DMA_PeripheralInc = 
                                DMA_PeripheralInc_Disable;         
                        SDDMA_InitStructure.DMA_MemoryInc = 
                                DMA_MemoryInc_Enable;   		   
                        SDDMA_InitStructure.DMA_PeripheralDataSize = 
                                DMA_PeripheralDataSize_Word;   
                        SDDMA_InitStructure.DMA_MemoryDataSize = 
                                DMA_MemoryDataSize_Word;	    
                        SDDMA_InitStructure.DMA_Mode = 
                                DMA_Mode_Normal;
                        SDDMA_InitStructure.DMA_Priority = 
                                DMA_Priority_High;
                        SDDMA_InitStructure.DMA_FIFOMode = 
                                DMA_FIFOMode_Enable; 
                        SDDMA_InitStructure.DMA_FIFOThreshold = 
                                DMA_FIFOThreshold_Full;
                        SDDMA_InitStructure.DMA_MemoryBurst = 
                                DMA_MemoryBurst_INC4;
                        SDDMA_InitStructure.DMA_PeripheralBurst = 
                                DMA_PeripheralBurst_INC4;
                        
                        DMA_Init(SD_SDIO_DMA_STREAM, &SDDMA_InitStructure);
                        
                        DMA_FlowControllerConfig(SD_SDIO_DMA_STREAM, 
                                                 DMA_FlowCtrl_Peripheral); 	
                        
                        DMA_Cmd(SD_SDIO_DMA_STREAM, ENABLE);
                        
                        /* Start DMA transfer */
                        SDIO_DMACmd(ENABLE);
                
                        /* Check Stream 3 transfer complete flag */
                        while ((uint32_t)DMA_GetFlagStatus(
                                       DMA2_Stream3,
                                       DMA_FLAG_TCIF3)== RESET);
                                   
#else /* STM32F1 DMA TX */	
                        SD_LowLevel_DMA_TxConfig(
                                (uint32_t *) data, 4 * (len / 4));
                        SDIO_DMACmd(ENABLE);
                        while ((SD_DMAEndOfTransferStatus() == RESET));
#endif /* STM32F40XX */
                }
        }

        /* transfer remaining bytes */
        remain = len % 4;
        if (remain > 0) {
                int offset = len - remain;
                uint32_t *data = (uint32_t *) data;
                uint32_t *tail32 = (uint32_t *) tail;

                owl_info("remain:%d offset:%d", remain, offset);

                for (i = 0; i < remain; i++)
                        tail[i] = data[offset + i];

                owl_assert(remain <= sizeof(tail));
                owl_assert(remain % 4 == 0);
                
                for (i = 0; i < remain / 4; ) {
                        uint32_t status;
                        status = SDIO->STA;
                        if (status & (SDIO_FLAG_DTIMEOUT | 
                                      SDIO_FLAG_DCRCFAIL |
                                      SDIO_FLAG_TXUNDERR | 
                                      SDIO_FLAG_STBITERR |
                                      SDIO_FLAG_RXOVERR)) {
                                owl_err("status:0x%08lx", status);
                                return OWL_ERR_CONN;
                        }

                        /* don't write if fifo is full */
                        if (status & SDIO_FLAG_TXFIFOF)
                                continue;

                        /* write 4 bytes at a time */
                        SDIO_WriteData(tail32[i++]);
                }
        }
        
        /* wait until transfer is done */
        owl_dbg("wait for txact");
        while (SDIO_GetFlagStatus(SDIO_FLAG_TXACT) != RESET);

        /* transfer is done */
        owl_assert(SDIO_GetFlagStatus(SDIO_FLAG_DATAEND) == SET);

        SDIO_ClearFlag(SDIO_FLAG_DATAEND);

        /* only hexdump if we're done downloading firmware */
        owl_hexdump("SDIOTX", (const char *) data, len);

        return len;
}
#pragma GCC diagnostic warning "-Wcast-qual"


/**
 * This function will be invoked when data should be received from the device.
 *
 * @param data should hold the read data upon return.
 * @param len is the number of bytes to read.
 *
 */
int 
owlsdio_read(void *bus, void *data, int len)
{
        owl_dbg("data:%p len:%d", data, len);
        owl_assert(len % 4 == 0);

        owl_unused(bus);

        uint32_t status;

        if (priv->flags & OWL_SDIO_FLAG_DMA) {
                (void) len;
                (void) data; /* DMA was started in owl_sdio_cmd() */

#if defined(STM32F40XX) /* STM32F4 DMA RX */
                /* Check Stream 3 transfer complete flag */        
                while ((uint32_t)DMA_GetFlagStatus(
                               DMA2_Stream3,
                               DMA_FLAG_TCIF3) == RESET);
#else
                while ((SD_DMAEndOfTransferStatus() == RESET));
#endif
                
        } else {
                int i = 0;
                uint32_t *data32 = (uint32_t *) data;
                
                while (i < len / 4) {

                        status = SDIO->STA;
                        if (status & (SDIO_FLAG_DTIMEOUT | 
                                      SDIO_FLAG_DCRCFAIL |
                                      SDIO_FLAG_TXUNDERR | 
                                      SDIO_FLAG_STBITERR |
                                      SDIO_FLAG_RXOVERR)) {
                                owl_err("status:0x%08lx", status);
                                return OWL_ERR_CONN;
                        }
                        
                        /* check if there is at least one word of 
                           data in the fifo */
                        if (!(status & SDIO_FLAG_RXDAVL)) {
                                owl_info("rxdavl not set");
                                continue;
                        }
                        
                        /* read one word of data */
                        data32[i++] = SDIO_ReadData();
                }
        }

        if (priv->flags & OWL_SDIO_FLAG_4BIT) {
                /* NOTE: In SDIO 4-bit mode, the CPU seems to 
                   interpret received data on DAT1 as an interrupt 
                   and will set the SDIOIT bit when receiving data 
                   (not when sending). If there is a _real_
                   interrupt from the device just after reading is completed,
                   the SDIOIT bit will be set again right after we clear it 
                   here since the irq line stays low as long as we don't 
                   ACK the IRQ on the device - so we will not lose 
                   any irq's */
                SDIO_ClearITPendingBit(SDIO_IT_SDIOIT);
        }

        do {
                status = SDIO->STA;
        } while(!(status & SDIO_FLAG_DATAEND) && (status & SDIO_FLAG_RXACT));

        SDIO_ClearFlag(SDIO_FLAG_DATAEND);

        owl_hexdump("SDIORX", (char *) data, len);

        return len;
}
