/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 *
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#define OWL_TRACE OWL_TRACE_NONE
#define OWL_FILE "owl_spi"

#include <owl/core/owl_err.h>
#include <owl/core/owl_debug.h>
#include <owl/core/owl_list.h>
#include <owl/util/cbuf.h>
#include <owl/drivers/owl_spi.h>
#include <owl/drivers/owl_gpio.h>


/* STM32 spi does not have support for multiple chip selects per bus, so
   we will reconfigure the bus when the chip select is changed */

struct spi_bus {
        int mode; /* master/slave */
        struct list_head list;
        SPI_TypeDef *addr;
        struct spi_device *last_device;
};

struct spi_device {
        struct spi_bus *bus;
        int cs;
        unsigned int freq;
        int xmode;
        int width;

#if OWL_CONFIG_SPI_SLAVE
        struct cbuf *cbuf;
#endif
        struct list_head list;
};

static struct list_head *devices = NULL;
static struct list_head *buses = NULL;

static void select_device(struct spi_bus *bus, struct spi_device *device);
static struct spi_bus *find_bus(SPI_TypeDef *addr);
static struct spi_device *find_device(int cs);


int 
owlspi_init(void *addr, int mode)
{
        struct spi_bus *bus;

/* Sanity check */
#if !OWL_CONFIG_SPI_SLAVE
        if (mode == OWL_SPI_MODE_SLAVE) {
                owl_err("missing config");
                return OWL_ERR_NOTSUPP;
        }
#endif

        owl_info("initializing spi bus at addr:%p to mode:%d", addr, mode);

        /* initialize bus list on first call */
        if (buses == NULL) {
                if ((buses = owl_os_alloc(sizeof(*buses))) == NULL) {
                        owl_err("could not alloc list");
                        return OWL_ERR_MEM;
                }

                list_init(buses);
        }

        if ((bus = find_bus(addr)) != NULL) {
                if (bus->mode == mode) {
                        owl_info("bus at addr:%p already initialized", addr);
                        return 0;

                } else {
                        owl_err("bus at addr:%p cannot change mode", addr);
                        return OWL_ERR_STATE;
                }
        }

        /* alloc bus context */
        if ((bus = owl_os_zalloc(sizeof(struct spi_bus))) == NULL) {
                owl_err("could not alloc bus");
                return OWL_ERR_MEM;
        }

        bus->addr = addr;
        bus->mode = mode;
        list_add_tail(&bus->list, buses);
        return 0;
}


int
owlspi_setup_cs(void *addr, int cs, int freq, int xmode, int width)
{
        struct spi_bus *bus;
        struct spi_device *device;

        /* Make sure that the bus is initialized */
        if ((bus = find_bus((SPI_TypeDef *) addr)) == NULL) {
                owl_err("bus at addr:%p not initialized", addr);
                return OWL_ERR_STATE;
        }

#if !OWL_CONFIG_SPI_SLAVE
        if (bus->mode == OWL_SPI_MODE_SLAVE) {
                owl_err("bus is already configured as slave");
                return OWL_ERR_PARAM;
        }
#endif

        owl_unused(cs);
        owl_info("setup cs_port:%x cs_pin:%x on addr:%p with freq:%d xmode:%d "
                 "width:%d",
                 STM32_PORT(cs), STM32_PIN(cs), addr, freq, xmode, width);

        switch (width) {
        case 8:
                break; /* ok */
        case 16:
                owl_err("width:%d not supported", width);
                return OWL_ERR_NOTSUPP;
        default:
                owl_err("invalid width:%d", width);
                return OWL_ERR_PARAM;
        };

        switch (xmode) {
        case OWL_SPI_TRANSFER_MODE_0:
        case OWL_SPI_TRANSFER_MODE_1:
        case OWL_SPI_TRANSFER_MODE_2:
        case OWL_SPI_TRANSFER_MODE_3:
                break;

        default:
                owl_err("invalid xmode:%d", xmode);
                return OWL_ERR_PARAM;
        };

        if (bus->mode == OWL_SPI_MODE_MASTER)
                /* CS */
                owlgpio_set(cs);

        /* initialize device list on first call */
        if (devices == NULL) {
                if ((devices = owl_os_alloc(sizeof(*devices))) == NULL) {
                        owl_err("could not alloc list");
                        return OWL_ERR_MEM;
                }

                list_init(devices);
        }

        /* alloc device context */
        if ((device = owl_os_zalloc(sizeof(struct spi_device))) == NULL) {
                owl_err("could not alloc bus");
                return OWL_ERR_MEM;
        }

        device->bus = bus;
        device->width = width;
        device->xmode = xmode;
        device->freq = freq;
        device->cs = cs;
        list_add_tail(&device->list, devices);

#if OWL_CONFIG_SPI_SLAVE
        if (bus->mode == OWL_SPI_MODE_SLAVE) {
                if ((device->cbuf = cbuf_create(1500)) == NULL) {
                        owl_err("could not alloc cbuf for rx");
                        return OWL_ERR_MEM;
                }

                /* Setup SPI for slave mode. 
                   Enable SPI here so its always enabled.*/
                owlspi_claim(bus->addr, device->cs);   
                SPI_Cmd(bus->addr, ENABLE);
        }
#endif
        return 0;
}


void
owlspi_claim(void *addr, int cs)
{
        struct spi_bus *bus;
        struct spi_device *device;

        owl_unused(bus);
        owl_vdbg("addr:%p cs:%d", addr, cs);

        bus = find_bus(addr);
        device = find_device(cs);

        owl_assert(bus);
        owl_assert(device);
    
        select_device(bus, device);

        if (bus->mode == OWL_SPI_MODE_MASTER) {
                /* Enable SPI */
                SPI_Cmd(addr, ENABLE);
                owlgpio_clr(cs);
        }
}

void
owlspi_release(void *addr, int cs)
{
        owl_unused(addr);
        owl_vdbg("addr:%p cs:%d", addr, cs);
     
        struct spi_bus *bus;
        
        bus = find_bus(addr);

        if (bus->mode == OWL_SPI_MODE_MASTER) {
                owlgpio_set(cs);
                /* Disable SPI  */
                SPI_Cmd(addr, DISABLE);
        }
}

int
owlspi_exchange(void *bus, const void *tx, void *rx, int len)
{
        SPI_TypeDef *spi = bus;
        const uint8_t *tx8 = tx;
        uint8_t *rx8 = rx;
        int pos = 0;

        owl_assert(len);
        owl_vdbg("bus:%p tx:%p rx:%p len:%d", bus, tx, rx, len);

        if (tx) {
                owl_hexdump("SPI TX", tx, len);
        }

        /* clock out/in the bytes */
        while (pos < len) {
                while (SPI_I2S_GetFlagStatus(spi, SPI_I2S_FLAG_TXE) 
                       == RESET);

                if (tx)
                        SPI_I2S_SendData(bus, tx8[pos] & 0xff);
                else
                        SPI_I2S_SendData(bus, 0xff);
                
                while (SPI_I2S_GetFlagStatus(spi, SPI_I2S_FLAG_RXNE)
                       == RESET);
                
                if (rx)
                        rx8[pos] = SPI_I2S_ReceiveData(spi);
                else
                        (void) SPI_I2S_ReceiveData(spi);
                pos++;
        }

        if (rx) {
                owl_hexdump("SPI RX", rx, len);
        }

        return len;
}

int
owlspi_write(void *bus, const void *buf, int len)
{
        SPI_TypeDef *spi = bus;
        
        struct spi_bus *spi_bus;
        struct spi_device *device;
        
        spi_bus = find_bus(spi);
        device = spi_bus->last_device;
         
        if (spi_bus->mode == OWL_SPI_MODE_MASTER) 
                return owlspi_exchange(bus, buf, NULL, len);
        
#if OWL_CONFIG_SPI_SLAVE
        else if (spi_bus->mode == OWL_SPI_MODE_SLAVE) {

                owl_unused(bus);
                const uint8_t *tx8 = buf;
                int pos = 0;

                while (pos < len) {
                        while (!(spi_bus->addr->SR & SPI_I2S_FLAG_TXE));
                     
                        SPI_I2S_SendData(spi_bus->addr, tx8[pos] & 0xff);
                        pos++;
                }
        }
#endif
        return len;
}

int
owlspi_read(void *bus, void *buf, int len)
{     
        SPI_TypeDef *spi = bus;
        struct spi_bus *spi_bus;
        int rx_len = 0;

        if ((spi_bus = find_bus(spi)) == NULL) {
                owl_dbg("spi bus not initialized");
                return 0;
        }

        if (spi_bus->mode == OWL_SPI_MODE_MASTER)
                rx_len = owlspi_exchange(bus, NULL, buf, len);
  
#if OWL_CONFIG_SPI_SLAVE
        if (spi_bus->mode == OWL_SPI_MODE_SLAVE) {
                struct spi_device *device;
                int avail;
                device = spi_bus->last_device;
        
                avail = cbuf_rdlen_tot(device->cbuf);
                if (len > avail)
                        len = avail;       

                rx_len = cbuf_get_data(device->cbuf, buf, len); 
        }
#endif
        return rx_len;
        
}

int
owlspi_exchange_ext(void *bus, const void *out, void *in, int len, int flags)
{
        owl_unused(flags);
        return owlspi_exchange(bus, out, in, len);
}

int 
owlspi_poll(void *bus)
{
        int len = 0;
        SPI_TypeDef *spi = bus;
        struct spi_bus *spi_bus;

        if ((spi_bus = find_bus(spi)) == NULL) {
                owl_dbg("spi bus not initialized");
                return 0;
        }
        
        if (spi_bus->mode == OWL_SPI_MODE_MASTER)
                return OWL_ERR_NOTSUPP;
        
#if OWL_CONFIG_SPI_SLAVE
        if (spi_bus->mode == OWL_SPI_MODE_SLAVE) {
                struct spi_device *device;
                device = spi_bus->last_device;
                len =  cbuf_rdlen_tot(device->cbuf);
        }
#endif
        return len;
} 

static void
select_device(struct spi_bus *bus, struct spi_device *device)
{
        SPI_InitTypeDef SPI_InitStructure;
        unsigned int br, div;

        owl_vdbg("bus:%p device:%p", bus, device);

        /* already configured for this device? */
        if (bus->last_device == device)
                return;

        /* Always called when SPI is DISABLED */
        
        /* Only width 8 is supported, this is checked in owlspi_setup_cs() */
        SPI_InitStructure.SPI_DataSize = SPI_DataSize_8b;
        switch (device->xmode) {
        case OWL_SPI_TRANSFER_MODE_0:
                SPI_InitStructure.SPI_CPOL = SPI_CPOL_Low;
                SPI_InitStructure.SPI_CPHA = SPI_CPHA_1Edge;
                break;

        case OWL_SPI_TRANSFER_MODE_1:
                SPI_InitStructure.SPI_CPOL = SPI_CPOL_Low;
                SPI_InitStructure.SPI_CPHA = SPI_CPHA_2Edge;
                break;

        case OWL_SPI_TRANSFER_MODE_2:
                SPI_InitStructure.SPI_CPOL = SPI_CPOL_High;
                SPI_InitStructure.SPI_CPHA = SPI_CPHA_1Edge;
                break;

        case OWL_SPI_TRANSFER_MODE_3:
                SPI_InitStructure.SPI_CPOL = SPI_CPOL_High;
                SPI_InitStructure.SPI_CPHA = SPI_CPHA_2Edge;
                break;

        default:
                owl_assert(0);
                owl_err("invalid xmode:%d", device->xmode);
                break;
        };

        /* SPI configuration */
        SPI_InitStructure.SPI_Direction = SPI_Direction_2Lines_FullDuplex;
        
        /* In Slave mode let the master control CS */
        if (bus->mode == OWL_SPI_MODE_MASTER) {
                SPI_InitStructure.SPI_Mode = SPI_Mode_Master;
                SPI_InitStructure.SPI_NSS = SPI_NSS_Soft;
        } else {
                SPI_InitStructure.SPI_Mode = SPI_Mode_Slave;
                SPI_InitStructure.SPI_NSS = SPI_NSS_Hard;
        }
        
        SPI_InitStructure.SPI_FirstBit = SPI_FirstBit_MSB;
        SPI_InitStructure.SPI_CRCPolynomial = 7;

        for (br = 0, div = 2; br < 8; br++, div <<= 1) {
                if (SystemCoreClock / div < device->freq) {
                        owl_dbg("prescaler:%d gives div:%d and baudrate:%d",
                                br, div, SystemCoreClock / div);
                        break;
                }
        }
        SPI_InitStructure.SPI_BaudRatePrescaler = br << 3;
        SPI_Init(bus->addr, &SPI_InitStructure);

        bus->last_device = device;

        /* SPI will be set to ENABLE by owlspi_claim() if master */
}


static struct spi_bus *
find_bus(SPI_TypeDef *addr)
{
        struct list_head *pos = NULL;
        struct spi_bus *bus = NULL;

        list_for_each(pos, buses) {
                bus = list_entry(pos, struct spi_bus, list);
                if (bus->addr == addr)
                        return bus;
        }

        return NULL;
}

static struct spi_device *
find_device(int cs)
{
        struct list_head *pos = NULL;
        struct spi_device *device = NULL;

        list_for_each(pos, devices) {
                device = list_entry(pos, struct spi_device, list);
                if (device->cs == cs)
                        return device;
        }

        return NULL;
}

#if OWL_CONFIG_SPI_SLAVE
static void
irq_handler(SPI_TypeDef *addr)
{
        uint8_t data;
        volatile uint8_t tmp; 

        struct spi_bus *spi_bus;
        struct spi_device *device;

        spi_bus = find_bus(addr);
        device = spi_bus->last_device;

        if (SPI_I2S_GetITStatus(spi_bus->addr,
                                SPI_I2S_IT_RXNE) == SET) {

                tmp = spi_bus->addr->DR;
                data = (uint8_t) tmp;
                cbuf_put_data(device->cbuf, &data, 1);
        }
}

void
SPI3_IRQHandler(void)
{
        irq_handler(SPI3);
}

void
SPI2_IRQHandler(void)
{
        irq_handler(SPI2);
}
#endif
