/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#define OWL_TRACE OWL_TRACE_MINIMAL
#define OWL_FILE "owl_uart"

#include <owl/core/owl_defs.h>
#include <owl/core/owl_debug.h>
#include <owl/core/owl_err.h>
#include <owl/core/owl_timer.h>
#include <owl/core/owl_os.h>
#include <owl/drivers/owl_uart.h>


struct uart_ctx {
        USART_TypeDef *port;
        char *ringbuf;
        int readpos;
        volatile int writepos;
        int size; /* RINGBUF_DEFAULT_SIZE */
        struct uart_ctx *next;
};

#define RINGBUF_DEFAULT_SIZE 16

static struct uart_ctx *head = NULL;

static struct uart_ctx *find_priv(USART_TypeDef *port);
static int read_ringbuf(struct uart_ctx *priv, char *out);
static void write_ringbuf(struct uart_ctx *priv);

int
owluart_init(void *uart,
             uint32_t baudrate,
             enum uart_databits databits,
             enum uart_parity parity,
             enum uart_stopbits stopbits,
             int rtscts)
{
        USART_InitTypeDef USART_InitStructure;
        struct uart_ctx *priv;
        USART_TypeDef *port = uart;

        owl_info("configuring uart with %lu %d-%s-%d rtscts:%d",
                 baudrate, databits, owluart_parity(parity),
                 stopbits, rtscts);

        USART_InitStructure.USART_BaudRate = baudrate;

#if OWL_CONFIG_API_CHECK
        switch (databits) {
        case UART_DATABITS_8:
                USART_InitStructure.USART_WordLength = USART_WordLength_8b;
                break;
        case UART_DATABITS_9:
                USART_InitStructure.USART_WordLength = USART_WordLength_9b;
                break;
        default:
                return OWL_ERR_NOTSUPP;
        }

        switch (parity) {
        case UART_PARITY_NONE:
                USART_InitStructure.USART_Parity = USART_Parity_No;
                break;
        case UART_PARITY_EVEN:
                USART_InitStructure.USART_Parity = USART_Parity_Even;
                break;
        case UART_PARITY_ODD:
                USART_InitStructure.USART_Parity = USART_Parity_Odd;
                break;
        default:
                return OWL_ERR_NOTSUPP;
        }

        switch (stopbits) {
        case UART_STOPBITS_1:
                USART_InitStructure.USART_StopBits = USART_StopBits_1;
                break;
        case UART_STOPBITS_2:
                USART_InitStructure.USART_StopBits = USART_StopBits_2;
                break;
        default:
                return OWL_ERR_NOTSUPP;
        }
#endif /* OWL_CONFIG_API_CHECK */
        
        if (rtscts)
                USART_InitStructure.USART_HardwareFlowControl =
                        USART_HardwareFlowControl_RTS_CTS;
        else
                USART_InitStructure.USART_HardwareFlowControl =
                        USART_HardwareFlowControl_None;
  
        USART_InitStructure.USART_Mode = USART_Mode_Rx | USART_Mode_Tx;
        USART_Init(port, &USART_InitStructure);

        /* if reinit */
        if ((priv = find_priv(port)) == NULL) {

                if ((priv = owl_os_zalloc(sizeof(*priv))) == NULL)
                        return OWL_ERR_MEM;

                priv->port = port;
                 /* add one since we never fill up the buffer to the last byte
                    to be able to see the difference between full and empty */
                priv->size = RINGBUF_DEFAULT_SIZE + 1;
                if ((priv->ringbuf = owl_os_alloc(priv->size)) == NULL) {
                        owl_os_free(priv);
                        return OWL_ERR_MEM;
                }

                if (head == NULL) {
                        owl_info("adding port %p to %p (head)", priv, head);
                        head = priv;

                } else {
                        struct uart_ctx *iter = head;
                        while (iter->next)
                                iter = iter->next;
                        owl_info("adding port %p to %p", priv, iter);
                        iter->next = priv;
                }
        } else {
                owl_info("reinit port");
        }

        /* init rest of fields since we might be reinitalized */
        priv->writepos = 0;
        priv->readpos = 0;

        /* NVIC configuration should be performed in owl_board.c */
        USART_ITConfig(port, USART_IT_RXNE, ENABLE);

        USART_Cmd(port, ENABLE);
        
        return 0;
}


void
owluart_outb(void *port, int c)
{
        /* write a character to the USART */
        USART_SendData(port, (uint8_t) c);

        /* Loop until the end of transmission */
        while (USART_GetFlagStatus(port, USART_FLAG_TC) == RESET);
}

int
owluart_inb(void *uart, unsigned int timeout)
{
        int err;
        char c;
        struct uart_ctx *priv;
        USART_TypeDef *port = uart;
        unsigned int ms = owltmr_get_ms();

        if ((priv = find_priv(port)) == NULL) {
                owl_err("port:%p not found", port);
                return OWL_ERR_PARAM;
        }

        for (;;) {
                if ((err = read_ringbuf(priv, &c)) == 0) {
                        owl_dbg("0x%x", c); 
                        return c;
                }
                
                if (timeout > 0 &&
                    owltmr_intvl_passed(ms, owltmr_get_ms(), timeout)) {
                        owl_err("got timeout (%d)", timeout);
                        return OWL_ERR_TIMEOUT;
                }
        }
}


int
owluart_write(void *uart, const void *buf, int len)
{
        USART_TypeDef *port = uart;
        const char *buf8 = buf;
        int pos = 0;

        while (pos < len) {
                USART_SendData(port, buf8[pos]);
                
                /* Loop until the end of transmission */
                while (USART_GetFlagStatus(port, USART_FLAG_TC) == RESET);
                pos++;
        }
        
        return len;
}

    
int
owluart_read(void *uart, void *buf, int len)
{
        struct uart_ctx *priv;
        USART_TypeDef *port = uart;
        int diff;
        int readlen = 0;
        int writepos;

        if ((priv = find_priv(port)) == NULL)
                return OWL_ERR_PARAM;

        /*
         * We carefully disable interrupts only when accessing members that
         * will be _changed_ by the isr or when changing members that will be
         * _accessed_ by the isr.
         *
         */
        
        /* writepos is written by irq_handler so make sure we're not interrupted
         * here
         */
        __disable_irq();
        writepos = priv->writepos;
        __enable_irq();

        /* should be safe to read from buffer with interrupts enabled since
         * the irq_handler will only add data to the buffer, and we already got
         * the writepos
         */
        
        diff = writepos - priv->readpos;
        

        if (!diff) {
                return 0;
            
        } else if (diff > 0) {
                int count = diff;
                readlen = len > count ? count : len;
                owl_dbg("buf:%p diff:%d len:%d count:%d readlen:%d readpos:%d",
                        buf, diff, len, count, readlen, priv->readpos);
                
                if (buf)
                        owl_os_memcpy(buf, priv->ringbuf + priv->readpos,
                                      readlen);
                
        } else if (diff < 0) {
                int count = priv->size + diff;
                int taillen = priv->size - priv->readpos;
                readlen = len > count ? count : len;
                owl_dbg("buf:%p diff:%d len:%d count:%d readlen:%d readpos:%d",
                        buf, diff, len, count, readlen, priv->readpos);
                
                if (buf) {
                        
                        if (readlen < taillen) {
                                owl_os_memcpy(buf, 
                                              priv->ringbuf + priv->readpos,
                                              readlen);
                                
                        } else {
                                owl_os_memcpy(buf, 
                                              priv->ringbuf + priv->readpos,
                                              taillen);
                                owl_os_memcpy(buf + taillen, 
                                              priv->ringbuf,
                                              readlen - taillen);
                                
                        }
                }
        }

        /* Disable interrupts to protect
         * A) the readpos update (readpos is read by irq_handler())
         * B) The rx irq reenable check. If the rx irq reenable check is
         *    unprotected we might get an interrupt which disables
         *    the rx irq between reenabling it and updating readpos.
         */
        __disable_irq();

        /* Reenable irq if it was disabled due to full ringbuf */
        if ((priv->writepos + 1) % priv->size == priv->readpos) 
                USART_ITConfig(port, USART_IT_RXNE, ENABLE);
        
        priv->readpos = (priv->readpos + readlen) % priv->size;
        __enable_irq();
        
        return readlen;
}

int 
owluart_exchange(void *uart, const void *tx, void *rx, int len)
{
        int err = 0;

        if (tx) {
                err = owluart_write(uart, tx, len);
        }
	if (rx) {
		while ((err = owluart_poll(uart)) < len);
		err = owluart_read(uart, rx, len);
	}

	return err;
}

int
owluart_poll(void *uart)
{
        struct uart_ctx *priv;
        USART_TypeDef *port = uart;
        int diff;

        if ((priv = find_priv(port)) == NULL)
                return 0;

        __disable_irq();
        diff = priv->writepos - priv->readpos;
        __enable_irq();

        if (diff < 0)
            diff += priv->size;
        
        return diff;
}

const char *
owluart_parity(enum uart_parity parity)
{
        switch (parity) {
        case UART_PARITY_NONE:
                return "N";
        case UART_PARITY_EVEN:
                return "E";
        case UART_PARITY_ODD:
                return "O";
        case UART_PARITY_MARK:
                return "M";
        case UART_PARITY_SPACE:
                return "S";
        };

        owl_assert(0);
        owl_err("invalid parity:%d", parity);
        return "";
}


int
owluart_set_bufsize(void *uart, int bufsize)
{
        struct uart_ctx *priv;
        void *p;
        USART_TypeDef *port = uart;
        if ((priv = find_priv(port)) == NULL) {
                owl_info("port not found uart:%p", uart);
                return OWL_ERR_PARAM;
        }

        if ((p = owl_os_alloc(bufsize)) == NULL)
                return OWL_ERR_MEM;

        /* add one since we never fill up the buffer to the last byte
           to be able to see the difference between full and empty */
        owl_info("setting ringbuf size to %d (+1)", bufsize);
        owl_os_free(priv->ringbuf);
        priv->readpos = 0;
        priv->writepos = 0;
        priv->size = bufsize + 1;
        priv->ringbuf = p;
        return 0;
}


static struct uart_ctx *
find_priv(USART_TypeDef *port)
{
        struct uart_ctx *iter;

        for (iter = head; iter; iter = iter->next) {
                if (iter->port == port)
                        return iter;
        }

        return NULL;
}

static void
irq_handler(USART_TypeDef *port)
{
        struct uart_ctx *iter;

        owl_vdbg("irq");
        iter = find_priv(port);
        if (NULL == iter)
                return;
        write_ringbuf(iter);
        
}

__attribute__((__interrupt__)) void
USART1_IRQHandler(void)
{
        irq_handler(USART1);
}

__attribute__((__interrupt__)) void
USART2_IRQHandler(void)
{
        irq_handler(USART2);
}

__attribute__((__interrupt__)) void
USART3_IRQHandler(void)
{
        irq_handler(USART3);
}

static void
write_ringbuf(struct uart_ctx *priv)
{
        USART_TypeDef *port = priv->port;

        owl_assert((priv->writepos + 1) % priv->size != priv->readpos);

        if (USART_GetITStatus(port, USART_IT_ORE) != RESET) {
                owl_err("error:OVRE");
        }
        
        if (USART_GetITStatus(port, USART_IT_NE) != RESET) {
                owl_err("error:NOISE");
        }
        
        if (USART_GetITStatus(port, USART_IT_FE) != RESET) {
                owl_err("error:FE");
        }
        
        if (USART_GetITStatus(port, USART_IT_PE) != RESET) {
                owl_err("error:PE");
        }

        if (USART_GetITStatus(port, USART_IT_RXNE) != RESET) {
                priv->ringbuf[priv->writepos] = USART_ReceiveData(port);

                owl_vdbg("got 0x%x writepos:%d++ readpos:%d size:%d",
                         priv->ringbuf[priv->writepos],
                         priv->writepos, priv->readpos, priv->size);
                priv->writepos = (priv->writepos + 1) % priv->size;

                /* ringbuf full? */
                if ((priv->writepos + 1) % priv->size == priv->readpos) {
                        /* Disable the USART Receive interrupt */
                        USART_ITConfig(port, USART_IT_RXNE, DISABLE);
                }
        }
}

static int
read_ringbuf(struct uart_ctx *priv, char *out)
{
        __disable_irq();
        if (priv->writepos == priv->readpos) {
                __enable_irq();
                return -1;

        } else {
                owl_vdbg("priv:0x%x writepos:%d readpos:%d",
                        priv, priv->writepos, priv->readpos);
                *out = priv->ringbuf[priv->readpos];
                owl_vdbg("got 0x%x", *out);
                priv->readpos = (priv->readpos + 1) % priv->size;
                __enable_irq();
                return 0;
        }
}


