/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 *
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */


#include <owl/core/owl_defs.h>

#include <wlp_api.h>
#include <wlp_ethernet.h>
#include <wlp_mcast.h>
#include <wlp_concurrent.h>

#include <lwip/opt.h>
#include <lwip/def.h>
#include <lwip/mem.h>
#include <lwip/pbuf.h>
#include <lwip/stats.h>
#include <lwip/sys.h>
#include <lwip/igmp.h>
#include <netif/etharp.h>
#include <netif/wlif.h>

#define IFNAME0 'w'
#define IFNAME1 'l'

static int
rx_cb(void *ctx, int len)
{
        struct netif *netif = ctx;
        struct pbuf *p;
        struct eth_hdr *ethhdr;
        void *stripped_pkt;
        int stripped_pkt_len;
        int wlp_hdr_len;
        int err;
        
        LWIP_DEBUGF(NETIF_DEBUG, ("rx_cb: len:%d\n", len));
        
        /* We allocate a continous pbuf */

        if (!len)
                len = 1600;
                        
        if ((p = pbuf_alloc(PBUF_RAW, len, PBUF_RAM)) == NULL) {
                LWIP_DEBUGF(NETIF_DEBUG, ("rx_cb: fail to alloc "
                                          "pbuf of len:%"S32_F"\n", len));
                return -1; /* rx_cb will be invoked again */
        }
        
        if ((err = wlp_rx(p->payload, p->len)) < 0) {
                LWIP_DEBUGF(NETIF_DEBUG, ("wlp_rx failed, err:%d\n", err));
                pbuf_free(p);
                return -1;

        }

        if ((err = wlp_process_rx(
                      p->payload, p->len,
                      &stripped_pkt, &stripped_pkt_len, NULL)) <= 0) {
                LWIP_DEBUGF(NETIF_DEBUG, ("not for lwip, err:%d\n", err));
                pbuf_free(p);
                return 0;
        }

        LINK_STATS_INC(link.recv);

        wlp_hdr_len = (char *) stripped_pkt - (char *) p->payload;
        pbuf_realloc(p, stripped_pkt_len + wlp_hdr_len);

        ethhdr = p->payload;

        switch (htons(ethhdr->type)) {
                /* IP or ARP packet? */
        case ETHTYPE_IP:
        case ETHTYPE_ARP:
#if PPPOE_SUPPORT
                /* PPPoE packet? */
        case ETHTYPE_PPPOEDISC:
        case ETHTYPE_PPPOE:
#endif /* PPPOE_SUPPORT */
                LWIP_DEBUGF(NETIF_DEBUG, ("rx_cb: to lwip\n"));
                if (netif->input(p, netif) != ERR_OK) {
                        LWIP_DEBUGF(NETIF_DEBUG, ("wlif_input: input error\n"));
                        pbuf_free(p);
                }
                break;

        default:
                LWIP_DEBUGF(NETIF_DEBUG, ("wlif_input: inval type:%04x\n",
                                          htons(ethhdr->type)));
                pbuf_free(p);
                break;
        }

        return 0;
}




/**
 * This function should do the actual transmission of the packet. The packet is
 * contained in the pbuf that is passed to the function. This pbuf
 * might be chained.
 *
 * @param netif the lwip network interface structure for this ethernetif
 * @param p the MAC packet to send (e.g. IP packet including MAC addresses and
 *        type)
 * @return ERR_OK if the packet could be sent
 *         an err_t value if the packet couldn't be sent
 *
 * @note Returning ERR_MEM here if a DMA queue of your MAC is full can lead to
 *       strange results. You might consider waiting for space in the DMA queue
 *       to become availale since the stack doesn't retry to send a packet
 *       dropped because of memory failure (except for the TCP timers).
 */
static err_t
low_level_output(struct netif *netif, struct pbuf *p)
{
        int iface, prev;
        int err;
        struct pbuf *q;

        iface = (int) netif->state;

        prev = wlp_set_iface(iface);
        
        if ((err = wlp_process_tx((char *) p->payload + WLP_HEADER_SIZE,
                                  p->len - WLP_HEADER_SIZE,
                                  p->tot_len - WLP_HEADER_SIZE,
                                  p->payload, 0)) < 0) {

                wlp_set_iface(prev);

                LWIP_DEBUGF(NETIF_DEBUG, ("wlp_process_tx, err:%d\n", err));
                if (err == WLP_ERR_MEM)
                        return ERR_MEM;
                else
                        return ERR_IF;
        }


        /* Send the data from the pbuf to the interface, one pbuf at a
         * time. The size of the data in each pbuf is kept in the ->len
         * variable.
         */
        for (q = p; q != NULL; q = q->next) {
                if ((err = wlp_tx(q->payload, q->len)) < 0) {
                        LWIP_DEBUGF(NETIF_DEBUG,
                                    ("wlp_tx failed err:%d\n", err));

                        /* Errors from wlp_tx() are fatal */
                        wlp_set_iface(prev);
                        return ERR_IF;
                }
        }
        
        wlp_set_iface(prev);

        LWIP_DEBUGF(NETIF_DEBUG, ("wlp_tx len:%d\n", p->tot_len));

        return ERR_OK; /* no one seems to check this anyway */
}

#if LWIP_IGMP
static err_t
igmp_mac_filter(struct netif *netif, ip_addr_t *group, u8_t action)
{
        struct wlp_mac_addr mcast_addr;
        uint32_t ipaddr = (uint32_t) group->addr;

        mcast_addr.octet[0]  = 0x01;
        mcast_addr.octet[1]  = 0x00;
        mcast_addr.octet[2]  = 0x5e;
        mcast_addr.octet[3]  = (ipaddr >> 8) & 0xff;
        mcast_addr.octet[3] &= ~(1 << 7);
        mcast_addr.octet[4]  = (ipaddr >> 16) & 0xff;
        mcast_addr.octet[5]  = (ipaddr >> 24) & 0xff;

        if (action == IGMP_ADD_MAC_FILTER)
                wlp_add_mcast(&mcast_addr);
        else
                wlp_del_mcast(&mcast_addr);

	return 0;	
}
#endif /* LWIP_IGMP */

/**
 * Should be called at the beginning of the program to set up the
 * network interface. It calls the function low_level_init() to do the
 * actual setup of the hardware.
 *
 * This function should be passed as a parameter to netif_add().
 *
 * @param netif the lwip network interface structure for this ethernetif
 * @return ERR_OK if the loopif is initialized
 *         ERR_MEM if private data couldn't be allocated
 *         any other err_t on error
 */
err_t
wlif_init(struct netif *netif)
{
        int iface, prev;
        int err;
        
        LWIP_ASSERT("netif != NULL", (netif != NULL));

#if LWIP_NETIF_HOSTNAME
        /* Initialize interface hostname */
        if (NULL == netif->hostname) {
                netif->hostname = "wlif";
        }
#endif /* LWIP_NETIF_HOSTNAME */

        netif->name[0] = IFNAME0;
        netif->name[1] = IFNAME1;

        /* We directly use etharp_output() here to save a function call.
         * You can instead declare your own function an call etharp_output()
         * from it if you have to do some checks before sending (e.g. if link
         * is available...) */
        netif->output = etharp_output;
        netif->linkoutput = low_level_output;
#if LWIP_IGMP
        netif_set_igmp_mac_filter(netif, igmp_mac_filter);
#endif /* LWIP_IGMP */
        
        /* device capabilities */
        netif->flags = NETIF_FLAG_BROADCAST | NETIF_FLAG_ETHARP |
                NETIF_FLAG_IGMP;
        /* NETIF_FLAG_LINK_UP must be set only when we have an wlan assoc */

        /* set MAC hardware address length */
        netif->hwaddr_len = ETHARP_HWADDR_LEN;

        iface = (int) netif->state;
        if ((prev = wlp_set_iface(iface)) < 0) {
                LWIP_DEBUGF(NETIF_DEBUG,
                            ("fail to set iface to %d, err:%d\n", iface, prev));
                return ERR_IF;
        }
        
        
	if ((err = wlp_get_hwaddr((struct wlp_mac_addr *) netif->hwaddr))) {
                LWIP_DEBUGF(NETIF_DEBUG,
                            ("fail to get mac address, err:%d\n", err));
                return ERR_IF;
        }
        
        /* maximum transfer unit */
#if defined(WLP_MTU_SIZE)
        netif->mtu = WLP_MTU_SIZE;
#else
        netif->mtu = 1500;
#endif

        wlp_set_rx_cb(rx_cb, netif);
        wlp_set_iface(prev);
        return ERR_OK;
}

/* Call this after a sequence of wlp_shutdown(), wlp_init() */
void
wlif_reinit(struct netif *netif)
{
        int iface, prev;

        iface = (int) netif->state;
        
        prev = wlp_set_iface(iface);
        wlp_set_rx_cb(rx_cb, netif);
        wlp_set_iface(prev);
}

void
wlif_poll(struct netif* netif)
{
        owl_unused(netif);
}

const int wlif_flags = 0;
