/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#define OWL_TRACE OWL_TRACE_ERR
#define OWL_FILE "hdg205"

#include <owl/core/owl_debug.h>
#include <owl/core/owl_defs.h>
#include <owl/core/owl_err.h>
#include <owl/components/hdg205.h>

#if defined(OWL_BOARD_HDG205_SHUTDOWN_PIN)
# include <owl/drivers/owl_board.h>
# include <owl/drivers/owl_gpio.h>
#endif /* defined(OWL_BOARD_HDG205_SHUTDOWN_PIN) */

#include <hdg205/wlp_hdg205.h>

int
hdg205_resume(void)
{
#if defined(OWL_BOARD_HDG205_SHUTDOWN_PIN)
        owlgpio_cfg(OWL_BOARD_HDG205_SHUTDOWN_PIN, OWL_GPIO_OUTPUT);
        owlgpio_clr(OWL_BOARD_HDG205_SHUTDOWN_PIN);
        owlboard_mdelay(1000);
#endif /* defined(OWL_BOARD_HDG205_SHUTDOWN_PIN) */
        return 0;
}

int
hdg205_suspend(void)
{
#if defined(OWL_BOARD_HDG205_SHUTDOWN_PIN)
        owlgpio_cfg(OWL_BOARD_HDG205_SHUTDOWN_PIN, OWL_GPIO_OUTPUT);
        owlgpio_set(OWL_BOARD_HDG205_SHUTDOWN_PIN);
        owlboard_mdelay(10);
#endif /* defined(OWL_BOARD_HDG205_SHUTDOWN_PIN) */
        return 0;
}


#if OWL_CONFIG_THREAD

# include <owl/util/rmutex.h>

/* Used by pico/thread/.c, see those files for details */
struct rmutex wlp_lock;

int
wlp_init(int (*exchange_cb)(void *ctx, const void *tx, void *rx, int len),
         void (*select_cb)(void *ctx, int select),
         void (*cmd_cb)(void *ctx, uint8_t idx, uint32_t arg,
                        uint32_t *rsp, const void *data, int len),
         int (*poll_cb)(void *ctx),
         void *ctx,
	 int flags)
{

	int err;
	
	owl_dbg("flags:0x%08x", flags);
	
	if ((err = rmutex_init(&wlp_lock))) {
		owl_err("fail to init rmutex, err:%d", err);
		return err;
	}

	if ((err = rmutex_lock(&wlp_lock)) < 0) {
                owl_err("destroyed while waiting, err:%d", err);
                return err;
        }

	if ((err = _wlp_init(
                     exchange_cb, select_cb, cmd_cb, poll_cb, ctx, flags))) {
		owl_err("wlp init failed, err:%d", err);
		rmutex_unlock(&wlp_lock);
		rmutex_destroy(&wlp_lock);
                return err;
        }
	
	rmutex_unlock(&wlp_lock);

	return 0;	
}

void
wlp_shutdown(void)
{
	int err;

	owl_dbg("enter");

	if ((err = rmutex_lock(&wlp_lock)) < 0) {
                owl_err("destroyed while waiting, err:%d", err);
                return;
        }
        
	_wlp_shutdown();

	/* Any other thread waiting on rmutex_lock() will return with
	   OWL_ERR_DESTROYED */
	rmutex_destroy(&wlp_lock);
}

void
wlp_set_error_cb(void (*error_cb)(void *ctx, int err), void *ctx)
{
	int err;

	owl_dbg("enter");

	if ((err = rmutex_lock(&wlp_lock)) < 0) {
                owl_err("destroyed while waiting, err:%d", err);
                return;
        }
        
	_wlp_set_error_cb(error_cb, ctx);

	rmutex_unlock(&wlp_lock);
}

int
wlp_program(int (*firmware_cb)(void *ctx, void *buf, int offset, int len),
            void *ctx)
{
	int err;

	owl_dbg("enter");

	if ((err = rmutex_lock(&wlp_lock)) < 0) {
                owl_err("destroyed while waiting, err:%d", err);
                return err;
        }
        
	err = _wlp_program(firmware_cb, ctx);

	rmutex_unlock(&wlp_lock);

	return err;
}

int
wlp_get_version(char *driver, int driver_len, char *firmware, int firmware_len)
{
	int err;

	owl_dbg("enter");

	if ((err = rmutex_lock(&wlp_lock)) < 0) {
                owl_err("destroyed while waiting, err:%d", err);
                return err;
        }
        
	err = _wlp_get_version(driver, driver_len, firmware, firmware_len);

	rmutex_unlock(&wlp_lock);

	return err;

}

#endif /* OWL_CONFIG_THREAD */
