/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#ifndef OWL_DEBUG_H
#define OWL_DEBUG_H

#include <owl/core/owl_defs.h>
#include <owl/core/owl_os.h>

/**
 * In each source files that uses owl_err(), owl_info(), owl_dbg() or
 * owl_assert(), define OWL_TRACE to the desired debug level before including
 * this file; e.g
 *
 * #define OWL_TRACE OWL_TRACE_INFO
 * #include <owl/core/owl_debug.h>
 *
 * Debug traces and assertions can be enabled by defining
 * OWL_DEBUG and/or OWL_ASSERT in the makefile.
 *
 * OWL_TRACE_ALWAYS can be defined in source files to force debugging.
 *
 */

/* Default trace level */
#ifndef OWL_TRACE
# define OWL_TRACE OWL_TRACE_ERR
#endif

#define OWL_TRACE_NONE    0
#define OWL_TRACE_MINIMAL 1
#define OWL_TRACE_ERR     2
#define OWL_TRACE_INFO    3
#define OWL_TRACE_DBG     4
#define OWL_TRACE_VDBG    5
#define OWL_TRACE_DATA    6

#define OWL_TRACE_DEBUG -1
#if OWL_TRACE == OWL_TRACE_DEBUG
# error "use OWL_TRACE_DBG instead of OWL_TRACE_DEBUG"
#endif

#ifndef OWL_FILE
# ifdef OWL_TRACE_NO_FILE
#  define OWL_FILE "(none)"
# else
#  define OWL_FILE __FILE__
# endif
#endif

#ifdef OWL_DEBUG_NO_TIME
# define _TRACE_PREFIX "%s:%d %s: "
# define _DEBUG_TIME
#else
extern uint32_t debug_time;
# define _TRACE_PREFIX "[%010d] %s:%d %s: "
# define _DEBUG_TIME debug_time,
#endif

#if defined(OWL_ASSERT)
# define owl_assert(cond)                                               \
        do {                                                            \
                if (!(cond)) {                                          \
                        for (;;);                                       \
                }                                                       \
        } while(0)

#elif defined(OWL_DEBUG)
# define owl_assert(cond)                                               \
        do {                                                            \
                if (!(cond)) {                                          \
                        owl_os_printf("--- ASSERTION FAILED ---\n\r");  \
                        owl_os_printf("-A- " _TRACE_PREFIX "%s\n\r",    \
                                      _DEBUG_TIME OWL_FILE, __LINE__,   \
                                      __func__, #cond);                 \
                }                                                       \
        } while(0)

#else
# define owl_assert(...) do {} while (0)
#endif

# define owl_ndbg(...) do {} while (0)

#if (defined(OWL_DEBUG) && OWL_TRACE >= OWL_TRACE_VDBG) ||              \
        (defined(OWL_TRACE_ALWAYS) && OWL_TRACE >= OWL_TRACE_VDBG)       
# ifdef OWL_DEBUG_NO_VAR_MACROS
#  define owl_vdbg(...) do {						\
	owl_os_printf("-V- " _TRACE_PREFIX,			\
		      _DEBUG_TIME OWL_FILE, __LINE__, __func__);	\
	owl_os_printf(__VA_ARGS__);					\
	owl_os_printf("\n");						\
    } while (0)
# else 
#  define owl_vdbg(format, args...)                                      \
        owl_os_printf("-V- " _TRACE_PREFIX format "\n\r", _DEBUG_TIME   \
                      OWL_FILE, __LINE__, __func__, ## args)
# endif
#else
# define owl_vdbg(...) do {} while (0)
#endif

#if (defined(OWL_DEBUG) && OWL_TRACE >= OWL_TRACE_DBG) ||               \
        (defined(OWL_TRACE_ALWAYS) && OWL_TRACE >= OWL_TRACE_DBG)       
# ifdef OWL_DEBUG_NO_VAR_MACROS
#  define owl_dbg(...) do {						\
	owl_os_printf("-D- " _TRACE_PREFIX,			\
		      _DEBUG_TIME OWL_FILE, __LINE__, __func__);	\
	owl_os_printf(__VA_ARGS__);					\
	owl_os_printf("\n");						\
    } while (0)
# else 
#  define owl_dbg(format, args...)                                     \
        owl_os_printf("-D- " _TRACE_PREFIX format "\n\r", _DEBUG_TIME \
                      OWL_FILE, __LINE__, __func__, ## args)
# endif
#else
# define owl_dbg(...) do {} while (0)
#endif

#if (defined(OWL_DEBUG) && OWL_TRACE >= OWL_TRACE_INFO) ||              \
        (defined(OWL_TRACE_ALWAYS) && OWL_TRACE >= OWL_TRACE_INFO)       
# ifdef OWL_DEBUG_NO_VAR_MACROS
#  define owl_info(...) do {                                             \
	owl_os_printf("-I- " _TRACE_PREFIX,		\
		      _DEBUG_TIME OWL_FILE, __LINE__, __func__);	\
	owl_os_printf(__VA_ARGS__);					\
        owl_os_printf("\n");						\
    } while (0)
# else 
#  define owl_info(format, args...)                                      \
        owl_os_printf("-I- " _TRACE_PREFIX format "\n\r", _DEBUG_TIME   \
                      OWL_FILE, __LINE__, __func__, ## args)
# endif
#else
# define owl_info(...) do {} while (0)
#endif

#if (defined(OWL_DEBUG) && OWL_TRACE >= OWL_TRACE_ERR) ||               \
        (defined(OWL_TRACE_ALWAYS) && OWL_TRACE >= OWL_TRACE_ERR)       
# ifdef OWL_DEBUG_NO_VAR_MACROS
#  define owl_err(...) do {                                             \
	owl_os_printf("-I- " _TRACE_PREFIX,			\
		      _DEBUG_TIME OWL_FILE, __LINE__, __func__);	\
	owl_os_printf(__VA_ARGS__);					\
        owl_os_printf("\n");						\
    } while (0)
# else 
#  define owl_err(format, args...)                                       \
        owl_os_printf("-E- " _TRACE_PREFIX format "\n\r", _DEBUG_TIME   \
                      OWL_FILE, __LINE__, __func__, ## args)
# endif
#elif (defined(OWL_DEBUG) && OWL_TRACE >= OWL_TRACE_MINIMAL) ||         \
        (defined(OWL_TRACE_ALWAYS) && OWL_TRACE >= OWL_TRACE_MINIMAL)       
# ifdef OWL_DEBUG_NO_VAR_MACROS
#  define owl_err(...) do {                                             \
	owl_os_printf("-E- " _TRACE_PREFIX "\n",			\
		      _DEBUG_TIME OWL_FILE, __LINE__, __func__);	\
    } while (0)
# else 
#  define owl_err(format, args...)                                       \
        owl_os_printf("-E- [%010d] %s:%d\n\r", _DEBUG_TIME              \
                      OWL_FILE, __LINE__, ## args)
# endif
#else
# define owl_err(...) do {} while (0)
#endif

void owl_hexdump_f(const char *title, const void *pos, int len);

#if (defined(OWL_DEBUG) && OWL_TRACE >= OWL_TRACE_DATA) ||      \
        (defined(OWL_TRACE_ALWAYS) && OWL_TRACE >= OWL_TRACE_DATA)       
# define owl_hexdump(...) owl_hexdump_f(__VA_ARGS__)
#else
# define owl_hexdump(...) do {} while (0)
#endif

#endif


