/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#ifndef OWL_DEFS_H
#define OWL_DEFS_H

#include <owl/core/owl_opt.h>

/* This is the only place where we should include owl_cc.h and owl_port.h */
#include <owl_cc.h>
#include "owl_port.h"

/* owl_cc.h should define the following types
 *
 * uint8_t
 * int8_t
 * uint16_t
 * int16_t
 * uint32_t
 * int32_t
 * ssize_t
 */
#define OWL_TRUE  1
#define OWL_FALSE 0

#ifndef LITTLE_ENDIAN
# define LITTLE_ENDIAN 0x1e /* LSB first */
#endif

#ifndef BIG_ENDIAN
# define BIG_ENDIAN    0xbe    /* MSB first */
#endif

#ifndef container_of
# define container_of(ptr, type, member)                         \
        ((type *)( (char *) ptr - offsetof(type,member)))
#endif /* !defined(container_of) */

#define owl_unused(var) (void) var
#define owl_array_size(a) (sizeof(a) / sizeof((a)[0]))

#define owl_align(addr, align) (((addr) / (align)) * (align))
#define owl_roundup(addr, align) (owl_align((addr) - 1, (align)) + (align))
#define owl_rounddown owl_align

#define owl_mbit(n) ((n) * 1024 * 128)
#define owl_mbyte(n) ((n) * 1024 * 1024)
#define owl_kbyte(n) ((n) * 1024)

#define owl_abs(x) ((x) < 0 ? -(x) : (x))
#define owl_min(x, y) ((x) < (y) ? (x) : (y))
#define owl_max(x, y) ((x) > (y) ? (x) : (y))

#define _in_range(c, lo, up)     ((uint8_t) c >= lo && (uint8_t) c <= up)
#define owl_isprint(c)          _in_range(c, 0x20, 0x7f)
#define owl_isdigit(c)          _in_range(c, '0', '9')
#define owl_isxdigit(c)         (owl_isdigit(c) || _in_range(c, 'a', 'f') || \
                                 _in_range(c, 'A', 'F'))
#define owl_islower(c)          _in_range(c, 'a', 'z')
#define owl_isupper(c)          _in_range(c, 'A', 'Z')
#define owl_isspace(c)          (c == ' ' || c == '\f' || c == '\n' || \
                                 c == '\r' || c == '\t' || c == '\v')

#if !defined(INT8_MIN)
# define INT8_MIN         (-127-1)
#endif
#if !defined(INT16_MIN)
# define INT16_MIN        (-32767-1)
#endif
#if !defined(INT32_MIN)
# define INT32_MIN        (-2147483647-1)
#endif
#if !defined(INT8_MAX)
# define INT8_MAX         +127
#endif
#if !defined(INT16_MAX)
# define INT16_MAX        +32767
#endif
#if !defined(INT32_MAX)
# define INT32_MAX        +2147483647
#endif
#if !defined(UINT8_MAX)
# define UINT8_MAX         255
#endif
#if !defined(UINT16_MAX)
# define UINT16_MAX        65535
#endif
#if !defined(UINT32_MAX)
# define UINT32_MAX        4294967295U
#endif

/* Posix time - Seconds since CLOCK_EPOCH
   Used by ntp.c and clock.c */
typedef uint32_t xtime_t;

/* A port of owl can provide its own owl_bswap16 and owl_bswap32 function. This
   is mainly useful when lots of calls to owl_bswapx are used. Then the code
   size could reduced by having a function rather than a macro for this.
   Also, when building a library that uses owl_defs, we don't want to rely on
   any other external functions than those declared in owl_os.h - so the
   owl_port.h used when building such libraries can have its own implementation
   of bswap */

#if defined(OWL_BSWAP16)
# define owl_bswap16 OWL_BSWAP16
#else
# define owl_bswap16(x)                        \
        (((((uint16_t) (x)) << 8) & 0xff00) |  \
         ((((uint16_t) (x)) >> 8) & 0xff))
#endif

#if defined(OWL_BSWAP32)
# define owl_bswap32 OWL_BSWAP32
#else
# define owl_bswap32(x) (((((uint32_t) (x)) << 24) & 0xff000000) |    \
                         ((((uint32_t) (x)) << 8) & 0xff0000) |       \
                         ((((uint32_t) (x)) >> 8) & 0xff00) |         \
                         ((((uint32_t) (x)) >> 24) & 0xff))
#endif

#define owl_bitset(flags, mask) ((flags) |= (mask))
#define owl_bitclr(flags, mask) ((flags) &= ~(mask))
#define owl_bit(flags, mask)    ((flags) & (mask))

#if BYTE_ORDER == LITTLE_ENDIAN
# define owl_htons(x)           owl_bswap16(x)
# define owl_ntohs(x)           owl_bswap16(x)
# define owl_htonl(x)           owl_bswap32(x)
# define owl_ntohl(x)           owl_bswap32(x)
# define owl_le16_to_cpu(x)     (x)
# define owl_le32_to_cpu(x)     (x)
# define owl_be16_to_cpu(x)     owl_bswap16(x)
# define owl_be32_to_cpu(x)     owl_bswap32(x)
# define owl_cpu_to_le16(x)     (x)
# define owl_cpu_to_le32(x)     (x)
# define owl_cpu_to_be16(x)     owl_bswap16(x)
# define owl_cpu_to_be32(x)     owl_bswap32(x)
#elif BYTE_ORDER == BIG_ENDIAN
# define owl_htons(x)           (x)
# define owl_ntohs(x)           (x)
# define owl_htonl(x)           (x)
# define owl_ntohl(x)           (x)
# define owl_le16_to_cpu(x)     owl_bswap16(x)
# define owl_le32_to_cpu(x)     owl_bswap32(x)
# define owl_be16_to_cpu(x)     (x)
# define owl_be32_to_cpu(x)     (x)
# define owl_cpu_to_le16(x)     owl_bswap16(x)
# define owl_cpu_to_le32(x)     owl_bswap32(x)
# define owl_cpu_to_be16(x)     (x)
# define owl_cpu_to_be32(x)     (x)
#else
# error "BYTE_ORDER not defined"
#endif

#endif
