/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#ifndef OWL_EVENTLOG_H
#define OWL_EVENTLOG_H

#include <owl/core/owl_defs.h>
#include <owl/core/owl_shell.h>
#include <owl/drivers/owl_board.h>

/**
 * The owl_eventlog_dbg(), owl_eventlog_info() and owl_eventlog_err() macros
 * should be used to capture events without affecting the timing. Therefore,
 * the implementation of these events requires very few cpu cycles by storing
 * the event data in a ram buffer. Up to OWL_CONFIG_EVENTLOG_SIZE events can be
 * stored in the buffer before it wraps around. Note that
 * OWL_CONFIG_EVENTLOG_SIZE must be a power of 2 due to the way wrapping is
 * implemented.
 * 
 * The captured events can be shown by the "events" command anytime. This will
 * show the capture time (relative to the first event in the log), delta time,
 * the function, associated string and integer.
 *
 * The event log level can be set in each source file by defining OWL_EVENTLOG.
 * The eventlog is enabled by defining OWL_CONFIG_EVENTLOG e.g. in the Makefile.
 *
 */

#define OWL_EVENTLOG_NONE    0
#define OWL_EVENTLOG_ERR     1
#define OWL_EVENTLOG_INFO    2
#define OWL_EVENTLOG_DBG     3
#define OWL_EVENTLOG_VDBG    4

#ifndef OWL_CONFIG_EVENTLOG_SIZE
# define OWL_CONFIG_EVENTLOG_SIZE 256 /* must be power of 2 */
#endif

#define OWL_EVENTLOG_LEVEL_MASK 0x0f
#define OWL_EVENTLOG_NODATA  0x80

struct owl_eventlog_data {
	uint32_t us;
	const char *desc;
	const char *func;
	const char *file;
        uint8_t level;
        int line;
	int data;
};

extern const struct owlsh_cmd cmd_eventlog;
extern struct owl_eventlog_data _edata[OWL_CONFIG_EVENTLOG_SIZE];
extern uint16_t _enext;

#ifndef OWL_FILE
# ifdef OWL_TRACE_NO_FILE
#  define OWL_FILE "(none)"
# else
#  define OWL_FILE __FILE__
# endif
#endif

/* requires 17 cy on ucr2 */
#define _owl_eventlog(s, i, l) do {                                     \
        struct owl_eventlog_data *e = &_edata[_enext];                  \
		e->desc = s;                                            \
		e->data = (int) i;                                      \
		e->func = __func__;                                     \
		e->file = OWL_FILE;                                     \
		e->line = __LINE__;                                     \
		e->us = owlboard_get_us();                              \
                e->level = l;                                           \
		_enext++;						\
                _enext &= ~OWL_CONFIG_EVENTLOG_SIZE;                    \
	} while(0)



/**
 * Write an event with an associated string and integer value at debug level.
 * The string and the integer will be shown when printing the event log with
 * the events command.
 *
 *
 * @param s is a string to add to this log entry
 * @param i is an integer value to add to this log entry
 */
#if OWL_CONFIG_EVENTLOG && (OWL_EVENTLOG >= OWL_EVENTLOG_VDBG)
# define owl_eventlog_vdbg(s, i) \
        _owl_eventlog(s, i, OWL_EVENTLOG_VDBG)
# define owl_eventstr_vdbg(s) \
        _owl_eventlog(s, 0, OWL_EVENTLOG_VDBG | OWL_EVENTLOG_NODATA)
#else
# define owl_eventlog_vdbg(s, i)
# define owl_eventstr_vdbg(s)
#endif

/**
 * Write an event with an associated string and integer value at debug level.
 * The string and the integer will be shown when printing the event log with
 * the events command.
 *
 * Use the dbg level for events that informs us about a specific state and that
 * occurs regularly.
 *
 * @param s is a string to add to this log entry
 * @param i is an integer value to add to this log entry
 */
#if OWL_CONFIG_EVENTLOG && (OWL_EVENTLOG >= OWL_EVENTLOG_DBG)
# define owl_eventlog_dbg(s, i) \
        _owl_eventlog(s, i, OWL_EVENTLOG_DBG)
# define owl_eventstr_dbg(s) \
        _owl_eventlog(s, 0, OWL_EVENTLOG_DBG | OWL_EVENTLOG_NODATA)
#else
# define owl_eventlog_dbg(s, i)
# define owl_eventstr_dbg(s)
#endif


/**
 * Write an event with an associated string and integer value at info level.
 * The string and the integer will be shown when printing the event log with
 * the events command.
 *
 * Use the info level for events that informs us about a specific state, but
 * does not apperar regularly.
 *
 * @param s is a string to add to this log entry
 * @param i is an integer value to add to this log entry
 */
#if OWL_CONFIG_EVENTLOG && (OWL_EVENTLOG >= OWL_EVENTLOG_INFO)
# define owl_eventlog_info(s, i) \
        _owl_eventlog(s, i, OWL_EVENTLOG_INFO)
# define owl_eventstr_info(s) \
        _owl_eventlog(s, 0, OWL_EVENTLOG_INFO | OWL_EVENTLOG_NODATA)
#else
# define owl_eventlog_info(s, i)
# define owl_eventstr_info(s)
#endif


/**
 * Write an event with an associated string and integer value at error level.
 * The string and the integer will be shown when printing the event log with
 * the events command.
 *
 * Use the error level for important events that we should certainly be notified
 * of.
 *
 * @param s is a string to add to this log entry
 * @param i is an integer value to add to this log entry
 */
#if OWL_CONFIG_EVENTLOG && (OWL_EVENTLOG >= OWL_EVENTLOG_ERR)
# define owl_eventlog_err(s, i) \
        _owl_eventlog(s, i, OWL_EVENTLOG_ERR)
# define owl_eventstr_err(s) \
        _owl_eventlog(s, 0, OWL_EVENTLOG_ERR | OWL_EVENTLOG_NODATA)
#else
# define owl_eventlog_err(s, i)
# define owl_eventstr_err(s)
#endif


#endif /* OWL_EVENTLOG_H */

