/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#define OWL_TRACE OWL_TRACE_MINIMAL
#define OWL_EVENTLOG OWL_EVENTLOG_NONE
#define OWL_FILE "owl_lwip"

#include <owl/core/owl_defs.h>
#include <owl/core/owl_debug.h>
#include <owl/core/owl_err.h>
#include <owl/core/owl_timer.h>
#include <owl/core/owl_eventlog.h>
#include <owl/core/owl_lwip.h>

#include <lwip/init.h>
#include <lwip/tcp_impl.h>
#include <lwip/ip_frag.h>
#if OWL_CONFIG_SOCKET_BSD
# include <lwip/tcpip.h>
#endif
#include <netif/etharp.h>
#include <netif/wlif.h>

struct owlip_ctx {

#if !OWL_CONFIG_SOCKET_BSD
        struct owltmr_tmo tcp_tmo;
        uint32_t tick;
#endif

        struct list_head ifaces;
};

static void ip_status_cb(struct netif *netif);
#if !OWL_CONFIG_SOCKET_BSD
static void tcp_tmr_cb(struct owltmr_tmo *tmo, void *ctx);
static void wifi_tmr_cb(struct owltmr_tmo *tmo, void *ctx);
#endif /* !OWL_CONFIG_SOCKET_BSD */

static struct owlip_ctx *priv = NULL;

int
owlip_init(void)
{
        owl_info("enter");

        /* Some legacy code might call this function multiple times - make sure
           we don't break that code */
        if (priv != NULL)
                return 0;
        
        if ((priv = owl_os_zalloc(sizeof(*priv))) == NULL)
                return OWL_ERR_MEM;

        list_init(&priv->ifaces);
        
        /* With OS 'owl', this is done in owl_os_init() */
#if !OWL_CONFIG_SOCKET_BSD
# if !OWL_CONFIG_OS_OWL
        lwip_init();
# endif
        owltmr_set_name(&priv->tcp_tmo, "tcp");
        
#else /* !OWL_CONFIG_SOCKET_BSD */
        tcpip_init(NULL, NULL);
#endif /* !OWL_CONFIG_SOCKET_BSD */

        return 0;
}

int
owlip_add(int id)
{
        struct iface_ctx *iface;
        struct ip_addr ipaddr, netmask, gw;
        struct netif *netif;
        
        owl_info("id:%d", id);

        /* Check that owlip_init() has been called */
        if (priv == NULL)
                return OWL_ERR_STATE;

        if ((iface = owlip_get_iface(id)) != NULL) {
                owl_info("already initialized");
                wlif_reinit(&iface->netif);
                return 0;
        }

        /* Allocate and initialize a new interface */
        if ((iface = owl_os_zalloc(sizeof(*iface))) == NULL)
                return OWL_ERR_MEM;
        
        iface->id = id;
        netif = &iface->netif;

        IP4_ADDR(&gw, 0,0,0,0);
        IP4_ADDR(&ipaddr, 0,0,0,0);
        IP4_ADDR(&netmask, 0,0,0,0);
        
        /* iface->id will be used by wlxif to know which Wi-Fi interface to use
           in the calls to wlp_set_iface() */
        if ((netif_add(
                     netif, &ipaddr, &netmask, &gw,
                     (void *) iface->id, wlif_init,
#if !OWL_CONFIG_SOCKET_BSD
		     ethernet_input
#else
		     tcpip_input
#endif /* !OWL_CONFIG_SOCKET_BSD */
		     )) == NULL) {
                owl_os_free(iface);
                return OWL_FAILURE;
        }

        netif_set_status_callback(netif, ip_status_cb);

        /* Set first interface as default */
        if (list_empty(&priv->ifaces)) {
                netif_set_default(netif);
        }

#if !OWL_CONFIG_SOCKET_BSD
        if (owl_bit(wlif_flags, WLIF_FLAG_POLL))
                owltmr_addtmo(0, 1, wifi_tmr_cb, iface);
#endif /* !OWL_CONFIG_SOCKET_BSD */
        
        list_add_tail(&iface->list, &priv->ifaces);
        return 0;
}

void
owlip_linkup(int id)
{
        struct iface_ctx *iface;
        struct netif *netif;
        owl_eventstr_info("linkup");

        if (priv == NULL)
                return;

        if ((iface = owlip_get_iface(id)) == NULL)
                return;

        netif = &iface->netif;
        netif_set_link_up(netif);
        
        if (owl_bit(iface->flags, FLAG_DHCP)) {

                if (iface->dhcp_state == DHCP_STATE_IDLE ||
                    iface->dhcp_state == DHCP_STATE_INFORMED) {
                        struct ip_addr ipaddr, netmask, gw;

#if OWL_CONFIG_HIBERNATE
                        if (owl_bit(iface->flags, FLAG_DHCP_INFORM)) {
                                netif_set_ipaddr(netif, &iface->ipaddr);
                                netif_set_netmask(netif, &iface->netmask);
                                netif_set_gw(netif, &iface->gw);
                                dns_setserver(0, &iface->dns);

                                owl_info("linkup, informing dhcp");
                                dhcp_inform(netif);
                                netif_set_up(netif);
                                iface->dhcp_state = DHCP_STATE_INFORMED;
                        } else
#endif /* OWL_CONFIG_HIBERNATE */
                        {
                                IP4_ADDR(&gw, 0,0,0,0);
                                IP4_ADDR(&ipaddr, 0,0,0,0);
                                IP4_ADDR(&netmask, 0,0,0,0);
                                netif_set_ipaddr(netif, &ipaddr);
                                netif_set_netmask(netif, &netmask);
                                netif_set_gw(netif, &gw);

                                owl_info("linkup, requesting dhcp");
                                dhcp_start(netif);

                                iface->dhcp_state = DHCP_STATE_RUN;
                        }
                }

        } else if (!ip_addr_isany(&netif->ip_addr)) {
                /* bring the interface up if a static ip is configured */

                owl_info("linkup, set interface up");
                netif_set_up(netif);

        } else {
                owl_info("linkup, netif still down");
        }

        owlip_start();
}

void
owlip_linkdown(int id)
{
        struct iface_ctx *iface;
        struct netif *netif;
        
        owl_eventstr_info("linkdown");

        if (priv == NULL)
                return;

        if ((iface = owlip_get_iface(id)) == NULL)
                return;

        netif = &iface->netif;
        netif_set_link_down(netif);
}

void
owlip_start(void)
{
#if !OWL_CONFIG_SOCKET_BSD
        owltmr_sched(&priv->tcp_tmo, TCP_TMR_INTERVAL, 1, tcp_tmr_cb, NULL);
#endif
}

void
owlip_stop(void)
{
#if !OWL_CONFIG_SOCKET_BSD
        owltmr_cancel(&priv->tcp_tmo);
#endif
}

struct iface_ctx *
owlip_get_iface(int id)
{
        struct list_head *pos = NULL;

        if (priv == NULL)
                return NULL;

        list_for_each(pos, &priv->ifaces) {
                struct iface_ctx *iface;
                iface = list_entry(pos, struct iface_ctx, list);
                if (iface->id == id)
                        return iface;
        }

        return NULL;
}

/* Used to be lwip_error and lived in owl_net.c */
int
owlip_error(err_t err)
{
        switch (err) {
        case ERR_OK:
                return 0;
        case ERR_ISCONN: /* fall through */
        case ERR_RTE:
                return OWL_ERR_STATE;
        case ERR_VAL:
                return OWL_ERR_PARAM;
        case ERR_MEM:
                return OWL_ERR_MEM;
        case ERR_USE:
                return OWL_ERR_INUSE;
        case ERR_INPROGRESS:
                return OWL_ERR_BUSY;
        default:
                owl_err("unknown lwip return code:%d", err);
                return OWL_FAILURE;
        };
}

static void
ip_status_cb(struct netif *netif)
{
        struct iface_ctx *iface;
        int id;

        id = (int) netif->state;
        iface = owlip_get_iface(id);
        owl_assert(iface != NULL);

        if (netif_is_up(netif)) {
                owl_eventstr_info("ip:1");
                owl_info("bound to %d.%d.%d.%d",
                         ip4_addr1(&netif->ip_addr),
                         ip4_addr2(&netif->ip_addr),
                         ip4_addr3(&netif->ip_addr),
                         ip4_addr4(&netif->ip_addr));

                if (netif->dhcp) {
                        owl_info("lease:%d renew:%d rebind:%d",
                                 netif->dhcp->offered_t0_lease,
                                 netif->dhcp->offered_t1_renew,
                                 netif->dhcp->offered_t2_rebind);
                }
                
#if OWL_CONFIG_HIBERNATE
                ip_addr_set(&iface->ipaddr, &netif->ip_addr);
                ip_addr_set(&iface->netmask, &netif->netmask);
                ip_addr_set(&iface->gw, &netif->gw);
                iface->dns = dns_getserver(0);
                owl_bitset(iface->flags, FLAG_DHCP_INFORM);
#endif /* OWL_CONFIG_HIBERNATE */

                if (iface->addr_cb)
                        iface->addr_cb(iface->addr_ctx, &netif->ip_addr);
                
        } else {
                owl_eventstr_info("ip:0");
                owl_info("ip status down");
                if (iface->addr_cb)
                        iface->addr_cb(iface->addr_ctx, NULL);
        }
}
            
#if !OWL_CONFIG_SOCKET_BSD
static void
tcp_tmr_cb(struct owltmr_tmo *tmo, void *ctx)
{
        owl_unused(tmo);
        owl_unused(ctx);

        priv->tick++;

        tcp_tmr();
        if (!(priv->tick % (5000 / TCP_TMR_INTERVAL)))
                etharp_tmr();
        if (!(priv->tick % (DHCP_FINE_TIMER_MSECS / TCP_TMR_INTERVAL)))
                dhcp_fine_tmr();
        if (!(priv->tick % (DHCP_COARSE_TIMER_MSECS / TCP_TMR_INTERVAL)))
                dhcp_coarse_tmr();
        if (!(priv->tick % (IP_TMR_INTERVAL / TCP_TMR_INTERVAL)))
                ip_reass_tmr();
#if LWIP_DNS == 1
        if (!(priv->tick % (DNS_TMR_INTERVAL / TCP_TMR_INTERVAL)))
                dns_tmr();
#endif
}

static void
wifi_tmr_cb(struct owltmr_tmo *tmo, void *ctx)
{
        struct iface_ctx *iface = ctx;

        owl_unused(tmo); /* unused */

        wlif_poll(&iface->netif);
}
#endif /* !OWL_CONFIG_SOCKET_BSD */

#if OWL_CONFIG_HIBERNATE

struct hibernate_entry {
        uint8_t flags;
        uint8_t id;
        struct ip_addr ipaddr;
        struct ip_addr netmask;
        struct ip_addr gw;
        struct ip_addr dns;
};

struct hibernate_ctx {
        unsigned int count;
        struct hibernate_entry entry[];
};

int
owlip_serialize(void *data, int len)
{
        struct hibernate_ctx *p = data;
        struct list_head *pos;
        int count = 0;
        int tot_len = 0;

        owl_info("data:%p len:%d", data, len);

        if (data == NULL)
                return OWL_ERR_PARAM;

        /* Serialize header */
        p = data;
        tot_len += sizeof(*p);
        if (tot_len > len) {
                owl_err("not enough room, current tot_len:%d len:%d",
                        tot_len, len);
                return OWL_ERR_SIZE;
        }

        /* Serialize interfaces */
        list_for_each(pos, &priv->ifaces) {
                struct hibernate_entry *entry;
                struct iface_ctx *iface;

                iface = list_entry(pos, struct iface_ctx, list);

                tot_len += sizeof(*entry);
                if (tot_len > len) {
                        owl_err("not enough room, current tot_len:%d len:%d",
                                tot_len, len);
                        return OWL_ERR_SIZE;
                }
                
                entry = &p->entry[count];
                
                entry->flags = 0;
                entry->id = iface->id;
                if (owl_bit(iface->flags, FLAG_DHCP_INFORM)) {
                        owl_bitset(entry->flags, FLAG_DHCP_INFORM);
                        ip_addr_set(&entry->ipaddr, &iface->ipaddr);
                        ip_addr_set(&entry->netmask, &iface->netmask);
                        ip_addr_set(&entry->gw, &iface->gw);
                        ip_addr_set(&entry->dns, &iface->dns);
                        
                        owl_dbg("ipaddr:%s", ip_ntoa(&entry->ipaddr));
                        owl_dbg("netmask:%s", ip_ntoa(&entry->netmask));
                        owl_dbg("gw:%s", ip_ntoa(&entry->gw));
                        owl_dbg("dns:%s", ip_ntoa(&entry->dns));
                }

                count++;
        }
        p->count = count;
        
        return tot_len;
}

int
owlip_deserialize(const void *data, int len)
{
        const struct hibernate_ctx *p;
        struct list_head *pos;
        int count;
        int tot_len = 0;
        
        owl_info("data:%p len:%d", data, len);
        
        if (data == NULL)
                return OWL_ERR_PARAM;

        /* Deserialize header */
        p = data;
        tot_len += sizeof(*p);
        if (tot_len > len) {
                owl_err("not enough data, current tot_len:%d len:%d",
                        tot_len, len);
                return OWL_ERR_SIZE;
        }
        
        for (count = 0; count < p->count; count++) {
                /* Find the iface with the corresponding id */
                list_for_each(pos, &priv->ifaces) {
                        const struct hibernate_entry *entry;
                        struct iface_ctx *iface;

                        iface = list_entry(pos, struct iface_ctx, list);
                        entry = &p->entry[count];
                        if (iface->id != entry->id)
                                continue;

                        /* Found */

                        tot_len += sizeof(*entry);
                        if (tot_len > len) {
                                owl_err("not enough data, current tot_len:%d "
                                        "len:%d", tot_len, len);
                                return OWL_ERR_SIZE;
                        }

                        if (owl_bit(entry->flags, FLAG_DHCP_INFORM)) {
                                owl_bitset(iface->flags, FLAG_DHCP_INFORM);
                                ip_addr_set(&iface->ipaddr, &entry->ipaddr);
                                ip_addr_set(&iface->netmask, &entry->netmask);
                                ip_addr_set(&iface->gw, &entry->gw);
                                ip_addr_set(&iface->dns, &entry->dns);
                                
                                owl_dbg("ipaddr:%s", ip_ntoa(&iface->ipaddr));
                                owl_dbg("netmask:%s", ip_ntoa(&iface->netmask));
                                owl_dbg("gw:%s", ip_ntoa(&iface->gw));
                                owl_dbg("dns:%s", ip_ntoa(&iface->dns));
                        }
                }
        }

        return tot_len;

}
#endif /* OWL_CONFIG_HIBERNATE */
