/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#define OWL_TRACE OWL_TRACE_ERR
#define OWL_FILE "owl_shell"

#include <owl/core/owl_defs.h>
#include <owl/core/owl_shell.h>
#include <owl/core/owl_debug.h>
#include <owl/core/owl_err.h>
#include <owl/core/owl_list.h>
#include <owl/core/owl_os.h>

#if OWL_CONFIG_SHELL_OUTPUT_QUEUE
# include <owl/core/owl_timer.h>

struct owlsh_buf {
        char str[OWL_CONFIG_SHELL_PRINTF_LEN];
        int len;
        int sent;
        struct list_head list;
};
#endif

/* Internal command representation */
struct owlsh_icmd {
        const struct owlsh_cmd *cmd;
        void *ctx;
        const char *instance;
        struct list_head list;
};

struct owlsh_ctx {
        void *data;                             /* context, passed to cmd_f */
        struct list_head icmds;
        const char *prompt;                     /* prompt */
        int (*write_f)(void *ctx, const void *data, int len);
        void *ctx;                              /* passed to write_f */
    
        char *buf;
        int buflen;
        int curpos;
        int escape;
#if OWL_CONFIG_SHELL_HISTORY
        char *history;
        int historylen;
#define NO_CHAR 0
#define FIRST_CHAR 1
#define SECOND_CHAR 2
        int historykey;
#endif /* OWL_CONFIG_SHELL_HISTORY */

        int flags;
        
        int argc;
        const char *argv[OWL_CONFIG_SHELL_MAXARGS];
        const char *instance; /* current instance of command */

        const struct owlsh_cmd *defcmd;
        void *defctx;
        
#if OWL_CONFIG_SHELL_OUTPUT_QUEUE
        struct owltmr_tmo *tmo;
        struct list_head outq; /* list of owlsh_buf */
#endif /* OWL_CONFIG_SHELL_OUTPUT_QUEUE */

};

static void parse_line(char *line, int *argc, const char **argv);
static struct owlsh_icmd *owlsh_find_cmd(struct owlsh_ctx *priv,
                                         const char *label,
                                         const char *instance);
#if OWL_CONFIG_SHELL_OUTPUT_QUEUE
static void timer_cb(struct owltmr_tmo *tmo, void *ctx);
static void process_outq(struct owlsh_ctx *priv);
#endif /* OWL_CONFIG_SHELL_OUTPUT_QUEUE */

void
owlsh_destroy(struct owlsh_ctx *priv)
{
        struct list_head *pos, *tmp;

        if (priv == NULL)
                return;

        owl_os_free(priv->buf);

        list_for_each_safe(pos, tmp, &priv->icmds) {
                struct owlsh_icmd *icmd;
                icmd = list_entry(pos, struct owlsh_icmd, list);
                list_del(&icmd->list);
                owl_os_free(icmd);
        }

#if OWL_CONFIG_SHELL_OUTPUT_QUEUE
        owltmr_deltmo(priv->tmo);
        list_for_each_safe(pos, tmp, &priv->outq) {
                struct owlsh_buf *buf;
                buf = list_entry(pos, struct owlsh_buf, list);
                list_del(&buf->list);
                owl_os_free(buf);
        }
#endif /* OWL_CONFIG_SHELL_OUTPUT_QUEUE */
        
        owl_os_free(priv);
}

struct owlsh_ctx *
owlsh_create(const char *prompt, 
             int (*write_f)(void *ctx, const void *data, int len), void *ctx,
             int flags)
{
        struct owlsh_ctx *priv;
        
        if (write_f == NULL)
                return NULL;

#if !OWL_CONFIG_SHELL_OUTPUT_COMPACT
        if (flags & OWLSH_COMPACT) {
                owl_err("compact output not enabled");
                return NULL;
        }
#endif

#if !OWL_CONFIG_SHELL_OUTPUT_FULL
        if (!(flags & OWLSH_COMPACT)) {
                owl_err("compact output required");
                return NULL;
        }
#endif
        
        priv = owl_os_alloc(sizeof(struct owlsh_ctx));
        if (priv == NULL)
                return NULL;
        owl_os_memset(priv, 0, sizeof(struct owlsh_ctx));
        
        priv->prompt = prompt ? prompt : "$";

        priv->buflen = OWL_CONFIG_SHELL_CMDLEN;
        priv->buf = owl_os_alloc(priv->buflen);

        if (priv->buf == NULL) {
                owlsh_destroy(priv);
                return NULL;
        }
#if OWL_CONFIG_SHELL_HISTORY
        priv->history = owl_os_zalloc(priv->buflen);
                
        if (priv->history == NULL) {
                owlsh_destroy(priv);
                return NULL;
        }
#endif /* OWL_CONFIG_SHELL_HISTORY */

        owl_os_memset(priv->buf, 0, priv->buflen);

        priv->flags = flags;
        priv->write_f = write_f;
        priv->ctx = ctx;

        list_init(&priv->icmds);
        
#if OWL_CONFIG_SHELL_OUTPUT_QUEUE
        list_init(&priv->outq);
        if ((priv->tmo = owltmr_addtmo(0, 1, timer_cb, priv)) == NULL) {
                owl_err("failed to add timer");
                owlsh_destroy(priv);
                return NULL;
        }
#endif /* OWL_CONFIG_SHELL_OUTPUT_QUEUE */

        
        if (!(priv->flags & OWLSH_NOPROMPT))
                owlsh_printf(priv, "%s ", priv->prompt);
        return priv;
}

int
owlsh_addcmd(struct owlsh_ctx *priv, const struct owlsh_cmd *cmd, void *ctx)
{
        return owlsh_addcmd_ext(priv, cmd, ctx, NULL);
}

int
owlsh_addcmd_ext(struct owlsh_ctx *priv, const struct owlsh_cmd *cmd,
                 void *ctx, const char *instance)
{
        struct owlsh_icmd *icmd;

        if ((icmd = owl_os_zalloc(sizeof(*icmd))) == NULL) {
                owl_err("could not alloc icmd");
                return OWL_ERR_MEM;
        }
        
        icmd->cmd = cmd;
        icmd->ctx = ctx;
        icmd->instance = instance;
        
        list_add_tail(&icmd->list, &priv->icmds);
        return 0;
}


void
owlsh_defcmd(struct owlsh_ctx *priv, const struct owlsh_cmd *cmd, void *ctx)
{
        priv->defcmd = cmd;
        priv->defctx = ctx;
}

void *
owlsh_data(struct owlsh_ctx *priv)
{
        return priv->data;
}

int
owlsh_compact(struct owlsh_ctx *priv)
{
        return priv->flags & OWLSH_COMPACT;
}

const char *
owlsh_instance(struct owlsh_ctx *priv)
{
        return priv->instance;
}

void
owlsh_key(struct owlsh_ctx *priv, char key)
{
        struct owlsh_icmd *icmd = NULL;

        if (priv == NULL)
                return;
        
        /* check escapechars */
        if (priv->escape) {
                /* save one char for the null termination */
                if (priv->curpos >= OWL_CONFIG_SHELL_CMDLEN - 1) {
                        priv->escape = 0;
                        return;
                }

                switch (key) {
                case 'r':
                        priv->buf[priv->curpos] = '\r';
                        break;
                        
                case 'n':
                        priv->buf[priv->curpos] = '\n';
                        break;

                case '\\':
                        priv->buf[priv->curpos] = '\\';
                        break;
                }
                
                priv->curpos++;
                priv->escape = 0;
                return;
        }

        switch (key) {
        case '\\':
                if (priv->flags & OWLSH_ESCAPECHARS) {
                        priv->escape = 1;
                        break;
                }
                
                /* fall through */

        case '\b':
                if (priv->curpos == 0)
                        break;

                if (!(priv->flags & OWLSH_NOECHO))
                        owlsh_printf(priv, "\b \b");
                priv->curpos--;
                break;

        case '\r':
        case '\n':
                priv->buf[priv->curpos] = '\0';

                if (!(priv->flags & OWLSH_NOECHO))
                        owlsh_printf(priv, "\r\n");

                owl_hexdump("SHELL RX", priv->buf, priv->curpos);

#if OWL_CONFIG_SHELL_HISTORY
                if (priv->curpos > 1) {
                        owl_os_memset(priv->history, 0, priv->buflen);
                        owl_os_memcpy(priv->history, priv->buf, priv->curpos);
                        priv->historylen = priv->curpos;
                }
#endif /* OWL_CONFIG_SHELL_HISTORY */

                parse_line(priv->buf, &priv->argc, priv->argv);
                icmd = owlsh_find_cmd(priv, priv->argv[0], priv->argv[1]);
                if (icmd != NULL) {
                        owl_info("invoking cmd, buf:'%s'", priv->buf);
                        priv->data = icmd->ctx;
                        /* Command matched on instance */
                        if (priv->argv[1] && icmd->instance &&
                            !owl_os_strcmp(priv->argv[1], icmd->instance)) {
                                priv->instance = icmd->instance;
                                priv->argv[1] = priv->argv[0];
                                priv->argc--;
                                icmd->cmd->cmd_f(priv, priv->argc,
                                                 &priv->argv[1]);
                        } else {
                                priv->instance = NULL;
                                icmd->cmd->cmd_f(priv, priv->argc, priv->argv);
                        }
                } else if (priv->defcmd) {
                        owl_info("invoking defcmd");
                        priv->data = priv->defctx;

                        priv->defcmd->cmd_f(priv, priv->argc, priv->argv);
                } else if (owl_os_strlen(priv->argv[0])) {
                        owl_info("unknown cmd, buf:'%s'", priv->buf);
                        if (owlsh_compact(priv))
                                owlsh_perr(priv, OWL_ERR_USAGE);
                        else
                                owlsh_printf(priv, "command not found:%s\r\n",
                                             priv->argv[0]);
                }

                priv->curpos = 0;

                if (!(priv->flags & OWLSH_NOPROMPT))
                        owlsh_printf(priv, "%s ", priv->prompt);
                break;

#if OWL_CONFIG_SHELL_HISTORY
        case 0x1b: /* first ascii character of up arrow */
                if (priv->historykey == NO_CHAR) {
                        priv->historykey = FIRST_CHAR;
                        break;
                } 
                /* fall through */
        
        case 0x5b: /* second ascii character of up arrow */
                if (priv->historykey == FIRST_CHAR) {
                        priv->historykey = SECOND_CHAR;
                        break;
                }
                /* fall through */
        
        case 0x41: { /* third ascii character of up arrow */
                int tmp_key =  priv->historykey;
                priv->historykey = NO_CHAR;

                if (tmp_key == SECOND_CHAR) {
                        if(priv->historylen) {
                                /* Remove everything on row */
                                while(priv->curpos) {
                                        owlsh_printf(priv, "\b \b");
                                        priv->curpos--;
                                }
                                
                                /* Print history to shell and 
                                   put history in shell buffer */
                                owlsh_printf(priv, priv->history);
                                owl_os_memcpy(priv->buf,
                                      priv->history,
                                              priv->historylen);
                                
                                priv->curpos = priv->historylen;
                        }
                        break;
                }
                /* fall through */
        }
#endif /* OWL_CONFIG_SHELL_HISTORY */
        default:
                /* save one char for the null termination */
                if (priv->curpos >= OWL_CONFIG_SHELL_CMDLEN - 1)
                        break;
                
                /* only store visible chars */
                if (key >= ' ' && key < 0x7f) {
                        if (!(priv->flags & OWLSH_NOECHO))
                                owlsh_printf(priv, "%c", key);
                        priv->buf[priv->curpos] = key;
                        priv->curpos++;
                }

                break;
        }
}

int
owlsh_input(struct owlsh_ctx *priv, const char *buf, int len)
{
        int i;
        
        for (i = 0; i < len; i++)
                owlsh_key(priv, buf[i]);

        return len;
}

void
owlsh_hexdump(struct owlsh_ctx *priv, const char *title, const void *pos,
              int len)
{
        const uint8_t *cpos = pos;
        char line[80], *p;
        int i, j;
        
        if (!owlsh_compact(priv))
                owlsh_printf(priv, "%s - hexdump(len=%d):\r\n", title, len);
        
	for (i = 0; i < len; i += 16) {
		p = line;

                p += owl_os_snprintf(p, line + sizeof(line) - p, "%04x  ", i);
                
		for (j = 0; j < 16; j++) {
                        if (i + j < len)
                                owl_os_snprintf(p, line + sizeof(line) - p,
                                                "%02x ", cpos[i + j]);
			else
				owl_os_strcpy(p, "   ");
			p += 3;
		}

                for (j = 0; j < 16; j++) {
                        if (i + j < len) {
                                if (owl_isprint(cpos[i + j]))
                                        *p++ = cpos[i + j];
                                else
                                        *p++ = '.';
                        }
                }
                
		*p = '\0';
		owlsh_printf(priv, "%s%s\r\n",
                             owlsh_compact(priv) ? "" : "   ", line);
	}
} 

int
owlsh_perr(struct owlsh_ctx *priv, int err)
{
        if (priv->flags & OWLSH_COMPACT)
                return owlsh_printf(
                        priv, "%s;%d\r\n", err >= 0 ? "ok" : "err", err);
        else
#if OWL_CONFIG_PERROR
                return owlsh_printf(
                        priv, "%s %d (%s)\r\n", err >= 0 ? "ok" : "err", err,
                        owl_error(err));
#else
                return owlsh_printf(priv, "%s %d\r\n", err >= 0 ? "ok" : "err",
                                    err);
#endif /* OWL_CONFIG_PERROR */
}

#if OWL_CONFIG_SHELL_OUTPUT_QUEUE
int
owlsh_printf(struct owlsh_ctx *priv, const char *fmt, ...)
{
        va_list args;
        struct owlsh_buf *buf;
        int len;
        
        va_start(args, fmt);

        if ((buf = owl_os_zalloc(sizeof(*buf))) == NULL) {
                owl_err("could not alloc buf");
                return OWL_ERR_MEM;
        }
        
        if ((len = owl_os_vsnprintf(
                     buf->str, sizeof(buf->str), fmt, args)) < 0) {
                owl_err("vsnprintf failed, len:%d", len);
                owl_os_free(buf);
                return OWL_ERR_MEM;
        }
        buf->len = len;

        owl_hexdump("SHELL TX", buf->str, buf->len);
        owl_info("response:'%s'", buf->str);
        list_add_tail(&buf->list, &priv->outq);
        process_outq(priv); /* buf might be free'd upon return */
        return len;
}

#else

int
owlsh_printf(struct owlsh_ctx *priv, const char *fmt, ...)
{
#if OWL_CONFIG_SHELL_PRINTF_STACK
        char str[OWL_CONFIG_SHELL_PRINTF_LEN];
#else /* OWL_CONFIG_SHELL_PRINTF_STACK */
        char *str = NULL;
#endif /* OWL_CONFIG_SHELL_PRINTF_STACK */
        va_list args;
        int len;

        va_start(args, fmt);

#if !OWL_CONFIG_SHELL_PRINTF_STACK
        if ((str = owl_os_alloc(OWL_CONFIG_SHELL_PRINTF_LEN)) == NULL)
                return OWL_ERR_MEM;
#endif /* !OWL_CONFIG_SHELL_PRINTF_STACK */
        
        if ((len = owl_os_vsnprintf(
                     str, OWL_CONFIG_SHELL_PRINTF_LEN, fmt, args)) < 0) {
#if !OWL_CONFIG_SHELL_PRINTF_STACK
                owl_os_free(str);
#endif /* !OWL_CONFIG_SHELL_PRINTF_STACK */
                return OWL_ERR_MEM;
        }

        owl_hexdump("SHELL TX", str, len);
        len = priv->write_f(priv->ctx, str, len);
#if !OWL_CONFIG_SHELL_PRINTF_STACK
        owl_os_free(str);
#endif /* !OWL_CONFIG_SHELL_PRINTF_STACK */
        return len;
}
#endif /* OWL_CONFIG_SHELL_OUTPUT_QUEUE */

/**
 * Caller responsible for freeing the result string.
 *
 */ 
char *
owlsh_joinargs(int argc, const char *argv[])
{
        int i;
        char *str = owl_os_zalloc(1);
        if (str == NULL)
                return NULL;

        /* join arg list */
        for (i = 0; i < argc; i++) {
                int sep;
                void *p;

                /* one byte extra for the null char and one byte for the
                 * extra space added unless this was the last token
                 */
                sep = i < argc - 1 ? 1 : 0;
                p = owl_os_realloc(
                        str,
                        owl_os_strlen(str) + owl_os_strlen(argv[i]) + 1 + sep);
                if (p == NULL) {
                        owl_os_free(str);
                        return NULL;
                }
                        
                str = p;
                owl_os_strcpy(str + owl_os_strlen(str), argv[i]); /* strcat */
                if (sep)
                        owl_os_strcpy(str + owl_os_strlen(str), " ");
        }

        return str;
}

static struct owlsh_icmd *
owlsh_find_cmd(struct owlsh_ctx *priv, const char *label, const char *instance)
{
        struct list_head *pos = NULL;
        struct owlsh_icmd *icmd = NULL;
        struct owlsh_icmd *iter = NULL;
        int ambiguous = 0;
        
        owl_vdbg("label:%s instance:%s", label, instance);

        /* Prevent match if we only have a single command */
        if (!owl_os_strlen(label))
                return NULL;

        list_for_each(pos, &priv->icmds) {
                iter = list_entry(pos, struct owlsh_icmd, list);

                /* exact match with a command that has multiple instances */
                if (!owl_os_strcmp(iter->cmd->label, label) &&
                    iter->instance &&
                    !owl_os_strcmp(iter->instance, instance))
                        return iter;

                /* exact match with single instance command */
                if (!owl_os_strcmp(iter->cmd->label, label) &&
                    iter->instance == NULL)
                        return iter;

                /* no partial match? */
                if (owl_os_strncmp(iter->cmd->label, label,
                                   owl_os_strlen(label)))
                        continue;

                /* ambiguous */
                if (icmd)
                        ambiguous = 1;
                
                icmd = iter;
        }
        
        return ambiguous ? NULL : icmd;
}


static void
parse_line(char *line, int *argc, const char **argv)
{
        char *tok;

        /* defaults */
        *argc = 0;
        argv[0] = ""; /* label */
        argv[1] = ""; /* instance */
        
        for (tok = owl_os_strtok(line, " ");
             tok;
             tok = owl_os_strtok(NULL, " ")) {
                argv[(*argc)++] = tok;
                if (*argc == OWL_CONFIG_SHELL_MAXARGS)
                        return;
        }
}

static void
help_f(struct owlsh_ctx *priv, int argc, const char *argv[])
{
        struct list_head *pos = NULL;
        struct owlsh_icmd *iter = NULL;

        owl_unused(argc);
        owl_unused(argv);
        
        list_for_each(pos, &priv->icmds) {
                char name[20];
                iter = list_entry(pos, struct owlsh_icmd, list);
                owl_os_snprintf(name, sizeof(name), "%s %s",
                                iter->cmd->label,
                                iter->instance ? iter->instance : "");
                owlsh_printf(priv, "%-20s %s\r\n", name,
                             iter->cmd->help ? iter->cmd->help : "");
        }
}


#if OWL_CONFIG_SHELL_OUTPUT_QUEUE
static void
timer_cb(struct owltmr_tmo *tmo, void *ctx)
{
        struct owlsh_ctx *priv = ctx;

        owl_unused(tmo);

        process_outq(priv);
}

static void
process_outq(struct owlsh_ctx *priv)
{
        struct list_head *pos, *tmp;
        struct owlsh_buf *buf;
        int err;

        list_for_each_safe(pos, tmp, &priv->outq) {
                buf = list_entry(pos, struct owlsh_buf, list);

                owl_vdbg("send buf:%p", buf);

                if ((err = priv->write_f(
                             priv->ctx, buf->str + buf->sent, buf->len)) < 0) {
                        owl_err("fail to send buf:%p offset:%d len:%d",
                                buf, buf->sent, buf->len);
                        return;
                }

                owl_assert(err <= buf->len);
                
                buf->len -= err;
                buf->sent += err;

                if (buf->len) {
                        /* data remains in the buffer, so try to transmit it on
                           next timer */
                        owl_dbg("buf:%p has len:%d more data", buf, buf->len);
                        return;
                        
                } else {
                        owl_vdbg("done sending buf:%p", buf);
                        list_del(&buf->list);
                        owl_os_free(buf);
                }
        }
}
#endif /* OWL_CONFIG_SHELL_OUTPUT_QUEUE */

const struct owlsh_cmd cmd_help = {
        "help", help_f, "print this information"
};
