/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#define OWL_TRACE OWL_TRACE_MINIMAL
#define OWL_FILE "timer"

#include <owl/core/owl_timer.h>
#include <owl/core/owl_debug.h>
#include <owl/core/owl_err.h>
#include <owl/core/owl_os.h>
#if OWL_CONFIG_SHELL
# include <owl/core/owl_shell.h>
#endif /* OWL_CONFIG_SHELL */

struct owltmr_ctx {
        struct list_head timeouts;
        struct list_head *iter;
        volatile unsigned int cur_ms;  /* current timestamp in ms */
#if OWL_CONFIG_TIMER_SEC
        volatile unsigned int cur_sec; /* current timestamp in sec */
        volatile unsigned int remainder; /* For cur_sec calc */
#endif /* OWL_CONFIG_TIMER_SEC */
#define FLAG_POLL               (1 << 0)
#define FLAG_SUSPEND_DISABLED   (1 << 1)
        int flags;
        
#if defined(OWL_DEBUG) || OWL_CONFIG_HIBERNATE
        unsigned int id; /* Next timer id */
#endif /* defined(OWL_DEBUG) || OWL_CONFIG_HIBERNATE */

        void (*suspend_cb)(void *ctx, unsigned int ms);
        void *suspend_ctx;

#if OWL_CONFIG_TIMER_STATS
        unsigned int poll_max; /* maximum time for owltmr_poll() */
        unsigned int cb_max; /* maximum time (ms) a single callback has been
                                exec */
#endif
};

#if OWL_CONFIG_HIBERNATE
struct hibernate_entry {
        char name[8];
        unsigned int flags;
        unsigned int ts;
        unsigned int intvl;
} OWL_PACKED;

struct hibernate_ctx {
        unsigned int cur_ms;
        unsigned int id;
#if OWL_CONFIG_TIMER_SEC
        unsigned int cur_sec; /* current timestamp in sec */
        unsigned int remainder; /* For cur_sec calc */
#endif /* OWL_CONFIG_TIMER_SEC */
        unsigned int count;
        struct hibernate_entry entry[];
} OWL_PACKED;
#endif /* OWL_CONFIG_HIBERNATAE */

#if defined(OWL_DEBUG) || OWL_CONFIG_HIBERNATE
# define tmo_name(tmo) (tmo)->name
#else /* defined(OWL_DEBUG) || OWL_CONFIG_HIBERNATE */
# define tmo_name(tmo) ""
#endif /* defined(OWL_DEBUG) || OWL_CONFIG_HIBERNATE */

static void suspend(void);
static void list_del_safe(struct list_head *entry, struct list_head **n);

static struct owltmr_ctx *priv = NULL;

int
owltmr_init(void)
{
        priv = owl_os_zalloc(sizeof(struct owltmr_ctx));
        if (priv == NULL) {
                owl_err("could not alloc priv");
                return OWL_ERR_MEM;
        }
        list_init(&priv->timeouts);
        return 0;
}

struct owltmr_tmo *
owltmr_alloc(void)
{
        struct owltmr_tmo *tmo = owl_os_zalloc(sizeof(struct owltmr_tmo));
        if (tmo == NULL) {
                owl_err("could not alloc tmo");
                return NULL;
        }
        owl_dbg("tmo:%p", tmo);
        return tmo;
}

void
owltmr_free(struct owltmr_tmo *tmo)
{
        if (tmo == NULL)
                return;

        owl_dbg("tmo:%p name:%s %s%s%s", tmo, tmo_name(tmo),
                (tmo->flags & TMO_PERIODIC) ? "PERIODIC " : "",
                (tmo->flags & TMO_SCHEDULED) ? "SCHEDULED " : "",
                (tmo->flags & TMO_CANCEL) ? "CANCEL " : "");
        
        if (owl_bit(tmo->flags, TMO_SCHEDULED))
                list_del_safe(&tmo->list, &priv->iter);

        owl_os_free(tmo);
}

void
owltmr_sched(struct owltmr_tmo *tmo, unsigned int ms, int periodic,
             owltmr_tmo_cb_t *cb, void *ctx)
{
        unsigned int ts;
        
#if defined(OWL_DEBUG) || OWL_CONFIG_HIBERNATE
        if (!owl_os_strlen(tmo->name)) {
                owl_os_snprintf(
                        tmo->name, sizeof(tmo->name) - 1,  "timer%d", priv->id);
                priv->id++;
        }
#endif /* defined(OWL_DEBUG) || OWL_CONFIG_HIBERNATE */

        owl_dbg("tmo:%p name:%s ms:%d periodic:%d cb:%p ctx:%p",
                tmo, tmo_name(tmo), ms, periodic, cb, ctx);
        
        owl_bitclr(tmo->flags, TMO_CANCEL);
        
        tmo->cb = cb;
        tmo->ctx = ctx;

#if OWL_CONFIG_TIMER_SEC
        if (owl_bit(tmo->flags, TMO_SEC))
                ts = priv->cur_sec;
        else
                ts = priv->cur_ms;
#else /* OWL_CONFIG_TIMER_SEC */
        ts = priv->cur_ms;
#endif /* OWL_CONFIG_TIMER_SEC */
        tmo->ts = ts;
        tmo->intvl = ms;
        if (periodic)
                owl_bitset(tmo->flags, TMO_PERIODIC);

        /* First time add the timer to the list */
        
        if (!owl_bit(tmo->flags, TMO_SCHEDULED))
                list_add_tail(&tmo->list, &priv->timeouts);

        owl_bitset(tmo->flags, TMO_SCHEDULED);
        owl_bitclr(tmo->flags, TMO_DISABLED);
}

void
owltmr_cancel(struct owltmr_tmo *tmo)
{
        if (!owl_bit(tmo->flags, TMO_SCHEDULED))
                return;

        owl_dbg("cancel tmo:%p name:%s", tmo, tmo_name(tmo));
        list_del_safe(&tmo->list, &priv->iter);
        owl_bitclr(tmo->flags, TMO_SCHEDULED);
        owl_vdbg("done");
}

int
owltmr_is_sched(struct owltmr_tmo *tmo)
{
        return owl_bit(tmo->flags, TMO_SCHEDULED);
}

struct owltmr_tmo *
owltmr_addtmo(unsigned int ms, int periodic, owltmr_tmo_cb_t *cb, void *ctx)
{
        struct owltmr_tmo *tmo;
        
        if ((tmo = owltmr_alloc()) == NULL) {
                owl_err("could not alloc tmo");
                return NULL;
        }

        owltmr_sched(tmo, ms, periodic, cb, ctx);
        return tmo;
}

void
owltmr_deltmo(struct owltmr_tmo *tmo)
{
        owltmr_free(tmo);
}

void
owltmr_modtmo(struct owltmr_tmo *tmo, unsigned int ms)
{
        owltmr_sched(
                tmo, ms, owl_bit(tmo->flags, TMO_PERIODIC),
                tmo->cb, tmo->ctx);
}

unsigned int
owltmr_get_ms(void)
{
        return priv->cur_ms;
}

#if OWL_CONFIG_TIMER_SEC
void
owltmr_get_sec(unsigned int *sec, unsigned int *ms)
{
        if (sec != NULL)
                *sec = priv->cur_sec;

        if (ms != NULL)
                *ms = priv->remainder;
}
#endif

void
owltmr_tick(unsigned int ms)
{
        if (priv == NULL)
                return;

        priv->cur_ms += ms;
#if OWL_CONFIG_TIMER_SEC
        priv->cur_sec += ms / 1000;
        priv->remainder += ms % 1000;
        priv->cur_sec += priv->remainder / 1000;
        priv->remainder %= 1000;
#endif /* OWL_CONFIG_TIMER_SEC */
}

void
owltmr_mdelay(unsigned int ms)
{
        unsigned int old = priv->cur_ms;
        owl_dbg("busy waiting until:%d now:%d [ms]",
                old + ms, priv->cur_ms);
        while (!owltmr_intvl_passed(old, priv->cur_ms, ms));
}

void
owltmr_poll(void)
{
        struct list_head *pos;

#if OWL_CONFIG_TIMER_STATS
        unsigned int poll_start, poll_time;
        poll_start = priv->cur_ms;
#endif /* OWL_CONFIG_TIMER_STATS */

        owl_vdbg("enter");

        owl_bitset(priv->flags, FLAG_POLL); 

        list_for_each_safe(pos, priv->iter, &priv->timeouts) {
                struct owltmr_tmo *iter;
                unsigned int ts;
                iter = list_entry(pos, struct owltmr_tmo, list);

                /* Timer is in list (for deserialize) but not used,
                   then skip this one */
                if (owl_bit(iter->flags, TMO_DISABLED))
                        continue;
                
#if OWL_CONFIG_TIMER_SEC
                if (owl_bit(iter->flags, TMO_SEC))
                        ts = priv->cur_sec;
                else
                        ts = priv->cur_ms;
#else /* OWL_CONFIG_TIMER_SEC */
                ts = priv->cur_ms;
#endif /* OWL_CONFIG_TIMER_SEC */
                if (!owltmr_intvl_passed(iter->ts, ts, iter->intvl))
                        continue;

                owl_vdbg("invoking tmo:%p name:%s", iter, tmo_name(iter));

                /* set flag before invoking cb() since owltmr_modtmo() might
                 * be invoked from the cb
                 */
                if (owl_bit(iter->flags, TMO_PERIODIC)) {
                        iter->ts = ts;
                } else {

                        /* Remove oneshot timers _before_ invoking the callback.
                           This allows us to free a one-show timer from its
                           callback. For periodic timers, the user have to
                           invoke owltmr_cancel() first. */
                        list_del(&iter->list);
                        owl_bitclr(iter->flags, TMO_SCHEDULED);
                }
                
                if (iter->cb) {
#if OWL_CONFIG_TIMER_STATS
                        unsigned int cb_start, cb_time;
                        cb_start = priv->cur_ms;
#endif /* OWL_CONFIG_TIMER_STATS */
                        iter->cb(iter, iter->ctx);
#if OWL_CONFIG_TIMER_STATS
                        cb_time = owltmr_get_intvl(cb_start, priv->cur_ms);
                        if (cb_time > priv->cb_max)
                                priv->cb_max = cb_time;
#endif /* OWL_CONFIG_TIMER_STATS */
                }
        }

        owl_bitclr(priv->flags, FLAG_POLL);

        priv->iter = NULL;

#if OWL_CONFIG_TIMER_STATS
        poll_time = owltmr_get_intvl(poll_start, priv->cur_ms);
        if (poll_time > priv->poll_max)
                priv->poll_max = poll_time;
#endif /* OWL_CONFIG_TIMER_STATS */
        
        if (!owl_bit(priv->flags, FLAG_SUSPEND_DISABLED))
                suspend();

}

int
owltmr_intvl_passed(unsigned int old, unsigned int new, unsigned int diff)
{
        if (diff && old == new)
                return 0;
        else if (owltmr_get_intvl(old, new) >= diff)
                return 1;
        else
                return 0;
}

unsigned int
owltmr_get_intvl(unsigned int old, unsigned int new)
{
        /* wrap */
        if (new < old)
                return ((unsigned int) (-1) - old) + new;
        
        /* no wrap */
        return new - old;
}

void
owltmr_set_suspend(int enable)
{
        if (enable)
                owl_bitclr(priv->flags, FLAG_SUSPEND_DISABLED);
        else
                owl_bitset(priv->flags, FLAG_SUSPEND_DISABLED);
}

void
owltmr_set_suspend_cb(void suspend_cb(void *ctx, unsigned int ms), void *ctx)
{
        priv->suspend_cb = suspend_cb;
        priv->suspend_ctx = ctx;
}

void
owltmr_set_flag(struct owltmr_tmo *tmo, int flag)
{
        owl_bitset(tmo->flags, flag);
}

void
owltmr_clr_flag(struct owltmr_tmo *tmo, int flag)
{
        owl_bitclr(tmo->flags, flag);
}

#if OWL_CONFIG_TIMER_SEC
void
owltmr_set_sec(struct owltmr_tmo *tmo, int enable)
{
        if (enable)
                owl_bitset(tmo->flags, TMO_SEC);
        else
                owl_bitclr(tmo->flags, TMO_SEC);
}
#endif /* OWL_CONFIG_TIMER_SEC */

#if defined(OWL_DEBUG) || OWL_CONFIG_HIBERNATE
void
owltmr_set_name(struct owltmr_tmo *tmo, const char *name)
{
        owl_os_strncpy(tmo->name, name, sizeof(tmo->name) - 1);
}
#endif /* defined(OWL_DEBUG) || OWL_CONFIG_HIBERNATE */

#if OWL_CONFIG_HIBERNATE

/* NOTE about timer hibernation.
 *
 * All SCHEDULED timers that has a name and the TMO_SERIALIZE_DISABLED bit
 * cleared will be serialized. Their schduled time, interval and flags will
 * stored.
 *
 * Upon deserialization, each timer configuration found in the hibernate file
 * will be matched up agains a timer which is currently in the priv->timouts
 * list. This means that timer must be either ADDED or SCHEDULED. ADDING a
 * timer is a method to ensure that the timer ends up in priv->timeouts
 * without being used yet. It will just be there to be matched up during
 * deserialization.
 *
 * Code that uses timers that should be serialized can perform e.g. the
 * following operations before deserialization starts to make sure that the
 * timer is included in serialization / deserialization.
 *
 * owltmr_set_name(&priv->foo_tmo, "foo");
 * owltmr_add(&priv->foo_tmo, foo_cb, NULL);
 *
 */

void
owltmr_add(struct owltmr_tmo *tmo, owltmr_tmo_cb_t *cb, void *ctx)
{
        owltmr_sched(tmo, 0, 0, cb, ctx);
        owltmr_set_flag(tmo, TMO_DISABLED);
}

void
owltmr_set_serialize(struct owltmr_tmo *tmo, int enable)
{
        if (enable)
                owl_bitclr(tmo->flags, TMO_SERIALIZE_DISABLED);
        else
                owl_bitset(tmo->flags, TMO_SERIALIZE_DISABLED);
}

int
owltmr_serialize(void *data, int len)
{
        struct hibernate_ctx *p;
        struct list_head *pos;
        int tot_len = 0;
        int count = 0;
        
        owl_info("data:%p len:%d", data, len);

        if (data == NULL)
                return OWL_ERR_PARAM;
        
        /* Serialize header */
        p = data;
        tot_len += sizeof(*p);
        if (tot_len > len) {
                owl_err("not enough room, current tot_len:%d len:%d",
                        tot_len, len);
                return OWL_ERR_SIZE;
        }
        p->cur_ms = priv->cur_ms;
        p->id = priv->id;
#if OWL_CONFIG_TIMER_SEC
        p->cur_sec = priv->cur_sec;
        p->remainder = priv->remainder;
        owl_info("serialized values cur_ms:%d cur_sec:%d remainder:%d id:%d",
                 p->cur_ms, p->cur_sec, p->remainder, p->id);
#else /* OWL_CONFIG_TIMER_SEC */
        owl_info("serialized values cur_ms:%d id:%d", p->cur_ms, p->id);
#endif /* OWL_CONFIG_TIMER_SEC */
        
        
        /* Serialize timeouts */
        list_for_each(pos, &priv->timeouts) {
                struct hibernate_entry *entry;
                struct owltmr_tmo *tmo;

                tmo = list_entry(pos, struct owltmr_tmo, list);

                /* Don't serialize timers which are specifically set to not
                   be serialized */
                if (owl_bit(tmo->flags, TMO_SERIALIZE_DISABLED))
                        continue;

                /* Don't serialize timers that doesnt have a name since we
                   wont be able to deserialize those timers properly since
                   we can't locate the proper timer object to update. */
                if (!owl_os_strlen(tmo->name))
                        continue;

                tot_len += sizeof(*entry);
                if (tot_len > len) {
                        owl_err("not enough room, current tot_len:%d len:%d",
                                tot_len, len);
                        return OWL_ERR_SIZE;
                }
                entry = &p->entry[count];
                owl_os_strcpy(entry->name, tmo->name);
                entry->flags = tmo->flags;
                entry->ts = tmo->ts;
                entry->intvl = tmo->intvl;
                
                owl_info("serialized tmo:%s flags:0x%x, ts:%u intvl:%u",
                        entry->name, entry->flags, entry->ts, entry->intvl);
                count++;
        }
        p->count = count;
        
        return tot_len;
}

int
owltmr_deserialize(const void *data, int len)
{
        const struct hibernate_ctx *p;
        struct list_head *pos;
        int count;
        int tot_len = 0;
        
        owl_info("data:%p len:%d", data, len);
        
        if (data == NULL)
                return OWL_ERR_PARAM;

        /* Deserialize header */
        p = data;
        tot_len += sizeof(*p);
        if (tot_len > len) {
                owl_err("not enough data, current tot_len:%d len:%d",
                        tot_len, len);
                return OWL_ERR_SIZE;
        }
        priv->cur_ms += p->cur_ms;
        debug_time += p->cur_ms;
        priv->id = p->id;
#if OWL_CONFIG_TIMER_SEC
        priv->cur_sec = p->cur_sec;
        priv->remainder = p->remainder;
        owl_info("deserialized cur_ms:%d cur_sec:%d remainder:%d id:%d",
                 priv->cur_ms, priv->cur_sec, p->remainder, priv->id);
#else /* OWL_CONFIG_TIMER_SEC */
        owl_info("deserialized cur_ms:%d id:%d", priv->cur_ms, priv->id);
#endif /* OWL_CONFIG_TIMER_SEC */
        
        for (count = 0; count < p->count; count++) {
                /* Find the timer with the corresponding name */
                list_for_each(pos, &priv->timeouts) {
                        const struct hibernate_entry *entry;
                        struct owltmr_tmo *tmo;

                        tmo = list_entry(pos, struct owltmr_tmo, list);
                        entry = &p->entry[count];
                        if (owl_os_strcmp(tmo->name, entry->name))
                                continue;

                        /* Found */

                        tot_len += sizeof(*entry);
                        if (tot_len > len) {
                                owl_err("not enough data, current tot_len:%d "
                                        "len:%d", tot_len, len);
                                return OWL_ERR_SIZE;
                        }
                        
                        tmo->flags = entry->flags;
                        tmo->ts = entry->ts;
                        tmo->intvl = entry->intvl;

                        owl_info("deserialized tmo:%s flags:0x%x, ts:%u "
                                 "intvl:%u will expire in:%d [ms]",
                                 tmo->name, tmo->flags, tmo->ts, tmo->intvl,
                                 tmo->ts + tmo->intvl - priv->cur_ms);
                }
        }

        return tot_len;
}
#endif /* OWL_CONFIG_HIBERNATE */

static void
suspend(void)
{
        struct list_head *pos;
        unsigned int next = 0;
        
        list_for_each(pos, &priv->timeouts) {
                unsigned int ts;
                unsigned int intvl;
                struct owltmr_tmo *iter;
                iter = list_entry(pos, struct owltmr_tmo, list);
                
                if (owl_bit(iter->flags, TMO_DISABLED))
                        continue;
                if (!owl_bit(iter->flags, TMO_SCHEDULED))
                        continue;
                if (owl_bit(iter->flags, TMO_CANCEL))
                        continue;
                if (owl_bit(iter->flags, TMO_IGNORE_SUSPEND))
                        continue;

#if OWL_CONFIG_TIMER_SEC
                if (owl_bit(iter->flags, TMO_SEC))
                        ts = priv->cur_sec;
                else
                        ts = priv->cur_ms;
#else /* OWL_CONFIG_TIMER_SEC */
                ts = priv->cur_ms;
#endif /* OWL_CONFIG_TIMER_SEC */

                /* A timer has already passed its deadline, so we should
                   not sleep. This might occur e.g. if we spend too long time
                   in a callback */
                if (owltmr_intvl_passed(iter->ts, ts, iter->intvl))
                        return;
                
                if (!(intvl = owltmr_get_intvl(ts, iter->ts + iter->intvl)))
			return; /* This timer just passed its deadline */

                owl_vdbg("tmo:%p name:%s intvl:%d expire at:%u in:%u [%s]",
                         iter, tmo_name(iter), iter->intvl,
                         iter->ts + iter->intvl, intvl,
                         owl_bit(iter->flags, TMO_SEC) ? "s" : "ms");

#if OWL_CONFIG_TIMER_SEC
                if (owl_bit(iter->flags, TMO_SEC)) {
                        /* The interval for this timer was in [seconds], so
                           convert it to milliseconds and handle any wrapping */
                        unsigned int ms = intvl * 1000;
                        unsigned int wrap = ms / 1000;
                        if (wrap != intvl)
                                intvl = -1;
                        else
                                intvl = ms;
                }
#endif /* OWL_CONFIG_TIMER_SEC */

                if (!next || intvl < next)
                        next = intvl;
        }

        owl_vdbg("ts:%u, next timeout in %u ms", priv->cur_ms, next);
        if (priv->suspend_cb)
                priv->suspend_cb(priv->suspend_ctx, next);
}

#if OWL_CONFIG_SHELL && OWL_CONFIG_TIMER_STATS
static void
timer_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        if (argc > 1) {
                if (!owl_os_strcmp("reset", argv[1])) {
                        priv->poll_max = 0;
                        priv->cb_max = 0;
                        
                } else {
                        goto usage;
                }
                
        } else {
                struct list_head *pos;
        
                owlsh_printf(sh, "Now:      %u\r\n",
                             (unsigned int) priv->cur_ms);
                owlsh_printf(sh, "Poll:      %u\r\n",
                             (unsigned int) priv->poll_max);
                owlsh_printf(sh, "Callback:  %u\r\n",
                             (unsigned int) priv->cb_max);
                owlsh_printf(sh, "\n\r");


                list_for_each(pos, &priv->timeouts) {
                        struct owltmr_tmo *iter;
                        unsigned int intvl, ts;
                        
                        iter = list_entry(pos, struct owltmr_tmo, list);
#if OWL_CONFIG_TIMER_SEC
                        if (owl_bit(iter->flags, TMO_SEC))
                                ts = priv->cur_sec;
                        else
                                ts = priv->cur_ms;
#else /* OWL_CONFIG_TIMER_SEC */
                        ts = priv->cur_ms;
#endif /* OWL_CONFIG_TIMER_SEC */
                        intvl = owltmr_get_intvl(ts, iter->ts + iter->intvl);
                        
                        owlsh_printf(
                                sh,
                                "name:%s intvl:%d expire at:%u in:%u [%s]\r\n",
                                tmo_name(iter), iter->intvl,
                                iter->ts + iter->intvl, intvl,
                                owl_bit(iter->flags, TMO_SEC) ? "s" : "ms");
                }                    
        }

        return;

usage:
        owlsh_printf(sh, "usage: timer [reset]\n\r");
}

const struct owlsh_cmd cmd_timer = {
        "timer", timer_f, "timer stats"
};
#endif /* OWL_CONFIG_SHELL && OWL_CONFIG_TIMER_STATS */


/**
 * list_del_safe - deletes entry from list.
 * @entry: the element to delete from the list.
 * @n: the next iterator position (parameter @n in list_for_each_safe).
 *
 * Note: Safe to call to delete another list entry while iterating with
 * list_for_each_safe(). This function is needed to handle the case where
 * a timer is deleted while we are iterating through the timer list.
 * list_for_each_safe() only supports removing the current position, not e.g.
 * the next position. list_del_safe() handles this for us.
 */
static void
list_del_safe(struct list_head *entry, struct list_head **n)
{
        if (*n == entry)
                *n = (*n)->next;

        list_del(entry);
}
