/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#ifndef OWL_TIMER_H
#define OWL_TIMER_H

#include <owl/core/owl_defs.h>
#include <owl/core/owl_list.h>

/**
 * @file owl_timer.h
 * @brief Generic timer implementation.
 * 
 * This module handles timeouts and delays in a platform-independent
 * way. It executes timer-based callback functions with "minimum-time-passed"
 * semantics : Any callback is guaranteed to wait for at least the specified
 * amount of time before being called. 
 *
 * In other words, a callback may be called later than requested, but
 * never earlier.
 *
 * The timers are driven by calls to \a owltmr_poll(). This function
 * will execute any pending callbacks whose timers has expired.
 *
 * Time-measurement is based on the information passed in calls to
 * \a owltmr_tick().
 *
 * The precision of the timers are based on the frequency of the calls
 * to \a owltmr_poll() and the frequency of the calls to \a owltmr_tick()
 * and the accuracy of the timestamp passed to it.
 *
 */

struct owltmr_tmo;
typedef void (owltmr_tmo_cb_t)(struct owltmr_tmo *, void *);

struct owltmr_tmo {
        owltmr_tmo_cb_t *cb;
        void *ctx;
#define TMO_PERIODIC (1 << 0) /* reschedules automatically after fire */
#define TMO_SCHEDULED (1 << 1) /* owltmr_sched() */
#define TMO_CANCEL    (1 << 2) /* pending owltmr_cancel() */
#define TMO_SERIALIZE_DISABLED    (1 << 3) /* see owltmr_set_serialize() */
#define TMO_SEC         (1 << 4) /* Uses second-resolution */
#define TMO_IGNORE_SUSPEND (1 << 5) /* This timeout will be ignored when
                                       calculating suspend time */
#define TMO_DISABLED    (1 << 6) /* Used when adding timouts that should be
                                    deserialized */
        unsigned int flags;
        unsigned int ts;        /* compare timestamp */
        unsigned int intvl;     /* interval in ms */
#if defined(OWL_DEBUG) || OWL_CONFIG_HIBERNATE
        char name[16];
#endif /* defined(OWL_DEBUG) || OWL_CONFIG_HIBERNATE */
        struct list_head list;
};


/**
 * @brief Initialize the timer module.
 *
 * @return
 *   - 0 on success
 *   - OWL_ERR_MEM
 */
int owltmr_init(void);

/**
 * @brief Add timer callback.
 * 
 * The callback will run in the execution context that calls \a owltmr_poll().
 * 
 * @param ms Timeout in milliseconds.
 * @param periodic 1 if the timer is periodic. 0 if it should only execute
 *        once.
 * @param cb Callback to execute when the timer expires.
 * @param ctx Opaque context that will be passed to the callback.
 * @return 
 *   - A pointer to a timer handle on success
 *   - NULL on failure
 */
struct owltmr_tmo *owltmr_addtmo(unsigned int ms, 
                                 int periodic,
                                 owltmr_tmo_cb_t *cb, 
                                 void *ctx);

/**
 * @brief Delete a timer added with owltmr_addtmo().
 *
 * @param tmo A timer handle to delete.
 *
 */
void owltmr_deltmo(struct owltmr_tmo *tmo);

/**
 * @brief Modify a timer.
 *
 * @param tmo A timer handle to modify.
 * @param ms New timeout in milliseconds.
 */
void owltmr_modtmo(struct owltmr_tmo *tmo, unsigned int ms);

/**
 * @brief Get the current timestamp in milliseconds.
 *
 * @return
 *    - The current time in milliseconds.
 */
unsigned int owltmr_get_ms(void);

#if OWL_CONFIG_TIMER_SEC
/**
 * @brief Get the current timestamp in seconds.
 *
 * @param sec will hold the number of seconds elapsed.
 * @param ms will hold the number of millisecond part of time elapsed.
 *
 */
void owltmr_get_sec(unsigned int *sec, unsigned int *ms);

/**
 * @brief Use second-resolution for the timeout to enable long timeouts.
 *      This is useful on platforms that only have 16 bit integer since
 *      otherwise the maximum timeout would be 65536 ms.
 *
 * @param tmo the timer handle to set resolution for.
 * @param enable specifices if second-resolution should be enabled or not.
 *
 */
void owltmr_set_sec(struct owltmr_tmo *tmo, int enable);
#endif /* OWL_CONFIG_TIMER_SEC */

/**
 * @brief Update the current time value.
 *
 * Tell the timer implementation what the current time is. 
 *
 * This function can be called in interrupt context.
 *
 * @param timestamp The time passed since the last call to \a owltmr_tick()
 *                  in milliseconds.
 */
void owltmr_tick(unsigned int ms);

/**
 * @brief Execute any expired timers.
 *
 * All registered, expired timer callbacks are executed by
 * this function.
 */
void owltmr_poll(void);

/**
 * Check if \a diff ms has passed between \a old and \a new. Handles
 * the case that \a new has wrapped and is now smaller than \a old.
 */
int owltmr_intvl_passed(unsigned int old, unsigned int new, unsigned int diff);

/**
 * Get the number of units passed between \a old and \a new. Handles
 * the case that \a new has wrapped and is now smaller than \a old.
 * The return value is still a postive number in this case.
 */
unsigned int owltmr_get_intvl(unsigned int old, unsigned int new);

/**
 * @brief Delay execution.
 *
 * Wait for the specified amount of time.
 *
 * This function cannot be used if owltmr_poll() is called in the
 * same executing context. Unless the millisecond time is updated
 * in another context (such as in an interrupt service routine) this
 * call will hang.
 * @param ms Time to delay in milliseconds.
 * 
 */
void owltmr_mdelay(unsigned int ms);

/**
 * @brief Allocate a timeout object to be scheduled later.
 *
 * Note that a struct owltmr_tmo can also be statically allocated.
 *
 * @return
 *   - A pointer to a timer handle on success
 *   - NULL on failure
 *
 */
struct owltmr_tmo *owltmr_alloc(void);

/**
 * @brief Free a timeout object allocated with owltmr_alloc()
 *
 */
void owltmr_free(struct owltmr_tmo *tmo);


/**
 * @brief Scehdule timer that was allocated with owltmr_alloc().
 *
 * Timer will only execute once. If the timeout is already sheduled, the
 * expiry time will be updated.
 *
 * @param tmo A timer handle to schedule.
 * @param ms Timeout in milliseconds.
 * @param periodic 1 if the timer is periodic. 0 if it should only execute
 *        once.
 * @param cb Callback to execute when the timer expires.
 * @param ctx Opaque context that will be passed to the callback.
 *
 */
void owltmr_sched(struct owltmr_tmo *tmo, unsigned int ms, int periodic,
                  owltmr_tmo_cb_t *cb, void *ctx);


/**
 * @brief Cancel a timer that was scheduled with with owltmr_sched().
 *
 *
 * @param tmo A timer handle to cancel.
 *
 */
void owltmr_cancel(struct owltmr_tmo *tmo);

/**
 * @brief Check if a timer is scheduled or not.
 *
 */
int owltmr_is_sched(struct owltmr_tmo *tmo);

/**
 * @brief Enable or disable the suspend functionality.
 *
 * This will set whether the callback registered with owltmr_set_suspend_cb()
 * will be invoked or not.
 *
 */
void owltmr_set_suspend(int enable);

/**
 * @brief Specificy a callback that will be invoked when the time to the
 *        next timeout is updated.
 *
 * This function can be used to suspend the MCU until the next timeout should
 * be handled. Note that if the MCU goes to sleep in such way that owltmr_tick()
 * wont get invoked at regular intervals, owltmr_tick() must be called when the
 * MCU suspends in order to invoke any expired timeout callbacks.
 *
 * @param suspend_cb is a function that will be invoked whenever the time to
 *              next timeout changes.
 * @param ctx is an opaque context pointer passed back to suspend_cb.
 *
 */
void owltmr_set_suspend_cb(void suspend_cb(void *ctx, unsigned int ms), void *ctx);

/**
 * @brief Set timer configuration flags.
 *
 * 
 */
void owltmr_set_flag(struct owltmr_tmo *tmo, int flag);

/**
 * @brief Clear timer configuration flags.
 *
 * 
 */
void owltmr_clr_flag(struct owltmr_tmo *tmo, int flag);

#if defined(OWL_DEBUG) || OWL_CONFIG_HIBERNATE
void owltmr_set_name(struct owltmr_tmo *tmo, const char *name);
#else /* defined(OWL_DEBUG) || OWL_CONFIG_HIBERNATE */
#define owltmr_set_name(...) do { } while (0)
#endif /* defined(OWL_DEBUG) || OWL_CONFIG_HIBERNATE */

#if OWL_CONFIG_HIBERNATE

/**
 * @bried Add a timer without scheduling it.
 *
 * This function should be called to allow deserialization of timers that
 * are not yet scheduled.
 *
 */
void owltmr_add(struct owltmr_tmo *tmo, owltmr_tmo_cb_t *cb, void *ctx);

/**
 * @brief Enable or disable hibernation of this timer callback.
 *
 * This will set whether the timout will be saved/restored with
 * owltmr_serialize() and owltmr_deserialize().
 *
 */
void owltmr_set_serialize(struct owltmr_tmo *tmo, int enable);

/**
 * Save the owl_timer state into a memory buffer. The data in the buffer can
 * later be used to restore the state with owltmr_deserialize().
 *
 * Note that no timer callbacks will be saved, since it is the responsibility
 * of the respective module that uses owl_timer to add the timeouts upon those
 * modules' serialization / deserialization.
 *
 *
 * @param data should point to a buffer that will hold the state of
 *      the module.
 * @param len is the maximum capacity of data.
 * @return the number of bytes stored, or error code if failure occurs.
 *
 */
int owltmr_serialize(void *data, int len);

/**
 * Restore owl_timer state obtained from a previous call to owltmr_serialize().
 *
 *
 * @param data points to state data.
 *
 * @return number of bytes consumed from data on success,
 *      error code on failure.
 *
 */
int owltmr_deserialize(const void *data, int len);

#else /* OWL_CONFIG_HIBERNATE */
# define owltmr_set_serialize(...) do { } while (0)
#endif /* OWL_CONFIG_HIBERNATE */

#if OWL_CONFIG_SHELL && OWL_CONFIG_TIMER_STATS
extern const struct owlsh_cmd cmd_timer;
#endif /* OWL_CONFIG_SHELL && OWL_CONFIG_TIMER_STATS */

#endif
