/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#ifndef OWL_GPIO_H
#define OWL_GPIO_H

#include <owl/core/owl_shell.h>

/**
 * @file owl_gpio.h
 *
 * @brief Basic GPIO-pin access.
 *
 * This file provides a simple GPIO-pin configuration and access API.
 *
 * The purpose of the API is to allow applications that only need to
 * do basic GPIO pin operations (testing and changing pin level) to do
 * so in a platform-independent fashion.
 *
 * The actual definition of the GPIO pin identifiers is platform-dependent.
 * The only requirement is that it be an integer value.
 *
 * For example: On the Atmel XMega CPUs a GPIO pin is defined as a
 * pin number on a port. To get the actual pin on the CPU the pin
 * number is added to a port offset. This is the value that should
 * be used to identify the pin to the owl_gpio interface.
 *
 * Consult the owl_gpio.c file in the relevant ports directory
 * for information on how to define pins for a particular platform.
 *
 */

#define OWL_GPIO_INPUT                  (1 << 0)
#define OWL_GPIO_OUTPUT                 (1 << 1)
#define OWL_GPIO_PULL_DOWN              (1 << 2)
#define OWL_GPIO_PULL_UP                (1 << 3)
#define OWL_GPIO_IRQ_FALLING_EDGE       (1 << 4)
#define OWL_GPIO_IRQ_RISING_EDGE        (1 << 5)
#define OWL_GPIO_IRQ_LEVEL_LOW          (1 << 6)
#define OWL_GPIO_IRQ_LEVEL_HIGH         (1 << 7)
#define OWL_GPIO_LOW                    (1 << 8)
#define OWL_GPIO_HIGH                   (1 << 9)

#define OWL_GPIO_PIN_NONE       -1

/**
 * Initialize GPIO system.
 *
 * @return 0 on success, -1 on failure.
 */
int owlgpio_init(void);

/**
 * Configure a GPIO pin.
 *
 * @param pin Pin to configure.
 * @param flags Bitmask of the GPIO configuration flags to use.
 *        Valid flags are
 *              OWL_GPIO_INPUT
 *              OWL_GPIO_OUTPUT
 *              OWL_GPIO_PULL_DOWN
 *              OWL_GPIO_PULL_UP
 *        OWL_GPIO_INPUT and OWL_GPIO_OUTPUT are mutually exclusive.
 *        OWL_GPIO_PULL_UP and OWL_GPIO_PULL_DOWN are mutually exclusive and are
 *        only applicable with OWL_GPIO_INPUT.
 * @return 
 *       - 0 on success
 *       - OWL_ERR_PARAM if the flags are invalid
 *       - OWL_ERR_RANGE if the GPIO pin number is invalid
 */
int owlgpio_cfg(int pin, int flags);

/**
 * Check GPIO pin state.
 *
 * @param pin Pin to check. The pin number must be valid.
 * @return 
 *       - 1 if the pin is high
 *       - 0 if the pin is low
 */
int owlgpio_is_high(int pin);


/**
 * Check GPIO pin state.
 *
 * @param pin Pin to check. The pin number must be valid.
 * @return 
 *       - 0 if the pin is high
 *       - 1 if the pin is low
 */
#define owlgpio_is_low(pin) (!owlgpio_is_high(pin))


/**
 * Toggle GPIO pin state.
 *
 * @param pin Pin to toggle. The pin number must be valid.
 */
#define owlgpio_toggle(pin)                                             \
    (owlgpio_is_high(pin) ? owlgpio_clr(pin) : owlgpio_set(pin))

/**
 * Set GPIO pin high.
 *
 * @param pin Pin to set.
 * @return 
 *       - 0 on success
 *       - OWL_ERR_RANGE if the GPIO pin number is invalid
 */
void owlgpio_set(int pin);

/**
 * Set GPIO pin low.
 *
 * @param pin Pin to clear.
 * @return 
 *       - 0 on success
 *       - OWL_ERR_RANGE if the GPIO pin number is invalid
 */
void owlgpio_clr(int pin);

/**
 * Convert a string to a GPIO pin number.
 *
 * @param str is the string to convert to pin numer
 * @return
 *      - pin number on success
 *      - OWL_ERR_RANGE if the string is invalid
 */
int owlgpio_parse(const char *str);

/**
 * Iterate through available pins. 
 *
 * @param pin is the iterator, to find the first pin use OWL_GPIO_PIN_NONE
 * @return
 *      - next pin number on success
 *      - OWL_GPIO_PIN_NONE if no more pins are available
 */
int owlgpio_next(int pin);

/**
 * Convert a GPIO pin number to string
 *
 * @param pin is the pin number to convert
 * @return
 *      - string value on success
 *      - NULL if the pin is invalid
 */
const char *owlgpio_str(int pin);


/**
 * Control GPIO pins.
 *
 */
extern const struct owlsh_cmd cmd_gpio;

#endif /* OWL_GPIO_H */
