/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#define OWL_TRACE OWL_TRACE_ERR
#define OWL_FILE "cbuf"

#include <owl/core/owl_debug.h>
#include <owl/core/owl_defs.h>
#include <owl/util/cbuf.h>


/* Circular buffer */


struct cbuf {
        int capacity;
        int rdpos;
        int wrpos;
        uint8_t data[];
};

struct cbuf *
cbuf_create(int size)
{
        struct cbuf *priv;
        owl_dbg("size:%d", size);
        size++;
        if ((priv = owl_os_zalloc(sizeof(*priv) + size)) == NULL) {
                owl_err("could not alloc cbuf");
                return NULL;
        }

        priv->capacity = size;
        return priv;
}

void
cbuf_destroy(struct cbuf *cbuf)
{
        owl_dbg("cbuf:%p", cbuf);
        if (cbuf != NULL)
                owl_os_free(cbuf);
}

void *
cbuf_wrptr(struct cbuf *cbuf)
{
        return cbuf->data + cbuf->wrpos;
}

int
cbuf_wrlen(struct cbuf *cbuf)
{
        if (cbuf->rdpos > cbuf->wrpos)
                return cbuf->rdpos - cbuf->wrpos - 1;
        else
                return cbuf->capacity - cbuf->wrpos;
}

void *
cbuf_rdptr(struct cbuf *cbuf)
{
        return cbuf->data + cbuf->rdpos;
}

int
cbuf_rdlen(struct cbuf *cbuf)
{
        if (cbuf->wrpos >= cbuf->rdpos)
                return cbuf->wrpos - cbuf->rdpos;
        else
                return cbuf->capacity - cbuf->rdpos;
}

void
cbuf_put(struct cbuf *cbuf, int len)
{
        owl_dbg("cbuf:%p len:%d", cbuf, len);
        owl_assert(len <= cbuf_wrlen_tot(cbuf));
        cbuf->wrpos = (cbuf->wrpos + len) % cbuf->capacity;
}

void
cbuf_get(struct cbuf *cbuf, int len)
{
        owl_dbg("cbuf:%p len:%d", cbuf, len);
        owl_assert(len <= cbuf_rdlen_tot(cbuf));
        cbuf->rdpos = (cbuf->rdpos + len) % cbuf->capacity;
}

/* Returns total number of available bytes (non-linear) */
int
cbuf_rdlen_tot(struct cbuf *cbuf)
{
	int len;

	len = cbuf->wrpos - cbuf->rdpos;
	if (len < 0)
		len += cbuf->capacity;

	return len;
}

/* Returns total number of available space (non-linear) */
int
cbuf_wrlen_tot(struct cbuf *cbuf)
{
	int len;

	len = cbuf->rdpos - cbuf->wrpos;
	if (len <= 0)
		len += cbuf->capacity;
	if (len > 0)
		len--;
	
	return len;
}


/* Wraps. Returns number of bytes written */
int
cbuf_put_data(struct cbuf *cbuf, const void *data, int len)
{
	int headlen, taillen;
	
        owl_dbg("cbuf:%p data:%p len:%d", cbuf, data, len);

	if (len > cbuf_wrlen_tot(cbuf))
		len = cbuf_wrlen_tot(cbuf);

	if (len > cbuf_wrlen(cbuf)) {
		headlen = cbuf_wrlen(cbuf);
		taillen = len - headlen;
	} else {
		headlen = len;
		taillen = 0;
	}

	owl_dbg("headlen:%d taillen:%d", headlen, taillen);

	owl_os_memcpy(cbuf_wrptr(cbuf), data, headlen);
	if (taillen)
		owl_os_memcpy(cbuf->data, (uint8_t *) data + headlen, taillen);
	
	cbuf_put(cbuf, len);
	return len;
}

/* Wraps. Returns number of bytes fetched */
int
cbuf_get_data(struct cbuf *cbuf, void *data, int len)
{
	len = cbuf_copy_data(cbuf, data, len);
	cbuf_get(cbuf, len);
	return len;
}


/* Wraps. Returns number of bytes fetched. Does not update the read pointer. */
int
cbuf_copy_data(struct cbuf *cbuf, void *data, int len)
{
	int headlen, taillen;
	
        owl_dbg("cbuf:%p data:%p len:%d", cbuf, data, len);

	if (len > cbuf_rdlen_tot(cbuf))
		len = cbuf_rdlen_tot(cbuf);

	if (len > cbuf_rdlen(cbuf)) {
		headlen = cbuf_rdlen(cbuf);
		taillen = len - headlen;
	} else {
		headlen = len;
		taillen = 0;
	}
	
	owl_dbg("headlen:%d taillen:%d", headlen, taillen);

	owl_os_memcpy(data, cbuf_rdptr(cbuf), headlen);
	if (taillen)
		owl_os_memcpy((uint8_t *) data + headlen, cbuf->data, taillen);

	return len;
}
