/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#define OWL_TRACE OWL_TRACE_ERR
#define OWL_FILE "iperf"

#include <owl/core/owl_err.h>
#include <owl/core/owl_debug.h>
#include <owl/core/owl_timer.h>
#if OWL_CONFIG_SHELL
#include <owl/core/owl_shell.h>
#endif /* OWL_CONFIG_SHELL */

#if OWL_CONFIG_SOCKET_BSD
#include <owl_port.h>
#else /* lwip */
# include <owl/core/owl_lwip.h>
# include <lwip/opt.h>
#endif /* OWL_CONFIG_SOCKET_BSD */

#include <owl/util/iperf.h>

#if OWL_CONFIG_SOCKET_BSD
 # if defined(NG_RTOS)
 #  define iperf_errno       ngOSGetErrno()
 #  define iperf_close(s)    closesocket(s)
 #  define IPERF_EAGAIN      NG_EAGAIN
 #  define IPERF_EWOULDBLOCK NG_EWOULDBLOCK
 #  define IPERF_ENOBUFS     NG_ENOBUFS
 #  define SHUT_RDWR         2 
 # else 
 #  define iperf_errno       errno
 #  define iperf_close(s)    close(s)
 #  define IPERF_EAGAIN      EAGAIN
 #  define IPERF_EWOULDBLOCK EWOULDBLOCK
 #  define IPERF_ENOBUFS     ENOBUFS
 # endif
#endif /* OWL_CONFIG_SOCKET_BSD */

struct iperf_ctx {
        char *buf;
        uint32_t buflen; /* size of buffers */
        uint32_t nbuf;   /* number of buffers, client mode */
        uint32_t bufleft; /* remaining bytes to transfer in current buf */
        
        uint32_t bufcnt;
        uint32_t recvd; /* Server mode */
        uint32_t start_ms;
        uint32_t duration; /* Client mode */

        uint32_t intvl_bufcnt;
        uint32_t intvl_recvd;
        int report_intvl;

        int client;
        int udp;

        void (*connect_cb)(void *ctx, int local_port, int remote_port,
                           char *local_ip, char *remote_ip);
        
        void (*report_cb)(void *ctx, int err, uint32_t bytes, 
			  uint32_t start_ts, uint32_t end_ts, int mss, 
			  uint32_t packets, uint32_t pkts_lost, 
			  uint32_t outoforder, int type);
	void *ctx;

#if OWL_CONFIG_SOCKET_BSD
        int sock;
        void (*state)(void *ctx);
        struct owltmr_tmo state_tmo;
#else /* lwip */
        struct tcp_pcb *tcp;
        struct udp_pcb *udp_pcb;
#endif /* OWL_CONFIG_SOCKET_BSD */

        char *local_ip;
        char *remote_ip;
        uint16_t local_port;
        uint16_t remote_port;

        /* For UDP */
        int32_t pkt_id;
        uint32_t pkts_lost;
        uint32_t pkts_outoforder;
        uint32_t intvl_pkts_lost;

        /* Server stats */
        uint32_t fin_cnt;
        uint32_t ack_fin;

        int bandwidth; /* Kbit/s */
        struct owltmr_tmo udp_tmo;
        struct owltmr_tmo report_tmo;

        uint8_t flags;
};

struct udp_server_hdr {
#define HEADER_VERSION1 0x80000000
        int32_t flags;
        int32_t total_len1; /* 64bit MSB */
        int32_t total_len2; /* 64bit LSB */
        int32_t stop_sec; /* end time */
        int32_t stop_usec;
        int32_t error_cnt; /* Packet loss */
        int32_t outorder_cnt; 
        int32_t datagrams; /* Nr of packets */
        int32_t jitter1;
        int32_t jitter2;
} OWL_PACKED;

struct udp_datagram {
    int32_t id;
    uint32_t sec;
    uint32_t usec;
    struct udp_server_hdr hdr;
} OWL_PACKED;

#define REPORT_KBITS  0
#define REPORT_MBITS  1
#define REPORT_KBYTES 2
#define REPORT_MBYTES 3

#if OWL_CONFIG_SOCKET_BSD
static void tcp_recv_cb(struct owltmr_tmo *tmo, void *ctx);
static void tcp_send_cb(struct owltmr_tmo *tmo, void *ctx);
static void udp_recv_cb(struct owltmr_tmo *tmo, void *ctx);

#else /* lwip */
static void tcp_send_data(struct iperf_ctx *priv);
static void tcp_err_cb(void *arg, err_t err);
static err_t tcp_connect_cb(void *arg, struct tcp_pcb *pcb, err_t err);
static err_t tcp_accept_cb(void *arg, struct tcp_pcb *pcb, err_t err);
static err_t tcp_recv_cb(void *arg, struct tcp_pcb *pcb, struct pbuf *p,
                         err_t err);

static void udp_recv_cb(void *arg, struct udp_pcb *upcb, struct pbuf *p,
                        struct ip_addr *addr, u16_t port);

#endif /* OWL_CONFIG_SOCKET_BSD */

static void report_timer_cb(struct owltmr_tmo *tmo, void *ctx);
static void udp_timer_cb(struct owltmr_tmo *tmo, void *ctx);
static int udp_send_data(struct iperf_ctx *priv, int len);
static int iperf_done(struct iperf_ctx *priv, int err);
static int iperf_advmss(struct iperf_ctx *priv);
static int iperf_rcvmss(struct iperf_ctx *priv);

int
iperf_client(char *ip, uint16_t port, int udp, 
             uint32_t nbuf, uint32_t buflen, int bandwidth, 
             int duration, int interval, uint8_t flags,
             void (*connect_cb)(void *ctx, int local_port, int remote_port,
                                char *local_ip, char *remote_ip),
             void (*report_cb)(void *ctx, int err, uint32_t bytes, 
                               uint32_t start_ts, uint32_t end_ts, int mss, 
                               uint32_t packets, uint32_t pkts_lost, 
			       uint32_t outoforder, int type),
	     void *ctx)
{
        int err;
        struct iperf_ctx *priv;
#if OWL_CONFIG_SOCKET_BSD
        struct sockaddr_in sin_remote;
#endif           
        owl_info("ip:%s port:%d protocol:%s nbuf:%d buflen:%d bandwidth:%d",
                 ip, port, udp ? "UDP" : "TCP" , 
                 nbuf, buflen, bandwidth);

        if (nbuf == 0 || buflen == 0 || port == 0)
                return OWL_ERR_PARAM;

        /* Allocate our private struct */
        if ((priv = owl_os_zalloc(sizeof(struct iperf_ctx))) == NULL) {
                owl_err("fail to alloc priv");
                return OWL_ERR_MEM;
        }

        priv->nbuf = nbuf;
        priv->buflen = buflen;
        priv->bufleft = buflen;
        priv->ctx = ctx;
        priv->report_intvl = interval;
        priv->bandwidth = bandwidth;
        priv->duration = duration * 1000;
        priv->client = 1;
        priv->udp = udp;
        priv->report_cb = report_cb;
        priv->connect_cb = connect_cb;
	priv->flags = flags;
        priv->remote_port = port;
        priv->remote_ip = owl_os_strdup(ip);
	
        /* Allocate a buffer to hold the data we're going to transmit */
        if ((priv->buf = owl_os_alloc(priv->buflen)) == NULL) {
                owl_err("fail to alloc buf (%d b)", priv->buflen);
                return OWL_ERR_MEM;
        }
        
#if OWL_CONFIG_SOCKET_BSD
        int optval = owl_kbyte(16);
	if (priv->udp) {
		if ((priv->sock = socket(
                             PF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0) {
                        owl_err("create socket failed, err:%d", iperf_errno);
                        
                        iperf_done(priv, iperf_errno);
                        return OWL_ERR_MEM;
                }

                if ((err = setsockopt(
                             priv->sock, SOL_SOCKET, SO_SNDBUF,
                             (char *)&optval, sizeof(optval))) < 0) {
                        owl_err("setsockopt failed, err:%d", iperf_errno);
			iperf_close(priv->sock);   
			iperf_done(priv, iperf_errno);
                        return OWL_FAILURE;
                }
                priv->pkt_id = -1;
                owltmr_sched(&priv->udp_tmo, 0, 1, udp_timer_cb, priv);
		
        } else { /* TCP */
		struct sockaddr local_addr;
                int len = sizeof(struct sockaddr);
		
                sin_remote.sin_family = AF_INET;
                inet_aton(priv->remote_ip, &sin_remote.sin_addr); 
                sin_remote.sin_port = owl_htons(priv->remote_port);
		
		if ((priv->sock = socket(
                             PF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) {
                        owl_err("socket failed, err:%d", iperf_errno);
                        iperf_done(priv, iperf_errno);
                        return OWL_ERR_MEM;
                }

                if ((err = setsockopt(
                             priv->sock, SOL_SOCKET, SO_SNDBUF,
                             (char *)&optval, sizeof(optval))) < 0) {
                        owl_err("setsockopt failed, err:%d", iperf_errno);
			iperf_close(priv->sock);
                        iperf_done(priv, iperf_errno);
                        return OWL_FAILURE;
                }
		
                /* Connect to remote server */
                if ((err = connect(priv->sock, 
                                   (struct sockaddr *) &sin_remote, 
                                   sizeof(sin_remote))) < 0) {
                        owl_err("could not connect, err:%d", iperf_errno);
                        iperf_done(priv ,iperf_errno);
                }             
		
                /* Get local ip and port */
                getsockname(
                        priv->sock, &local_addr, &len);
                
                priv->local_port = owl_ntohs(
                        ((struct sockaddr_in *) &local_addr)->sin_port);
                
                priv->local_ip = owl_os_strdup(
                        inet_ntoa(((struct sockaddr_in *) 
                                   &local_addr)->sin_addr));
                
                if (priv->connect_cb)
                        priv->connect_cb(priv->ctx, priv->local_port,
                                         priv->remote_port, priv->local_ip, 
                                         priv->remote_ip);
                
                owltmr_sched(&priv->report_tmo, priv->report_intvl * 1000, 
                             1, report_timer_cb, priv);
                
                owltmr_sched(&priv->state_tmo, 0, 1, tcp_send_cb, priv);
        }
#else /* lwip */
        if (priv->udp) {
                if ((priv->udp_pcb = udp_new()) == NULL) {
                        owl_err("fail to alloc udp pcb");
                        return OWL_ERR_MEM;
                }
                owl_dbg("udp_new:%p", priv->udp_pcb);

                /* Needed for server stats */
                udp_recv(priv->udp_pcb, udp_recv_cb, priv);
       
                priv->pkt_id = -1;
                owltmr_sched(&priv->udp_tmo, 0, 1, udp_timer_cb, priv);
                
        } else { /* TCP */
                struct ip_addr remote_ip; 
                ip_aton(priv->remote_ip, &remote_ip);

                if ((priv->tcp = tcp_new()) == NULL) {
                        owl_err("fail to alloc tcp pcb");
                        return iperf_done(priv, OWL_ERR_MEM);
                }
                owl_dbg("tcp_new:%p", priv->tcp); 
                
                tcp_arg(priv->tcp, priv);
                tcp_err(priv->tcp, tcp_err_cb);
                tcp_recv(priv->tcp, tcp_recv_cb);
                
                owl_dbg("tcp_connect:%p", priv->tcp);
                if ((err = tcp_connect(
                             priv->tcp, &remote_ip, priv->remote_port,
                             tcp_connect_cb)) != ERR_OK) {
                        owl_err("fail to conn err:%d", err);
                        return iperf_done(priv, owlip_error(err));
                }
                /* Wait for tcp_connect_cb() or tcp_err_cb() */
        }
#endif /* OWL_CONFIG_SOCKET_BSD */
        return 0;
}

int
iperf_server(uint16_t port, int udp, 
             int interval, uint32_t nbuf, uint32_t buflen, uint8_t flags, 
             void (*connect_cb)(void *ctx, int local_port, int remote_port,
                                char *local_ip, char *remote_ip),
             void (*report_cb)(void *ctx, int err, uint32_t bytes, 
                               uint32_t start_ts, uint32_t end_ts, int mss,
                               uint32_t packets, uint32_t pkts_lost, 
			       uint32_t outoforder, int type),
	     void *ctx)
{
        int err;
        struct iperf_ctx *priv;
#if OWL_CONFIG_SOCKET_BSD
        struct sockaddr_in sin_local;
#else
	struct tcp_pcb *lpcb;
#endif /* OWL_CONFIG_SOCKET_BSD */

	owl_info("port:%d protocol:%s report_intvl:%d", 
		 port, udp ? "UDP" : "TCP", interval);

        /* Allocate our private struct */
        if ((priv = owl_os_zalloc(sizeof(struct iperf_ctx))) == NULL) {
                owl_err("fail to alloc priv");
                return iperf_done(priv, OWL_ERR_MEM);
        }
	
        priv->local_port = port;
        priv->report_cb = report_cb;
        priv->connect_cb = connect_cb;
	priv->report_intvl = interval;
        priv->client = 0;
        priv->buflen = buflen;
	priv->nbuf = nbuf;
	priv->ctx = ctx;
        priv->udp = udp;
        priv->flags = flags;
        
#if OWL_CONFIG_SOCKET_BSD
        sin_local.sin_family = AF_INET;
        sin_local.sin_addr.s_addr = (INADDR_ANY);
        sin_local.sin_port = owl_htons(port);
        
        priv->local_ip = owl_os_strdup(inet_ntoa(sin_local.sin_addr));
        int optval = 
	        owl_kbyte(priv->nbuf ? priv->nbuf : 128); 
        
	/* We need a buffer to store incoming data */
        if ((priv->buf = owl_os_zalloc(priv->buflen)) == NULL) {
                owl_err("fail to alloc buf (%d b)", priv->buflen);
                return iperf_done(priv, OWL_ERR_MEM); 
        }
	
        if (priv->udp) {
		if ((priv->sock = socket(
                             PF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0) {
                        owl_err("socket failed, err:%d", iperf_errno);
                        iperf_done(priv, iperf_errno);
                        return OWL_ERR_MEM;
		}

		if ((err = setsockopt(
		             priv->sock, SOL_SOCKET, SO_RCVBUF,
			     (char *)&optval, sizeof(optval))) < 0) {
                        owl_err("setsockopt failed, err:%d", iperf_errno);
			iperf_close(priv->sock);
                        iperf_done(priv, iperf_errno);
                        return OWL_FAILURE;
                }

                if ((err = bind(priv->sock,
                                (struct sockaddr *) &sin_local,
                                sizeof(sin_local))) < 0) {
                        owl_err("could not bind socket, err:%d", 
				iperf_errno);
			iperf_close(priv->sock);
                        iperf_done(priv, iperf_errno);
                }
		
                owltmr_sched(&priv->state_tmo, 0, 1, udp_recv_cb, priv);
                priv->bufcnt = -1;
                priv->intvl_bufcnt = -1;
                priv->pkt_id = -1;

        } else { /* TCP */
		struct sockaddr_in sin_remote;
                int new_sock;
                int size = sizeof(sin_remote);
		
		if ((priv->sock = socket(
                             PF_INET, SOCK_STREAM, 0)) < 0) {
                        owl_err("socket failed, err:%d", iperf_errno);
                        iperf_done(priv, iperf_errno);
                        return OWL_ERR_MEM;
                }
		
                if ((err = bind(priv->sock,
                                (struct sockaddr *) &sin_local, 
                                sizeof(sin_local))) < 0) {
                        owl_err("fail to bind, err:%d",	iperf_errno);
			iperf_close(priv->sock);
                        iperf_done(priv, iperf_errno);
                }

                if ((err = listen(priv->sock, 0)) < 0) {
			owl_err("fail to listen, err:%d", iperf_errno);
			iperf_close(priv->sock);
                        iperf_done(priv, iperf_errno);
		}
                if ((new_sock = accept(
                             priv->sock, 
                             (struct sockaddr *) &sin_remote, &size)) < 0) {
                        iperf_done(priv, iperf_errno);
                        return OWL_FAILURE;
                }

                priv->remote_ip = owl_os_strdup(
                        inet_ntoa(sin_remote.sin_addr));
                
                priv->remote_port = owl_ntohs(sin_remote.sin_port);
		iperf_close(priv->sock);
		priv->sock = new_sock;
		
		if ((err = setsockopt(
		             priv->sock, SOL_SOCKET, SO_RCVBUF,
			     (char *)&optval, sizeof(optval))) < 0) {
                        owl_err("setsockopt failed, err:%d", iperf_errno);
			iperf_close(priv->sock);
                        iperf_done(priv, iperf_errno);
                        return OWL_FAILURE;
                }
		priv->start_ms = owltmr_get_ms();
                
                if (priv->connect_cb)
                        priv->connect_cb(priv->ctx, 
                                         priv->local_port,
                                         priv->remote_port, 
                                         priv->local_ip,
                                         priv->remote_ip);
		
		owltmr_sched(
                        &priv->report_tmo, priv->report_intvl * 1000, 1, 
                        report_timer_cb, priv);
                
                owltmr_sched(&priv->state_tmo, 0, 1, tcp_recv_cb, priv);
		
        }
#else /* lwip */
        if (priv->udp) {
                if ((priv->udp_pcb = udp_new()) == NULL) {
                        owl_err("fail to alloc udp pcb");
                        return iperf_done(priv, OWL_ERR_MEM);
                }

                if ((err = udp_bind(
                             priv->udp_pcb, IP_ADDR_ANY, port)) != ERR_OK) {
                        owl_err("fail to udp bind err:%d", err);
                        return iperf_done(priv, err);
                }

                udp_recv(priv->udp_pcb, udp_recv_cb, priv);
                priv->bufcnt = -1;
                priv->intvl_bufcnt = -1;
                priv->pkt_id = -1;
           
        } else { /* TCP */
                if ((priv->tcp = tcp_new()) == NULL) {
                        owl_err("fail to alloc tcp pcb");
                        return iperf_done(priv, OWL_ERR_MEM);
                }
                owl_dbg("tcp_new:%p", priv->tcp); 
        
                if ((err = tcp_bind(
                             priv->tcp, IP_ADDR_ANY, port)) != ERR_OK) {
                        owl_err("bind failed:%d", err);
                        tcp_abort(priv->tcp);
                        priv->tcp = NULL;
                        return iperf_done(priv, owlip_error(err));
                }
                
                if ((lpcb = tcp_listen(priv->tcp)) == NULL) {
                        owl_err("fail to listen");
                        tcp_abort(priv->tcp);
                        priv->tcp = NULL;
                        return iperf_done(priv, OWL_ERR_MEM);
                }
                
                /* priv->tcp was free'd by tcp_listen() */
                priv->tcp = lpcb;
                tcp_arg(priv->tcp, priv);
                tcp_accept(priv->tcp, tcp_accept_cb);
                /* Wait for tcp_accept_cb() */
        }
#endif /* OWL_CONFIG_SOCKET_BSD */
        return 0;
}

#if OWL_CONFIG_SOCKET_BSD
static void 
tcp_recv_cb(struct owltmr_tmo *tmo, void *ctx)
{
        struct iperf_ctx *priv = ctx;
        int err;

        owl_unused(tmo);
	
	if((err = recv(priv->sock, priv->buf, 
		       priv->buflen, MSG_DONTWAIT)) > 0) {
		priv->recvd += err;
		priv->bufcnt++;
		priv->intvl_recvd += err;
		priv->intvl_bufcnt++;
        } else if (err == 0) {
                /* Remote end closed the connection */
                iperf_done(priv, 0);
        } else if (iperf_errno == IPERF_EAGAIN || 
		   iperf_errno == IPERF_EWOULDBLOCK) {
                /* No data, try again */
                return;
        } else {
                /* fail */
                iperf_done(priv, iperf_errno);
        }
}

static void 
tcp_send_cb(struct owltmr_tmo *tmo, void *ctx)
{
        struct iperf_ctx *priv = ctx;
        int err;
        int intvl;
        int len;

        owl_unused(tmo);
       
        owl_dbg("bufcnt:%d bufleft:%d", priv->bufcnt, priv->bufleft);
        
        len = priv->bufleft;
        intvl = owltmr_get_intvl(priv->start_ms, owltmr_get_ms());
        
        /* all sent */
        if (priv->bufcnt == priv->nbuf ||
            intvl >= priv->duration) {
                iperf_done(priv, 0);
                return;
	}
	if ((err = send(priv->sock, priv->buf ,len, MSG_DONTWAIT)) < 0) {
                if (iperf_errno == IPERF_EAGAIN || 
		    iperf_errno == IPERF_EWOULDBLOCK) {
			return;
                } else { 
                        owl_err("failed to write err:%d", iperf_errno);
                        iperf_done(priv, iperf_errno);
                }
        }
        owl_dbg("wrote %d bytes", err);
        priv->bufleft -= err;
        
        if (priv->bufleft == 0) {
                priv->bufleft = priv->buflen;
                priv->bufcnt++;
                priv->intvl_bufcnt++;
        }
}

static void
udp_recv_cb(struct owltmr_tmo *tmo, void *ctx)
{
        struct iperf_ctx *priv = ctx;
        struct udp_datagram *udp_buf = (struct udp_datagram *) priv->buf;
        int32_t pkt_id;
        int err = 0;

        owl_unused(tmo);

        /* Start */
        if (priv->pkt_id < 0) {
                struct sockaddr_in sin_remote;
                int len = sizeof(sin_remote);

                /* get remote ip and port for server report */
                if ((err = recvfrom(priv->sock, priv->buf, 
				    priv->buflen, 0, 
				    (struct sockaddr *) &sin_remote, 
                                    &len)) < 0) {
			owl_err("failed to recv on sock:%d", priv->sock);
			iperf_done(priv, iperf_errno);
			return;
                }
                
		priv->start_ms = owltmr_get_ms();

                /* Store remote ip and port */
                priv->remote_port = owl_ntohs(sin_remote.sin_port);
                priv->remote_ip = owl_os_strdup(
                        inet_ntoa(sin_remote.sin_addr));
                
                if (priv->connect_cb)
                        priv->connect_cb(priv->ctx, 
                                         priv->local_port,
                                         priv->remote_port, 
                                         priv->local_ip,
                                         priv->remote_ip);
                
                owltmr_sched(&priv->report_tmo, priv->report_intvl * 1000, 
                             1, report_timer_cb, priv);

        } else {
                if ((err = recv(priv->sock, priv->buf, 
                                priv->buflen, MSG_DONTWAIT)) < 0) {
			if (iperf_errno == IPERF_EAGAIN || 
			    iperf_errno == IPERF_EWOULDBLOCK) {
                                /* No data received, try again */
                                return;
                        } else {
                                /* failure */
                                iperf_done(priv, iperf_errno);
                                return;
                        }
                }
        }

        priv->recvd += err;
        priv->intvl_recvd += err;
        priv->bufcnt++;
        priv->intvl_bufcnt++;
        
	pkt_id = (int32_t) owl_ntohl(udp_buf->id);

        /* Do not store negative id:s when checking packet loss.
           check of negative id:s (end transm.) is done further below. */
        if (pkt_id < 0) {
                /* Dont calculate packet loss at end of session
                   if remote side is not using H&D iperf version (bug) */
                if (owl_bit(priv->flags, IPERF_REMOTE_VERSION_HD))   
                        pkt_id = -pkt_id;
                else
                        goto done;
        }

        /* Check for packet-loss / out-of-order */
        if (pkt_id != priv->pkt_id + 1) {
                if (pkt_id < priv->pkt_id + 1) {
                        priv->pkts_outoforder++;
                } else {
                        uint32_t lost = pkt_id - priv->pkt_id - 1;
                        priv->intvl_pkts_lost += lost;
                        priv->pkts_lost += lost; 
                        priv->intvl_bufcnt += lost;
                        priv->bufcnt += lost;
                }
        }

done :
        /* Client is sending negative id:s = end of session */
        if ((int) owl_ntohl(udp_buf->id) < 0) 
                iperf_done(priv, 0);
        else 
                priv->pkt_id = pkt_id;
}

#else /* lwip */

static void
udp_recv_cb(void *arg, struct udp_pcb *upcb, struct pbuf *p,
            struct ip_addr *addr, u16_t port)
{
        struct iperf_ctx* priv = arg;
        struct udp_datagram *udp_buf = (struct udp_datagram *) p->payload;
        int32_t pkt_id;

        owl_dbg("p:%p p->tot_len:%d", p, p->tot_len);

        /* If in UDP client mode and receiving,
           the server has probably sent us stats. */
        if (priv->client) {
                if (!priv->ack_fin) {
                        priv->buf = (char *) udp_buf;
                        priv->ack_fin = 1;
                        iperf_done(priv, 0);
                } else { 
                
                }
                return;
        }

        pkt_id = (int32_t) owl_ntohl(udp_buf->id);

        /* Start */
        if (priv->pkt_id < 0) {
                priv->start_ms = owltmr_get_ms();
                        
                /* Save remote ip and port so we can send back iperf msgs 
                   to host */
                priv->remote_ip = owl_os_strdup(ip_ntoa(addr));
                priv->remote_port = port;
                
                if (priv->connect_cb)
                        priv->connect_cb(priv->ctx, 
                                         priv->local_port,
                                         priv->remote_port, 
                                         priv->local_ip,
                                         priv->remote_ip);
      
                owltmr_sched(&priv->report_tmo, priv->report_intvl * 1000, 
                             1, report_timer_cb, priv);
        }

        priv->recvd += p->tot_len;
        priv->intvl_recvd += p->tot_len;
        priv->bufcnt++;
        priv->intvl_bufcnt++;
        
        /* Do not store negative id:s when checking packet loss.
           check of negative id:s (end transm.) is done further below. */
        if (pkt_id < 0) {
                /* Dont calculate packet loss at end of session
                   if remote side is not using H&D iperf version (bug) */
                if (owl_bit(priv->flags, IPERF_REMOTE_VERSION_HD))   
                        pkt_id = -pkt_id;
                else
                        goto done;
        }
        /* Check for packet-loss / out-of-order */
        if (pkt_id != priv->pkt_id + 1) {
                if (pkt_id < priv->pkt_id + 1) {
                        priv->pkts_outoforder++;
                } else {
                        uint32_t lost = pkt_id - priv->pkt_id - 1;
                        priv->intvl_pkts_lost += lost;
                        priv->pkts_lost += lost; 
                        priv->intvl_bufcnt += lost;
                        priv->bufcnt += lost;
                }
        }

done :
        /* Client is sending negative id:s = end of session */
        if ((int) owl_ntohl(udp_buf->id) < 0) 
                iperf_done(priv, 0);
        else 
                priv->pkt_id = pkt_id;
        
        pbuf_free(p);
}

static void
tcp_send_data(struct iperf_ctx *priv)
{
        err_t err;
        int intvl;
        
        owl_dbg("bufcnt:%d bufleft:%d tcp_sndbuf:%d",
                priv->bufcnt, priv->bufleft, tcp_sndbuf(priv->tcp));

        while (tcp_sndbuf(priv->tcp)) {

                int len = priv->bufleft;
                intvl = owltmr_get_intvl(priv->start_ms, owltmr_get_ms());
                
                /* all sent */
                if (priv->bufcnt == priv->nbuf ||
                    intvl >= priv->duration) {
                        iperf_done(priv, 0);
                        return;
                }
                
                /* We cannot send more data than space available in the send
                   buffer. */
                if (len > tcp_sndbuf(priv->tcp))
                        len = tcp_sndbuf(priv->tcp);

                if ((err = tcp_write(
                             priv->tcp, priv->buf, len, 0)) != ERR_OK) {
                        /* out of pbufs */
                        owl_info("fail to write len:%d err:%d", len, err);
                        return;
                }

                tcp_output(priv->tcp);
                owl_dbg("wrote %d bytes", len);

                priv->bufleft -= len;
                
                if (priv->bufleft == 0) {
                        priv->bufleft = priv->buflen;
                        priv->bufcnt++;
                        priv->intvl_bufcnt++;
                }
        }
}

static err_t
tcp_sent_cb(void *arg, struct tcp_pcb *pcb, u16_t len)
{
        struct iperf_ctx *priv = arg;

        (void) pcb;
        (void) len;
        
        tcp_send_data(priv);
        return ERR_OK;
}

static void 
tcp_err_cb(void *arg, err_t err)
{
        struct iperf_ctx *priv = arg;
        (void) err;
        owl_info("err:%d", err);
        priv->tcp = NULL; /* free'd by lwip upon return */
        iperf_done(priv, OWL_ERR_CONN);
}


static err_t 
tcp_connect_cb(void *arg, struct tcp_pcb *pcb, err_t err)
{
        struct iperf_ctx *priv = arg;

        (void) pcb;
        (void) err;
        
        owl_assert(priv->client);
        owl_info("iperf connected, start send");
    
        priv->local_port = pcb->local_port;

        priv->local_ip = owl_os_strdup(ip_ntoa(&pcb->local_ip));

        if (priv->connect_cb)
                priv->connect_cb(priv->ctx, priv->local_port, 
                                 priv->remote_port,
                                 priv->local_ip, 
                                 priv->remote_ip);

        priv->start_ms = owltmr_get_ms();
        
        owltmr_sched(&priv->report_tmo, priv->report_intvl * 1000, 1, 
                     report_timer_cb, priv);

        tcp_sent(pcb, tcp_sent_cb);
        tcp_send_data(priv);

        return ERR_OK;
}


static err_t
tcp_accept_cb(void *arg, struct tcp_pcb *pcb, err_t err)
{
        struct iperf_ctx *priv = arg;

        (void) err;

        owl_dbg("pcb:%p", pcb);
        
        if (tcp_close(priv->tcp) != ERR_OK) {
                owl_err("fail to close listeing pcb, aborting pcb");
                tcp_abort(priv->tcp);
        }
        
        priv->remote_ip = owl_os_strdup(ip_ntoa(&pcb->remote_ip));
        priv->remote_port = pcb->remote_port;
        priv->local_ip = owl_os_strdup(ip_ntoa(&pcb->local_ip));

        if (priv->connect_cb)
                priv->connect_cb(priv->ctx, priv->local_port, 
                                 priv->remote_port,
                                 priv->local_ip, 
                                 priv->remote_ip);
        
        priv->tcp = pcb;
        tcp_arg(pcb, priv);
        tcp_recv(pcb, tcp_recv_cb);
        tcp_err(pcb, tcp_err_cb);

        priv->start_ms = owltmr_get_ms();

        owltmr_sched(&priv->report_tmo, priv->report_intvl * 1000, 1, 
                     report_timer_cb, priv);
        return ERR_OK;
}


static err_t 
tcp_recv_cb(void *arg, struct tcp_pcb *pcb, struct pbuf *p, err_t err)
{
        struct iperf_ctx* priv = arg;

        (void) err;

        owl_dbg("p:%p p->tot_len:%d", p, p ? p->tot_len : 0);

        if (priv->client) {
                owl_info("ignoring rx data");
                return ERR_OK;
        }
        
        /* p will be NULL when remote end is done */
        if (p == NULL) {
                iperf_done(priv, 0);
                return ERR_OK;
        }

        priv->recvd += p->tot_len;
        priv->bufcnt++;
        priv->intvl_recvd += p->tot_len;
        priv->intvl_bufcnt++;

        tcp_recved(pcb, p->tot_len);
        pbuf_free(p);

        return ERR_OK;
}
#endif /* !OWL_CONFIG_SOCKET_BSD */

#if __GNUC__
#pragma GCC diagnostic warning "-Wcast-qual"
#endif

static int
udp_send_data(struct iperf_ctx *priv, int len)
{
        int err = 0;
#if OWL_CONFIG_SOCKET_BSD
        struct sockaddr_in sin_remote;
        sin_remote.sin_family = AF_INET;
        inet_aton(priv->remote_ip, &sin_remote.sin_addr);
        sin_remote.sin_port = owl_htons(priv->remote_port);

	if((err = sendto(priv->sock, priv->buf, len, MSG_DONTWAIT, 
                         (struct sockaddr *) &sin_remote, 
                         sizeof(sin_remote))) < 0) {
                if (iperf_errno == IPERF_EAGAIN || 
		    iperf_errno == IPERF_EWOULDBLOCK ||
		    iperf_errno == IPERF_ENOBUFS) {
			return OWL_ERR_AGAIN;
                } else {
                        iperf_done(priv, iperf_errno);
                }
        }
        return err;
        
#else /* lwip */
        struct pbuf *p;
        struct ip_addr remote_ip;
        ip_aton(priv->remote_ip, &remote_ip);

        if ((p = pbuf_alloc(PBUF_TRANSPORT, len, PBUF_ROM)) == NULL) {
                owl_dbg("fail to alloc udp pbuf of len:%d", priv->buflen);
                return OWL_ERR_MEM;
        }
        p->payload = priv->buf;
        
        if ((err = udp_sendto(priv->udp_pcb, p, &remote_ip,
                              priv->remote_port)) != ERR_OK) {
                owl_dbg("fail to send udp pbuf %d", err);
        }
        pbuf_free(p);
        
        return owlip_error(err);
#endif /* OWL_CONFIG_SOCKET_BSD */
}

static void
report_timer_cb(struct owltmr_tmo *tmo, void *ctx)
{
        struct iperf_ctx *priv = ctx;
        uint32_t bytes;
        
        uint32_t ms = 
                owltmr_get_intvl(priv->start_ms, owltmr_get_ms());
        
        if (priv->client) 
                bytes = priv->intvl_bufcnt * priv->buflen;
        else
                bytes = priv->intvl_recvd;

        if (priv->report_cb)
                priv->report_cb(priv->ctx, 0, bytes, 
				ms - (priv->report_intvl * 1000), 
				ms, 0, priv->intvl_bufcnt, 
				priv->intvl_pkts_lost, 0,
				IPERF_REPORT_TYPE_INTERVAL);
        
        priv->intvl_bufcnt = 0;
        priv->intvl_recvd = 0;
        priv->intvl_pkts_lost = 0;
}

static void
udp_timer_cb(struct owltmr_tmo *tmo, void *ctx)
{
        struct iperf_ctx *priv = ctx;
        struct udp_datagram *udp_data = (struct udp_datagram *) priv->buf;
        int err;
        int intvl;
        
        owl_unused(tmo);

        /* Update packet id and time in datagram to be sent to server */
        if (priv->pkt_id > 0 && priv->fin_cnt == 0) {
                udp_data->id = owl_htonl(priv->pkt_id);

                udp_data->sec = owl_htonl(
                        (owltmr_get_ms() - priv->start_ms) / 1000);
		udp_data->usec = owl_htonl(
			 (owltmr_get_ms() - priv->start_ms) * 1000);	   
        }

        if (priv->pkt_id < 0) {
                priv->start_ms = owltmr_get_ms();

                owltmr_sched(&priv->report_tmo, priv->report_intvl * 1000, 
                             1, report_timer_cb, priv);
#if OWL_CONFIG_SOCKET_BSD
                struct sockaddr local_addr;
                int len = sizeof(struct sockaddr);
		
		priv->pkt_id = 0;
		udp_data->id = owl_htonl(priv->pkt_id);

                udp_data->sec = owl_htonl(
                        (owltmr_get_ms() - priv->start_ms) / 1000);
		udp_data->usec = owl_htonl(
                        (owltmr_get_ms() - priv->start_ms) * 1000);
		
		/* Send first packet with ID:0 and store local ip and port */
                if ((err = udp_send_data(priv, priv->buflen)) < 0)
                        return;

                /* Get local ip and port */
                getsockname(
                        priv->sock, &local_addr, &len);
                
                priv->local_port = owl_ntohs(
                        ((struct sockaddr_in *) &local_addr)->sin_port);
                
                priv->local_ip = owl_os_strdup(
                        inet_ntoa(((struct sockaddr_in *) 
                                   &local_addr)->sin_addr));
                
                if (priv->connect_cb)
                        priv->connect_cb(
                                priv->ctx, 
                                priv->local_port, priv->remote_port, 
                                priv->local_ip, priv->remote_ip);
		goto done;
#else /* lwip */
                priv->local_port = priv->udp_pcb->local_port;
                priv->local_ip = owl_os_strdup(
                        ip_ntoa(&priv->udp_pcb->local_ip));

                if (priv->connect_cb)
                        priv->connect_cb(priv->ctx, priv->local_port, 
                                         priv->remote_port,
                                         priv->local_ip,
                                         priv->remote_ip);
#endif /* OWL_CONFIG_SOCKET_BSD */
        }

        /* all sent */
        if (priv->bufcnt == priv->nbuf ||
            owltmr_intvl_passed(priv->start_ms, 
                                owltmr_get_ms(),
                                priv->duration)) {
         
                iperf_done(priv, 0);
                return;
        }
        
        /* Bandwidth limit */
        if (priv->bandwidth && priv->bufcnt) {
                uint32_t sent, bandwidth;

                intvl = owltmr_get_intvl(priv->start_ms, owltmr_get_ms());

                if (!intvl)
                        return;

                sent = priv->bufcnt * priv->buflen * 8 * 1000;
                bandwidth = sent / intvl;
                
                if (bandwidth > priv->bandwidth) {
                        owl_err("bandwidth:%d exceeded limit:%d",
                                bandwidth, priv->bandwidth);
                        return;
                }
        }

        if ((err = udp_send_data(priv, priv->buflen)) < 0)
                return;
        
#if OWL_CONFIG_SOCKET_BSD
done:
#endif /* OWL_CONFIG_SOCKET_BSD */

        priv->pkt_id++;
        priv->bufcnt++;
        priv->intvl_bufcnt++;
}

static int
iperf_done(struct iperf_ctx *priv, int err)
{
        uint32_t ms = 0;
        int i = 0;

        owl_dbg("err:%d", err);
        
        owltmr_cancel(&priv->report_tmo);

        if (err != 0) {
                if (priv->report_cb)
                        priv->report_cb(priv->ctx, err, 0, 0, 0, 0, 0, 0, 0, 
					IPERF_REPORT_TYPE_ERR);
                goto done;
        }

        ms = owltmr_get_intvl(priv->start_ms, owltmr_get_ms());

	if (!priv->udp) {
                uint32_t bytes = 0;
                int tcp_mss = 0;
		
                /* End report */
                if (priv->client) {
                        bytes = priv->bufcnt * priv->buflen;
			tcp_mss = iperf_advmss(priv);
		} else {
                        bytes = priv->recvd;
			tcp_mss = iperf_rcvmss(priv);
		}

		if (priv->report_cb)
                        priv->report_cb(priv->ctx, 0, bytes, 0,
					ms, tcp_mss, 0, 0, 
					0, IPERF_REPORT_TYPE_DONE);
                
        } else if (priv->udp && priv->client) {
                int new;
                int intvl = 0;
                
                struct udp_datagram *udp_data = 
                        (struct udp_datagram *) priv->buf;
                
                /* Print end stats */
                if (priv->report_cb && priv->fin_cnt == 0) {
                        priv->report_cb(priv->ctx, 0,
					priv->bufcnt *
					priv->buflen,
					0, ms, 0, priv->bufcnt, 0, 0,
					IPERF_REPORT_TYPE_DONE);
                }
                
                /* Send ACK_FIN 10 times */
                if (priv->fin_cnt == 0)
                        udp_data->id = owl_htonl((-priv->pkt_id));
                
#if OWL_CONFIG_SOCKET_BSD
                while(priv->fin_cnt < 10 && !priv->ack_fin) {
                        int len = 0;
			
                        udp_send_data(priv, sizeof(struct udp_datagram));
                        
                        /* Wait 250ms for server status */
                        ms = owltmr_get_ms();
                        
                        do {
                                new = owltmr_get_ms();
                        } while ((intvl = 
                                  owltmr_get_intvl(ms, new)) < 250);
                        
                        if ((len = recv(
                                     priv->sock, 
                                     priv->buf,
                                     sizeof(struct udp_datagram), 
                                     MSG_DONTWAIT)) > 0) {
				
                                priv->ack_fin = 1;
                        } 
                        priv->fin_cnt++;
                }
#else /* lwip */
                if (priv->fin_cnt < 10 && !priv->ack_fin) {
                        udp_send_data(
                                priv, sizeof(struct udp_datagram));
                        
                        /* Wait 250ms for server status */
                        ms = owltmr_get_ms();
                        
                        do {
                                new = owltmr_get_ms();
                        } while ((intvl = 
                                  owltmr_get_intvl(ms, new)) < 250);
                        
                        priv->fin_cnt++;
  
                         /* iperf_done will be called again
                            from udp_timer_cb */
                        return 0;
                }
#endif /* OWL_CONFIG_SOCKET_BSD */
		if (owl_htonl(udp_data->hdr.flags) != HEADER_VERSION1) 
                        err = OWL_FAILURE;
		
                /* server report */
                if (priv->report_cb)
                        priv->report_cb(
				priv->ctx, err,
				owl_ntohl(udp_data->hdr.total_len2),
				0,
				owl_ntohl(udp_data->hdr.stop_sec) * 1000 +
				owl_ntohl(udp_data->hdr.stop_usec) / 1000,
				0,
				owl_ntohl(udp_data->hdr.datagrams), 
				owl_ntohl(udp_data->hdr.error_cnt),
				0,
				IPERF_REPORT_TYPE_SERVER);

        } else if (priv->udp && !priv->client) {
                int new;
                int intvl = 0;

#if OWL_CONFIG_SOCKET_BSD
                owl_os_memset(priv->buf, 0, sizeof(struct udp_datagram));
#else /* lwip */
                if ((priv->buf = owl_os_zalloc(
                             sizeof(struct udp_datagram))) == NULL) {
                        owl_err("could not alloc ACK_FIN");
                }
#endif /* OWL_CONFIG_SOCKET_BSD */

                struct udp_datagram *udp = 
                        (struct udp_datagram *) priv->buf;
                

                /* Print endstats */
                if (priv->report_cb)
                        priv->report_cb(priv->ctx, 0, priv->recvd, 0, ms, 0, 
					priv->bufcnt, priv->pkts_lost,
					priv->pkts_outoforder, 
					IPERF_REPORT_TYPE_DONE);
                
                udp->hdr.flags = owl_htonl(HEADER_VERSION1);
                udp->hdr.total_len1 = 0;
                udp->hdr.total_len2 = owl_htonl(priv->recvd);
                udp->hdr.stop_sec = owl_htonl(ms / 1000);
                udp->hdr.stop_usec = owl_htonl((ms % 1000) * 1000);
                udp->hdr.outorder_cnt = owl_htonl(priv->pkts_outoforder);
                udp->hdr.error_cnt = owl_htonl(priv->pkts_lost);
                udp->hdr.datagrams = owl_htonl(priv->bufcnt);
                
                ms = owltmr_get_ms();
                
                /* Try 10 times with 250ms interval*/
                for (i = 0; i < 10; i++) {
                        udp_send_data(priv, sizeof(*udp));
                        
                        do { 
                                new = owltmr_get_ms();
                        } while ((intvl = owltmr_get_intvl(ms, new)) < 250);
                }
        }
done:
#if OWL_CONFIG_SOCKET_BSD 
        owltmr_cancel(&priv->state_tmo);
	shutdown(priv->sock, SHUT_RDWR);
	iperf_close(priv->sock);
	
#else /* lwip */
        if (priv->tcp) {
                tcp_arg(priv->tcp, NULL);
                tcp_sent(priv->tcp, NULL);
                tcp_recv(priv->tcp, NULL);
                tcp_close(priv->tcp);
        }
        
        if (priv->udp) {
                udp_remove(priv->udp_pcb);
        }

#endif /* OWL_CONFIG_SOCKET_BSD */
        owltmr_cancel(&priv->udp_tmo);
        
        if (priv->remote_ip)
                owl_os_free(priv->remote_ip);
        if (priv->local_ip)
                owl_os_free(priv->local_ip);
        if (priv->buf)
                owl_os_free(priv->buf);
        
        owl_os_free(priv);
        return err;
}

static int
iperf_advmss(struct iperf_ctx *priv)
{
	/* TBD: get mss when using NG_RTOS */
#if OWL_CONFIG_SOCKET_BSD
 #if defined(NG_RTOS)
	return ngTcp_MinMss;
 #else
	struct tcp_info tcp_info;
	int info_len = sizeof(struct tcp_info);  
	
	getsockopt(priv->sock, IPPROTO_TCP, TCP_INFO,
			 &tcp_info, &info_len);

	return tcp_info.tcpi_advmss;
 #endif /* NG_RTOS */
 
#else /* lwip */
	return priv->tcp->mss;
#endif /* OWL_CONFIG_SOCKET_BSD */
}

static int
iperf_rcvmss(struct iperf_ctx *priv)
{
	/* TBD: get mss when using NG_RTOS */
#if OWL_CONFIG_SOCKET_BSD
 #ifndef NG_RTOS 
	struct tcp_info tcp_info;
	int info_len = sizeof(struct tcp_info);  
	
	getsockopt(priv->sock, IPPROTO_TCP, TCP_INFO,
			 &tcp_info, &info_len);

	return tcp_info.tcpi_rcv_mss;
 #endif /* !NG_RTOS */

#else /* lwip */
	return (int) TCP_MSS;
#endif /* OWL_CONFIG_SOCKET_BSD */
	return 0;
}

#if OWL_CONFIG_SHELL
/* Store some info if shell command is used */
static struct shell_ctx {
        int client;
        int report_format;
        int udp;
	int tcp_mss;
} shell_ctx;

static void
connect_cb(void *ctx, int local_port, int remote_port,
           char *local_ip, char *remote_ip)
{
        struct owlsh_ctx *sh = ctx;

        owlsh_printf(sh, "local %s port:%d connected ",local_ip, local_port);
        owlsh_printf(sh, "with %s port %d\r\n", remote_ip, remote_port);
        
        owlsh_printf(sh, "Interval         "
                     "Transfer      Bandwidth    ");
        
        if (!shell_ctx.client && shell_ctx.udp) 
                owlsh_printf(sh, "Lost/Total Datagrams");
        
        owlsh_printf(sh, "\n\r");
}

static void
report_cb(void *ctx, int err, uint32_t bytes, uint32_t start_ts, 
          uint32_t end_ts, int mss, uint32_t packets, uint32_t pkts_lost, 
	  uint32_t outoforder, int type)
{
        struct owlsh_ctx *sh = ctx;
        uint32_t bandwidth = 0;
        uint32_t bandwidth_dec = 0;
        uint32_t bytes_dec = 0;
        char *format_bw = NULL;
        char *format_len = NULL;
        uint32_t ms = end_ts - start_ts;

        owl_dbg("ms:%d bytes:%d packets:%d", ms, bytes, packets);
        
        if (type == IPERF_REPORT_TYPE_ERR) {
                owlsh_printf(sh, "err: %s\r\n", owl_error(err));
                return;
        }
        
        if (shell_ctx.report_format == REPORT_KBITS) {
		bandwidth = (((bytes * 8) / 1024) * 1000) / ms;
		bandwidth_dec = ((((bytes * 1000 * 8) / 1024) * 1000) / 
				 ms) % 1000;
                format_bw = "KBits";
        } else if (shell_ctx.report_format == REPORT_MBITS) {
		bandwidth = ((((bytes * 8) / (1024 * 1024)) * 1000) / ms);
		bandwidth_dec = ((((bytes * 1000 * 8) / (1024 * 1024)) * 1000) 
				 / ms)  % 1000;
                format_bw = "MBits";
        } else if (shell_ctx.report_format == REPORT_KBYTES) {
                bandwidth = ((bytes / 1024) * 1000) / ms; 
                bandwidth_dec = ((((bytes * 1000) / 1024) * 1000) / 
                                 ms) % 1000; 
                format_bw = "KBytes";
        } else if (shell_ctx.report_format == REPORT_MBYTES) {
                bandwidth = ((bytes / (1024 * 1024)) * 1000) / ms;
                bandwidth_dec = ((((bytes * 1000) / ( 1024 * 1024)) * 1000) / 
                                 ms) % 1000;
                format_bw = "MBytes";
        }
        
        if (bytes < (1024 * 1024)) {
                bytes_dec = ((bytes * 100) / 1024) % 100;
                bytes /= 1024;
                format_len = "Kbytes";
        
        } else if (bytes < (1024 * 1024 * 1024)) {
                bytes_dec = ((bytes * 100) / (1024 * 1024)) % 100;
                bytes /= (1024 * 1024);
                format_len = "Mbytes";

        } else {
                bytes_dec = ((bytes * 100) / (1024 * 1024 * 1024)) % 100;
                bytes /= (1024 * 1024 * 1024);
                format_len = "Gbytes"; 
        }

        if (shell_ctx.udp && !shell_ctx.client) {
                owlsh_printf(sh,"%d.%d - %d.%d sec %d.%02d %s  %d.%03d %s/sec "
                             "%d / %d\r\n", 
                             (int) start_ts / 1000, (int) start_ts % 1000,
			     (int) end_ts / 1000, (int) end_ts % 1000, 
			     (int) bytes, (int) bytes_dec, format_len,
                             (int) bandwidth, (int) bandwidth_dec, format_bw,
                             (int) pkts_lost, (int) packets);
	
		if (type == IPERF_REPORT_TYPE_DONE)
			owlsh_printf(sh, "Received %d packets out of order"
                                     "\r\n", (int) outoforder);

	} else if (shell_ctx.udp && shell_ctx.client) { 
		if (type == IPERF_REPORT_TYPE_SERVER) {
			if (err == OWL_FAILURE) {
				owlsh_printf(sh, "\r\n WARNING: did not "
					     "receive ACK of last datagram" 
                                             " after 10 tries\r\n");
			} else {
				owlsh_printf(sh, "\r\n Server report:\r\n");
				owlsh_printf(sh,"%d.%d - %d.%d sec %d.%02d %s  "
					     "%d.%03d %s/sec %d / %d\r\n", 
					     (int) start_ts / 1000,
					     (int) start_ts % 1000,
					     (int) end_ts / 1000, 
					     (int) end_ts % 1000,
					     (int) bytes, (int) bytes_dec, 
					     format_len, (int) bandwidth, 
					     (int) bandwidth_dec, format_bw, 
					     (int) pkts_lost, (int) packets);
			}
		} else {
			owlsh_printf(sh,"%d.%d - %d.%d sec %d.%02d %s  %d.%03d "
				     "%s/sec \r\n", 
				     (int) start_ts / 1000,
				     (int) start_ts % 1000,
				     (int) end_ts / 1000, 
				     (int) end_ts % 1000,
				     (int) bytes, (int) bytes_dec, 
				     format_len, (int) bandwidth, 
				     (int) bandwidth_dec, format_bw);
		}

		if (type == IPERF_REPORT_TYPE_DONE)
			owlsh_printf(sh, "Sent %d datagrams\r\n",
                                     (int) packets);
		
        } else { /* tcp */
                owlsh_printf(sh,"%d.%d - %d.%d sec %d.%02d %s  "
                             "%d.%03d %s/sec \r\n", 
			     (int) start_ts / 1000,
			     (int) start_ts % 1000,
                             (int) end_ts / 1000, 
			     (int) end_ts % 1000,
			     (int) bytes, (int) bytes_dec, 
                             format_len, (int) bandwidth, 
                             (int) bandwidth_dec, format_bw);

		if (type == IPERF_REPORT_TYPE_DONE && mss && shell_ctx.tcp_mss)
			owlsh_printf(sh, "MSS size %d bytes\r\n", mss);
	}
}

void
iperf_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int i;
        uint16_t port = 5001;          /* -p */
        uint32_t nbuf = 1024;          /* -n */
        uint32_t buflen = 1024;        /* -l */
        char ip[16];

        int err = 0;
        int bandwidth = 0;
        int duration = 10;
        int interval = 1;
        uint8_t flags = 0;

        shell_ctx.client = 1;
        shell_ctx.udp = 0;
        shell_ctx.report_format = REPORT_MBITS;

        void (*report)(void *ctx, int err, uint32_t bytes, 
                       uint32_t start_ts, uint32_t end_ts, int mss,
                       uint32_t packets, uint32_t pkts_lost,
                       uint32_t outoforder, int type) = report_cb;

        if (argc < 2)
                goto err_parse;

	for (i = 1; i < argc; i++) {
                if (owl_os_strcmp(argv[i], "-h") == 0) {
                        /* -h help */
                        goto err_parse;

                } else if (owl_os_strcmp(argv[i], "-c") == 0) {
                        /* -c <client> */

			shell_ctx.client = 1;

                } else if (owl_os_strcmp(argv[i], "-s") == 0) {
                        /* -s <server> */

			shell_ctx.client = 0;

                } else if (owl_os_strcmp(argv[i], "-u") == 0) {
                        /* -u <udp> */

			shell_ctx.udp = 1;

                } else if (owl_os_strcmp(argv[i], "-p") == 0) {
                        /* -p <port> */
			if (++i >= argc)
				goto err_parse;
                        
                        port = owl_os_atoi(argv[i]);
                        
                } else if (owl_os_strcmp(argv[i], "-n") == 0) {
                        
                        /* -n <nbuf> */
			if (++i >= argc)
				goto err_parse;
                        
                        nbuf = owl_os_atoi(argv[i]);

		} else if (owl_os_strcmp(argv[i], "-l") == 0) {
			/* -l <buflen> */
			if (++i >= argc)
				goto err_parse;
                        
                        buflen = owl_os_atoi(argv[i]);
                        
		} else if (owl_os_strcmp(argv[i], "-t") == 0) {
                        /* time */
                        if (++i >= argc)
				goto err_parse;

                        duration = owl_os_atoi(argv[i]);
                        
                } else if (owl_os_strcmp(argv[i], "-b") == 0) {
			/* bandwidth limit */
			if (++i >= argc)
				goto err_parse;
                        
                        bandwidth = owl_os_atoi(argv[i]);
                        
                } else if (owl_os_strcmp(argv[i], "-f") == 0) {
                        /* format */

                        if (++i >= argc)
				goto err_parse;

                        if (owl_os_strchr(argv[i], 'k'))
                                shell_ctx.report_format = REPORT_KBITS;
                        else if (owl_os_strchr(argv[i], 'm'))
                                shell_ctx.report_format = REPORT_MBITS;
                        else if (owl_os_strchr(argv[i], 'K'))
                                shell_ctx.report_format = REPORT_KBYTES;
                        else if (owl_os_strchr(argv[i], 'M'))
                                shell_ctx.report_format = REPORT_MBYTES;
                        else
                                goto err_parse;

                } else if (owl_os_strcmp(argv[i], "-i") == 0) {
                        /* <interval> */
                        if (++i >= argc)
				goto err_parse;

                        interval = owl_os_atoi(argv[i]);

                } else if (owl_os_strcmp(argv[i], "-q") == 0) {
                        /* no report cb that will print stuff */
                        report = NULL;
                      
                } else if (owl_os_strcmp(argv[i], "-m") == 0) {
                        /* -m <print MSS> */
                        
                         shell_ctx.tcp_mss = 1;

                } else if (owl_os_strcmp(argv[i], "-a") == 0) {
                        /* -a <remote iperf version = H&D> */
                        
			flags |= IPERF_REMOTE_VERSION_HD;
                        
                } else {
                        owl_os_memcpy(ip, argv[i], sizeof(ip));
                }
        }
        if (shell_ctx.client)
                err = iperf_client(ip, port, shell_ctx.udp, nbuf, buflen, 
                                   bandwidth, duration, interval, flags, 
                                   connect_cb, report, sh); 
        else if (!shell_ctx.client)
                err = iperf_server(port, shell_ctx.udp, interval, nbuf, 
				   buflen, flags, connect_cb, report, sh);
        if (err < 0)
                owlsh_printf(sh, "%s %d (%s)\n\r", 
                             err >= 0 ? "ok" : "err", err, owl_error(err));
        return;

err_parse:
        owlsh_printf(sh, "usage: iperf "
                     "[-s|-c host] [options]]\n\r");
        owlsh_printf(sh, "Server or Client:\n\r");
        owlsh_printf(sh, "-p, --port      "
                     "#server port to listen on/connect to\n\r");
        owlsh_printf(sh, "-f, --format    "
                     "[kmKM]  format to report: Kbits, Mbits, "
                     "KBytes, MBytes\n\r");
        owlsh_printf(sh, "-i, --interval  "
                     "#seconds between periodic bandwidth reports\n\r");
        owlsh_printf(sh, "-h, --help      show this message and quit\n\r");
        owlsh_printf(sh, "Server specific:\n\r");
        owlsh_printf(sh, "-s, --server    run in server mode\n\r");
        owlsh_printf(sh, "-a, --using H&D modified iperf version in client"
                     "mode\n\r");
        owlsh_printf(sh, "Client specific:\n\r");
        owlsh_printf(sh, "-c, --client    <host>    "
                     "run in client mode, connecting to <host>\n\r");
        owlsh_printf(sh, "-u, --udp       use UDP rather than TCP\n\r");
        owlsh_printf(sh, "-b, --bandwidth #[KM][/#] "
                     "target bandwidth in bits/sec (0 for unlimited)\n\r");
        owlsh_printf(sh, "-t, --time      #time in seconds to transmit "
                     "for\n\r");
        owlsh_printf(sh, "-n, --bytes     ");
        owlsh_printf(sh, "#[KM] number of bytes to transmit. In server\r\n");
        owlsh_printf(sh, "mode, this sets the receive socket SO_RCVBUF\r\n");
        owlsh_printf(sh, "size (BSD ONLY) in Kbytes \n\r");
        owlsh_printf(sh, "-l, --len       "
                     "#[KM] length of buffer to read or write\n\r");
        owlsh_printf(sh, "-m, --print-mss   "
                     "#prints TCP maximum segment size\n\r");
}

const struct owlsh_cmd cmd_iperf = {
        "iperf", iperf_f, "iperf"
};
#endif /* OWL_CONFIG_SHELL */
