/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#ifndef IPERF_BSD_H
#define IPERF_BSD_H

/* FLAGS */
/* There is a bug in iperf v2.0.5
   resulting in non-detectable packet loss
   at end of session in server mode.
   Use this flag if H&D iperf implementation is used on remote side. */
#define IPERF_REMOTE_VERSION_HD   (1 << 0)

/* variable "type" passed to report_cb can be any of these defines.
   "Type" specifies which kind of event that triggered the callback */
#define IPERF_REPORT_TYPE_INTERVAL 0
#define IPERF_REPORT_TYPE_DONE 1
#define IPERF_REPORT_TYPE_SERVER 2
#define IPERF_REPORT_TYPE_ERR -1

/**
 * Start iperf in client mode.
 * 
 * @param ip: remote address.
 * @param port: remote port.
 * @param udp: if > 0 using udp, else using tcp.
 * @param nbuf: number of buflen sized bufs to send.
 * @param buflen: size of buffer to send.
 * @param bandwidth: bandwidth limit for sending in udp mode (bits/sec).
 * @param duration: session time limit in sec.
 * @param interval: report interval.
 * @param flags: flagfield which can be any of the flag options above.
 * @param connect_cb:          
 *        functionpointer to callback invoked when a connection has been made.
 *        param ctx is the opaque pointer set by @param ctx.
 *        param local_port contains the local port iperf is using.
 *        param remote_port contains the remote port that iperf is connected to.
 *        param local_ip contains the local ip in the form of a string.
 *        param remote_ip contains the remote ip in the form of a string.
 * @param report_cb:
 *        functionpointer to callback invoked when an reportevent has been 
 *        triggered. This also functions as a error callback.
 *        param ctx is the opaque pointer set by @param ctx.
 *        param err will contain any error message during session.
 *        If err < 0 report_cb will be invoked and the session will be 
 *        terminated.
 *        (see owl_err.h).
 *        param bytes contains the number of successfully sent bytes.
 *        param start_ts contains the start timestamp in ms for the 
 *        report interval.
 *        param end_ts contains the end timestamp in ms for the 
 *        report interval. 
 *        param mss contains the remote mss if tcp is used.
 *        param packets not used in client mode.
 *        param packets_lost not used in client mode.
 *        param outoforder not used in client mode.
 *        param type specifies which kind of event that triggered the 
 *        report callback. This can be any of the defines 
 *        IPERF_REPORT_TYPE_XXX. 
 * @param ctx: Opaque pointer that will be passed to the callbacks.
 * 
 **/
int
iperf_client(char *ip, uint16_t port, int udp, 
             uint32_t nbuf, uint32_t buflen, int bandwidth, int duration, 
             int interval, uint8_t flags,
             void (*connect_cb)(void *ctx, int local_port, int remote_port,
                                char *local_ip, char *remote_ip),
             void (*report_cb)(void *ctx, int err, uint32_t bytes, 
                               uint32_t start_ts, uint32_t end_ts, int mss,
                               uint32_t packets, uint32_t pkts_lost, 
			       uint32_t outoforder, int type),
             void *ctx);

/**
 * Start iperf in server mode.
 * 
 * @param port: local port to use for incoming connections.
 * @param udp: if > 0 using udp, else using tcp.
 * @param interval: report interval.
 * @param nbuf: if OWL_CONFIG_SOCKET_BSD=1 this can be used
 *        to increase the buffer for incoming data. (In Kbytes, default is 128).
 *        Otherwise nbuf should be set to 0 and is not used.
 * @param buflen: size of incoming data buffer.
 * @param flags: flagfield which can be any of the flag options above.
 * @param connect_cb:          
 *        functionpointer to callback invoked when a connection has been made.
 *        param ctx is the opaque pointer set by @param ctx.
 *        param local_port contains the local port iperf is using.
 *        param remote_port contains the remote port that iperf is connected to.
 *        param local_ip contains the local ip in the form of a string.
 *        param remote_ip contains the remote ip in the form of a string.
 * @param report_cb:
 *        functionpointer to callback invoked when an reportevent has been 
 *        triggered. This also functions as a error callback.
 *        param ctx is the opaque pointer set by @param ctx.
 *        param err will contain any error message during session.
 *        If err < 0 report_cb will be invoked and the session will be 
 *        terminated.
 *        (see owl_err.h).
 *        param bytes contains the number of successfully received bytes.
 *        param start_ts contains the start timestamp in ms for the 
 *        report interval.
 *        param end_ts contains the end timestamp in ms for the 
 *        report interval. 
 *        param mss contains the local mss if tcp is used.
 *        param packets contains the number of received packets during the 
 *        report interval if udp is used.
 *        param packets_lost contains the number of lost packets during the 
 *        report interval if udp is used.
 *        param outoforder contains any number of packets that has been
 *        received out of order in udp mode. Out of order packets is only
 *        valid when param type = IPERF_REPORT_TYPE_DONE.
 *        param type specifies which kind of event that triggered the 
 *        report callback. This can be any of the defines 
 *        IPERF_REPORT_TYPE_XXX.
 * @param ctx: Opaque pointer that will be passed to the callbacks.
 * 
 **/
int
iperf_server(uint16_t port, int udp, 
             int interval, uint32_t nbuf, uint32_t buflen, uint8_t flags, 
             void (*connect_cb)(void *ctx, int local_port, int remote_port,
                                char *local_ip, char *remote_ip),
             void (*report_cb)(void *ctx, int err, uint32_t bytes, 
                               uint32_t start_ts, uint32_t end_ts, int mss,
                               uint32_t packets, uint32_t pkts_lost,
			       uint32_t outoforder, int type),
             void *ctx);

#if OWL_CONFIG_SHELL
extern const struct owlsh_cmd cmd_iperf;
#endif /* OWL_CONFIG_SHELL */

#endif
