/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 *
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#ifndef BOOTFILE_H
#define BOOTFILE_H

#if !OWL_CONFIG_HOST
# include <owl/core/owl_defs.h>
# include <owl/core/owl_shell.h>
#endif /* !OWL_CONFIG_HOST */



/*
 * NOTE: The bootfile stuff is shared by the application and the bootloader,
 * so make sure that you know what you are doing before changing anything!
 *
 */

/*
 * The bootloader will read a struct bootfile_hdr and a number of
 * bootfile_desc's upon start. Each bootfile_desc describe an
 * operation that the bootloader should perform. Currently, there is
 * only one operation supported, that is to copy data from one mtd
 * device to another. Essentially this can be used to upgrade the
 * application image running in the internal flash memory of the
 * MCU. The internal flash and the external flash are both mtd
 * devices.
 *
 * The application should manage the transfer of the new firmware file
 * in some way so it can be stored in the filesystem. bootfile_xxx()
 * can then be invoked to write the bootfile to point to the
 * particular file. The bootloader will then read the descriptor and
 * perform the update (bootfile_execute()). Then the bootloader will
 * clear the bootfile_hdr to prevent from "upgrading" on every boot
 * (bootfile_erase()).
 *
 * Several different type of descriptors could potentially be used to
 * instruct the bootloader to perform various tasks. For the sake of
 * simplicity, the size of a bootfile_desc is fixed to 32 bytes, no
 * matter what type it is.
 *
 */


/* Sanity check OWL_CONFIG_BOOTFILE_ADDR.
 *
 * NOTE: Must be kept in sync with the fat[], ext and extcount provided
 *       to fixedfs_create()
 *
 *       The value of OWL_CONFIG_BOOTFILE_ADDR depends on the board since e.g
 *       on the SPB860 the wifi and mcu shares the exteral flash. The wifi
 *       firmware must live at addr [ 0, 262143 ] and the bootfile must be
 *       block-aligned.
 *
 *       The bootfile is assumed to always live in the BOOTFILE_DEV_EXTFLASH
 *       device.
 */


#define BOOTFILE_MAGIC 0xb00110ad

#define BOOTFILE_DEV_INTFLASH 0
#define BOOTFILE_DEV_EXTFLASH 1

struct bootfile_hdr {
        uint32_t magic; /* BOOTLOADER_MAGIC */
        uint32_t count;
        uint32_t adler32; /* checksum of bootfile_desc[] */
};

struct bootfile_desc_mtdcopy {
        uint8_t src_dev; /* source device (internal or external flash) */
        uint8_t dst_dev; /* destination device (internal or external flash) */
        uint8_t pad[2];
        uint32_t src_addr;     /* mtd address in source device */
        uint32_t dst_addr;     /* mtd address in destination device */
        uint32_t size;    /* number of bytes to copy */
        uint32_t adler32; /* checksum for the data chunk */
};

struct bootfile_desc {
#define BOOTFILE_DESC_MTDCOPY 0 /* mostly for future use */
        uint8_t type;
        uint8_t reserved1[3];

        union {
                struct bootfile_desc_mtdcopy mtdcopy;
                char data[28];
        }; /* sizeof: 28 */
}; /* sizeof: 32 */

struct bootfile {
        struct bootfile_hdr hdr;
        struct bootfile_desc desc[];
};


#if !OWL_CONFIG_HOST

#include <owl/util/mtd/mtd.h>

/**
 * Execute the descriptors. Validation of all descriptors will be perfomed
 * before any of them will be invoked.
 *
 *
 * @param intflash is the MTD device for internal flash, corresponding to
 *              BOOTFILE_DEV_INTFLASH.
 * @param extflash is the MTD device for external flash, corresponding to
 *              BOOTFILE_DEV_EXTFLASH. The bootfile is assumed to live here at
 *              offset OWL_CONFIG_BOOTFILE_ADDR.
 */
int bootfile_execute(struct mtd_dev *intflash, struct mtd_dev *extflash);


/**
 * Check if a bootfile exists.
 *
 * @param mtd is the mtd device for the external flash (where the bootfile
 *              is assumed to live).
 * @return 1 if the boofile exists, 0 if it does not exists and an error code
 *              otherwise.
 */
int bootfile_exists(struct mtd_dev *mtd);

/**
 * Erase the bootfile. The bootfile lives in the first eraseblock, so just
 * erase that one and we're done.
 *
 * @param mtd is the mtd device for the external flash (where the bootfile
 *              is assumed to live).
 */
int bootfile_erase(struct mtd_dev *mtd);






struct bootfile_ctx;


/**
 * Initialize bootfile desc for entries from fs files. The call to
 * bootfile_init() should be followed by a number of calls to
 * bootfile_update() and a final call to bootfile_finish() after all the
 * files have been added.
 *
 * @param ctx shoulis the bootfile context created through bootfile_init()
 *
 */
struct bootfile_ctx *bootfile_init(void);

/**
 * Add bootfile desc entries for a fs file to the bootfile.
 * The bootfile context will be freed upon failure.
 *
 * @param priv is the bootfile context created through bootfile_init()
 * @param filename is the file for which bootfile desc's should be added
 * @param dev is the destination device
 * @param addr is the destination address
 *
 *
 */
int bootfile_update(struct bootfile_ctx *priv,
                    const char *filename, int dev, uint32_t addr);

/**
 * Finish the bootfile write operation and free the bootfile context.
 *
 */
int bootfile_finish(struct bootfile_ctx *priv);

/**
 * Abort the bootfile write operation and do not write a proper bootfile hdr.
 *
 */
void bootfile_abort(struct bootfile_ctx *priv);


extern const struct owlsh_cmd cmd_bootfile;


#endif /* !OWL_CONFIG_HOST */

#endif /* BOOTFILE_H */
