/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 *
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#define OWL_TRACE OWL_TRACE_DBG
#define OWL_FILE "httpd"

#include <owl/core/owl_defs.h>
#include <owl/core/owl_list.h>
#include <owl/core/owl_timer.h>
#include <owl/core/owl_err.h>
#include <owl/core/owl_debug.h>

#include <pico/httpd/httpd.h>
#include <pico/httpd/httpd_404.h>

#include <wlp_api.h>
#include <wlp_socket.h>

struct httpd_document {
        httpd_cb_t *f;
        void *ctx;

        enum http_method method;
        const char *path;

        struct list_head list;
};

struct httpd_conn {
        httpd_cb_t *f;
        void *ctx;

        
        char reqbuf[1024];
        uint16_t reqlen; /* Offset into reqbuf */
        
        char rspbuf[OWL_CONFIG_HTTPD_RSPLEN];
        uint16_t offset; /* Offset into rspbuf */
        int seq;
        int left;

        int sock;
        struct httpd_ctx *httpd;
};

struct httpd_ctx {
        uint16_t port;
        int lsock;

        struct list_head documents;
};

static void listen_cb(void *ctx, int lsockid, int sockid,
                      const struct ip_addr *ip, uint16_t port);
static void conn_cb(void *ctx, int sockid,
                    const struct ip_addr *ip, uint16_t port, int connected);
static void sent_cb(void *ctx, int sockid, int len);
static void recv_cb(void *ctx, int sockid,
                    const struct ip_addr *ip, uint16_t port,
                    void *data, int len);
static int http_got_header(struct httpd_conn *conn);
static struct httpd_document *find_document(struct list_head *documents,
                                            char *req, int len);
static void close_conn(struct httpd_conn *conn);
static void send_data(struct httpd_conn *conn);
static void httpd_f(struct owlsh_ctx *sh, int argc, const char *argv[]);

struct httpd_ctx *
httpd_create(void)
{
        struct httpd_ctx *priv;

        owl_dbg("enter");

        if ((priv = owl_os_zalloc(sizeof(struct httpd_ctx))) == NULL)
                return NULL;

        list_init(&priv->documents);


        return priv;
}


int
httpd_start(struct httpd_ctx *priv, uint16_t port)
{
        int err;
        owl_dbg("port:%d", port);
        
        /* Create a listening socket for the HTTP server */
        if ((priv->lsock = wlp_socket(WLP_SOCK_STREAM, 0)) < 0) {
                owl_err("Failed to open TCP socket");
                return priv->lsock;
        }
        
        if ((err = wlp_bind(priv->lsock, IP_ADDR_ANY, port))) {
                owl_err("wlp_bind failed err: %d", err);
                return err;
        }
        
        if ((err = wlp_listen(priv->lsock, 1, listen_cb, priv))) {
                owl_err("wlp_listen failed err: %d", err);
                return err;
        }

        return 0;
}

void
httpd_stop(struct httpd_ctx *priv)
{
        wlp_close(priv->lsock);
}

struct httpd_document *
httpd_add_document(struct httpd_ctx *priv,
                   enum http_method method, const char *path,
                   httpd_cb_t *cb, void *ctx)
{
        struct httpd_document *doc;

        if ((doc = owl_os_zalloc(sizeof(*doc))) == NULL)
                return NULL;

        doc->method = method;
        doc->path = path;
        doc->f = cb;
        doc->ctx = ctx;

        owl_info("adding document:%p", doc);

        list_add_tail(&doc->list, &priv->documents);
        return doc;
}

static void
listen_cb(void *ctx, int lsockid, int sockid,
          const struct ip_addr *ip, uint16_t port)
{
        struct httpd_ctx *priv = ctx;
        struct httpd_conn *conn;

        /* Allocate memory for the structure that holds the state of the
           connection. */
        if ((conn = owl_os_zalloc(sizeof(*conn))) == NULL) {
                owl_err("out of mem");
                wlp_close(sockid);
                return;
        }
        owl_info("conn:%p", conn);

        conn->httpd = priv;
        conn->sock = sockid;
        
        wlp_set_recv_cb(sockid, recv_cb, conn);
        wlp_set_sent_cb(sockid, sent_cb, conn);
        wlp_set_conn_cb(sockid, conn_cb, conn);
                        

        wlp_accept(lsockid);
}

static void
conn_cb(void *ctx, int sockid,
        const struct ip_addr *ip, uint16_t port, int connected)
{
        struct httpd_conn *conn = ctx;

        owl_info("conn:%p sockid:%d connncted:%d", conn, sockid, connected);

        /* Socket already closed */

        /* XXX support this */
        
        owl_os_free(conn);
}

static void
sent_cb(void *ctx, int sockid, int len)
{
        struct httpd_conn *conn = ctx;

        owl_dbg("sockid:%d len:%d", sockid, len);
        owl_unused(len);

        if (conn->left) {

                send_data(conn);

        } else {
                int err;
                
                /* Request more data from the file */
                
                if (conn->f) {
                        owl_info("requesting more data seq:%d", conn->seq);
                        err = conn->f(conn->ctx, conn->seq, NULL, 0,
                                      conn->rspbuf, sizeof(conn->rspbuf));
                        
                        owl_info("got %d bytes from handler:%p",
                                 err, conn->f);

                        if (err > 0) {
                                
                                conn->left = err;
                                conn->offset = 0;
                                conn->seq++;
                                send_data(conn);

                        } else {
                                close_conn(conn);
                        }
                }
        }
}

/**
 * When data is received from the user agent, this function will be called.
 *
 * If the complete header has not yet been recieved, we will keep the pbuf and
 * wait for more data.
 *
 * As soon as this function is called with a buffer that contains the final
 * part of the header, we will parse the header and try to match the method
 * and path with our list of documents.
 *
 * If no document is found, the 404 handler will be used.
 *
 */
static void
recv_cb(void *ctx, int sockid, const struct ip_addr *ip, uint16_t port,
        void *data, int len)
{
        struct httpd_conn *conn = ctx;
        struct httpd_document *doc = NULL;
        int sndlen;
        
        if (data == NULL) {
                owl_info("data was NULL, closing conn:%p", conn);
                close_conn(conn);
                return;
        }
        
        owl_dbg("conn:%p sockid:%d len:%d", conn, sockid, len);
        
        /* Inform TCP that we have taken the data. */
        wlp_recved(sockid, len);

        if (len > sizeof(conn->reqbuf) - conn->reqlen)
                len = sizeof(conn->reqbuf) - conn->reqlen;
        
        owl_os_memcpy(conn->reqbuf + conn->reqlen, data, len);
        conn->reqlen += len;

        if (conn->f) {
                /* There is already a handler associated with this connection.
                 * The handler has already been invoked with the header,
                 * so invoke it again with more data.
                 *
                 * Set the seq to 0 and reqbuf to NULL in this call to
                 * indicate that this is more data coming from the original
                 * request.
                 */
                goto recv_data;
                
        } else if (!http_got_header(conn)) {

                /* Check if we have received a complete header */
                owl_info("header not complete, waiting for more data");
                return;
        }

        if ((doc = find_document(
                     &conn->httpd->documents, conn->reqbuf,
                     conn->reqlen)) == NULL) {
                
                /* 404 */
                owl_info("no document found, 404");
                conn->f = httpd_404;
                conn->ctx = NULL;

        } else {
                owl_dbg("set handler:%p", doc->f);
                conn->f = doc->f;
                conn->ctx = doc->ctx;
        }


        conn->seq = 0;
        
recv_data:
        owl_hexdump("HTTP RX", conn->reqbuf, conn->reqlen);

        /* This is the firest call to conn->f */
        sndlen = conn->f(conn->ctx, 0,
                         conn->reqbuf, conn->reqlen,
                         conn->rspbuf, sizeof(conn->rspbuf));
        
        /* A handler should wait for all expected data before sending a
           reply. */
        if (conn->left || conn->seq) {
                owl_err("handler received data during sending of reply, "
                        "conn->seq:%d conn->left:%d conn:%p",
                        conn->seq, conn->left, conn);
        }

        if (sndlen > 0) {
                owl_info("handler:%p replied with %d bytes", conn->f, sndlen);
                conn->left = sndlen;
                conn->seq++;
                send_data(conn);

        } else if (len == OWL_ERR_AGAIN) {
                owl_dbg("handler:%p is waiting for more data", conn->f);

        } else {
                /* error or done from handler */
                owl_info("handler:%p returned %d, close conn", conn->f, sndlen);
                close_conn(conn);
        }

}

/**
 * Return 1 if the connection have received a complete HTTP header.
 * We do some ugly string matching to find an empty line [CRLF] in the
 * received HTTP data.
 *
 */
static int
http_got_header(struct httpd_conn *conn)
{
        int offset;

        for (offset = 0; offset < conn->reqlen; offset++) {

                if (!owl_os_memcmp(
                            conn->reqbuf + offset, "\r\n\r\n", 4)) {
                        owl_info("found http header at offset:%d", offset);
                        return 1;
                }
        }

        owl_info("http header not complete");

        return 0;
}

/**
 * Find a document that matches the http request
 *
 */
static struct httpd_document *
find_document(struct list_head *documents, char *req, int len)
{
        struct list_head *pos = NULL;
        struct httpd_document *doc = NULL;

        owl_dbg("req:%p len:%d", req, len);

        list_for_each(pos, documents) {
                doc = list_entry(pos, struct httpd_document, list);

                owl_dbg("trying to match document with path:%s", doc->path);

                if (len < 5)  {
                        owl_info("header too short len:%d", len);
                        return NULL;
                }

                if (doc->method != HTTP_METHOD_GET) {
                        owl_info("doc has invalid method doc->method:%d",
                                 doc->method);
                        continue;
                }

                if (len < owl_os_strlen(doc->path) + 4) {
                        owl_info("header too short to fit document path "
                                 "len:%d doc->path:'%s'", len, doc->path);
                        continue;
                }

                if (!owl_os_strncmp(req, "GET ", 4) &&
                    !owl_os_strncmp(
                            doc->path, req + 4, owl_os_strlen(doc->path))) {
                        owl_dbg("found matching doc with path:%s", doc->path);
                        return doc;
                }

                if (!owl_os_strncmp(req, "GET / ", 6) &&
                    !owl_os_strcmp(doc->path, "/index.html")) {
                        owl_dbg("found default doc");
                        return doc;
                }

        }

        owl_info("no document found");
        return NULL;
}

static void
close_conn(struct httpd_conn *conn)
{
        owl_dbg("close conn:%p", conn);

        wlp_close(conn->sock);
        owl_os_free(conn);
}

static void
send_data(struct httpd_conn *conn)
{
        owl_dbg("left:%d", conn->left);

        while (conn->left) {
                int err;
                
                err = wlp_send(
                        conn->sock, conn->rspbuf + conn->offset, conn->left);

                if (err > 0) {
                        owl_dbg("wrote %d bytes", err);
                        owl_hexdump("HTTP TX", conn->rspbuf + conn->offset, err);
                        conn->offset += err;
                        conn->left -= err;
                } else {
                        owl_err("fail to write:%d", err);
                        return;
                }
        }
}

static void
httpd_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        struct httpd_ctx *httpd = owlsh_data(sh);
        int port = 80;
        int err;
        
        if (argc < 2) {
                owlsh_usage(sh, "usage: httpd <start|stop> [port]\r\n");
                return;
        }

        if (!owl_os_strcmp("start", argv[1])) {

                if (argc > 2)
                        port = owl_os_atoi(argv[2]);
                err = httpd_start(httpd, port);
                owlsh_perr(sh, err);
                
        } else if (!owl_os_strcmp("stop", argv[1])) {
                httpd_stop(httpd);
                owlsh_perr(sh, 0);
        }
}

const struct owlsh_cmd cmd_httpd = { "httpd", httpd_f, "httpd start/stop" };

