/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#ifndef OWL_HTTPD_H
#define OWL_HTTPD_H

#include <owl/core/owl_shell.h>

/**
 * HTTPd will parse incoming HTTP headers and pass it to the matching registered
 * handler, see httpd_add_handler().
 *
 * See the description of httpd_cb_t for more information on how the
 * handlers should operate.
 *
 */

enum http_method {
        HTTP_METHOD_GET = 0, /* GET */
        HTTP_METHOD_POST,    /* POST */
};



/**
 * HTTP callback function for handling a HTTP request.
 *
 * This function will be called multiple times, with or without additional data
 * from the HTTP request.
 *
 * If the handler fails in any way, an error code (other than OWL_ERR_AGAIN)
 * should be returned. The handler will not be called again for this request.
 *
 * A handler will be called the first time when a complete HTTP header has
 * been received and parsed by the HTTP server.
 *
 * There are two general types of handlers:
 *
 * 1) Handlers which expect HTTP content (i.e. something following the blank
 *    line (CRLF) after the HTTP header.
 * 2) Handlers which expects only the HTTP header and no content.
 *
 * For case 1), the handler need to inform the HTTP server if it requires
 * more data before any response can be sent. This is performed by returning
 * OWL_ERR_AGAIN. Once the complete HTTP request is received, the handler
 * can start filling the rspbuf and returning a length > 0. When 0 is returned,
 * the HTTP server assumes that the complete reply has been sent.
 *
 * For case 2), the handler should fill the rspbuf for each call, returning > 0
 * until done (return 0).
 *
 * Note that the HTTP header is only passed to the handler during the first
 * call, then the header will be free'd by the HTTP server. Also, the
 * raw data buffer (p) will only be provided during the first call (or when
 * additional data is provided for a handler that returned OWL_ERR_AGAIN).
 *
 * Also note that a handler must always make sure that the complete request is
 * obtained before sending any reply.
 *
 * @param ctx will be passed back as registered during httpd_add_cb()
 * @param seq increased by one for each call for each session, starts at zero
 * @param req is the pointer to the complete HTTP request
 * @param p is pbuf containing the raw request
 * @param rspbuf is a pointer to a memory location which should hold a (partial)
 *              reply.
 * @param rsplen is the size of the rsp buffer.
 * @param local is connection-specific local data. This is actual a void
 *              pointer-to-pointer which the document can use to store local
 *              state across multiple calls. Compared to static vars, the local
 *              parameters will ensure that each connection has its own state,
 *              even though the same document is accessed by multiple
 *              connections.
 *
 * @return the number of bytes written to the rsp buffer.
 */
typedef int (httpd_cb_t)(void *ctx, int seq, const char *reqbuf, uint16_t reqlen,
                         char *rspbuf, uint16_t rsplen);

struct httpd_document;

/**
 * Allocate memory for the HTTP server.
 *
 */
struct httpd_ctx *httpd_create(void);


/**
 * Start listening for incoming connections on the port set during
 * httpd_create().
 *
 *
 */ 
int httpd_start(struct httpd_ctx *priv, uint16_t port);


/**
 * Stop listening for incoming connections
 *
 */
void httpd_stop(struct httpd_ctx *priv);



struct httpd_document *
httpd_add_document(struct httpd_ctx *priv,
                      enum http_method method, const char *path,
                      httpd_cb_t *cb, void *ctx);

extern const struct owlsh_cmd cmd_httpd;

#endif
