/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */


/* wlp_ipaddr.h implementation */

#define OWL_TRACE OWL_TRACE_ERR
#define OWL_FILE "pico/lwip/ipaddr"

#include <owl/core/owl_defs.h>
#include <owl/core/owl_debug.h>
#include <owl/core/owl_err.h>
#include <owl/core/owl_timer.h>
#include <owl/core/owl_lwip.h>

#include <wlp_api.h>
#include <wlp_ipaddr.h>
#include <wlp_concurrent.h>

void
wlp_set_ipaddr_cb(void (*addr_cb)(void *ctx, const struct ip_addr *addr),
                  void *ctx)
{
        int id = wlp_get_iface();
        struct iface_ctx *iface = owlip_get_iface(id);

        if (iface == NULL)
                return;
        
        iface->addr_cb = addr_cb;
        iface->addr_ctx = ctx;
}

int
wlp_set_ipaddr(const struct ip_addr *ip, const struct ip_addr *netmask,
               const struct ip_addr *gw, const struct ip_addr *dns)
{
        int id = wlp_get_iface();
        struct iface_ctx *iface = owlip_get_iface(id);
        struct netif *netif;

        if (iface == NULL)
                return WLP_ERR_STATE;

        netif = &iface->netif;
        
        wlp_set_dhcp(0);

        if (iface->dhcp_state != DHCP_STATE_IDLE) {
                owl_info("setting static ip, release dhcp");
                dhcp_release(netif);
                dhcp_stop(netif);
                iface->dhcp_state = DHCP_STATE_IDLE;
                owl_bitclr(iface->flags, FLAG_DHCP);
        }

        if (ip && !ip_addr_isany(ip))
                netif_set_ipaddr(netif, (struct ip_addr *) ip);
        if (netmask && !ip_addr_isany(netmask))
                netif_set_netmask(netif, (struct ip_addr *) netmask);
        if (gw && !ip_addr_isany(gw))
                netif_set_gw(netif, (struct ip_addr *) gw);
#if LWIP_DNS == 1
        if (dns && !ip_addr_isany(dns))
                dns_setserver(0, (struct ip_addr *) dns);
#endif

        if (netif_is_link_up(netif))
                netif_set_up(netif);

        return 0;
}

int
wlp_get_ipaddr(struct ip_addr *ip, struct ip_addr *netmask,
               struct ip_addr *gw, struct ip_addr *dns)
{
        int id = wlp_get_iface();
        struct iface_ctx *iface = owlip_get_iface(id);
        struct netif *netif;

        if (iface == NULL)
                return WLP_ERR_STATE;

        netif = &iface->netif;
        
        *ip = netif->ip_addr;
        *netmask = netif->netmask;
        *gw = netif->gw;
        *dns = dns_getserver(0);

        if (netif_is_up(netif))
                return 0;
        else
                return WLP_ERR_CONN;
}

int
wlp_set_dhcp(int enable)
{
        int id = wlp_get_iface();
        struct iface_ctx *iface = owlip_get_iface(id);
        struct netif *netif;

        owl_info("enable:%d", enable);

        if (iface == NULL)
                return WLP_ERR_STATE;

        netif = &iface->netif;

        if (enable)
                owl_bitset(iface->flags, FLAG_DHCP);
        else
                owl_bitclr(iface->flags, FLAG_DHCP);

        if (netif_is_link_up(netif)) {

                if (!owl_bit(iface->flags, FLAG_DHCP) &&
                    iface->dhcp_state != DHCP_STATE_IDLE) {
                        owl_info("link is up and dhcp disabled, relea dhcp");
                        dhcp_release(netif);
                        dhcp_stop(netif);
                        iface->dhcp_state = DHCP_STATE_IDLE;
                }

                else if (owl_bit(iface->flags, FLAG_DHCP) &&
                         iface->dhcp_state == DHCP_STATE_IDLE) {
                        owl_info("link is up and dhcp enabled, request dhcp");
                        netif_set_down(netif); /* static ip no longer valid */
                        dhcp_start(netif);
                        iface->dhcp_state = DHCP_STATE_RUN;
                }
                
        } else {

                /*
                 * We need to set the interface to DOWN here if dhcp is enabled
                 * while the link is down since we don't really have an IP
                 * address anymore. The scenario below would fail if
                 * netif_set_down() is omitted below.
                 *
                 *   $ dhcp on
                 *   ok 0
                 *   $ linkup angr foobar123
                 *   ok 0
                 *   $ link_cb link:1
                 *   addr_cb ip:192.168.2.114
                 *   $ dhcp off
                 *   ip status down
                 *   $ ifconfig 1.2.3.4
                 *   bound to 1.2.3.4
                 *   ok 0
                 *   $ linkdown
                 *   ok 0
                 *   $ link_cb link:0
                 *   $ dhcp on
                 *   ip status down             <--- netif_set_down()
                 *   ok 0
                 *   $ linkup angr foobar123
                 *   ok 0
                 *   $ link_cb link:1
                 *   bound to 192.168.2.114     <--- netif_set_up()
                 */

                /* any static ip no longer valid */
                if (owl_bit(iface->flags, FLAG_DHCP))
                        netif_set_down(netif); 
        }

        return 0;
}

int
wlp_get_dhcp(int *enabled)
{
        int id = wlp_get_iface();
        struct iface_ctx *iface = owlip_get_iface(id);

        owl_info("enter");

        if (iface == NULL)
                return WLP_ERR_STATE;

        if (owl_bit(iface->flags, FLAG_DHCP))
                *enabled = 1;
        else
                *enabled = 0;

        return 0;
}
