/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 *
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#define OWL_TRACE OWL_TRACE_DATA
#define OWL_FILE "pico/shell/ipaddr"

#include <owl/core/owl_defs.h>
#include <owl/core/owl_shell.h>
#include <owl/core/owl_err.h>
#include <owl/core/owl_debug.h>

#include <wlp_ipaddr.h>
#include <wlp_inet.h>

#include <pico/wlp_shell.h>

/*
 * Evaluate the owl pico API (wlp_ipaddr) using the wlp_shell.
 *
 * The functions provided by the wlp_api can be invoked from a command line,
 * using a one-to-one mapping between commands and api function wherever
 * possible.
 *
 *
 *
 *
 */

static void set_ipaddr_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void get_ipaddr_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void set_dhcp_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void get_dhcp_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static int parse_int(const char *str, int *out);

const struct owlsh_cmd cmd_set_ipaddr = {
        "set_ipaddr", set_ipaddr_f, "wlp_set_ipaddr(ip, mask, gw, dns)",
};
const struct owlsh_cmd cmd_get_ipaddr = {
        "get_ipaddr", get_ipaddr_f, "wlp_get_ipaddr()"
};
const struct owlsh_cmd cmd_set_dhcp = {
        "set_dhcp", set_dhcp_f, "wlp_set_dhcp(enable)"
};
const struct owlsh_cmd cmd_get_dhcp = {
        "get_dhcp", get_dhcp_f, "wlp_get_dhcp()"
};


static void
set_ipaddr_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        struct ip_addr ip, mask, gw, dns;

        mask = ip_addr_any;
        gw = ip_addr_any;
        dns = ip_addr_any;

        if (argc < 2)
                goto err_parse;
        if (!ip_aton(argv[1], &ip))
                goto err_parse;
        if (argc >= 3 && !ip_aton(argv[2], &mask))
                goto err_parse;
        if (argc >= 4 && !ip_aton(argv[3], &gw))
                goto err_parse;
        if (argc >= 5 && !ip_aton(argv[4], &dns))
                goto err_parse;

        res = wlp_set_ipaddr(&ip, &mask, &gw, &dns);
        owlsh_perr(sh, res);
        return;

err_parse:
        owlsh_printf(sh, "usage: set_ipaddr <ip> [mask] [gw] [dns]\r\n");
}

static void
get_ipaddr_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        struct ip_addr ip, mask, gw, dns;

        owl_unused(argc);
        owl_unused(argv);
        
        res = wlp_get_ipaddr(&ip, &mask, &gw, &dns);
        owlsh_printf(sh, "%s %d (", res >= 0 ? "ok" : "err", res);
        owlsh_printf(sh, "ip:%s ", ip_ntoa(&ip));
        owlsh_printf(sh, "mask:%s ", ip_ntoa(&mask));
        owlsh_printf(sh, "gw:%s ", ip_ntoa(&gw));
        owlsh_printf(sh, "dns:%s)\r\n", ip_ntoa(&dns));
}

static void
set_dhcp_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        int dhcp;

        if (argc < 2)
                goto err_parse;
        if (!parse_int(argv[1], &dhcp))
                goto err_parse;

        res = wlp_set_dhcp(dhcp);
        owlsh_perr(sh, res);
        return;

err_parse:
        owlsh_printf(sh, "usage: set_dhcp <0|1>\r\n");
}

static void
get_dhcp_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        int enabled;

        owl_unused(argc);
        owl_unused(argv);

        res = wlp_get_dhcp(&enabled);
        owlsh_printf(sh, "%s %d (enabled:%d)\r\n",
                     res >= 0 ? "ok" : "err", res, enabled);
}

static int
parse_int(const char *str, int *out)
{
        char *end;
        *out = owl_os_strtol(str, &end, 0);
        if (end == str)
                return 0;

        return 1;
}
