/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 *
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#define OWL_TRACE OWL_TRACE_DATA
#define OWL_FILE "pico/shell/link"

#include <owl/core/owl_defs.h>
#include <owl/core/owl_shell.h>
#include <owl/core/owl_err.h>
#include <owl/core/owl_debug.h>

#include <wlp_api.h>

#include <pico/wlp_shell.h>

/*
 * Evaluate the owl pico API (wlp_api) using the wlp_shell.
 *
 * The functions provided by the wlp_api can be invoked from a command line,
 * using a one-to-one mapping between commands and api function wherever
 * possible.
 *
 *
 *
 *
 */

static void scan_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void linkup_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void linkdown_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void set_linkctx_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void get_linkctx_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void set_wps_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void set_mode_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void get_mode_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void set_ps_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void conf_ps_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void conf_ap_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void set_channel_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void get_hwaddr_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void get_network_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void get_rate_f(struct owlsh_ctx *sh, int argc, const char *argv[]);

static int parse_ssid(const char *str, struct wlp_ssid *ssid);
static int parse_int(const char *str, int *out);
static int32_t parse_int32(const char *str, int32_t *out);
static const char *rate_str(uint8_t rate);

/* Set in linkctx_cb() */
static int linkctx_len = 0;

const struct owlsh_cmd cmd_linkup = {
        "linkup", linkup_f, "wlp_linkup(ssid, key, auth)"
};
const struct owlsh_cmd cmd_linkdown = {
        "linkdown", linkdown_f, "wlp_linkdown()"
};
const struct owlsh_cmd cmd_set_linkctx = {
        "set_linkctx", set_linkctx_f, "wlp_set_link_context()"
};
const struct owlsh_cmd cmd_get_linkctx = {
        "get_linkctx", get_linkctx_f, "wlp_get_link_context()"
};
const struct owlsh_cmd cmd_set_wps = {
        "set_wps", set_wps_f, "wlp_set_wps()"
};
const struct owlsh_cmd cmd_set_mode = {
        "set_mode", set_mode_f, "wlp_set_mode()"
};
const struct owlsh_cmd cmd_get_mode = {
        "get_mode", get_mode_f, "wlp_get_mode()"
};
const struct owlsh_cmd cmd_set_ps = {
        "set_ps", set_ps_f, "wlp_set_ps()"
};
const struct owlsh_cmd cmd_conf_ps = {
        "conf_ps", conf_ps_f, "wlp_conf_ps()"
};
const struct owlsh_cmd cmd_conf_ap = {
        "conf_ap", conf_ap_f, "wlp_conf_ap()"
};
const struct owlsh_cmd cmd_set_channel = {
        "set_channel", set_channel_f, "wlp_set_channel()"
};
const struct owlsh_cmd cmd_get_hwaddr = {
        "get_hwaddr", get_hwaddr_f, "wlp_get_hwaddr()"
};
const struct owlsh_cmd cmd_get_network = {
        "get_network", get_network_f, "wlp_get_network()"
};
const struct owlsh_cmd cmd_get_rate = {
        "get_rate", get_rate_f, "wlp_get_rate()"
};
const struct owlsh_cmd cmd_scan = {
        "scan", scan_f, "wlp_scan(ssid)"
};

static void
linkctx_cb(void *ctx, const void *linkctx, int len)
{
        owl_info("got linkctx:%p len:%d", linkctx, len);
        owl_hexdump("linkctx", linkctx, len);
        linkctx_len = len;
        owl_os_memcpy(ctx, linkctx, len);
}

static void
linkup_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        struct wlp_ssid ssid;
        const char *auth;
        owl_os_memset(&ssid, 0, sizeof(ssid));
        if (argc >= 4) {
                /* linkup <ssid> <key> <auth> */
                enum wlp_auth_mode auth_mode;

                if (!parse_ssid(argv[1], &ssid))
                        goto err_parse;
                auth = argv[3];
                if (owl_os_strcmp(auth, "wpa2") == 0)
                        auth_mode = WLP_AUTH_WPA2_PSK;
                else if (owl_os_strcmp(auth, "wpa") == 0)
                        auth_mode = WLP_AUTH_WPA_PSK;
                else if (owl_os_strcmp(auth, "wep") == 0)
                        auth_mode = WLP_AUTH_WEP;
                else
                        goto err_parse;
                res = wlp_linkup(&ssid, argv[2], auth_mode);
        } else if (argc >= 3) {
                /* linkup <ssid> <key> */
                if (!parse_ssid(argv[1], &ssid))
                        goto err_parse;
                res = wlp_linkup(&ssid, argv[2], WLP_AUTH_AUTO);
        } else if (argc >= 2) {
                /* linkup <ssid> */
                if (!parse_ssid(argv[1], &ssid))
                        goto err_parse;

                res = wlp_linkup(&ssid, NULL, WLP_AUTH_AUTO);
        } else if (argc >= 1) {
                /* linkup */
                res = wlp_linkup(NULL, NULL, WLP_AUTH_AUTO);
        } else {
                goto err_parse;
        }

        owlsh_perr(sh, res);
        return;

err_parse:
        owlsh_printf(sh, "usage: linkup [ssid] [key] [wep|wpa|wpa2]\r\n");
}

static void
linkdown_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;

        (void)argc;
        (void)argv;
        res = wlp_linkdown();
        owlsh_perr(sh, res);
}

static void
set_linkctx_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        void *ctx = owlsh_data(sh);
        
        (void)argc;
        (void)argv;
        res = wlp_set_link_context(ctx, linkctx_len);
        owlsh_perr(sh, res);
}

static void
get_linkctx_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        void *ctx = owlsh_data(sh);
        
        (void)argc;
        (void)argv;
        res = wlp_get_link_context(linkctx_cb, ctx);
        owlsh_perr(sh, res);
}

static void
set_wps_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        int enable;

        if (argc < 2)
                goto err_parse;
        if (!parse_int(argv[1], &enable))
                goto err_parse;

        res = wlp_set_wps(enable);
        owlsh_perr(sh, res);
        return;

err_parse:
        owlsh_printf(sh, "usage: set_wps <0|1>\r\n");
}

static void
set_mode_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        uint8_t mode;

        if (argc < 2)
                goto err_parse;
        if (!owl_os_strcmp(argv[1], "sta"))
                mode = WLP_MODE_STA;
        else if (!owl_os_strcmp(argv[1], "ap"))
                mode = WLP_MODE_AP;
        else if (!owl_os_strcmp(argv[1], "adhoc"))
                mode = WLP_MODE_ADHOC;
        else if (!owl_os_strcmp(argv[1], "rf"))
                mode = WLP_MODE_RFTEST;
        else
                goto err_parse;

        res = wlp_set_mode(mode);
        owlsh_perr(sh, res);
        return;

err_parse:
        owlsh_printf(sh, "usage: set_mode <sta|ap|adhoc|rf|wps>\r\n");

}


static void
get_mode_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        uint8_t mode;
        const char *modestr = NULL;

        (void)argc;
        (void)argv;

        res = wlp_get_mode(&mode); /* always success */

        switch (mode) {
        case WLP_MODE_STA: modestr = "sta"; break;
        case WLP_MODE_AP: modestr = "ap"; break;
        case WLP_MODE_ADHOC: modestr = "adhoc"; break;
        case WLP_MODE_RFTEST: modestr = "rf"; break;        
        default: break;
        };

        owlsh_printf(sh, "%s %d (%s)\r\n",
                     res >= 0 ? "ok" : "err", res,
                     res >= 0 ? modestr : "N/A");
}


static void
set_channel_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        int channel;

        if (argc < 2)
                goto err_parse;
        if (!parse_int(argv[1], &channel))
                goto err_parse;

        res = wlp_set_channel(channel, NULL, NULL);
        owlsh_perr(sh, res);
        return;

err_parse:
        owlsh_printf(sh, "usage: set_channel <channel>\r\n");

}

static void
set_ps_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        int enable;

        if (argc < 2)
                goto err_parse;
        if (!parse_int(argv[1], &enable))
                goto err_parse;
        res = wlp_set_ps(enable);

        owlsh_perr(sh, res);
        return;

err_parse:
        owlsh_printf(sh, "usage: set_ps <0|1>\r\n");

}

static void
conf_ps_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        int32_t ps_traffic_timeout, ps_delay, listen_interval;
        int use_ps_poll, rx_all_dtim;

        if (argc != 6)
                goto err_parse;

        if (!parse_int(argv[1], &use_ps_poll))
                goto err_parse;
        if (!parse_int32(argv[2], &ps_traffic_timeout))
                goto err_parse;
        if (!parse_int32(argv[3], &ps_delay))
                goto err_parse;
        if (!parse_int(argv[4], &rx_all_dtim))
                goto err_parse;
        if (!parse_int32(argv[5], &listen_interval))
                goto err_parse;

        res = wlp_conf_ps((uint8_t)use_ps_poll,
                          (uint32_t)ps_traffic_timeout,
                          (uint32_t)ps_delay,
                          (uint8_t)rx_all_dtim,
                          (uint32_t)listen_interval);

        owlsh_perr(sh, res);
        return;

err_parse:
        owlsh_printf(sh, "usage: conf_ps <use_ps_poll> <ps_traffic_timeout>\r\n");
        owlsh_printf(sh, "               <ps_delay> <rx_all_dtim> <listen_interval>\r\n");

}

static void
conf_ap_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;

        int hidden;
        int beacon_intvl;
        int inact_tmo;
        const char *country = NULL;

        if (argc < 4)
                goto err_parse;

        if (!parse_int(argv[1], &hidden))
                goto err_parse;
        if (!parse_int(argv[2], &beacon_intvl))
                goto err_parse;
        if (!parse_int(argv[3], &inact_tmo))
                goto err_parse;
        if (argc > 4)
                country = argv[4];

        res = wlp_conf_ap(hidden, beacon_intvl, inact_tmo, country);

        owlsh_perr(sh, res);
        return;

err_parse:
        owlsh_printf(sh, "usage: conf_ap <hidden> <beacon_intvl>\r\n");
        owlsh_printf(sh, "               <inact_tmo> [country]>\r\n");

}

static void
get_hwaddr_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        struct wlp_mac_addr hwaddr;

        (void)argc;
        (void)argv;

        res = wlp_get_hwaddr(&hwaddr);

        /* out vars always valid */
        owlsh_printf(sh, "%s %d (%02x:%02x:%02x:%02x:%02x:%02x)\r\n",
                     res >= 0 ? "ok" : "err", res,
                     hwaddr.octet[0], hwaddr.octet[1], hwaddr.octet[2],
                     hwaddr.octet[3], hwaddr.octet[4], hwaddr.octet[5]);
}

static void
get_network_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        struct wlp_network net;
        char ssid_str[WLP_SSID_MAX_LENGTH + 1];

        (void)argc;
        (void)argv;

        res = wlp_get_network(&net);

        owlsh_printf(sh, "%s %d", res >= 0 ? "ok" : "err", res);

        /* out vars always invalid if return < 0 */
        if (res) {
                owlsh_printf(sh, "\r\n");
                return;
        }
        
        owl_os_memcpy(ssid_str, net.ssid.ssid, net.ssid.len);
        ssid_str[net.ssid.len] = 0;

        owlsh_printf(sh,
                     " (%s %02x:%02x:%02x:%02x:%02x:%02x ",
                     net.ssid.len ? ssid_str : "N/A",
                     net.bssid.octet[0], net.bssid.octet[1],
                     net.bssid.octet[2], net.bssid.octet[3],
                     net.bssid.octet[4], net.bssid.octet[5]);

        if (net.rssi != WLP_RSSI_UNKNOWN)
                owlsh_printf(sh, "rssi:%d ", net.rssi);
        else
                owlsh_printf(sh, "rssi:N/A ");

        if (net.snr != WLP_SNR_UNKNOWN)
                owlsh_printf(sh, "snr:%d", net.snr);
        else
                owlsh_printf(sh, "snr:N/A");

        owlsh_printf(sh, ")\r\n");
}

static void
get_rate_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        uint8_t txrate = WLP_RATE_UNKNOWN;
        uint8_t rxrate = WLP_RATE_UNKNOWN;
        
        (void)argc;
        (void)argv;

        res = wlp_get_rate(&txrate, &rxrate);
        
        owlsh_printf(sh, "%s %d (", res >= 0 ? "ok" : "err", res);
        owlsh_printf(sh, "tx:%s ", rate_str(txrate));
        owlsh_printf(sh, "rx:%s", rate_str(rxrate));
        owlsh_printf(sh, ")\r\n");
}

static void
scan_cb(void *ctx, struct wlp_network *net)
{
	struct owlsh_ctx *sh = ctx;

        if (net == NULL) {
                owlsh_perr(sh, 0);
                return;
        }
        
        if (owlsh_compact(sh)) {

#if OWL_CONFIG_SHELL_OUTPUT_COMPACT
                char ssid_buf[WLP_SSID_MAX_LENGTH + 1];
                const char *auth = "";
                
                /* <ssid len>;<ssid>;<hwaddr>;
                   <none|wep|wpa|wpa-psk|wpa2|wpa2-psk>;<rssi> */

                        
                owl_os_memset(ssid_buf, 0, sizeof(ssid_buf));
                owl_os_memcpy(ssid_buf, net->ssid.ssid, net->ssid.len);

                switch (net->auth) {
                case WLP_AUTH_AUTO:
                        auth = "none";
                        break;
                case WLP_AUTH_WEP:
                        auth = "wep";
                        break;
                case WLP_AUTH_WPA_PSK:
                        auth = "wpa-psk";
                        break;
                case WLP_AUTH_WPA2_PSK:
                        auth = "wpa2-psk";
                        break;
                case WLP_AUTH_WPA:
                        auth = "wpa";
                        break;
                case WLP_AUTH_WPA2:
                        auth = "wpa2";
                        break;
                default:
                        owl_err("invalid auth:%d", auth);
                        owl_assert(0);
                }

                owlsh_printf(
                        sh,
                        "%d;%s;%02x:%02x:%02x:%02x:%02x:%02x;%s;%d\r\n",
                        net->ssid.len, ssid_buf,
                        net->bssid.octet[0], net->bssid.octet[1],
                        net->bssid.octet[2], net->bssid.octet[3],
                        net->bssid.octet[4], net->bssid.octet[5],
                        auth, net->rssi);
                        
#endif /* OWL_CONFIG_SHELL_OUTPUT_COMPACT */
        } else  {
                
#if OWL_CONFIG_SHELL_OUTPUT_FULL
                char buf[WLP_SSID_MAX_LENGTH + 1];

                /* bssid */
                owlsh_printf(sh, "%02x:%02x:%02x:%02x:%02x:%02x  ",
                             net->bssid.octet[0], net->bssid.octet[1],
                             net->bssid.octet[2], net->bssid.octet[3],
                             net->bssid.octet[4], net->bssid.octet[5]);

                /* type, channel, rssi, snr */
                owlsh_printf(sh, "%s  %02d %02d  ",
                             owl_bit(net->flags, WLP_NET_FLAG_ADHOC) ?
                             "ADHOC" : "INFRA",
                             net->channel, (int) net->rssi);

                /* snr */
                if (net->snr != WLP_SNR_UNKNOWN)
                        owlsh_printf(sh, "%02d   ", net->snr);
                else
                        owlsh_printf(sh, "N/A  ");

                /* auth */
                switch (net->auth) {
                case WLP_AUTH_AUTO:
                        owlsh_printf(sh, "NONE      ");
                        break;
                case WLP_AUTH_WEP:
                        owlsh_printf(sh, "WEP       ");
                        break;
                case WLP_AUTH_WPA_PSK:
                        owlsh_printf(sh, "WPA-PSK   ");
                        break;
                case WLP_AUTH_WPA2_PSK:
                        owlsh_printf(sh, "WPA2-PSK  ");
                        break;
                case WLP_AUTH_WPA:
                        owlsh_printf(sh, "WPA       ");
                        break;
                case WLP_AUTH_WPA2:
                        owlsh_printf(sh, "WPA2      ");
                        break;
                }

                /* ssid */
                owl_os_memset(buf, 0, sizeof buf);
                owl_os_memcpy(buf, net->ssid.ssid, net->ssid.len);
                owlsh_printf(sh, "%s", buf);
                owlsh_printf(sh, "\r\n");
#endif /* OWL_CONFIG_SHELL_OUTPUT_FULL */
        }
}

static void
scan_f(struct owlsh_ctx *sh, int argc,  const char *argv[])
{
        int res;
        struct wlp_ssid ssid;

        if (argc > 1) {
		if (!parse_ssid(argv[1], &ssid))
			goto err_parse;
		res = wlp_scan(&ssid, scan_cb, sh);
        } else {
		res = wlp_scan(NULL, scan_cb, sh);
        }

        if (res < 0)
                owlsh_perr(sh, res);

#if OWL_CONFIG_SHELL_OUTPUT_FULL
        if (!owlsh_compact(sh)) {
                owlsh_printf(sh,
                             "BSSID              TYPE   CH RSSI SNR  "
                             "AUTH      SSID\r\n");
                owlsh_printf(sh, "-----------------------------------"
                             "----------------------------------\r\n");
        }
#endif
	return;

err_parse:
        owlsh_printf(sh, "usage: scan [ssid]\r");
}

static int
parse_int(const char *str, int *out)
{
        char *end;
        *out = owl_os_strtol(str, &end, 0);
        if (end == str)
                return 0;

        return 1;
}

static int32_t
parse_int32(const char *str, int32_t *out)
{
        char *end;
        *out = owl_os_strtol(str, &end, 0);
        if (end == str)
                return 0;

        return 1;
}

static int
parse_ssid(const char *str, struct wlp_ssid *ssid)
{
        ssid->len = owl_os_strlen(str);
        owl_os_memcpy(ssid->ssid, str, ssid->len);
        return 1;
}

static const char *
rate_str(uint8_t rate)
{
        switch (rate) {
        case WLP_RATE_1MBIT:    return "1";
        case WLP_RATE_2MBIT:    return "2";
        case WLP_RATE_5_5MBIT:  return "5.5";
        case WLP_RATE_11MBIT:   return "11";
        case WLP_RATE_6MBIT:    return "6";
        case WLP_RATE_9MBIT:    return "9";
        case WLP_RATE_12MBIT:   return "12";
        case WLP_RATE_18MBIT:   return "18";
        case WLP_RATE_24MBIT:   return "24";
        case WLP_RATE_36MBIT:   return "36";
        case WLP_RATE_48MBIT:   return "48";
        case WLP_RATE_54MBIT:   return "54";
        case WLP_RATE_6_5MBIT:  return "6.5";
        case WLP_RATE_13MBIT:   return "13";
        case WLP_RATE_19_5MBIT: return "19.5";
        case WLP_RATE_26MBIT:   return "26";
        case WLP_RATE_39MBIT:   return "39";
        case WLP_RATE_52MBIT:   return "52";
        case WLP_RATE_58_5MBIT: return "58.5";
        case WLP_RATE_65MBIT:   return "65";
        default: return "N/A";
        }
}
