/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 *
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#define OWL_TRACE OWL_TRACE_DATA
#define OWL_FILE "pico/shell/mcast"

#include <owl/core/owl_defs.h>
#include <owl/core/owl_shell.h>
#include <owl/core/owl_err.h>
#include <owl/core/owl_debug.h>

#include <wlp_api.h>
#include <wlp_mcast.h>

#include <pico/wlp_shell.h>

/*
 * Evaluate the owl pico API (wlp_api) using the wlp_shell.
 *
 * The functions provided by the wlp_api can be invoked from a command line,
 * using a one-to-one mapping between commands and api function wherever
 * possible.
 *
 *
 *
 *
 */

static void add_mcast_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void del_mcast_f(struct owlsh_ctx *sh, int argc, const char *argv[]);

static int parse_hexstr(const char *str, uint8_t *data, int len);

const struct owlsh_cmd cmd_add_mcast = {
        "add_mcast", add_mcast_f, "wlp_add_mcast()"
};

const struct owlsh_cmd cmd_del_mcast = {
        "del_mcast", del_mcast_f, "wlp_del_mcast()"
};

static void
add_mcast_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        struct wlp_mac_addr hwaddr;
        int res;
        
        if (argc < 2)
                goto err_parse;

        if (!owl_os_strcmp(argv[1], "all")) {
                
                /* All multicast will be received */
                res = wlp_add_mcast(NULL);
                
        } else {
                
                if ((res = parse_hexstr(
                             argv[1], hwaddr.octet,
                             sizeof(hwaddr.octet))) < 0) {
                        goto err_parse;
                }
                
                /* Only multicast from the given address will be receivd.
                   Can be called multiple times to add more addresses. */
                res = wlp_add_mcast(&hwaddr);
        }

        owlsh_perr(sh, res);
        return;
        
err_parse:
        owlsh_printf(sh, "usage: add_mcast <all|hwaddr>\r\n");
}

static void
del_mcast_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        struct wlp_mac_addr hwaddr;
        int res;

        if (argc < 2)
                goto err_parse;

        if (!owl_os_strcmp(argv[1], "all")) {

                /* No multicast will be received (delete all addresses) */
                res = wlp_del_mcast(NULL);
                
        } else {
                
                if ((res = parse_hexstr(
                             argv[1], hwaddr.octet,
                             sizeof(hwaddr.octet))) < 0) {
                        goto err_parse;
                }

                /* Remove an address that has previously been added with
                   wlp_add_mcast() */
                res = wlp_del_mcast(&hwaddr);
        }

        owlsh_perr(sh, res);
        return;
        
err_parse:
        owlsh_printf(sh, "usage: del_mcast <all|hwaddr>\r\n");
}

static int
parse_hexstr(const char *str, uint8_t *data, int len)
{

		
	int i;
	char buf[3];
	int strlen = owl_os_strlen(str);
	
	if (strlen % 2)
		return OWL_ERR_PARAM;

        /* Fixed-size hexstr */
	if (strlen / 2 != len)
		return OWL_ERR_SIZE;
	
	buf[2] = 0;
	for (i = 0; i < strlen; i += 2) {
		if (!owl_isxdigit(str[i]) ||
		    !owl_isxdigit(str[i + 1])) {
			return OWL_ERR_PARAM;
		}
                        
		buf[0] = str[i];
		buf[1] = str[i + 1];
		data[i / 2] = owl_os_strtol(buf, NULL, 16);
	}

	return strlen / 2;
}
