/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 *
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#define OWL_TRACE OWL_TRACE_ERR
#define OWL_FILE "rftest"

#include <owl/core/owl_defs.h>
#include <owl/core/owl_shell.h>
#include <owl/core/owl_err.h>
#include <owl/core/owl_debug.h>
#include <owl/core/owl_timer.h>

#include <pico/wlp_shell.h>

#include <wlp_rftest.h>

static void start_tx_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void stop_tx_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void get_rxstats_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void clr_rxstats_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void rxstats_cb(void *ctx, struct wlp_rfstats *rxstats);

static uint8_t parse_rate(const char *rate);

const struct owlsh_cmd cmd_start_tx = {
        "start_tx", start_tx_f, "wlp_start_tx()"
};

const struct owlsh_cmd cmd_stop_tx = {
        "stop_tx", stop_tx_f, "wlp_stop_tx()"
};

const struct owlsh_cmd cmd_get_rxstats = {
        "get_rxstats", get_rxstats_f, "wlp_get_rxstats()"
};

const struct owlsh_cmd cmd_clr_rxstats = {
        "clr_rxstats", clr_rxstats_f, "wlp_clr_rxstats()"
};

static void
start_tx_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        uint32_t count;
        uint16_t interval;
        uint8_t rate;
        uint16_t size;

        if (argc < 5) 
                goto err_parse;

        count = owl_os_atoi(argv[1]);
        interval = owl_os_atoi(argv[2]);
        size = owl_os_atoi(argv[3]);
        rate = parse_rate(argv[4]);

        res = wlp_start_tx(count, interval, size, rate);
        owlsh_perr(sh, res);
        return;
        
err_parse:
        owlsh_printf(sh, "usage: start_tx <count> <interval> <size> <rate>\r\n");
}

static void
stop_tx_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;

        res = wlp_stop_tx();
        owlsh_perr(sh, res);
}

static void
get_rxstats_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        uint8_t rate;

        if (argc < 2) 
                goto err_parse;

        rate = parse_rate(argv[1]);
        
        res = wlp_get_rxstats(rate, rxstats_cb, sh);
        owlsh_perr(sh, res);
        return;

err_parse:
        owlsh_printf(sh, "usage: get_rxstats <rate>\r\n");
}

static void
clr_rxstats_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        res = wlp_clr_rxstats();
        owlsh_perr(sh, res);
}

static void
rxstats_cb(void *ctx, struct wlp_rfstats *rxstats)
{
        struct owlsh_ctx *sh = ctx;
        
        owlsh_printf(
                sh, " num_frames:         %lu\r\n",
                (unsigned long) rxstats->num_frames);
        owlsh_printf(
                sh, " num_mcast:          %lu\r\n",
                (unsigned long) rxstats->num_mcast);
        owlsh_printf(
                sh, " num_ucast_to_us:    %lu\r\n",
                (unsigned long) rxstats->num_ucast_to_us);
        owlsh_printf(
                sh, " num_ucast_to_other: %lu\r\n",
                (unsigned long) rxstats->num_ucast_to_other);
        owlsh_printf(
                sh, " num_bad_crc:        %lu\r\n",
                (unsigned long) rxstats->num_bad_crc);
        
        if (rxstats->num_frames) {
                
                owlsh_printf(
                        sh, " rssi_dbm:           %ld (%ld/%lu)\r\n",
                        (long) (rxstats->rssi_dbm / rxstats->num_frames),
                        (long) rxstats->rssi_dbm,
                        (unsigned long) rxstats->num_frames);

                owlsh_printf(
                        sh, " snr_db:             %ld (%ld/%lu)\r\n",
                        (long) (rxstats->snr_db / rxstats->num_frames),
                        (long) rxstats->snr_db,
                        (unsigned long) rxstats->num_frames);

                owlsh_printf(
                        sh, " dc_re:              %ld (%ld/%lu)\r\n",
                        (long) (rxstats->dc_re / rxstats->num_frames),
                        (long) rxstats->dc_re,
                        (unsigned long) rxstats->num_frames);

                owlsh_printf(
                        sh, " dc_im:              %ld (%ld/%lu)\r\n",
                        (long) (rxstats->dc_im / rxstats->num_frames),
                        (long) rxstats->dc_im,
                        (unsigned long) rxstats->num_frames);

                owlsh_printf(
                        sh, " freq_error:         %ld (%ld/%lu)\r\n",
                        (long) (rxstats->freq_error / rxstats->num_frames),
                        (long) rxstats->freq_error,
                        (unsigned long) rxstats->num_frames);

                owlsh_printf(
                        sh, " delayspread:        %ld (%ld/%lu)\r\n",
                        (long) (rxstats->delayspread / rxstats->num_frames),
                        (long) rxstats->delayspread,
                        (unsigned long) rxstats->num_frames);

        } else {
                owlsh_printf(sh, " rssi_dbm:           N/A\r\n");
                owlsh_printf(sh, " snr_db:             N/A\r\n");
                owlsh_printf(sh, " dc_re:              N/A\r\n");
                owlsh_printf(sh, " dc_im:              N/A\r\n");
                owlsh_printf(sh, " freq_error:         N/A\r\n");
                owlsh_printf(sh, " delayspread:        N/A\r\n");
        }
}

static uint8_t
parse_rate(const char *rate)
{
        if (!owl_os_strcmp(rate, "1"))
                return WLP_RATE_1MBIT;
        else if (!owl_os_strcmp(rate, "2"))
                return WLP_RATE_2MBIT;
        else if (!owl_os_strcmp(rate, "5.5"))
                return WLP_RATE_5_5MBIT;
        else if (!owl_os_strcmp(rate, "11"))
                return WLP_RATE_11MBIT;
        else if (!owl_os_strcmp(rate, "6"))
                return WLP_RATE_6MBIT;
        else if (!owl_os_strcmp(rate, "9"))
                return WLP_RATE_9MBIT;
        else if (!owl_os_strcmp(rate, "12"))
                return WLP_RATE_12MBIT;
        else if (!owl_os_strcmp(rate, "18"))
                return WLP_RATE_18MBIT;
        else if (!owl_os_strcmp(rate, "24"))
                return WLP_RATE_24MBIT;
        else if (!owl_os_strcmp(rate, "36"))
                return WLP_RATE_36MBIT;
        else if (!owl_os_strcmp(rate, "48"))
                return WLP_RATE_48MBIT;
        else if (!owl_os_strcmp(rate, "54"))
                return WLP_RATE_54MBIT;
        else if (!owl_os_strcmp(rate, "6.5"))
                return WLP_RATE_6_5MBIT;
        else if (!owl_os_strcmp(rate, "13"))
                return WLP_RATE_13MBIT;
        else if (!owl_os_strcmp(rate, "19.5"))
                return WLP_RATE_19_5MBIT;
        else if (!owl_os_strcmp(rate, "26"))
                return WLP_RATE_26MBIT;
        else if (!owl_os_strcmp(rate, "39"))
                return WLP_RATE_39MBIT;
        else if (!owl_os_strcmp(rate, "52"))
                return WLP_RATE_52MBIT;
        else if (!owl_os_strcmp(rate, "58.5"))
                return WLP_RATE_58_5MBIT;
        else if (!owl_os_strcmp(rate, "65"))
                return WLP_RATE_65MBIT;
        else
                return WLP_RATE_UNKNOWN;
}
