/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 *
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#define OWL_TRACE OWL_TRACE_DBG
#define OWL_FILE "pico/shell/socket"

#include <owl/core/owl_defs.h>
#include <owl/core/owl_shell.h>
#include <owl/core/owl_err.h>
#include <owl/core/owl_debug.h>

#include <wlp_socket.h>
#include <wlp_inet.h>

/*
 * Evaluate the owl pico API (wlp_socket) using the wlp_shell.
 *
 * The functions provided by the wlp_api can be invoked from a command line,
 * using a one-to-one mapping between commands and api function wherever
 * possible.
 *
 *
 *
 *
 */

static void socket_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void bind_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void listen_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void connect_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void close_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void send_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void sendto_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void get_hostbyname_f(struct owlsh_ctx *sh, int argc,
                             const char *argv[]);


static void recv_cb(void *ctx, int sockid,
                    const struct ip_addr *addr, uint16_t port,
                    void *data, int len);
static void lookup_cb(void *ctx, const struct ip_addr *ip);
static void listen_cb(void *ctx, int lsockid, int sockid,
                      const struct ip_addr *ip, uint16_t port);
static void conn_cb(void *ctx, int sockid,
                    const struct ip_addr *addr, uint16_t port, int connected);
static int parse_uint16(const char *str, uint16_t *out);
static int parse_int(const char *str, int *out);

const struct owlsh_cmd cmd_socket = {
        "socket", socket_f, "wlp_socket(type, proto)"
};
const struct owlsh_cmd cmd_bind = {
        "bind", bind_f, "wlp_bind(sock, ip, port)"
};
const struct owlsh_cmd cmd_listen = {
        "listen", listen_f, "wlp_listen(sockid, backlog)"
};
const struct owlsh_cmd cmd_connect = {
        "connect", connect_f, "wlp_connect(sockid, ip, port)"
};
const struct owlsh_cmd cmd_close = {
        "close", close_f, "wlp_close(sockid)"
};
const struct owlsh_cmd cmd_send = {
        "send", send_f, "wlp_send(sockid, data)"
};
const struct owlsh_cmd cmd_sendto = {
        "sendto", sendto_f, "wlp_sendto(sockid, data, ip, port)"
};
const struct owlsh_cmd cmd_get_hostbyname = {
        "get_hostbyname", get_hostbyname_f, "wlp_get_hostbyname(dnsname)"
};

static void
socket_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int sockid;
        int type;
        int protocol = 0;

        if (argc < 2)
                goto err_parse;

        if (!owl_os_strcmp(argv[1], "stream"))
                type = WLP_SOCK_STREAM;
        else if (!owl_os_strcmp(argv[1], "dgram"))
                type = WLP_SOCK_DGRAM;
        else if (!owl_os_strcmp(argv[1], "raw"))
                type = WLP_SOCK_RAW;
        else
                goto err_parse;

        if (argc > 2 && !parse_int(argv[2], &protocol))
                goto err_parse;

        sockid = wlp_socket(type, protocol);
        wlp_set_conn_cb(sockid, conn_cb, sh);
        wlp_set_recv_cb(sockid, recv_cb, sh);
        owlsh_perr(sh, sockid);
        return;

err_parse:
        owlsh_printf(sh, "usage: socket <stream|dgram|raw> [protocol]\r\n");

}
static void
bind_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        int sockid;
        struct ip_addr ip;
        uint16_t port;

        if (argc < 4)
                goto err_parse;
        if (!parse_int(argv[1], &sockid))
                goto err_parse;
        if (!ip_aton(argv[2], &ip))
                goto err_parse;
        if (!parse_uint16(argv[3], &port))
                goto err_parse;

        res = wlp_bind(sockid, &ip, port);
        owlsh_perr(sh, res);
        return;

err_parse:
        owlsh_printf(sh, "usage: bind <sockid> <ip> <port>\r\n");

}

static void
listen_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        int sockid;
        int backlog = 1;

        if (argc < 2)
                goto err_parse;
        if (!parse_int(argv[1], &sockid))
                goto err_parse;
        if (argc > 2 && !parse_int(argv[2], &backlog))
                goto err_parse;


        res = wlp_listen(sockid, backlog, listen_cb, sh);

        owlsh_perr(sh, res);
        return;

err_parse:
        owlsh_printf(sh, "usage: listen <sockid> [backlog]\r\n");

}

static void
connect_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        int sockid;
        struct ip_addr ip;
        uint16_t port;

        if (argc < 4)
                goto err_parse;
        if (!parse_int(argv[1], &sockid))
                goto err_parse;
        if (!ip_aton(argv[2], &ip))
                goto err_parse;
        if (!parse_uint16(argv[3], &port))
                goto err_parse;

        res = wlp_connect(sockid, &ip, port);

        owlsh_perr(sh, res);
        return;

err_parse:
        owlsh_printf(sh, "usage: connect <sockid> <ip> <port>\r\n");
}

static void
close_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        int sockid;

        if (argc < 2)
                goto err_parse;
        if (!parse_int(argv[1], &sockid))
                goto err_parse;

        res = wlp_close(sockid);

        owlsh_perr(sh, res);
        return;

err_parse:
        owlsh_printf(sh, "usage: close <sockid>\r\n");
}

static void
send_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        int sockid;
        char *data;
        int len;
        int pos = 0;

        if (argc < 3)
                goto err_parse;

        if (!parse_int(argv[1], &sockid))
                goto err_parse;

        if ((data = owlsh_joinargs(argc - 2, argv + 2)) == NULL) {
                res = OWL_ERR_MEM;
                goto out;
        }

        len = owl_os_strlen(data);
        owlsh_printf(sh, "sending %d bytes\r\n", len);
        while (pos < len) {

                if ((res = wlp_send(sockid, data + pos, len - pos)) < 0) {
                        owl_os_free(data);
                        goto out;

                } else {
                        owlsh_printf(sh, "wrote:%d remain:%d\r\n",
                                     res, len - pos - res);
                }
                pos += res;
        }
        owl_os_free(data);
        res = 0;

out:
        owlsh_perr(sh, res);
        return;

err_parse:
        owlsh_printf(sh, "usage: send <sockid> <data ...>\r\n");
}

static void
sendto_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        int sockid;
        struct ip_addr ip;
        uint16_t port;
        char *data;
        int len;
        int pos = 0;

        if (argc < 4)
                goto err_parse;

        if (!parse_int(argv[1], &sockid))
                goto err_parse;

        if (!ip_aton(argv[2], &ip))
                goto err_parse;

        if (!parse_uint16(argv[3], &port))
                goto err_parse;

        if ((data = owlsh_joinargs(argc - 4, argv + 4)) == NULL) {
                res = OWL_ERR_MEM;
                goto out;
        }

        len = owl_os_strlen(data);
        owlsh_printf(sh, "sending %d bytes to %s:%d\r\n", len,
                     ip_ntoa(&ip), port);

        while (pos < len) {

                if ((res = wlp_sendto(sockid, data + pos, len - pos,
                                      &ip, port)) < 0) {
                        owl_os_free(data);
                        goto out;

                } else {
                        owlsh_printf(sh, "wrote:%d remain:%d\r\n",
                                     res, len - pos - res);
                }
                pos += res;
        }
        owl_os_free(data);
        res = 0;

out:
        owlsh_perr(sh, res);
        return;

err_parse:
        owlsh_printf(sh, "usage: sendto <sockid> <ip> <port> <data ...>\r\n");
}

static void
get_hostbyname_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;

        if (argc < 2)
                goto err_parse;

        res = wlp_get_hostbyname(argv[1], lookup_cb, sh);
        owlsh_perr(sh, res);
        return;

err_parse:
        owlsh_printf(sh, "usage: get_hostbyname <dnsname>\r\n");
}

static void
recv_cb(void *ctx, int sockid, const struct ip_addr *ip, uint16_t port,
        void *data, int len)
{
        struct owlsh_ctx *sh = ctx;

        owlsh_printf(sh, "recv_cb sockid:%d addr:%s:%d len:%d %s\r\n",
                     sockid, ip_ntoa(ip), port, len,
                     len ? "" : "(connection lost)");

        wlp_recved(sockid, len);
        owlsh_hexdump(sh, "RX", data, len);
}

static void
lookup_cb(void *ctx, const struct ip_addr *ip)
{
        struct owlsh_ctx *sh = ctx;
        if (ip == NULL)
                owlsh_printf(sh, "lookup_cb: host not found\r\n");
        else
                owlsh_printf(sh, "lookup_cb ip:%s\r\n", ip_ntoa(ip));
}

static void
listen_cb(void *ctx, int lsockid, int sockid,
          const struct ip_addr *ip, uint16_t port)
{
        struct owlsh_ctx *sh = ctx;

        owlsh_printf(sh, "listen_cb lsockid:%d sockid:%d ip:%s port:%d",
                     lsockid, sockid, ip_ntoa(ip), port);

        wlp_set_conn_cb(sockid, conn_cb, sh);
        wlp_set_recv_cb(sockid, recv_cb, sh);
        wlp_accept(lsockid);

        owlsh_perr(sh, sockid);
}

static void
conn_cb(void *ctx, int sockid,
        const struct ip_addr *ip, uint16_t port, int connected)
{
        struct owlsh_ctx *sh = ctx;
        owlsh_printf(sh, "conn_cb sockid:%d addr:%s:%d connected:%d\r\n",
                     sockid, ip_ntoa(ip), port, connected);
}

static int
parse_uint16(const char *str, uint16_t *out)
{
        char *end;
        int val;
        val = owl_os_strtol(str, &end, 0);
        if (end == str)
                return 0;

        if (val > 65535)
                return 0;

        if (val < 0)
                return 0;

        *out = val;
        return 1;
}

static int
parse_int(const char *str, int *out)
{
        char *end;
        *out = owl_os_strtol(str, &end, 0);
        if (end == str)
                return 0;

        return 1;
}
