/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 *
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#define OWL_TRACE OWL_TRACE_INFO
#define OWL_FILE "wlp_apitest"

#include <owl/core/owl_defs.h>
#include <owl/core/owl_shell.h>
#include <owl/core/owl_err.h>
#include <owl/core/owl_debug.h>
#include <owl/core/owl_timer.h>

#include <pico/wlp_apitest.h>

#include <wlp_api.h>
#include <wlp_ipaddr.h>
#include <wlp_inet.h>
#include <wlp_concurrent.h>

struct apitest_ctx;

struct apitest_ctx {
        int iter;
        int count;
        struct wlp_ssid ssid;
        char key[WLP_MAX_PASS_LEN];

#define TEST_SCAN_DISCONNECTED (1 << 0)
#define TEST_LINKUP_AP         (1 << 1)
#define TEST_LINKUP_STA        (1 << 2)
#define TEST_SCAN_CONNECTED    (1 << 3)
        int mask;

        int link_tmo;
        int dhcp_tmo;
        struct ip_addr ip;

        struct owlsh_ctx *sh;
        struct owltmr_tmo tmo;
        void (*state)(struct apitest_ctx *priv, int ev);
#if !WLP_CONFIG_STACK_OFFLOAD
        int iface;
#endif /* !WLP_CONFIG_STACK_OFFLOAD */
};

#define EVENT_TIMEOUT         0
#define EVENT_SCAN_COMPLETE   1
#define EVENT_LINKUP          2
#define EVENT_LINKDOWN        3
#define EVENT_IPADDR          4

static void apitest_f(struct owlsh_ctx *sh, int argc, const char *argv[]);

static void sched_timer(struct apitest_ctx *priv, int sec);
static int parse_ssid(const char *str, struct wlp_ssid *ssid);
#if defined(OWL_DEBUG) && OWL_TRACE >= OWL_TRACE_INFO
static const char *event_str(int ev);
#endif /* defined(OWL_DEBUG) && OWL_TRACE >= OWL_TRACE_INFO */

static void apitest_next(struct apitest_ctx *priv);
static void apitest_stop(struct apitest_ctx *priv, int err);
static void apitest_scan_disconnected(struct apitest_ctx *priv);
static void apitest_linkup_ap(struct apitest_ctx *priv);
static void apitest_linkdown_ap(struct apitest_ctx *priv);
static void apitest_linkup_sta(struct apitest_ctx *priv);
static void apitest_linkdown_sta(struct apitest_ctx *priv);
static void apitest_scan_connected(struct apitest_ctx *priv);

static void state_next(struct apitest_ctx *priv, int ev);
static void state_scan_disconnected(struct apitest_ctx *priv, int ev);
static void state_linkup_ap(struct apitest_ctx *priv, int ev);
static void state_linkdown_ap(struct apitest_ctx *priv, int ev);
static void state_linkup_sta(struct apitest_ctx *priv, int ev);
static void state_dhcp(struct apitest_ctx *priv, int ev);
static void state_scan_connected(struct apitest_ctx *priv, int ev);
static void state_linkdown_sta(struct apitest_ctx *priv, int ev);

static void timer_cb(struct owltmr_tmo *tmo, void *ctx);
static void link_cb(void *ctx, int link);
static void addr_cb(void *ctx, const struct ip_addr *ip);
static void scan_cb(void *ctx, struct wlp_network *net);

/* Don't include owl_lwip.h */
extern void owlip_linkup(int id);
extern void owlip_linkdown(int id);

const struct owlsh_cmd cmd_apitest = {
        "apitest", apitest_f, "wlp_api test"
};


static void
apitest_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        struct apitest_ctx *priv = NULL;
        
        if (argc < 3)
                goto err_parse;

        if ((priv = owl_os_zalloc(sizeof(*priv))) == NULL)
                return;

        priv->sh = sh;
        
        priv->count = owl_os_atoi(argv[1]);
        priv->mask = owl_os_strtol(argv[2], NULL, 0);
        if (argc > 3)
                parse_ssid(argv[3], &priv->ssid);
        if (argc > 4)
                owl_os_strcpy(priv->key, argv[4]);

#if !WLP_CONFIG_STACK_OFFLOAD
        /* Remember the interface to use so we can pass it to owlip_linkup()
           etc */
        priv->iface = wlp_get_iface();
#endif /* !WLP_CONFIG_STACK_OFFLOAD */
        
        wlp_set_link_cb(link_cb, priv);
        wlp_set_ipaddr_cb(addr_cb, priv);
        wlp_set_dhcp(1);

        apitest_next(priv);
        return;

err_parse:
        owl_os_free(priv);
        owlsh_usage(sh, "usage: apitest <count> <mask> [ssid] [key]\r\n");
}

static int
parse_ssid(const char *str, struct wlp_ssid *ssid)
{
        ssid->len = owl_os_strlen(str);
        owl_os_memcpy(ssid->ssid, str, ssid->len);
        return 1;
}

static void
sched_timer(struct apitest_ctx *priv, int sec)
{
        owltmr_sched(&priv->tmo, sec * 1000, 0, timer_cb, priv);
}

#if defined(OWL_DEBUG) && OWL_TRACE >= OWL_TRACE_INFO
static const char *
event_str(int ev)
{
        switch (ev) {
        case EVENT_TIMEOUT:
                return "TIMEOUT";
        case EVENT_SCAN_COMPLETE:
                return "SCAN_COMPLETE";
        case EVENT_LINKUP:
                return "LINKUP";
        case EVENT_LINKDOWN:
                return "LINKDOWN";
        case EVENT_IPADDR:
                return "IPADDR";
        default:
                return "N/A";
        };
}
#endif /* defined(OWL_DEBUG) && OWL_TRACE >= OWL_TRACE_INFO */

/* Tests ---------------------------------------------------------------- */


static void
apitest_next(struct apitest_ctx *priv)
{

        if (priv->iter < priv->count) {
                owlsh_printf(priv->sh, "\r\n");
                owlsh_printf(
                        priv->sh, "APITEST              [%04d]\r\n",
                        priv->iter + 1);
                owlsh_printf(
                        priv->sh, "---------------------------\r\n");
                
                sched_timer(priv, 0);
                priv->state = state_next;
                priv->iter++;
                
        } else {
                
                apitest_stop(priv, 0);
        }
}

static void
apitest_stop(struct apitest_ctx *priv, int err)
{
        owlsh_printf(
                priv->sh, "\r\nSTATS\r\n");
        owlsh_printf(
                priv->sh, "-------------------\r\n");
        owlsh_printf(
                priv->sh, "ERROR:         %d\r\n", err);
        owlsh_printf(
                priv->sh, "COUNT:         %04d\r\n", priv->count);
        owlsh_printf(
                priv->sh, "ITERATIONS:    %04d\r\n", priv->iter);
        owlsh_printf(
                priv->sh, "MASK:          0x%02x\r\n", priv->mask);
        owlsh_printf(
                priv->sh, "LINK TIMEOUTS: %04d\r\n", priv->link_tmo);
        owlsh_printf(
                priv->sh, "DHCP TIMEOUTS: %04d\r\n", priv->dhcp_tmo);

        /* Restore wlp state */
        wlp_set_link_cb(NULL, NULL);
        wlp_set_ipaddr_cb(NULL, NULL);
        wlp_set_mode(WLP_MODE_STA);
        wlp_linkdown();
        

        /* Dealloc */
        owltmr_cancel(&priv->tmo);
        owl_os_free(priv);
}

static void
apitest_scan_disconnected(struct apitest_ctx *priv)
{
        int err;

        owlsh_printf(priv->sh, "  SCAN (DISCONNECTED)\r\n");
        if ((err = wlp_scan(NULL, scan_cb, priv))) {
                apitest_stop(priv, err);
                return;
        }
        
        sched_timer(priv, 15);
        priv->state = state_scan_disconnected;
}

static void
apitest_linkup_ap(struct apitest_ctx *priv)
{
        int err;

        owlsh_printf(priv->sh, "  SET_MODE AP\r\n");
        if ((err = wlp_set_mode(WLP_MODE_AP))) {
                apitest_stop(priv, err);
                return;
        }

        owlsh_printf(priv->sh, "  LINKUP\r\n");
        if ((err = wlp_linkup(&priv->ssid, NULL, WLP_AUTH_AUTO))) {
                apitest_stop(priv, err);
                return;
        }
                
        sched_timer(priv, 3);
        priv->state = state_linkup_ap;
}

static void
apitest_linkdown_ap(struct apitest_ctx *priv)
{
        int err;
                
        owlsh_printf(priv->sh, "  LINKDOWN\r\n");
        if ((err = wlp_linkdown())) {
                apitest_stop(priv, err);
                return;
        }

        sched_timer(priv, 1);
        priv->state = state_linkdown_ap;
}

static void
apitest_linkup_sta(struct apitest_ctx *priv)
{
        int err;
        
        owlsh_printf(priv->sh, "  SET_MODE STA\r\n");
        if ((err = wlp_set_mode(WLP_MODE_STA))) {
                apitest_stop(priv, err);
                return;
        }

        owlsh_printf(priv->sh, "  LINKUP\r\n");
        if ((err = wlp_linkup(&priv->ssid, priv->key, WLP_AUTH_AUTO))) {
                apitest_stop(priv, err);
                return;
        }
                
        sched_timer(priv, 20);
        priv->state = state_linkup_sta;
}
                
static void
apitest_dhcp(struct apitest_ctx *priv)
{
        owlsh_printf(priv->sh, "  DHCP REQUEST\r\n");
        sched_timer(priv, 20);
        priv->state = state_dhcp;
}

static void
apitest_linkdown_sta(struct apitest_ctx *priv)
{
        int err;

        owlsh_printf(priv->sh, "  LINKDOWN\r\n");
        if ((err = wlp_linkdown())) {
                apitest_stop(priv, err);
                return;
        }

        sched_timer(priv, 5);
        priv->state = state_linkdown_sta;
}

static void
apitest_scan_connected(struct apitest_ctx *priv)
{
        int err;
        
        owlsh_printf(priv->sh, "  SCAN (CONNECTED)\r\n");
        if ((err = wlp_scan(NULL, scan_cb, priv))) {
                apitest_stop(priv, err);
                return;
        }
        
        sched_timer(priv, 60);
        priv->state = state_scan_connected;
}

/* State Machine ------------------------------------------------------ */

static void
state_next(struct apitest_ctx *priv, int ev)
{
        owl_info("ev:%s", event_str(ev));
        
        switch (ev) {
        case EVENT_TIMEOUT:
                if (owl_bit(priv->mask, TEST_SCAN_DISCONNECTED)) {
                        apitest_scan_disconnected(priv);
                } else if (owl_bit(priv->mask, TEST_LINKUP_AP)) {
                        apitest_linkup_ap(priv);
                } else if (owl_bit(priv->mask, TEST_LINKUP_STA)) {
                        apitest_linkup_sta(priv);
                } else {
                        apitest_next(priv);
                }
                return;

        default:
                apitest_stop(priv, OWL_ERR_STATE);
                return;
        };
}

static void
state_scan_disconnected(struct apitest_ctx *priv, int ev)
{
        owl_info("ev:%s", event_str(ev));
        
        switch (ev) {
        case EVENT_LINKDOWN:
        case EVENT_LINKUP:
                /* Consume linkdown and linkup events. We might get these in
                   the rare occasion that we lost the link just before
                   state_linkdown_sta is entered. EVENT_LINKDOWN will then
                   be received in state_linkdown_sta but wlp_linkdown() has
                   not yet been processed since the EVENT_LINKDOWN belongs to
                   the link loss. However, this state will be entered since
                   we got EVENT_LINKDOWN. After the link is lost, the device
                   will reconnect, resulting in EVENT_LINKUP followed by
                   EVENT_LINKDOWN (since wlp_linkdown() is now processed).
                   These two events will be consumed here. */
                return;
                
        case EVENT_SCAN_COMPLETE:
                if (owl_bit(priv->mask, TEST_LINKUP_AP)) {
                        apitest_linkup_ap(priv);
                } else if (owl_bit(priv->mask, TEST_LINKUP_STA)) {
                        apitest_linkup_sta(priv);
                } else {
                        apitest_next(priv);
                }
                return;

        case EVENT_TIMEOUT:
                apitest_stop(priv, OWL_ERR_TIMEOUT);
                return;

        default:
                apitest_stop(priv, OWL_ERR_STATE);
                return;
        };
}

static void
state_linkup_ap(struct apitest_ctx *priv, int ev)
{
        owl_info("ev:%s", event_str(ev));
        
        switch (ev) {
        case EVENT_LINKUP:
        case EVENT_TIMEOUT:
                apitest_linkdown_ap(priv);
                return;

        default:
                apitest_stop(priv, OWL_ERR_STATE);
                return;
        };
}

static void
state_linkdown_ap(struct apitest_ctx *priv, int ev)
{
        owl_info("ev:%s", event_str(ev));
        
        switch (ev) {

        case EVENT_LINKUP:
                /* Ignore and wait for timeout */
                return;
                
        case EVENT_LINKDOWN:
        case EVENT_TIMEOUT:
                if (owl_bit(priv->mask, TEST_LINKUP_STA)) {
                        apitest_linkup_sta(priv);
                } else {
                        apitest_next(priv);
                }
                return;

        default:
                apitest_stop(priv, OWL_ERR_STATE);
                return;
        };
}


static void
state_linkup_sta(struct apitest_ctx *priv, int ev)
{
        owl_info("ev:%s", event_str(ev));
        
        switch (ev) {
        case EVENT_LINKDOWN:
                return;
                
        case EVENT_LINKUP:
                apitest_dhcp(priv);
                return;

        case EVENT_TIMEOUT:
                priv->link_tmo++;
                /* Make sure we call wlp_linkdown() */
                apitest_linkdown_sta(priv);

                /* Jump to next, otherwise we will get a timeout in
                   state_linkdown_sta() since the link is already down */
                apitest_next(priv);
                return;

        default:
                apitest_stop(priv, OWL_ERR_STATE);
                return;
        };
}

static void
state_dhcp(struct apitest_ctx *priv, int ev)
{
        owl_info("ev:%s", event_str(ev));
        
        switch (ev) {
        case EVENT_IPADDR:
                owlsh_printf(priv->sh, "  DHCP IP:%s\r\n", ip_ntoa(&priv->ip));
                if (owl_bit(priv->mask, TEST_SCAN_CONNECTED)) {
                        apitest_scan_connected(priv);
                } else {
                        apitest_linkdown_sta(priv);
                }
                return;
                
        case EVENT_TIMEOUT:
                owlsh_printf(priv->sh, "  DHCP TIMEOUT\r\n");
                priv->dhcp_tmo++;
                apitest_linkdown_sta(priv);
                return;

        default:
                apitest_stop(priv, OWL_ERR_STATE);
                return;
        };

}

static void
state_scan_connected(struct apitest_ctx *priv, int ev)
{
        owl_info("ev:%s", event_str(ev));
        
        switch (ev) {
        case EVENT_SCAN_COMPLETE:
                apitest_linkdown_sta(priv);
                return;
                
        case EVENT_LINKDOWN:
        case EVENT_LINKUP:
                /* Wait for EVENT_SCAN_COMPLETE */
                return;
                
        case EVENT_TIMEOUT:
                apitest_stop(priv, OWL_ERR_TIMEOUT);
                return;

        default:
                apitest_stop(priv, OWL_ERR_STATE);
                return;
        };
}

static void
state_linkdown_sta(struct apitest_ctx *priv, int ev)
{
        owl_info("ev:%s", event_str(ev));
        
        switch (ev) {
        case EVENT_LINKUP:
                /* This might occur if the link goes back again before
                   wlp_linkdown() has been processed, e.g. on HDG205.
                   Wait for EVENT_LINKDOWN in this case */
                return;
                
        case EVENT_LINKDOWN:
                apitest_next(priv);
                return;

        case EVENT_TIMEOUT:
                apitest_stop(priv, OWL_ERR_TIMEOUT);
                return;

        default:
                apitest_stop(priv, OWL_ERR_STATE);
                return;
        };
}
        
/* Events ---------------------------------------------------------------- */

static void
timer_cb(struct owltmr_tmo *tmo, void *ctx)
{
	struct apitest_ctx *priv = ctx;
	priv->state(priv, EVENT_TIMEOUT);
}

static void
link_cb(void *ctx, int link)
{
	struct apitest_ctx *priv = ctx;

        owl_info("link:%d", link);
        
        if (link > 0) {
#if !WLP_CONFIG_STACK_OFFLOAD
                owlip_linkup(priv->iface);
#endif /* !WLP_CONFIG_STACK_OFFLOAD */
                priv->state(priv, EVENT_LINKUP);
        } else {
#if !WLP_CONFIG_STACK_OFFLOAD
                owlip_linkdown(priv->iface);
#endif /* !WLP_CONFIG_STACK_OFFLOAD */
                priv->state(priv, EVENT_LINKDOWN);
        }
}

static void
addr_cb(void *ctx, const struct ip_addr *ip)
{
	struct apitest_ctx *priv = ctx;

        if (ip == NULL) {
                owl_info("ip:NULL");
        } else {
                owl_info("ip:%s", ip_ntoa(ip));
                ip_addr_set(&priv->ip, ip);
                priv->state(priv, EVENT_IPADDR);
        }
}

static void
scan_cb(void *ctx, struct wlp_network *net)
{
	struct apitest_ctx *priv = ctx;
        
        if (net == NULL) {
                priv->state(priv, EVENT_SCAN_COMPLETE);
        }                
}
