/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#ifndef WLP_SHELL_H
#define WLP_SHELL_H

#include <owl/core/owl_shell.h>


/**
 * Enable or disable WPS (Wi-Fi Protected Setup).
 *
 * When enabled, pin or push button methods can be used to connect to an
 * access-point. To use the pin method, the pin should be passed as @key to
 * linkup after enabling WPS. To use the push-button method, linkup
 * should be called without a key after enabling WPS.
 *
 */ 
extern const struct owlsh_cmd cmd_set_wps;

/**
 * Set the Wi-Fi mode to AP, STA or ADHOC.
 *
 * set_mode will configure the Wi-Fi device to operate either as a
 * station or as an access point. The current mode can be obtained through
 * get_mode.
 *
 */
extern const struct owlsh_cmd cmd_set_mode;

/**
 * Check if the device is currently in AP, STA or ADHOC mode.
 *
 * get_mode will be return the current operation mode of the Wi-Fi device.
 *
 */
extern const struct owlsh_cmd cmd_get_mode;

/**
 * Enable or disable WiFi power save.
 *
 * Enable/disable legacy power save mode (as opposed to WMM power save
 * which is not supported in the current driver). In legacy power save
 * mode, the device will power down when idle. When connected, the
 * device will wake up to receive beacon frames and any buffered data
 * from the AP. The response time when legacy power save is enabled
 * might therefore be as long as the AP beacon interval (commonly 100
 * ms). However, the throughput should not be affected.
 *
 */
extern const struct owlsh_cmd cmd_set_ps;

/**
 * Configure WiFi power save parameters.
 *
 * This call is not necessary to enable WiFi power save.
 * If "set_ps 1" is called before calling conf_ps, then default power
 * save parameters are used.
 *
 * Beware that setting @rx_all_dtim to 0 and @listen_interval to a
 * large value (> 10) can make the device very difficult to
 * reach. When @rx_all_dtim is 0 you risk loosing multicast data
 * because that data is transferred after a DTIM beacon. This will be
 * particularly noticable when another entitity on the same network
 * segment tries to reach the device and does not know which MAC
 * address the device has. In this case an ARP packet is normally sent
 * to the broadcast address and this will be unlikely to be heard
 * by the device.
 * 
 * @use_ps_poll: Use PS-Poll frames to retrieve buffered data.
 *       Any changes to this parameter will take effect upon next
 *       connect or when power save is enabled through
 *       set_ps.
 *       Note: To retrieve one buffered packet, the ps poll scheme
 *       needs one ps poll packet to the AP instead of two null
 *       packets in the power management bit scheme. Ps poll avoids
 *       the overhead of traffic monitoring time in active mode as
 *       well. But since each ps poll request can make the AP release
 *       only one buffered packet, it is not the optimal scheme for
 *       applications with heavy downlink traffic.
 * @ps_traffic_timeout Timeout in [ms] to wait for more buffered
 *      data from AP. This setting has no effect if use_ps_poll is
 *      1. Any changes to this parameter will take effect
 *      immediately.
 * @ps_delay Power save will de delayed ps_delay [ms] after
 *      connecting to an AP.
 * @rx_all_dtim If set to 1, then STA will wake up to listen to
 *       every beacon containing DTIM (delivery traffic indication
 *       messages) when connected. The actual DTIM interval is
 *       configured in the AP.  If the DTIM interval, as configured
 *       in the AP, is larger than @listen_interval, the STA will
 *       wakeup according to the @listen_interval parameter.  Note
 *       that if @rx_all_dtim is set to 0 and @listen_interval is
 *       set to something larger than 1 you risk missing reception of
 *       broadcast/multicast data because the AP will transmit that
 *       in conjunction with a DTIM beacon.
 * @listen_interval The Listen Interval field is used to
 *       indicate to the AP how often a STA in power save mode wakes
 *       to listen to beacon frames. The value of this parameter is
 *       expressed in units of Beacon Interval. An AP may use the
 *       Listen Interval information in determining the lifetime of
 *       frames that it buffers for a STA.  Any changes to this
 *       parameter will take effect upon next association.
 */
extern const struct owlsh_cmd cmd_conf_ps;


/**
 * Configure the parameters that will be used when the device is in access
 * point mode.
 *
 * @hidden can be set to non-zero in order to configure the device
 *               with hidden SSID. This means that the device will not
 *               broadcast its SSID, but it will still respond to clients
 *               that already knows the SSID.
 * @beacon_intvl specifies interval in which beacons are transmitted
 *               by the device. This parameter does not have any effect if
 *               hidden is 0
 * @inact_tmo specifies the idle time in minutes before the device
 *               will disconnect an inactive client.
 * @country specifies the region that should be advertised by the device.
 *               Will not be changed if NULL.
 */
extern const struct owlsh_cmd cmd_conf_ap;


/**
 * Set the channel. Only supported in access point or adhoc mode.
 *
 * This function is only supported when get_mode returns "ap" or "adhoc".
 * set_channel must be invoked before linkup to have effect.
 *
 */
extern const struct owlsh_cmd cmd_set_channel;

/**
 * Connect the wifi link. Upon successful return, the wifi device will
 * try to establish and maintain a connection to the network specified by the
 * parameters. When the link status (connected, disconnected) changes,
 * the shell will be notified of the new status with the message "link:0"
 * or"link:1". Information about the current network can be obtained with
 * the command get_network. Note again that successful return means that
 * the wifi device will now _initate_ a link using the desired
 * parameters, not that the link is already established.
 *
 * If linkup us invoked when already connected to an access point,
 * the wifi device will try to switch to the network configuration according
 * to the provided parameters.
 *
 * If the @auth parameter is provided, it will set a requrement for the
 * _minimal_ authentication level that is allowed.
 *
 * In access point mode, this command will start an access point with the
 * given ssid and key parameters. See set_mode for more information on
 * how to set access point mode.
 */
extern const struct owlsh_cmd cmd_linkup;

/**
 * Disconnect the wifi link. The wifi device will no longer try to
 * maintain the network connection. Any link status message will be
 * printerd when disconnection is completed, if currently connected to an
 * access point. Note again that successful return means that the wifi
 * device will now _initate_ a link disconnection, not that the link is
 * already disconnected.
 *
 * If linkdown is invoked when already disconnected, any ongoing connection
 * attempts will be stopped.
 *
 * In access point mode, this function will stop the access point and it will
 * no longer transmit beacons. All clients will be disconnected. See
 * set_mode for more information on how to set access point mode.
 */
extern const struct owlsh_cmd cmd_linkdown;

/**
 * Set context for the link. The context is only valid for the next
 * linkup call.
 *
 */
extern const struct owlsh_cmd cmd_set_linkctx;

/**
 * Retrieve information about the link, such as the PMK. This can be used to
 * later speed up the linkup process by passing the same context to
 * set_linkctx.
 */
extern const struct owlsh_cmd cmd_get_linkctx;

/**
 * Return the 802.3 MAC address of the network interface.
 *
 */
extern const struct owlsh_cmd cmd_get_hwaddr;

/**
 * Get the description of the currently associated network, if any.
 *
 * If associated, the device will update the RSSI and SNR information
 * each time this function is called by listening for beacon frames from
 * the current AP.
 * 
 * In access point mode, this function will return the description of the
 * device when the link is up, e.g. first client connected. See set_mode
 * for more information on how to set access point mode. In this mode,
 * the RSSI and SNR values will be set to 0.
 */
extern const struct owlsh_cmd cmd_get_network;

extern const struct owlsh_cmd cmd_get_rate;

/**
 * Scan for networks. If @ssid is provided, an active scan will be
 * performed by sending PROBE REQ to the specified ssid. This can be used
 * to find hidden networks.
 */
extern const struct owlsh_cmd cmd_scan;

/**
 * Send a RAW frame using the Wi-Fi device. The frame can only be sent
 * when the Wi-Fi device is not connected to any access point.
 *
 * To simplify usage, a header for a QOS or beacon frame will be built, but
 * additional data can be provided on the command line.
 *
 * If @data contains less bytes than @len, the the last byte in @data
 * will be repeated to obtain @len bytes.
 *
 */
extern const struct owlsh_cmd cmd_rawtx;

extern const struct owlsh_cmd cmd_set_ipaddr;
extern const struct owlsh_cmd cmd_get_ipaddr;
extern const struct owlsh_cmd cmd_set_dhcp;
extern const struct owlsh_cmd cmd_get_dhcp;
extern const struct owlsh_cmd cmd_set_dhcpd;

extern const struct owlsh_cmd cmd_socket;
extern const struct owlsh_cmd cmd_bind;
extern const struct owlsh_cmd cmd_listen;
extern const struct owlsh_cmd cmd_connect;
extern const struct owlsh_cmd cmd_close;
extern const struct owlsh_cmd cmd_send;
extern const struct owlsh_cmd cmd_sendto;
extern const struct owlsh_cmd cmd_get_hostbyname;

extern const struct owlsh_cmd cmd_get_stats;
extern const struct owlsh_cmd cmd_loopback;
extern const struct owlsh_cmd cmd_heartbeat;

extern const struct owlsh_cmd cmd_start_tx;
extern const struct owlsh_cmd cmd_stop_tx;
extern const struct owlsh_cmd cmd_get_rxstats;
extern const struct owlsh_cmd cmd_clr_rxstats;

extern const struct owlsh_cmd cmd_add_mcast;
extern const struct owlsh_cmd cmd_del_mcast;

extern const struct owlsh_cmd cmd_get_iface;
extern const struct owlsh_cmd cmd_set_iface;

#endif
