/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#define OWL_TRACE OWL_TRACE_NONE
#define OWL_FILE "wlp_ttcp"

#include <owl/core/owl_shell.h>
#include <owl/core/owl_err.h>
#include <owl/core/owl_debug.h>
#include <owl/core/owl_timer.h>

#include <wlp_socket.h>
#include <wlp_inet.h>

#include <pico/wlp_ttcp.h>

struct ttcp_ctx {
        int sockid;
        char *buf;
        uint16_t buflen;
        uint16_t bufpos;
        uint32_t bufcnt;
        uint32_t start_ms;
        uint32_t stop_ms;
        int send;

        uint32_t recvd; /* recv mode only */
        int acceptid;   /* recv mode only */
	uint32_t bufidx;

	struct owltmr_tmo tmo;

        void (*progress_cb)(void *ctx, int progress);
        void (*done_cb)(void *ctx, int err, uint32_t bytes, uint32_t ms);
        void *ctx;
};

static void timer_cb(struct owltmr_tmo *tmo, void *ctx); 
static void recv_cb(void *ctx, int sockid,
                    const struct ip_addr *ip, uint16_t port,
                    void *data, int len);
static void listen_cb(void *ctx, int lsockid, int sockid,
                      const struct ip_addr *ip, uint16_t port);
static void conn_cb(void *ctx, int sockid,
                    const struct ip_addr *ip, uint16_t port, int connected);
static void sent_cb(void *ctx, int sockid, int len);
static int ttcp_done(struct ttcp_ctx *priv, int err);

/**
 * Start a ttcp connection for transmission. The ttcp transfer will
 * proceed asynchronously upon successful return.
 * 
 *
 * @return sockid if started successfully.
 */
int
ttcp_connect(const struct ip_addr *ip, uint16_t port, uint16_t bufcnt,
             uint16_t buflen,
             void (*progress_cb)(void *ctx, int progress),
             void (*done_cb)(void *ctx, int err, uint32_t bytes, uint32_t ms),
             void *ctx)
{
        int err;
        struct ttcp_ctx *priv;
        
        owl_info("ip:%s port:%d bufcnt:%d buflen:%d",
                 ip_ntoa(ip), port, bufcnt, buflen);

        if (bufcnt == 0 || buflen == 0 || port == 0)
                return OWL_ERR_PARAM;

        /* Allocate our private ttcp struct */
        if ((priv = owl_os_zalloc(sizeof(struct ttcp_ctx))) == NULL) {
                owl_info("fail to alloc priv");
                return OWL_ERR_MEM;
        }

        priv->bufcnt = bufcnt;
        priv->buflen = buflen;
        priv->progress_cb = progress_cb;
        priv->done_cb = done_cb;
        priv->ctx = ctx;
        priv->send = 1;
        
        /* Allocate a buffer to hold the data we're going to transmit */
        if ((priv->buf = owl_os_alloc(priv->buflen)) == NULL) {
                owl_info("fail to alloc buf (%d b)", priv->buflen);
                return ttcp_done(priv, OWL_ERR_MEM);
        }
        
        /* Create a TCP socket */
        if ((priv->sockid = wlp_socket(WLP_SOCK_STREAM, 0)) < 0) {
                owl_info("fail to create sock err:%d", priv->sockid);
                return ttcp_done(priv, priv->sockid);
        }

        /* Connect the socket to the remote ip and port. The conn_cb
         * passed to wlp_connect() will be invoked when the connection
         * attempt is completed. The result will be passed in the
         * "connected" parameter to conn_cb.
         */
        if ((err = wlp_connect(priv->sockid, ip, port)) < 0) {
                owl_info("fail to connect err:%d", err);
                return ttcp_done(priv, err);
        }

        wlp_set_conn_cb(priv->sockid, conn_cb, priv);
        wlp_set_recv_cb(priv->sockid, recv_cb, priv);
	wlp_set_sent_cb(priv->sockid, sent_cb, priv);

        /* Wait for conn_cb() */
        return priv->sockid;
}


/**
 * Start a ttcp connection for reception. The ttcp transfer will
 * proceed asyncrously upon successful return.
 * 
 *
 * @return sockid if started successfully.
 */
int
ttcp_listen(uint16_t port, uint16_t buflen,
            void (*progress_cb)(void *ctx, int progress),
            void (*done_cb)(void *ctx, int err, uint32_t bytes, uint32_t ms),
            void *ctx)
{
        int err;
        struct ttcp_ctx *priv;
        
        owl_info("port:%d buflen:%d", port, buflen);

        /* Allocate our private ttcp struct */
        if ((priv = owl_os_zalloc(sizeof(struct ttcp_ctx))) == NULL) {
                owl_info("fail to alloc priv");
                return ttcp_done(priv, OWL_ERR_MEM);
        }

        priv->buflen = buflen;
        priv->progress_cb = progress_cb;
        priv->done_cb = done_cb;
        priv->ctx = ctx;

        /* Allocate a buffer to hold the data we're going to recv */
        if ((priv->buf = owl_os_alloc(priv->buflen)) == NULL) {
                owl_err("fail to alloc buf (%d b)", priv->buflen);
                return ttcp_done(priv, OWL_ERR_MEM);
        }

        /* Create a TCP socket */
        if ((priv->sockid = wlp_socket(WLP_SOCK_STREAM, 0)) < 0) {
                owl_err("fail to create sock err:%d", priv->sockid);
                return ttcp_done(priv, priv->sockid);
        }

        /* Bind the socket to the specified port */
        if ((err = wlp_bind(priv->sockid, IP_ADDR_ANY, port)) < 0) {
                owl_err("fail to bind sock err:%d", err);
                return ttcp_done(priv, err);
        }

        /* Start listening, listen_cb() will be invoked when we get an
         * incoming connection
         */
        if ((err = wlp_listen(priv->sockid, 1, listen_cb, priv)) < 0) {
                owl_err("fail to listen err:%d", err);
                return ttcp_done(priv, err);
        }

        /* Wait for listen_cb() */
        return priv->sockid;
}

static void
conn_cb(void *ctx, int sockid,
        const struct ip_addr *ip, uint16_t port, int connected)
{
        struct ttcp_ctx *priv = ctx;

        owl_unused(sockid);
        
        owl_info("connected:%d", connected);
        
        /* In send mode: connection attempt failed or lost connection
         * In recv mode: lost connection
         */
        if (connected <= 0) {
                ttcp_done(priv, connected);
                return;
        }
        
        owl_assert(priv->send);
        owl_info("ttcp connected, start send");

        priv->start_ms = owltmr_get_ms();
	owltmr_sched(&priv->tmo, 0, 0, timer_cb, priv);
}

static void
timer_cb(struct owltmr_tmo *tmo, void *ctx)
{
	int err;
	struct ttcp_ctx *priv = ctx;

        owl_unused(tmo);
        
	while (priv->bufidx < priv->bufcnt) {
                int len;

                len = priv->buflen - priv->bufpos;
                
                if ((err = wlp_send(
                             priv->sockid, priv->buf + priv->bufpos, len)) <= 0)
                        return;

                if (err < len) {
                        priv->bufpos += err;
                        continue;
                } else {
                        priv->bufpos = 0;
                }

                /* print some progress information */
                if (priv->progress_cb)
                        priv->progress_cb(priv->ctx, priv->bufidx);

                priv->bufidx++;
	}

        priv->stop_ms = owltmr_get_ms();
        ttcp_done(priv, 0);
}

static void
sent_cb(void *ctx, int sockid, int len)
{
	struct ttcp_ctx *priv = ctx;

        owl_unused(sockid);
        owl_unused(len);

        owl_info("enter");
	owltmr_sched(&priv->tmo, 0, 0, timer_cb, priv);
}

static void
listen_cb(void *ctx, int lsockid, int sockid,
          const struct ip_addr *ip, uint16_t port)
{
        struct ttcp_ctx *priv = ctx;

        priv->acceptid = sockid;
                
        wlp_set_conn_cb(priv->acceptid, conn_cb, priv);
        wlp_set_recv_cb(priv->acceptid, recv_cb, priv);
        wlp_accept(lsockid);

        /* conn accepted, incoming data will be handled in the recv_cb() */
        owl_info("sockid:%d acceptid:%d", lsockid, priv->acceptid);
        
        priv->start_ms = owltmr_get_ms();
}

static void
recv_cb(void *ctx, int sockid,
        const struct ip_addr *ip, uint16_t port,
        void *data, int len)
{
        struct ttcp_ctx *priv = ctx;
        int err;

        owl_dbg("sockid:%d len:%d", sockid, len);
        
        /* if remote end shuts down when we're in send mode, we'll
         * get this information from wlp_send() anyway
         */
        if (priv->send) {
                owl_info("ignoring rx data");
                return;
        }

        /* rx done when remote end shuts down connection */
        if (len == 0) {
                owl_info("done len:0");
                priv->stop_ms = owltmr_get_ms();
                ttcp_done(priv, 0);
                return;
        }

        err = len;
        wlp_recved(sockid, err);
        
        owl_dbg("recv:%d", err);
        priv->bufcnt++;
        priv->recvd += err;
        /* print some progress information */
        if (priv->progress_cb)
                priv->progress_cb(priv->ctx, priv->bufcnt);
}

static int
ttcp_done(struct ttcp_ctx *priv, int err)
{
        owl_info("err:%d", err);

        uint32_t bytes = priv->send ?
                (uint32_t) (priv->bufcnt * priv->buflen) : priv->recvd;
        uint32_t ms = owltmr_get_intvl(priv->start_ms, priv->stop_ms);
        
        if (priv->done_cb)
                priv->done_cb(priv->ctx, err, bytes, ms);
        
        owl_os_free(priv->buf);
        wlp_close(priv->sockid);
        if (!priv->send)
                wlp_close(priv->acceptid);
        owl_os_free(priv);
        return err;
}

static void
progress_cb(void *ctx, int i)
{
        struct owlsh_ctx *sh = ctx;
        owlsh_printf(sh, ".");
        if (i > 0 && i % 80 == 0)
                owlsh_printf(sh, "\n\r");
}

static void
done_cb(void *ctx, int err, uint32_t bytes, uint32_t ms)
{
        struct owlsh_ctx *sh = ctx;
        owlsh_printf(sh, "\n\r");
        
        if (err)
                owlsh_printf(sh, "ttcp err:%d (%s)\n\r", err, owl_error(err));
                
        else if (ms > 0)
                owlsh_printf(sh, "ttcp ok: %lu bytes processed in %lu ms, "
                             "%lu KB/s\n\r",
                             (long unsigned) bytes,
                             (long unsigned) ms,
                             (long unsigned) (bytes / ms));
        else
                owlsh_printf(sh, "ttcp ok: %lu bytes processed in 0 ms\n\r",
                             (unsigned long) bytes);
}

static void
ttcp_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int i;
        uint16_t port = 5000;          /* -p */
        uint16_t bufcnt = 1024;          /* -n */
        uint16_t buflen = 1024;        /* -l */
        int connect = 0;              /* -t , -r */
        struct ip_addr ip = ip_addr_any;
        int err;

	/* parse arguments, ugly */
	for (i = 1; i < argc; i++) {
                
                if (owl_os_strcmp(argv[i], "-p") == 0) {
                        /* -p <port> */
			if (++i >= argc)
				goto err_parse;
                        
                        port = owl_os_atoi(argv[i]);
                        
                } else if (owl_os_strcmp(argv[i], "-n") == 0) {
                        
                        /* -n <bufcnt> */
			if (++i >= argc)
				goto err_parse;
                        
                        bufcnt = owl_os_atoi(argv[i]);

		} else if (owl_os_strcmp(argv[i], "-l") == 0) {
			/* -l <buflen> */
			if (++i >= argc)
				goto err_parse;
                        
                        buflen = owl_os_atoi(argv[i]);

		} else if (owl_os_strcmp(argv[i], "-t") == 0) {
                        /* transmit */
                        connect = 1;

                } else if (owl_os_strcmp(argv[i], "-r") == 0) {
                        /* receieve */
                        connect = 0;

		} else if (!ip_aton(argv[i], &ip)) {
                        /* ip address */
                        goto err_parse;
                }
        }

        if (connect && ip_addr_cmp(&ip, &ip_addr_any))
                goto err_parse;
        
        if (connect)
                err = ttcp_connect(&ip, port, bufcnt, buflen,
                                   progress_cb, done_cb, sh);
        else
                err = ttcp_listen(port, buflen, progress_cb, done_cb, sh);


        if (err < 0)
                owlsh_printf(sh, "%s %d (%s)\n\r", err >= 0 ? "ok" : "err", err,
                             owl_error(err));
        else if (connect)
                owlsh_printf(sh, "ttcp connect %s:%d bufcnt:%d buflen:%d "
                             "sockid:%d\n\r",
                             ip_ntoa(&ip), port, bufcnt, buflen, err);
        else 
                owlsh_printf(sh, "ttcp listen port:%d buflen:%d lsockid:%d\n\r",
                             port, buflen, err);

        return;
        
err_parse:
        owlsh_printf(sh, "usage: ttcp [-trpnl]\n\r");
        
}

const struct owlsh_cmd cmd_ttcp = { "ttcp", ttcp_f, "ttcp throughput test" };
