/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 *
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <stdint.h>
#include <errno.h>
#include <assert.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <linux/elf.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <arpa/inet.h>

#include <owl/util/mtd/archive.h>
#include <owl/util/mtd/bootfile.h>


#define MAX_ENTRIES 64

struct archive_entry {
        struct archive_hdr hdr;
        const char *fname; /* filename */
        char *buf;
        int len;
};

int verbose = 0;

#define MOD_ADLER 65521

#define arx_err(format, args...)                                       \
        fprintf(stderr, "-E- %s:%d %s: " format "\n", __FILE__, __LINE__, \
                      __func__, ## args)
#define arx_dbg(format, args...) do {                                   \
                if (verbose) {                                          \
                        printf("-D- %s:%d %s: " format "\n", __FILE__,  \
                               __LINE__, __func__, ## args);            \
                }                                                       \
        } while(0)



uint32_t 
arx_adler32(void *buf1, size_t len1, void *buf2, size_t len2)
{
        uint32_t a = 1;
        uint32_t b = 0;
        uint8_t *data = buf1;
        size_t index;

        for (index = 0; index < len1; ++index) {
                a = (a + data[index]) % MOD_ADLER;
                b = (b + a) % MOD_ADLER;
        }

        if (buf2) {
                data = buf2;
                for (index = 0; index < len2; ++index) {
                        a = (a + data[index]) % MOD_ADLER;
                        b = (b + a) % MOD_ADLER;
                }
        }

        return (b << 16) | a;
}


void
arx_usage(void)
{
        const char usage[] =
                "Usage: archive [OPTION...]\n"
                "  -c <archive>                         create\n"
                "  -a <file> <dev> <addr>               add file\n"
                "  -t <archive>                         list contents\n"
                "  -v                                   verbose\n\n"
                "Examples:\n"
                "  archive -c hwtest-spb860-r4566.arx\n"
                "          -a hwtest-spb860.img 1 524288\n"
                "          -a raw_flashimage.bin 1 655360\n"
                "  archive -t hwtest-spb860-r4566.arx\n\n";

        printf("%s", usage);
}

int
arx_create(char *filename, struct archive_entry *entry, int count, uint32_t tag)
{
        int tot_len = 0;
        FILE *file;
        int i, err;
        uint32_t adler32;

        for (i = 0; i < count; i++) {
                int fd2;
                struct stat stat_buf;

                if ((err = stat(entry[i].fname, &stat_buf)) < 0) {
                        arx_err("could not stat:%s, err:%d",
                                entry[i].fname, err);
                        return 1;
                }

                entry[i].len = stat_buf.st_size;
                arx_dbg("entry[%d].len:%d", i, entry[i].len);
                        
                if ((fd2 = open(entry[i].fname, O_RDONLY)) < 0) {
                        arx_err("failed to open:%s, err:%d",
                                entry[i].fname, fd2);
                        return 1;
                }
                
                if ((entry[i].buf = malloc(entry[i].len)) == NULL) {
                        arx_err("could not alloc buf for file:%s len:%d",
                                entry[i].fname, entry[i].len);
                        return 1;
                }

                if ((err = read(fd2, entry[i].buf,
                                entry[i].len)) != entry[i].len) {
                        arx_err("failed to read file:%s len:%d err:%d",
                                entry[i].fname, entry[i].len, err);
                        return 1;
                }

                close(fd2);
                tot_len += entry[i].len;
        }

        /* Update the archive hdr, including tot_len */
        for (i = 0; i < count; i++) {  
                entry[i].hdr.tot_len = htonl(tot_len);
                entry[i].hdr.addr = htonl(entry[i].hdr.addr);
                entry[i].hdr.len = htonl(entry[i].len);
                entry[i].hdr.pad[0] = 0;
                entry[i].hdr.pad[1] = 0;
                entry[i].hdr.magic = htonl(ARCHIVE_MAGIC);
                entry[i].hdr.flags = 0;
                entry[i].hdr.dev = BOOTFILE_DEV_EXTFLASH;
                entry[i].hdr.tag = htonl(tag);
                entry[i].hdr.adler32 = 0;

                /* Compute checksum */
                adler32 = arx_adler32(&entry[i].hdr, sizeof(entry[i].hdr),
                                      entry[i].buf, entry[i].len);
                entry[i].hdr.adler32 = htonl(adler32);

                tot_len -= entry[i].len;
        }
       
        if (tot_len) {
                arx_err("invalid length tot_len:%d", tot_len);
                return 1;
        }
        
        arx_dbg("opening outfile:%s", filename);        
        if ((file = fopen(filename, "w")) < 0) {
                arx_err("fail to open outfile:%s", filename);
                return 1;
        }
        
        for (i = 0; i < count; i++) {
                if ((err = fwrite(&entry[i].hdr, 1, sizeof(entry[i].hdr),
                                  file)) != sizeof(entry[i].hdr)) {
                        arx_err("fail to write hdr for file:%s err:%d",
                                entry[i].fname, err);
                        return 1;
                }

                arx_dbg("wrote %d bytes of hdr for file:%s",
                        err, entry[i].fname);
                
                if ((err = fwrite(entry[i].buf, 1,
                                  entry[i].len, file)) != entry[i].len) {
                        
                        arx_err("fail to write buf, "
                                "len:%d file:%s err:%d",
                                entry[i].len, entry[i].fname, err);
                        return 1;
                }

                arx_dbg("wrote %d bytes of data for file:%s",
                        err, entry[i].fname);
        }

        fclose(file);
        return 0;
}


int arx_list(char *filename)
{
        FILE *file;
        void *buf;
        int err;
        uint32_t adler32_file;
        uint32_t adler32_real;
        int len;
        
        arx_dbg("opening file:%s", filename);        
        if ((file = fopen(filename, "r")) < 0) {
                arx_err("fail to open file:%s", filename);
                return 1;
        }

        for (;;) {
                struct archive_hdr hdr;
                if ((err = fread(&hdr, sizeof(hdr), 1, file)) != 1) {
                        if (feof(file))
                                return 0;
                        
                        arx_err("fail to read hdr, err:%d", err);
                        return 1;
                }

                len = ntohl(hdr.len);

                if ((buf = malloc(len)) == NULL) {
                        arx_err("fail to alloc buf of len:%d", len);
                        return 1;
                }
                
                if ((err = fread(buf, len, 1, file)) != 1) {
                        arx_err("fail to read, err:%d", err);
                        return 1;
                }

                adler32_file = ntohl(hdr.adler32);
                hdr.adler32 = 0;
                adler32_real = arx_adler32(&hdr, sizeof(hdr), buf, len);
                free(buf);

                hdr.tot_len = ntohl(hdr.tot_len);
                hdr.addr = ntohl(hdr.addr);
                hdr.len = ntohl(hdr.len);
                hdr.magic = ntohl(hdr.magic);
                hdr.flags = ntohl(hdr.flags);
                hdr.tag = ntohl(hdr.tag);

                printf("magic:           0x%x\n", hdr.magic);
                printf("flags:           0x%x\n", hdr.flags);
                printf("tag:             0x%x\n", hdr.tag);
                printf("adler32 (file):  0x%x\n", adler32_file);
                printf("adler32 (real)   0x%x\n", adler32_real);
                printf("dev:             0x%x\n", hdr.dev);
                printf("addr:            0x%x\n", hdr.addr);
                printf("len:             %d\n", hdr.len);
                printf("tot_len:         %d\n", hdr.tot_len);
                printf("pad[0]:          %d\n", hdr.pad[0]);
                printf("pad[1]:          %d\n", hdr.pad[1]);
                printf("pad[2]:          %d\n\n", hdr.pad[2]);
        }

        return 0;
}

int
main(int argc, char *argv[])
{
        struct archive_entry entry[MAX_ENTRIES];
        char *filename;
        int count = 0; /* num entries */
        int err;
        char c;
        uint32_t tag = 0;
        
        /* operations */
        int create = 0;
        int list = 0;

        

        while ((c = getopt(argc, argv, "hc:vt:ab:")) != -1) {
                switch (c) {
                case 'a':
                        if (count > MAX_ENTRIES - 1) {
                                arx_err("max %d entries allowed",
                                        MAX_ENTRIES);
                                arx_usage();
                                return 1;
                        }
                        
                        if (optind + 3 > argc) {
                                arx_err("option -a requires 3 arguments");
                                return 1;
                        }
                        
                        entry[count].fname = argv[optind++];
                        entry[count].hdr.dev = strtol(argv[optind++],
                                                          NULL, 0);
                        entry[count].hdr.addr = strtol(argv[optind++],
                                                           NULL, 0);
                        count++;
                        break;
                        
                case 'v':
                        verbose++;
                        break;

                case 'b':
                        tag = strtol(optarg, NULL, 0);
                        break;

                case 'c':
                        create = 1;
                        filename = optarg;
                        break;
                        
                case 't':
                        list = 1;
                        filename = optarg;
                        break;

                case 'h':
                default:
                        arx_usage();
                        return 1;
                }
        }
        
        arx_dbg("create:%d list:%d verbose:%d entries:%d",
                create, list, verbose, count);

        if (create) {
                if ((err = arx_create(filename, entry, count, tag)))
                        return err;
                return arx_list(filename);
                
        } else if (list) {
                return arx_list(filename);
        }

        return 0;
}
