/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 *
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <stdint.h>
#include <errno.h>
#include <assert.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <linux/elf.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <arpa/inet.h>

#include <owl/util/mtd/bootfile.h>





#define MAX_DESC 64
struct bootfile_param {
        uint32_t src_dev;
        uint32_t dst_dev;
        uint32_t src_addr;
        uint32_t dst_addr;
        char *filename;
        char *buf;
        int len;
};
struct bootfile_ctx {
        int count;
        struct bootfile_param param[MAX_DESC];
};

int verbose = 0;

#define pr_err(format, args...)                                         \
        fprintf(stderr, "-E- %s:%d %s: " format "\n", __FILE__, __LINE__, \
                __func__, ## args)
#define pr_dbg(format, args...) do {                                    \
                if (verbose) {                                          \
                        printf("-D- %s:%d %s: " format "\n", __FILE__,  \
                               __LINE__, __func__, ## args);            \
                }                                                       \
        } while(0)



#ifndef LITTLE_ENDIAN
# define LITTLE_ENDIAN 0x0a0b0c0d /* LSB first */
#endif

#ifndef BIG_ENDIAN
# define BIG_ENDIAN 0x0d0c0b0a /* MSB first */
#endif
int byte_order = LITTLE_ENDIAN;

static uint16_t
bswap16(uint16_t x)
{
        if (byte_order == BIG_ENDIAN)
                return ((((x) << 8) & 0xff00) | (((x) >> 8) & 0xff));
        else
                return x;
}


static uint32_t
bswap32(uint32_t x)
{
        if (byte_order == BIG_ENDIAN)
                return  ((((x) << 24) & 0xff000000) | (((x) << 8) & 0xff0000) |
                         (((x) >> 8) & 0xff00) | (((x) >> 24) & 0xff));
        else
                return x;
}
        


#define MOD_ADLER 65521

static uint32_t 
adler32(void *buf, size_t len)
{
        uint32_t a = 1;
        uint32_t b = 0;
        uint8_t *data = buf;
        size_t index;


        for (index = 0; index < len; ++index) {
                a = (a + data[index]) % MOD_ADLER;
                b = (b + a) % MOD_ADLER;
        }

        return (b << 16) | a;
}






static void
usage(void)
{
        const char usage[] =
                "Usage: bootfile [OPTION...]\n"
                "  -c <bootfile>                        create\n"
                "  -a <file>\n"
                "     <src dev> <src addr>\n"
                "     <dst dev> <dst addr>              add file\n"
                "  -t <bootfile>                        list contents\n"
                "  -e <be|le>                           specify target endian\n"
                "  -v                                   verbose\n\n"
                "Examples:\n"
                "  bootfile -c bootfile.bin\n"
                "           -a hwtest-spb860.img\n"
                "              1 524288\n"
                "              0 0\n"
                "           -a raw_flashimage.bin\n"
                "              1 655360\n"
                "              1 0\n\n"
                "  bootfile -t bootfile.bin\n\n";

        printf("%s", usage);
}

int
bootfile_create(struct bootfile_ctx *priv, char *filename)
{
        FILE *file;
        int i, err;

        struct bootfile *b;
        int b_size = sizeof(b->hdr) + priv->count * sizeof(b->desc[0]);
        if ((b = calloc(1, b_size)) == NULL) {
                pr_err("could not alloc priv of size:%d", b_size);
                return 1;
        }
        
        for (i = 0; i < priv->count; i++) {
                struct bootfile_desc *desc = &b->desc[i];
                struct bootfile_desc_mtdcopy *mtdcopy = &b->desc[i].mtdcopy;
                struct bootfile_param *param = &priv->param[i];
                
                int fd2;
                struct stat stat_buf;

                if ((err = stat(param->filename, &stat_buf)) < 0) {
                        pr_err("could not stat:%s, err:%d",
                               param->filename, err);
                        return 1;
                }

                param->len = stat_buf.st_size;
                pr_dbg("desc:%d size:%d", i, param->len);
                
                if ((fd2 = open(param->filename, O_RDONLY)) < 0) {
                        pr_err("failed to open:%s, err:%d",
                               param->filename, fd2);
                        return 1;
                }
                
                if ((param->buf = malloc(param->len)) == NULL) {
                        pr_err("could not alloc buf for file:%s len:%d",
                               param->filename, param->len);
                        return 1;
                }

                if ((err = read(fd2, param->buf, param->len)) != param->len) {
                        pr_err("failed to read file:%s len:%d err:%d",
                               param->filename, param->len, err);
                        return 1;
                }

                close(fd2);

                desc->type = BOOTFILE_DESC_MTDCOPY;
                mtdcopy->src_dev = param->src_dev;
                mtdcopy->dst_dev = param->dst_dev;
                mtdcopy->src_addr = bswap32(param->src_addr);
                mtdcopy->dst_addr = bswap32(param->dst_addr);
                mtdcopy->size = bswap32(param->len);
                mtdcopy->adler32 = bswap32(adler32(param->buf, param->len));
        }

        b->hdr.count = bswap32(priv->count);
        b->hdr.magic = bswap32(BOOTFILE_MAGIC);
        b->hdr.adler32 = bswap32(adler32(&b->desc[0],
                                         priv->count * sizeof(b->desc[0])));
        
        
        pr_dbg("opening outfile:%s", filename);        
        if ((file = fopen(filename, "w")) < 0) {
                pr_err("fail to open outfile:%s", filename);
                return 1;
        }
        
        if ((err = fwrite(b, 1, b_size, file)) != b_size) {
                pr_err("fail to write bootfile, err:%d", err);
                return 1;
        }
        
        pr_dbg("wrote %d bytes", err);

        fclose(file);
        return 0;
}


int bootfile_list(char *filename)
{
        FILE *file;
        void *buf;
        int err;
        int len;
        struct stat stat_buf;
        struct bootfile *b;
        int b_size;
        int count;
        int i;
        
        if ((err = stat(filename, &stat_buf)) < 0) {
                pr_err("could not stat:%s, err:%d", filename, err);
                return 1;
        }

        b_size = stat_buf.st_size;
        pr_dbg("size:%d", b_size);
        
        if ((b = malloc(b_size)) == NULL) {
                pr_err("could not alloc bootfile");
                return 1;
        }
        
        pr_dbg("opening file:%s", filename);        
        if ((file = fopen(filename, "r")) < 0) {
                pr_err("fail to open file:%s", filename);
                return 1;
        }

        if ((err = fread(b, b_size, 1, file)) != 1) {
                pr_err("fail to read fat, err:%d", err);
                return 1;
        }
        fclose(file);
        
        b->hdr.magic = bswap32(b->hdr.magic);
        b->hdr.count = bswap32(b->hdr.count);
        b->hdr.adler32 = bswap32(b->hdr.adler32);

        count = (b_size - sizeof(b->hdr)) / sizeof(b->desc[0]);

        if (b->hdr.count != count) {
                pr_err("unexpected count:%d, expected:%d", b->hdr.count, count);
                return 1;
        }
        
        printf("magic:          0x%x\n", b->hdr.magic);
        printf("count:          0x%x\n", b->hdr.count);
        printf("adler32 (file): 0x%x\n", b->hdr.adler32);
        printf("adler32 (real): 0x%x\n\n",
               adler32((void *) &b->desc[0], count * sizeof(b->desc[0])));
        
        for (i = 0; i < count; i++) {
                struct bootfile_desc *desc = &b->desc[i];
                struct bootfile_desc_mtdcopy *mtdcopy = &b->desc[i].mtdcopy;

                mtdcopy->src_addr = bswap32(mtdcopy->src_addr);
                mtdcopy->dst_addr = bswap32(mtdcopy->dst_addr);
                mtdcopy->size = bswap32(mtdcopy->size);
                mtdcopy->adler32 = bswap32(mtdcopy->adler32);

                printf("type:            %d\n", desc->type);
                printf("src_dev:         %d\n", mtdcopy->src_dev);
                printf("src_addr:        0x%x\n", mtdcopy->src_addr);
                printf("dst_dev:         %d\n", mtdcopy->dst_dev);
                printf("dst_addr:        0x%x\n", mtdcopy->dst_addr);
                printf("size:            0x%x\n", mtdcopy->size);
                printf("adler32:         0x%x\n\n", mtdcopy->adler32);
        }

        return 0;
}

int
main(int argc, char *argv[])
{
        struct bootfile_ctx *priv;
        char *filename;
        int err;
        char c;
        
        /* operations */
        int create = 0;
        int list = 0;

        if ((priv = calloc(1, sizeof(*priv))) == NULL) {
                pr_err("could not alloc priv");
                return 1;
        }

        while ((c = getopt(argc, argv, "e:hc:vt:as")) != -1) {
                switch (c) {
                case 'a':
                        if (priv->count > MAX_DESC - 1) {
                                pr_err("max %d descs allowed", MAX_DESC);
                                usage();
                                return 1;
                        }

                                                
                        if (optind + 5 > argc) {
                                pr_err("option -a requires 5 arguments");
                                return 1;
                        }
                        
                        priv->param[priv->count].filename = argv[optind++];
                        priv->param[priv->count].src_dev =
                                strtol(argv[optind++], NULL, 0);
                        priv->param[priv->count].src_addr =
                                strtol(argv[optind++], NULL, 0);
                        priv->param[priv->count].dst_dev =
                                strtol(argv[optind++], NULL, 0);
                        priv->param[priv->count].dst_addr =
                                strtol(argv[optind++], NULL, 0);
                        priv->count++;
                        break;                     
                        
                case 'v':
                        verbose++;
                        break;

                case 'e':
                        if (!strcmp("be", optarg))
                                byte_order = BIG_ENDIAN;
                        else if (!strcmp("le", optarg))
                                byte_order = LITTLE_ENDIAN;
                        else {
                                usage();
                                return 1;
                        }

                        break;

                case 'c':
                        create = 1;
                        filename = optarg;
                        break;
                        
                case 't':
                        list = 1;
                        filename = optarg;
                        break;

                case 'h':
                default:
                        usage();
                        return 1;
                }
        }
        
        pr_dbg("create:%d list:%d verbose:%d count:%d",
               create, list, verbose, priv->count);

        if (create) {
                if ((err = bootfile_create(priv, filename)))
                        return err;
                
                return bootfile_list(filename);
                
        } else if (list) {
                return bootfile_list(filename);
        }

        return 0;
}
