/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 *
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <stdint.h>
#include <errno.h>
#include <assert.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <linux/elf.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <arpa/inet.h>

#include <owl/util/mtd/fixedfs.h>



struct fixedfs_desc {
        char *filename;
        char *buf;
        int len;
        char *fsname;
};

#define MAX_BLOCKS 64
#define MAX_FILES 64
struct fixedfs_ctx {
        int slot[MAX_BLOCKS];
        int block_count;
        
        struct fixedfs_desc desc[MAX_FILES];
        int desc_count;
};

int verbose = 0;

#define pr_err(format, args...)                                         \
        fprintf(stderr, "-E- %s:%d %s: " format "\n", __FILE__, __LINE__, \
                __func__, ## args)
#define pr_dbg(format, args...) do {                                    \
                if (verbose) {                                          \
                        printf("-D- %s:%d %s: " format "\n", __FILE__,  \
                               __LINE__, __func__, ## args);            \
                }                                                       \
        } while(0)



#ifndef LITTLE_ENDIAN
# define LITTLE_ENDIAN 0x0a0b0c0d /* LSB first */
#endif

#ifndef BIG_ENDIAN
# define BIG_ENDIAN 0x0d0c0b0a /* MSB first */
#endif
int byte_order = LITTLE_ENDIAN;

static uint16_t
bswap16(uint16_t x)
{
        if (byte_order == BIG_ENDIAN)
                return ((((x) << 8) & 0xff00) | (((x) >> 8) & 0xff));
        else
                return x;
}


static uint32_t
bswap32(uint32_t x)
{
        if (byte_order == BIG_ENDIAN)
                return  ((((x) << 24) & 0xff000000) | (((x) << 8) & 0xff0000) |
                         (((x) >> 8) & 0xff00) | (((x) >> 24) & 0xff));
        else
                return x;
}
        


#define MOD_ADLER 65521

static uint32_t 
adler32(void *buf, size_t len)
{
        uint32_t a = 1;
        uint32_t b = 0;
        uint8_t *data = buf;
        size_t index;


        for (index = 0; index < len; ++index) {
                a = (a + data[index]) % MOD_ADLER;
                b = (b + a) % MOD_ADLER;
        }

        return (b << 16) | a;
}






static void
usage(void)
{
        const char usage[] =
                "Usage: fixedfs [OPTION...]\n"
                "  -c <fs>                              create\n"
                "  -a <file> <name>                     add file\n"
                "  -t <fs>                              list contents\n"
                "  -s [ blocks ]                         specify blocks\n"
                "  -e <be|le>                           specify target endian\n"
                "  -v                                   verbose\n\n"
                "Examples:\n"
                "  fixedfs -c fs.bin\n"
                "          -s 256000 128000 64000\n"
                "          -a ar4100.bin firmware\n"
                "          -a bootfile.bin bootfile\n"
                "          -a web.bin web\n"
                "  fixedfs -t fs.bin\n\n";

        printf("%s", usage);
}

int
fixedfs_create(struct fixedfs_ctx *priv, char *filename)
{
        FILE *file;
        int i, err;

        struct fixedfs_fat *fat;
        int fat_size =
                sizeof(fat->hdr) + priv->block_count * sizeof(fat->blocks[0]);
        
        if (priv->desc_count > priv->block_count) {
                pr_err("invalid desc_count:%d, block_count:%d",
                       priv->desc_count, priv->block_count);
                return 1;
        }
        
        if ((fat = malloc(fat_size)) == NULL) {
                pr_err("could not alloc fat");
                return 1;
        }
        memset(fat, 0, fat_size);
        
        /* Fill in desc->buf and desc->len for priv->desc */
        for (i = 0; i < priv->desc_count; i++) {
                struct fixedfs_desc *desc = &priv->desc[i];
                
                int fd2;
                struct stat stat_buf;

                if (desc->filename == NULL) {
                        desc->len = 0;
                        continue;
                }
                if ((err = stat(desc->filename, &stat_buf)) < 0) {
                        pr_err("could not stat:%s, err:%d",
                               desc->filename, err);
                        return 1;
                }

                desc->len = stat_buf.st_size;
                pr_dbg("desc[%d].len:%d", i, desc->len);
                        
                if ((fd2 = open(desc->filename, O_RDONLY)) < 0) {
                        pr_err("failed to open:%s, err:%d",
                               desc->filename, fd2);
                        return 1;
                }
                
                if ((desc->buf = malloc(desc->len)) == NULL) {
                        pr_err("could not alloc buf for file:%s len:%d",
                               desc->filename, desc->len);
                        return 1;
                }

                if ((err = read(fd2, desc->buf, desc->len)) != desc->len) {
                        pr_err("failed to read file:%s len:%d err:%d",
                               desc->filename, desc->len, err);
                        return 1;
                }

                close(fd2);
        }

        for (i = 0; i < priv->block_count; i++) {
                struct fixedfs_block *f = &fat->blocks[i];
                if (i < priv->desc_count) {
                        struct fixedfs_desc *desc = &priv->desc[i];
                        strcpy(f->name, desc->fsname);
                        if (desc->len)
                                f->adler32 = bswap32(adler32(desc->buf, desc->len));
                        else
                                desc->len = priv->slot[i];
                        if (desc->len > priv->slot[i]) {
                                pr_err("slot:%d to small:%d, need:%d",
                                       i, priv->slot[i], desc->len);
                                return 1;
                        }
                        f->size = bswap32(desc->len);
                        /* desc->hdr.name alrady filled in */
                        f->flags = 0;
                        f->next = BLOCK_NONE;
                        f->prev = BLOCK_NONE;
                        
                } else {
                        f->flags = FIXEDFS_FREE;
                        f->next = BLOCK_NONE;
                        f->prev = BLOCK_NONE;
                }
        }

        fat->hdr.adler32 = bswap32(
                adler32((void *) &fat->blocks[0],
                        priv->block_count * sizeof(fat->blocks[0])));
        fat->hdr.magic = bswap32(FIXEDFS_MAGIC);
        fat->hdr.blkcnt = bswap32(priv->block_count);
        
        pr_dbg("opening outfile:%s", filename);        
        if ((file = fopen(filename, "w")) < 0) {
                pr_err("fail to open outfile:%s", filename);
                return 1;
        }
        
        if ((err = fwrite(fat, 1, fat_size, file)) != fat_size) {
                pr_err("fail to write fat, err:%d", err);
                return 1;
        }
        
        pr_dbg("wrote %d bytes of fat", err);

        fclose(file);
        return 0;
}


int fixedfs_list(char *filename)
{
        FILE *file;
        void *buf;
        int err;
        uint32_t adler32_file;
        uint32_t adler32_real;
        int len;
        int fat_size;
        struct stat stat_buf;
        struct fixedfs_fat *fat;
        int count;
        int i;
        
        if ((err = stat(filename, &stat_buf)) < 0) {
                pr_err("could not stat:%s, err:%d", filename, err);
                return 1;
        }

        fat_size = stat_buf.st_size;
        pr_dbg("fat_size:%d", fat_size);
        
        if ((fat = malloc(fat_size)) == NULL) {
                pr_err("could not alloc fat");
                return 1;
        }
        
        pr_dbg("opening file:%s", filename);        
        if ((file = fopen(filename, "r")) < 0) {
                pr_err("fail to open file:%s", filename);
                return 1;
        }

        if ((err = fread(fat, fat_size, 1, file)) != 1) {
                pr_err("fail to read fat, err:%d", err);
                return 1;
        }
        fclose(file);
        
        count = (fat_size - sizeof(fat->hdr)) / sizeof(struct fixedfs_block);

        fat->hdr.magic = bswap32(fat->hdr.magic);
        fat->hdr.adler32 = bswap32(fat->hdr.adler32);
        fat->hdr.blkcnt = bswap32(fat->hdr.blkcnt);
        
        printf("magic:          0x%x\n", fat->hdr.magic);
        printf("count:          %d\n", fat->hdr.blkcnt);
        printf("adler32 (file): 0x%x\n", fat->hdr.adler32);
        printf("adler32 (real): 0x%x\n\n",
               adler32((void *) &fat->blocks[0],
                       count * sizeof(struct fixedfs_block)));
        
        for (i = 0; i < count; i++) {
                struct fixedfs_block *f = &fat->blocks[i];

                f->size = bswap32(f->size);
                f->adler32 = bswap32(f->adler32);

                printf("slot:            %d\n", i);
                printf("name:            %s\n", f->name);
                printf("flags:           0x%x\n", f->flags);
                printf("size:            0x%x\n", f->size);
                printf("adler32:         0x%x\n\n", f->adler32);
                printf("next:            %d\n", f->next);
                printf("prev:            %d\n", f->prev);
        }

        return 0;
}

int
main(int argc, char *argv[])
{
        struct fixedfs_ctx *priv;
        char *filename;
        int err;
        char c;
        
        /* operations */
        int create = 0;
        int list = 0;

        if ((priv = calloc(1, sizeof(*priv))) == NULL) {
                pr_err("could not alloc priv");
                return 1;
        }

        while ((c = getopt(argc, argv, "e:hc:vt:abs")) != -1) {
                switch (c) {
                case 'a':
                        if (priv->desc_count > MAX_FILES - 1) {
                                pr_err("max %d files allowed", MAX_FILES);
                                usage();
                                return 1;
                        }
                        
                        if (optind + 2 > argc) {
                                pr_err("option -a requires 2 arguments");
                                return 1;
                        }
                        
                        priv->desc[priv->desc_count].filename = argv[optind++];
                        priv->desc[priv->desc_count].fsname = argv[optind++];
                        priv->desc_count++;
                        break;
                case 'b':
                        if (priv->desc_count > MAX_FILES - 1) {
                                pr_err("max %d files allowed", MAX_FILES);
                                usage();
                                return 1;
                        }
                        
                        if (optind + 1 > argc) {
                                pr_err("option -b requires 1 argument");
                                return 1;
                        }
                        
                        priv->desc[priv->desc_count].fsname = argv[optind++];
                        priv->desc_count++;
                        break;

                case 's':

                        for (priv->block_count = 0;
                             priv->block_count < MAX_BLOCKS;
                             priv->block_count++) {
                                int size;
                                char *endptr = NULL;
                                if (optind + 1 > argc)
                                        break;
                                
                                size = strtol(argv[optind], &endptr, 0);
                                if (endptr == argv[optind]) {
                                        break;
                                }
                                pr_dbg("found slot of size:%d", size);
                                priv->slot[priv->block_count] = size;
                                optind++;
                        }
                        
                        if (priv->block_count == MAX_BLOCKS) {
                                pr_err("max %d blocks allowed", MAX_BLOCKS);
                                return 1;
                        }

                        break;
                        
                case 'e':
                        if (!strcmp("be", optarg))
                                byte_order = BIG_ENDIAN;
                        else if (!strcmp("le", optarg))
                                byte_order = LITTLE_ENDIAN;
                        else {
                                usage();
                                return 1;
                        }

                        break;
    
                case 'v':
                        verbose++;
                        break;

                case 'c':
                        create = 1;
                        filename = optarg;
                        break;
                        
                case 't':
                        list = 1;
                        filename = optarg;
                        break;

                case 'h':
                default:
                        usage();
                        return 1;
                }
        }
        
        pr_dbg("create:%d list:%d verbose:%d blocks:%d files:%d",
               create, list, verbose, priv->block_count, priv->desc_count);

        if (create) {
                if ((err = fixedfs_create(priv, filename)))
                        return err;
                return fixedfs_list(filename);
                
        } else if (list) {
                return fixedfs_list(filename);
        }

        return 0;
}
