/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#define OWL_TRACE OWL_TRACE_ERR
#define OWL_FILE "wlp_api"

#include <owl/core/owl_debug.h>
#include <owl/core/owl_os.h>

#if WLP_CONFIG_STACK_OFFLOAD
# include <owl/core/owl_list.h>
# include <wlp_inet.h>
# include <wlp_ipaddr.h>
# include <wlp_socket.h>
#endif /* WLP_CONFIG_STACK_OFFLOAD */

#include <wlp_api.h>
#include <wlp_ethernet.h>

#include <hdg205/wlp_proto.h>
#include <hdg205/wlp_hpi.h>
#include <hdg205/wlp_hdg205.h>

/* TX Window must be more than TXWIN_DATA to send a data frame, this is to
   avoid that a running data transer prevents us from sending mgmt commands */
#define TXWIN_DATA 2

struct wlp_ctx {

        void (*loop_f)(void *ctx, void *data, int len);
        void *loop_ctx;

        void (*heartbeat_f)(void *ctx);
        void *heartbeat_ctx;
        
        void (*yield_f)(void *ctx);
        void *yield_ctx;

        void (*scan_f)(void *ctx, struct wlp_network *net);
        void *scan_ctx;
        
        void (*link_f)(void *ctx, int link);
        void *link_ctx;

        void (*linkctx_f)(void *ctx, const void *linkctx, int len);
        void *linkctx_ctx;
        
        void (*chan_f)(void *ctx, int status);
        void *chan_ctx;

        int (*rx_f)(void *ctx, int len);
        void *rx_ctx;

        int (*program_f)(void *ctx, void *buf, int offset, int len);
        void *program_ctx;

        void (*get_rxstats_f)(void *ctx, struct wlp_rfstats *rxstats);
        void *get_rxstats_ctx;

        void (*error_f)(void *ctx, int err);
        void *error_ctx;

#define FLAG_INIT   (1 << 0)
#define FLAG_LINK   (1 << 1)
#define FLAG_POLL   (1 << 2)
        uint8_t flags;

        /* Last rx and tx rate */
        uint8_t txrate;
        uint8_t rxrate;

        /* TX window size */
        int txwin;
        int txlen;
        int txpad;
        
        volatile int irq; /* accessed from interrupt context */

        struct wlp_mac_addr hwaddr;
        struct wlp_network net;

        int mode;

        /* Firmware upgrade */
        char version[16];
        int fwlen;
        int fwpos;

#if WLP_CONFIG_STACK_OFFLOAD
        struct ip_addr ip;
        struct ip_addr netmask;
        struct ip_addr gw;
        struct ip_addr dns;

        int dhcp;
        
        void (*addr_f)(void *ctx, const struct ip_addr *addr);
        void *addr_ctx;

        void (*lookup_f)(void *ctx, const struct ip_addr *ip);
        void *lookup_ctx;

        struct list_head sockets;
#endif /* WLP_CONFIG_STACK_OFFLOAD */
};

#if WLP_CONFIG_STACK_OFFLOAD
struct wlp_sock { 
	int16_t id; 
        uint16_t sndbuf;
        uint8_t type;
        
        struct list_head list;
        
        void (*listen_f)(void *ctx, int lsockid, int sockid,
                         const struct ip_addr *ip, uint16_t port); 
	void (*conn_f)(void *ctx, int sockid,
                       const struct ip_addr *addr, uint16_t port,
                       int connected); 
        void (*recv_f)(void *ctx, int sockid,
                       const struct ip_addr *addr, uint16_t port,
                       void *data, int len);
	void (*sent_f)(void *ctx, int sockid, int len);
	void *recv_ctx; /* callback context for rx */ 
	void *sent_ctx;
	void *ctx; /* callback context for listen and conn */ 
};

const struct ip_addr ip_addr_any;
#endif /* WLP_CONFIG_STACK_OFFLOAD */ 


static struct wlp_ctx *priv = NULL;

static void process_init_cfm(struct wlp_msg *msg);
static void process_loopback_cfm(struct wlp_msg *msg);
static void process_heartbeat_cfm(struct wlp_msg *msg);
static void process_archive_cfm(struct wlp_msg *msg);
static void process_data_cfm(struct wlp_msg *msg);
static void process_set_channel_ind(struct wlp_msg *msg);
static void process_get_linkctx_ind(struct wlp_msg *msg);
static void process_scan_ind(struct wlp_msg *msg);
static void process_linkup_ind(struct wlp_msg *msg);
static void process_linkdown_ind(struct wlp_msg *msg);
static void process_get_rxstats_ind(struct wlp_msg *msg);

static int archive_send(void);
static int archive_read(void *buf, int offset, int len);
static void archive_done(int err);

static struct wlp_msg *msg_alloc(int id, int len);
static void msg_free(struct wlp_msg *msg);
static int msg_send(struct wlp_msg *msg);

static void notify_error(int err);

#if defined(OWL_DEBUG) && OWL_TRACE >= OWL_TRACE_INFO
static const char *dbg_ssid(char *ssid, int ssid_len);
#endif

#if WLP_CONFIG_STACK_OFFLOAD
static void process_set_ipaddr_ind(struct wlp_msg *msg);
static void process_get_hostbyname_ind(struct wlp_msg *msg);
static void process_connect_ind(struct wlp_msg *msg);
static void process_listen_ind(struct wlp_msg *msg);
static void process_recv_ind(struct wlp_msg *msg);
static void process_sent_ind(struct wlp_msg *msg);

static int sock_nextid(void);
static struct wlp_sock *sock_alloc(int type, int sockid);
static void sock_free(struct wlp_sock *sock);
static struct wlp_sock *sock_find(int sockid);
#endif /* WLP_CONFIG_STACK_OFFLOAD */

/* wlp_hdg205.h --------------------------------------------------------- */

int
_wlp_init(int (*exchange_cb)(void *ctx, const void *tx, void *rx, int len),
          void (*select_cb)(void *ctx, int select),
          void (*cmd_cb)(void *ctx, uint8_t idx, uint32_t arg, uint32_t *rsp,
                         const void *data, int len),
          int (*poll_cb)(void *ctx),
          void *ctx,
          int flags)
{
        int err;
        struct wlp_msg *msg;
        struct wlp_init_req *req;
        
        owl_dbg("flags:%d exchange_cb:%p ctx:%p", flags, exchange_cb, ctx);

        if ((priv = owl_os_zalloc(sizeof(*priv))) == NULL)
                return WLP_ERR_MEM;
   
        if (owl_bit(flags, WLP_HPI_POLL))
                owl_bitset(priv->flags, FLAG_POLL);

        if ((err = hpi_init(exchange_cb, select_cb, cmd_cb, poll_cb, 
                            ctx, flags)))
                return err;
        
        /* Initialize to 1 to be able to send INIT REQ, will be set to the
           actual size when INIT CFM is recvd */
        priv->txwin = 1;
        priv->txrate = WLP_RATE_UNKNOWN;
        priv->rxrate = WLP_RATE_UNKNOWN;
        
        if ((msg = msg_alloc(
                     WLP_TYPE_REQ | WLP_ID_INIT, sizeof(*req))) == NULL)
                return WLP_ERR_MEM;
                        
        req = &msg->u.req.u.init;
#if WLP_CONFIG_STACK_OFFLOAD
        req->flags = HDG205_STACK_OFFLOAD;
        list_init(&priv->sockets);
#endif /* WLP_CONFIG_STACK_OFFLOAD */
        
#if WLP_CONFIG_HPI == WLP_CONFIG_HPI_SDIO
        owl_bitset(req->flags, HDG205_SDIO);
#endif
        if ((err = msg_send(msg)))
                return err;

        while (!owl_bit(priv->flags, FLAG_INIT))
                wlp_poll();

        return 0;
}

void
_wlp_set_error_cb(void (*error_cb)(void *ctx, int err), void *ctx)
{
        priv->error_f = error_cb;
        priv->error_ctx = ctx;
}

int
_wlp_program(int (*program_cb)(void *ctx, void *buf, int offset, int len),
             void *ctx)
{
        int err;
        char version[16];

        if (program_cb == NULL)
                return 0;

        priv->program_f = program_cb;
        priv->program_ctx = ctx;
        priv->fwpos = 0;
        
        if ((err = archive_read(
                     version, WLP_FIRMWARE_VERSION_OFFSET,
                     sizeof(version))) < 0) {
                owl_err("fail to get fw version, err:%d", err);
                return err;
        }
        
        if (!owl_os_strcmp(version, priv->version)) {
                owl_info("firmware versions match, version:%s", version);
                return WLP_ERR_IGNORED;
        }
        
        owl_info("program %s->%s", priv->version, version);

        if ((err = archive_read(
                     &priv->fwlen, WLP_FIRMWARE_LENGTH_OFFSET,
                     sizeof(priv->fwlen))) < 0) {
                owl_err("fail to get fw length, err:%d", err);
                return err;
        }

        return archive_send();
}

int
_wlp_reset(void)
{
        struct wlp_msg *msg;
        if ((msg = msg_alloc(WLP_TYPE_REQ | WLP_ID_RESET, 0)) == NULL)
                return WLP_ERR_MEM;
        
        return msg_send(msg);
}

int
_wlp_get_version(char *driver, int driver_len,
                 char *firmware, int firmware_len)
{
        if (priv == NULL)
                return WLP_ERR_STATE;

        /* Updated manually */
        owl_os_strncpy(driver, "1.3.2", driver_len - 1);
        owl_os_strncpy(firmware, priv->version, firmware_len - 1);
        return 0;
}

void
_wlp_shutdown(void)
{
        if (owl_bit(priv->flags, FLAG_LINK)) {
                if (priv->link_f)
                        priv->link_f(priv->link_ctx, 0);
        }
        
        hpi_shutdown();
        owl_os_free(priv);
}

/* wlp_mgmt.h --------------------------------------------------------- */

void
_wlp_set_yield_cb(void (*yield_cb)(void *ctx), void *ctx)
{
        owl_dbg("yield_cb:%p ctx:%p", yield_cb, ctx);
        
        priv->yield_f = yield_cb;
        priv->yield_ctx = ctx;
}

void
_wlp_irq(void)
{
        if (priv == NULL)
                return;
        
        priv->irq = 1;
}

int
_wlp_poll(void)
{
        if (priv == NULL)
                return WLP_ERR_STATE;
        
        if (priv->flags & FLAG_POLL)
                hpi_poll();
        
        if (!priv->irq)
                return 0;
        
        if (priv->rx_f) {
                
		return priv->rx_f(priv->rx_ctx, 0);
                
        } else {
                void *buf;
                int err = 0;
                
                if ((buf = owl_os_alloc(WLP_MAX_PKT_LEN)) == NULL) {
			owl_err("could not alloc rxbuf");
                        return WLP_ERR_MEM;
                }
		
                if ((err = wlp_rx(buf, WLP_MAX_PKT_LEN)) < 0)
                        goto out;
                
                if ((err = wlp_process_rx(
			     buf, WLP_MAX_PKT_LEN, NULL, NULL, NULL)) < 0)
                        goto out;

        out:
                owl_os_free(buf);
                return err;
        }
}

int
_wlp_heartbeat(void (*cb)(void *ctx), void *ctx)
{
        struct wlp_msg *msg;

        owl_dbg("enter");

        priv->heartbeat_f = cb;
        priv->heartbeat_ctx = ctx;
        
        if ((msg = msg_alloc(WLP_TYPE_REQ | WLP_ID_HEARTBEAT, 0)) == NULL)
                return WLP_ERR_MEM;

        return msg_send(msg);
}

/* wlp_concurrent.h ------------------------------------------------------- */

int
_wlp_get_iface(void)
{
        return 0;
}

int
_wlp_set_iface(int iface)
{
        if (iface >= WLP_IFACES)
                return WLP_ERR_NOTSUPP;

        return 0; /* Previous interface */
}

/* wlp_loopback.h --------------------------------------------------------- */

void
_wlp_set_loopback_cb(void (*loop_cb)(void *ctx, void *data, int len), void *ctx)
{
        priv->loop_f = loop_cb;
        priv->loop_ctx = ctx;
}

int
_wlp_loopback(void *data, int len)
{
        int err;
        struct wlp_msg *msg;
        struct wlp_loopback_req *req;
        
        if ((msg = msg_alloc(
                     WLP_TYPE_REQ | WLP_ID_LOOPBACK,
                     sizeof(*req) + len)) == NULL)
                return WLP_ERR_MEM;

        req = &msg->u.req.u.loopback;
        req->len = owl_htons(len);
        owl_os_memcpy(req->data, data, len);
        
        if ((err = msg_send(msg)))
                return err;

        return 0;
}

/* wlp_link.h --------------------------------------------------------- */

void
_wlp_set_link_cb(void (*link_cb)(void *ctx, int link), void *ctx)
{
        owl_dbg("link_cb:%p ctx:%p", link_cb, ctx);
        
        priv->link_f = link_cb;
        priv->link_ctx = ctx;
}

int
_wlp_scan(struct wlp_ssid *ssid,
             void (*scan_cb)(void *ctx, struct wlp_network *net), void *ctx)
{
        int err;
        struct wlp_msg *msg;
        struct wlp_scan_req *req;
        
        if ((msg = msg_alloc(WLP_TYPE_REQ | WLP_ID_SCAN, sizeof(*req))) == NULL)
                return WLP_ERR_MEM;

        req = &msg->u.req.u.scan;
        if (ssid)
                owl_os_memcpy(&req->ssid, ssid, sizeof(req->ssid));
        
        if ((err = msg_send(msg)))
                return err;

        priv->scan_f = scan_cb;
        priv->scan_ctx = ctx;
        return 0;
}

int
_wlp_linkup(struct wlp_ssid *ssid, const char *key, enum wlp_auth_mode auth)
{
        struct wlp_msg *msg;
        struct wlp_linkup_req *req;

        owl_dbg("ssid:%s key:%s auth:%d", ssid, key ? key : "null", auth);
        
        if (key && auth != WLP_AUTH_WEP) {
                int len = owl_os_strlen(key);
                if (len < 8 || len > 63) {
                        return WLP_ERR_PARAM;
                }
        }
        
        if ((msg = msg_alloc(
                     WLP_TYPE_REQ | WLP_ID_LINKUP, sizeof(*req))) == NULL)
                return WLP_ERR_MEM;

        req = &msg->u.req.u.linkup;
        
        if (ssid)
                owl_os_memcpy(&req->ssid, ssid, sizeof(req->ssid));
        if (key)
                owl_os_strncpy(req->key, key, sizeof(req->key) - 1);

        req->auth = (uint8_t) auth;

        return msg_send(msg);
}

int
_wlp_linkdown(void)
{
        struct wlp_msg *msg;

        owl_dbg("enter");
        
        if ((msg = msg_alloc(WLP_TYPE_REQ | WLP_ID_LINKDOWN, 0)) == NULL)
                return WLP_ERR_MEM;

        return msg_send(msg);
}

int
_wlp_get_hwaddr(struct wlp_mac_addr *hwaddr)
{
        owl_dbg("hwaddr:%p", hwaddr);
        
        if (hwaddr == NULL)
                return WLP_ERR_PARAM;
        
        owl_os_memcpy(hwaddr, &priv->hwaddr, sizeof(*hwaddr));
        owl_dbg("hwaddr:%02x%02x%02x%02x%02x%02x",
                hwaddr->octet[0], hwaddr->octet[1],
                hwaddr->octet[2], hwaddr->octet[3],
                hwaddr->octet[4], hwaddr->octet[5]);

        return 0;
}


int
_wlp_get_network(struct wlp_network *net)
{
        owl_dbg("net:%p", net);

        if (owl_bit(priv->flags, FLAG_LINK)) {
		if (net != NULL)
			owl_os_memcpy(net, &priv->net, sizeof(*net));
                return 0;
        } else {
                return WLP_ERR_CONN;
        }
}

int
_wlp_get_rate(uint8_t *tx, uint8_t *rx)
{
        if (tx != NULL)
                *tx = priv->txrate;

        if (rx != NULL)
                *rx = priv->rxrate;

        return 0;
}

int
_wlp_set_wps(int enable)
{
        struct wlp_msg *msg;
        struct wlp_set_wps_req *req;

        owl_dbg("enable:%d", enable);
                
        if ((msg = msg_alloc(
                     WLP_TYPE_REQ | WLP_ID_SET_WPS, sizeof(*req))) == NULL)
                return WLP_ERR_MEM;

        req = &msg->u.req.u.set_wps;
        req->enable = enable;
        
        return msg_send(msg);
}

int
_wlp_get_link_context(void (*linkctx_cb)(void *ctx, const void *linkctx,
                                         int len),
		      void *ctx)
{
        int err;
        struct wlp_msg *msg;

        if (priv->linkctx_f)
                return WLP_ERR_PARAM;
                
        if ((msg = msg_alloc(WLP_TYPE_REQ | WLP_ID_GET_LINKCTX, 0)) == NULL)
                return WLP_ERR_MEM;
        
        if ((err = msg_send(msg)))
                return err;

        priv->linkctx_f = linkctx_cb;
        priv->linkctx_ctx = ctx;
        return 0;
}

int
_wlp_set_link_context(const void *linkctx, int len)
{
        struct wlp_msg *msg;
        struct wlp_set_linkctx_req *req;

        if ((msg = msg_alloc(
                     WLP_TYPE_REQ | WLP_ID_SET_LINKCTX,
                     sizeof(*req) + len)) == NULL)
                return WLP_ERR_MEM;

        req = &msg->u.req.u.set_linkctx;
        req->len = owl_htons(len);
        owl_os_memcpy(req->data, linkctx, len);
        
        return msg_send(msg);
}

int
_wlp_set_mode(uint8_t mode)
{
        struct wlp_msg *msg;
        struct wlp_set_mode_req *req;

        owl_dbg("mode:%d", mode);

        if (owl_bit(priv->flags, FLAG_LINK))
                return WLP_ERR_STATE;
        
        if ((msg = msg_alloc(
                     WLP_TYPE_REQ | WLP_ID_SET_MODE, sizeof(*req))) == NULL)
                return WLP_ERR_MEM;
        
        req = &msg->u.req.u.set_mode;
        req->mode = mode;
        priv->mode = mode;
        return msg_send(msg);
}

int
_wlp_get_mode(uint8_t *mode)
{
        owl_dbg("mode:%p", mode);

        *mode = priv->mode;
        return 0;
}

int
_wlp_set_baudrate(uint32_t baudrate)
{
        struct wlp_msg *msg;
        struct wlp_set_baudrate_req *req;

        owl_dbg("baudrate:%d", baudrate);

        if ((msg = msg_alloc(
                     WLP_TYPE_REQ | WLP_ID_SET_BAUDRATE, sizeof(*req))) == NULL)
                return WLP_ERR_MEM;

        req = &msg->u.req.u.set_baudrate;
        req->baudrate = owl_htonl(baudrate);
        
        return msg_send(msg);
}

int
_wlp_set_channel(uint8_t chan,
                 void (*chan_cb)(void *ctx, int status), void *ctx)
{
        int err;
        struct wlp_msg *msg;
        struct wlp_set_channel_req *req;

        owl_dbg("chan:%d", chan);

        if ((msg = msg_alloc(
                     WLP_TYPE_REQ | WLP_ID_SET_CHANNEL, sizeof(*req))) == NULL)
                return WLP_ERR_MEM;

        req = &msg->u.req.u.set_channel;
        req->channel = chan;
        
        if ((err = msg_send(msg)))
                return err;

        priv->chan_f = chan_cb;
        priv->chan_ctx = ctx;
        return 0;
}

int
_wlp_conf_ap(int hidden, int beacon_intvl, int inact_tmo,
                const char *country)
{
        struct wlp_msg *msg;
        struct wlp_conf_ap_req *req;

        if ((msg = msg_alloc(
                     WLP_TYPE_REQ | WLP_ID_CONF_AP, sizeof(*req))) == NULL)
                return WLP_ERR_MEM;

        req = &msg->u.req.u.conf_ap;
        req->hidden = hidden;
        req->beacon_intvl = owl_htons(beacon_intvl);
        req->inact_tmo = owl_htons(inact_tmo);
        owl_os_strcpy(req->country, country);

        return msg_send(msg);
}

int
_wlp_set_ps(int enable)
{
        struct wlp_msg *msg;
        struct wlp_set_ps_req *req;

        if ((msg = msg_alloc(
                     WLP_TYPE_REQ | WLP_ID_SET_PS, sizeof(*req))) == NULL)
                return WLP_ERR_MEM;

        req = &msg->u.req.u.set_ps;
        req->enable = enable;

        return msg_send(msg);
}

int
_wlp_conf_ps(uint8_t ps_poll, uint32_t traffic_tmo, uint32_t ps_delay,
	     uint8_t rx_all_dtim, uint16_t listen_intvl)
{
        struct wlp_msg *msg;
        struct wlp_conf_ps_req *req;
        
        owl_dbg("ps_delay:%d traffic_timout:%d "
                "listen_intvl:%d ps_poll:%d rx_all_dtim:%d)",
                ps_delay, traffic_tmo, listen_intvl, ps_poll,
                rx_all_dtim);

        if ((msg = msg_alloc(
                     WLP_TYPE_REQ | WLP_ID_CONF_PS, sizeof(*req))) == NULL)
                return WLP_ERR_MEM;
        
        req = &msg->u.req.u.conf_ps;        
        req->ps_delay = owl_htonl(ps_delay);
        req->traffic_tmo = owl_htonl(traffic_tmo);
        req->listen_intvl = owl_htons(listen_intvl);
        req->ps_poll = ps_poll;
        req->rx_all_dtim = rx_all_dtim;

        return msg_send(msg);
}

/* wlp_rftest.h ------------------------------------------------------------ */

int
_wlp_start_tx(uint32_t count, uint16_t intvl, uint16_t size, uint8_t rate)
{
        struct wlp_msg *msg;
        struct wlp_start_tx_req *req;

        if ((msg = msg_alloc(
                     WLP_TYPE_REQ | WLP_ID_START_TX, sizeof(*req))) == NULL)
                return WLP_ERR_MEM;
        
        req = &msg->u.req.u.start_tx;        
        req->count = owl_htonl(count);
        req->intvl = owl_htons(intvl);
        req->size = owl_htons(size);
        req->rate = rate;

        return msg_send(msg);
}

int
_wlp_stop_tx(void)
{
        struct wlp_msg *msg;
        if ((msg = msg_alloc(WLP_TYPE_REQ | WLP_ID_STOP_TX, 0)) == NULL)
                return WLP_ERR_MEM;
        
        return msg_send(msg);
}

int
_wlp_get_rxstats(uint8_t rate,
                 void (*rxstats_cb)(void *ctx, struct wlp_rfstats *rxstats),
                 void *ctx)
{
        struct wlp_msg *msg;
        struct wlp_get_rxstats_req *req;

        if ((msg = msg_alloc(
                     WLP_TYPE_REQ | WLP_ID_GET_RXSTATS, sizeof(*req))) == NULL)
                return WLP_ERR_MEM;

        req = &msg->u.req.u.get_rxstats;        
        req->rate = rate;
        priv->get_rxstats_f = rxstats_cb;
        priv->get_rxstats_ctx = ctx;

        return msg_send(msg);
}

int
_wlp_clr_rxstats(void)
{
        struct wlp_msg *msg;
        if ((msg = msg_alloc(WLP_TYPE_REQ | WLP_ID_CLR_RXSTATS, 0)) == NULL)
                return WLP_ERR_MEM;
        
        return msg_send(msg);
}

/* wlp_mcast.h ----------------------------------------------------------- */

int
_wlp_add_mcast(struct wlp_mac_addr *hwaddr)
{
        struct wlp_msg *msg;
        struct wlp_add_mcast_req *req;

        if ((msg = msg_alloc(
                     WLP_TYPE_REQ | WLP_ID_ADD_MCAST, sizeof(*req))) == NULL)
                return WLP_ERR_MEM;

        req = &msg->u.req.u.add_mcast;
        
        if (hwaddr != NULL) {
                
                owl_info("hwaddr:%02x:%02x:%02x:%02x:%02x:%02x",
                         hwaddr->octet[0], hwaddr->octet[1],
                         hwaddr->octet[2], hwaddr->octet[3],
                         hwaddr->octet[4], hwaddr->octet[5]);

                owl_os_memcpy(&req->hwaddr, hwaddr, sizeof(req->hwaddr));
                
        } else {

                owl_info("hwaddr:NULL");
                owl_bitset(req->flags, HDG205_MCAST_ALL);

        }

        return msg_send(msg);
}

int
_wlp_del_mcast(struct wlp_mac_addr *hwaddr)
{
        struct wlp_msg *msg;
        struct wlp_del_mcast_req *req;

        if ((msg = msg_alloc(
                     WLP_TYPE_REQ | WLP_ID_DEL_MCAST, sizeof(*req))) == NULL)
                return WLP_ERR_MEM;
        
        req = &msg->u.req.u.del_mcast;        

        if (hwaddr != NULL) {
                
                owl_info("hwaddr:%02x:%02x:%02x:%02x:%02x:%02x",
                         hwaddr->octet[0], hwaddr->octet[1],
                         hwaddr->octet[2], hwaddr->octet[3],
                         hwaddr->octet[4], hwaddr->octet[5]);

                owl_os_memcpy(&req->hwaddr, hwaddr, sizeof(req->hwaddr));
                
        } else {

                owl_info("hwaddr:NULL");
                owl_bitset(req->flags, HDG205_MCAST_ALL);

        }

        return msg_send(msg);
}

/* wlp_ethernet.h --------------------------------------------------------- */

void
_wlp_set_rx_cb(int (*rx_cb)(void *ctx, int len), void *ctx)
{
        priv->rx_f = rx_cb;
        priv->rx_ctx = ctx;
}

int
_wlp_process_tx(void *ethhdr, int len, int totlen, void *p,
                uint16_t vlanid_prio)
{
        struct wlp_msg *msg = p;
        struct wlp_data_req *req = &msg->u.req.u.data;
        
        owl_vdbg("ethhdr:%p len:%d totlen:%d p:%p txwin:%d",
                 ethhdr, len, totlen, p, priv->txwin);

        /* Make sure we don't use up the tx window with data frames to be
           able to handle management requests such as wlp_scan() etc during
           data transfers. */
        if (priv->txwin <= TXWIN_DATA)
                return WLP_ERR_BUSY;

	/* Caller can set p to NULL to test if transmission is allowed or not */
        if (p == NULL)
		return 0;

        /* totlen is the size of the ethernet frame, NOT including any
           wlp header. We need to remember the number of padding bytes to
           be able to add them to the last call to wlp_tx() */
        priv->txlen = hpi_process_tx(
                p, totlen + sizeof(msg->hdr) + sizeof(*req));

        /* hpi_process_tx() will return the total number of bytes that will
           be transmitted after HPI padding is added. To compute the number of
           padding bytes, we will subtract the "payload" part here. */ 
        priv->txpad = priv->txlen - totlen - sizeof(msg->hdr) - sizeof(*req);
        
        /* Don't apply padding to frames from the IP stack, since they might
           be fragmented anyway. We can not ensure that each fragment is
           size-aligned so that must be handled on a lower level, if required */
        msg->hdr.pad = 0;
        msg->hdr.len = totlen + sizeof(*req);
        msg->hdr.id = WLP_TYPE_REQ | WLP_ID_DATA;
        msg->hdr.res = 0;

        /* This field was introduced in firmware 1.1.8 so make sure to upgrade
           before any data frames are sent! */
        req->vlanid_prio = owl_htons(vlanid_prio);
        
        owl_vdbg("msg:%p type:%d id:%d len:%d pad:%d res:%d vlanid:%d prio:%d",
                 msg, WLP_TYPE(msg->hdr.id), WLP_ID(msg->hdr.id),
                 msg->hdr.len, msg->hdr.pad, msg->hdr.res,
                 req->vlanid_prio >> 3, req->vlanid_prio & 0x7);
        
        msg->hdr.len = owl_htons(msg->hdr.len);
        msg->hdr.id = owl_htons(msg->hdr.id);
        
        priv->txwin--;
        return 0;
}

int
_wlp_tx(void *data, int len)
{
        int err;

        owl_vdbg("data:%p len:%d priv->txlen:%d priv->txpad:%d",
                 data, len, priv->txlen, priv->txpad);
        owl_hexdump("TX", data, len);

        priv->txlen -= len;
        /* Only padding remains? Then transfer the padding after this last
           chunk so we don't required any more calls to wlp_tx() */
        if (priv->txlen == priv->txpad)
                len += priv->txpad;
        
        if ((err = hpi_tx(data, len)) < 0) {
                owl_err("tx failed, err:%d", err);
                notify_error(err);
                return err;
        }

        return 0;
}

int
_wlp_rx(void *data, int len)
{
        struct wlp_msg *msg = data;
        int err;
        
        owl_vdbg("data:%p len:%d", data, len);

        priv->irq = 0;

        if ((err = hpi_rx(data, len)) < 0) {
                notify_error(err);
                return err;
        }

        msg->hdr.len = owl_ntohs(msg->hdr.len);
        msg->hdr.id = owl_ntohs(msg->hdr.id);

        return 0;
}

int
_wlp_process_rx(void *data, int len, void **stripped, int *stripped_len,
                uint16_t *vlanid_prio)
{       
        struct wlp_msg *msg = data;

        owl_vdbg("msg:%p type:%d id:%d len:%d pad:%d res:%d",
                 msg, WLP_TYPE(msg->hdr.id), WLP_ID(msg->hdr.id), msg->hdr.len,
                 msg->hdr.pad, msg->hdr.res);
        
        msg->hdr.len -= msg->hdr.pad;

        if (WLP_TYPE(msg->hdr.id) == WLP_TYPE_CFM)
                priv->txwin++;       
        
        switch (msg->hdr.id) {
        case WLP_TYPE_CFM | WLP_ID_INIT:
                process_init_cfm(msg);
                return 0;
        case WLP_TYPE_CFM | WLP_ID_LOOPBACK:
                process_loopback_cfm(msg);
                return 0;
        case WLP_TYPE_CFM | WLP_ID_HEARTBEAT:
                process_heartbeat_cfm(msg);
                return 0;
        case WLP_TYPE_CFM | WLP_ID_ARCHIVE:
                process_archive_cfm(msg);
                return 0;
        case WLP_TYPE_CFM | WLP_ID_DATA:
                process_data_cfm(msg);
                return 0;
        case WLP_TYPE_IND | WLP_ID_SET_CHANNEL:
                process_set_channel_ind(msg);
                return 0;
        case WLP_TYPE_IND | WLP_ID_GET_LINKCTX:
                process_get_linkctx_ind(msg);
                return 0;
        case WLP_TYPE_IND | WLP_ID_LINKUP:
                process_linkup_ind(msg);
                return 0;
        case WLP_TYPE_IND | WLP_ID_LINKDOWN:
                process_linkdown_ind(msg);
                return 0;
        case WLP_TYPE_IND | WLP_ID_SCAN:
                process_scan_ind(msg);
                return 0;
        case WLP_TYPE_IND | WLP_ID_DATA: {                
                struct wlp_data_ind *ind;
                uint16_t vlanid_prio_snr;
                
                ind = &msg->u.ind.u.data;
                
                if (stripped)
                        *stripped = ind->data;
                if (stripped_len)
                        *stripped_len = msg->hdr.len - sizeof(*ind);

                owl_vdbg("*stripped:%p *stripped_len:%d",
                         stripped ? *stripped : NULL,
                         stripped_len ? *stripped_len : 0);
                        
                vlanid_prio_snr = owl_ntohs(ind->vlanid_prio);

                if (ind->rssi_snr) {
                        /* The highest bit in vlanid_prio is used to indicate
                           whether we have an updated SNR value in the rssi_snr
                           field. If '1', then the field rssi_snr contains a
                           new snr value, otherwise it contains a rssi value. */

                        if (owl_bit(vlanid_prio_snr, HDG205_SNR_UPDATED))
                                priv->net.snr = ind->rssi_snr;
                        else
                                priv->net.rssi = ind->rssi_snr;
                }

                if (ind->rate != WLP_RATE_UNKNOWN)
                        priv->rxrate = ind->rate;

                if (vlanid_prio != NULL) {
                        owl_bitclr(vlanid_prio_snr, HDG205_SNR_UPDATED);
                        *vlanid_prio = vlanid_prio_snr;
                }
                    
                return *stripped_len; /* To be processed by IP stack */
        }
        case WLP_TYPE_IND | WLP_ID_GET_RXSTATS:
                process_get_rxstats_ind(msg);
                return 0;

#if WLP_CONFIG_STACK_OFFLOAD
        case WLP_TYPE_IND | WLP_ID_SET_IPADDR:
                process_set_ipaddr_ind(msg);
                return 0;
        case WLP_TYPE_IND | WLP_ID_GET_HOSTBYNAME:
                process_get_hostbyname_ind(msg);
                return 0;
        case WLP_TYPE_IND | WLP_ID_CONNECT:
                process_connect_ind(msg);
                return 0;
        case WLP_TYPE_IND | WLP_ID_LISTEN:
                process_listen_ind(msg);
                return 0;
        case WLP_TYPE_IND | WLP_ID_RECV:
                process_recv_ind(msg);
                return 0;
        case WLP_TYPE_IND | WLP_ID_SENT:
                process_sent_ind(msg);
                return 0;
#endif /* WLP_CONFIG_STACK_OFFLOAD */
        };
        
        return 0;
}

#if WLP_CONFIG_STACK_OFFLOAD

/* wlp_ipaddr.h ----------------------------------------------------------- */

void
_wlp_set_ipaddr_cb(void (*addr_cb)(void *ctx, const struct ip_addr *addr),
                   void *ctx)
{
        priv->addr_f = addr_cb;
        priv->addr_ctx = ctx;
}

int
_wlp_set_ipaddr(const struct ip_addr *ip, const struct ip_addr *netmask,
                const struct ip_addr *gw, const struct ip_addr *dns)
{
        struct wlp_msg *msg;
        struct wlp_set_ipaddr_req *req;
        int err;
        
        owl_dbg("ip:%d.%d.%d.%d mask:%d.%d.%d.%d "
                "gw:%d.%d.%d.%d dns:%d.%d.%d.%d",
                ip4_addr1(ip), ip4_addr2(ip),
                ip4_addr3(ip), ip4_addr4(ip),
                ip4_addr1(netmask), ip4_addr2(netmask),
                ip4_addr3(netmask), ip4_addr4(netmask),
                ip4_addr1(gw), ip4_addr2(gw),
                ip4_addr3(gw), ip4_addr4(gw),
                ip4_addr1(dns), ip4_addr2(dns),
                ip4_addr3(dns), ip4_addr4(dns));

        if ((msg = msg_alloc(
                     WLP_TYPE_REQ | WLP_ID_SET_IPADDR, sizeof(*req))) == NULL)
                return WLP_ERR_MEM;
        
        req = &msg->u.req.u.set_ipaddr;

        req->ip = ip4_addr_get_u32(ip);
        req->netmask = ip4_addr_get_u32(netmask);
        req->gw = ip4_addr_get_u32(gw);
        req->dns = ip4_addr_get_u32(dns);
        
        if ((err = msg_send(msg)))
                return err;

        ip_addr_set(&priv->ip, ip);
        ip_addr_set(&priv->netmask, netmask);
        ip_addr_set(&priv->gw, gw);
        ip_addr_set(&priv->dns, dns);

        return 0;
}

int
_wlp_get_ipaddr(struct ip_addr *ip, struct ip_addr *netmask,
                struct ip_addr *gw, struct ip_addr *dns)
{
        if (ip != NULL)
                ip_addr_set(ip, &priv->ip);
        
        if (gw != NULL)
                ip_addr_set(gw, &priv->gw);
        
        if (netmask != NULL)
                ip_addr_set(netmask, &priv->netmask);
        
        if (dns != NULL)
                ip_addr_set(dns, &priv->dns);

        return 0;
}

int
_wlp_set_dhcp(int enable)
{
        struct wlp_msg *msg;
        struct wlp_set_dhcp_req *req;

        owl_dbg("enable:%d", enable);
        
        if ((msg = msg_alloc(
                     WLP_TYPE_REQ | WLP_ID_SET_DHCP, sizeof(*req))) == NULL)
                return WLP_ERR_MEM;

        req = &msg->u.req.u.set_dhcp;        
        req->enable = enable;

        priv->dhcp = enable;
        
        return msg_send(msg);
}

int
_wlp_get_dhcp(int *enabled)
{
        owl_dbg("enabled:%p", enabled);

        if (enabled)
                *enabled = priv->dhcp;

        return 0;
}

static void
process_set_ipaddr_ind(struct wlp_msg *msg)
{
        struct wlp_set_ipaddr_ind *ind = &msg->u.ind.u.set_ipaddr;
        
        ip4_addr_set_u32(&priv->ip, ind->ip);
        ip4_addr_set_u32(&priv->netmask, ind->netmask);
        ip4_addr_set_u32(&priv->gw, ind->gw);
        ip4_addr_set_u32(&priv->dns, ind->dns);

        owl_dbg("ip:%d.%d.%d.%d mask:%d.%d.%d.%d "
                "gw:%d.%d.%d.%d dns:%d.%d.%d.%d",
                ip4_addr1(&priv->ip), ip4_addr2(&priv->ip),
                ip4_addr3(&priv->ip), ip4_addr4(&priv->ip),
                ip4_addr1(&priv->netmask), ip4_addr2(&priv->netmask),
                ip4_addr3(&priv->netmask), ip4_addr4(&priv->netmask),
                ip4_addr1(&priv->gw), ip4_addr2(&priv->gw),
                ip4_addr3(&priv->gw), ip4_addr4(&priv->gw),
                ip4_addr1(&priv->dns), ip4_addr2(&priv->dns),
                ip4_addr3(&priv->dns), ip4_addr4(&priv->dns));

        if (priv->addr_f)
                priv->addr_f(priv->addr_ctx, &priv->ip);
}

/* wlp_socket.h ----------------------------------------------------------- */

int
_wlp_socket(int type, int proto)
{
        struct wlp_msg *msg;
        struct wlp_sock *sock;
        struct wlp_socket_req *req;
        int err;
        
	owl_dbg("type:%d proto:%d", type, proto);

        if ((msg = msg_alloc(
                     WLP_TYPE_REQ | WLP_ID_SOCKET, sizeof(*req))) == NULL)
                return WLP_ERR_MEM;
         
        if ((sock = sock_alloc(type, 0)) == NULL) {
                msg_free(msg);
                return WLP_ERR_MEM;
        }

        req = &msg->u.req.u.socket;
        req->type = type;
        req->proto = proto;
        req->xid = owl_htons(sock->id);

        if ((err = msg_send(msg))) {
                sock_free(sock);
                return err;
        }

        return sock->id;
}

int
_wlp_set_conn_cb(int sockid,
                 void (*conn_cb)(void *ctx, int sockid,
                                 const struct ip_addr *addr, uint16_t port,
                                 int connected),
                 void *ctx)
{
        struct wlp_sock *sock;

	owl_dbg("sockid:%d conn_cb:%p ctx:%p", sockid, conn_cb, ctx);

        if ((sock = sock_find(sockid)) == NULL) {
                return WLP_ERR_PARAM;
        }

        sock->conn_f = conn_cb;
        sock->ctx = ctx;
        
        return 0;
}

int
_wlp_set_recv_cb(int sockid,
                 void (*recv_cb)(void *ctx, int sockid,
                                 const struct ip_addr *addr, uint16_t port,
                                 void *data, int len),
                 void *ctx)
{
        struct wlp_sock *sock;

        owl_dbg("sockid:%d recv_cb:%p ctx:%p", sockid, recv_cb, ctx);
        
        if ((sock = sock_find(sockid)) == NULL) {
                return WLP_ERR_PARAM;
        }

        sock->recv_f = recv_cb;
        sock->recv_ctx = ctx;
        
        return 0;
}

int
_wlp_set_sent_cb(int sockid,
                 void (*sent_cb)(void *ctx, int sockid, int len),
                 void *ctx)
{
        struct wlp_sock *sock;

        owl_dbg("sockid:%d conn_cb:%p ctx:%p", sockid, sent_cb, ctx);
        
        if ((sock = sock_find(sockid)) == NULL) {
                return WLP_ERR_PARAM;
        }

        sock->sent_f = sent_cb;
        sock->sent_ctx = ctx;
        
        return 0;
}

int
_wlp_bind(int sockid, const struct ip_addr *ip, uint16_t port)
{
        struct wlp_msg *msg;
        struct wlp_bind_req *req;
 
	owl_dbg("sockid:%d ip:%s port:%d",
                sockid, ip ? ip_ntoa(ip) : "NULL", port); 

        if ((msg = msg_alloc(WLP_TYPE_REQ | WLP_ID_BIND, sizeof(*req))) == NULL)
                return WLP_ERR_MEM;

        req = &msg->u.req.u.bind;
        req->xid = owl_htons(sockid); 
        req->ip = ip4_addr_get_u32(ip);
	req->port = owl_htons(port); 

        return msg_send(msg);
}

int
_wlp_listen(int sockid, int backlog,
            void (*listen_cb)(void *ctx, int lsockid, int sockid,
                              const struct ip_addr *ip, uint16_t port),
            void *ctx)
{
        struct wlp_msg *msg;
        struct wlp_listen_req *req;
        struct wlp_sock *sock;
        int err;
        
	owl_dbg("sockid:%d backlog:%d listen_cb:%p ctx:%p", 
		sockid, backlog, listen_cb, ctx); 

	if ((sock = sock_find(sockid)) == NULL) 
		return WLP_ERR_PARAM; 

        if ((msg = msg_alloc(
                     WLP_TYPE_REQ | WLP_ID_LISTEN, sizeof(*req))) == NULL)
                return WLP_ERR_MEM;

        req = &msg->u.req.u.listen;
	req->xid = owl_htons(sockid);
	req->backlog = backlog;

        if ((err = msg_send(msg)))
                return err;

	sock->listen_f = listen_cb; 
	sock->ctx = ctx; 
        return 0;
}

int
_wlp_connect(int sockid, const struct ip_addr *ip, uint16_t port)
{
        struct wlp_msg *msg;
        struct wlp_connect_req *req;
        struct wlp_sock *sock;
 
	owl_dbg("sockid:%d ip:%s port:%d", 
		sockid, ip ? ip_ntoa(ip) : "NULL", port); 
        		 
	if ((sock = sock_find(sockid)) == NULL) 
		return WLP_ERR_PARAM; 

        if ((msg = msg_alloc(
                     WLP_TYPE_REQ | WLP_ID_CONNECT, sizeof(*req))) == NULL)
                return WLP_ERR_MEM;

        req = &msg->u.req.u.connect;
	req->xid = owl_htons(sockid); 
	req->ip = ip4_addr_get_u32(ip);
	req->port = owl_htons(port);

        return msg_send(msg);
}

int
_wlp_accept(int sockid)
{
        struct wlp_msg *msg;
        struct wlp_accept_req *req;
        
	owl_dbg("sockid:%d", sockid); 
        		 
	if (sock_find(sockid) == NULL) 
		return WLP_ERR_PARAM; 

        if ((msg = msg_alloc(
                     WLP_TYPE_REQ | WLP_ID_ACCEPT, sizeof(*req))) == NULL)
                return WLP_ERR_MEM;

        req = &msg->u.req.u.accept;
	req->xid = owl_htons(sockid); 

        return msg_send(msg);
}

int
_wlp_close(int sockid)
{
        struct wlp_msg *msg;
        struct wlp_close_req *req;
        struct wlp_sock *sock;
        int err;
        
	owl_dbg("sockid:%d", sockid); 

        if (sockid < 0)
                return WLP_ERR_IGNORED;
        
	if ((sock = sock_find(sockid)) == NULL) 
		return WLP_ERR_PARAM; 

        if ((msg = msg_alloc(
                     WLP_TYPE_REQ | WLP_ID_CLOSE, sizeof(*req))) == NULL)
                return WLP_ERR_MEM;

        req = &msg->u.req.u.close;
	req->xid = owl_htons(sockid); 

        if ((err = msg_send(msg)))
                return err;

        sock_free(sock);
        return 0;
}

int
_wlp_recved(int sockid, int16_t len)
{
        struct wlp_msg *msg;
        struct wlp_recved_req *req;
        struct wlp_sock *sock;
        
	owl_dbg("sockid:%d len:%d", sockid, len); 

	if ((sock = sock_find(sockid)) == NULL) 
		return WLP_ERR_PARAM; 

        if ((msg = msg_alloc(
                     WLP_TYPE_REQ | WLP_ID_RECVED, sizeof(*req))) == NULL)
                return WLP_ERR_MEM;

        req = &msg->u.req.u.recved;
	req->xid = owl_htons(sockid); 
	req->len = owl_htons(len); 

        return msg_send(msg);
}

int
_wlp_send(int sockid, const char *buf, int16_t len)
{
        struct wlp_msg *msg;
        struct wlp_send_req *req;
        struct wlp_sock *sock;
        int err;
        
	owl_dbg("sockid:%d buf:%p len:%d", sockid, buf, len); 

	if ((sock = sock_find(sockid)) == NULL) 
		return WLP_ERR_PARAM; 

        switch (sock->type) {
        case WLP_SOCK_STREAM:
                if (!sock->sndbuf)
                        return 0;
                if (len > sock->sndbuf)
                        len = sock->sndbuf;
                if (len > 1460)
                        len = 1460;
                break;
                
        case WLP_SOCK_DGRAM:
        case WLP_SOCK_RAW:
                if (len > 1460)
                        return WLP_ERR_SIZE;
                break;
        };
        
        if ((msg = msg_alloc(
                     WLP_TYPE_REQ | WLP_ID_SEND, sizeof(*req) + len)) == NULL)
                return WLP_ERR_MEM;

        req = &msg->u.req.u.send;
	req->xid = owl_htons(sockid);
        req->len = owl_htons(len);
        owl_os_memcpy(req->data, buf, len);

        if ((err = msg_send(msg)) < 0) {
                return err;
        } else {
                if (sock->type == WLP_SOCK_STREAM)
                        sock->sndbuf -= len;
                return len;
        }
}

int
_wlp_sendto(int sockid, const char *buf, int16_t len,
            const struct ip_addr *ip, uint16_t port)
{
        struct wlp_msg *msg;
        struct wlp_sendto_req *req;
        struct wlp_sock *sock;
        int err;

        owl_dbg("sockid:%d buf:%p len:%d ip:%s port:%d", 
		sockid, buf, len, ip ? ip_ntoa(ip) : "NULL", port); 
        
	if ((sock = sock_find(sockid)) == NULL) 
		return WLP_ERR_PARAM; 

        if (len > 1460)
                return WLP_ERR_SIZE;

        if ((msg = msg_alloc(
                     WLP_TYPE_REQ | WLP_ID_SENDTO, sizeof(*req) + len)) == NULL)
                return WLP_ERR_MEM;

        req = &msg->u.req.u.sendto;
	req->xid = owl_htons(sockid);
        req->len = owl_htons(len);
        req->ip = ip4_addr_get_u32(ip);
	req->port = owl_htons(port); 
        owl_os_memcpy(req->data, buf, len);

        if ((err = msg_send(msg)) < 0)
                return err;
        else
                return len;
}

int
_wlp_get_peeraddr(int sockid, struct ip_addr *peer)
{
        return WLP_ERR_NOTSUPP;
}

int
_wlp_get_hostbyname(const char *host,
                    void (*lookup_cb)(void *ctx, const struct ip_addr *ip),
                    void *ctx)
{
        struct wlp_msg *msg;
        struct wlp_get_hostbyname_req *req;
        int err;

        owl_dbg("host:%s", host);
        
	if (host == NULL) 
		return WLP_ERR_PARAM; 
	 
	if (owl_os_strlen(host) > WLP_HOSTNAME_MAX_LENGTH) 
		return WLP_ERR_PARAM; 

	if (priv->lookup_f) 
		return WLP_ERR_BUSY;

        if ((msg = msg_alloc(
                     WLP_TYPE_REQ | WLP_ID_GET_HOSTBYNAME,
                     sizeof(*req))) == NULL)
                return WLP_ERR_MEM;

        req = &msg->u.req.u.get_hostbyname;
	owl_os_strcpy(req->host, host); 

        if ((err = msg_send(msg)))
                return err;
        
        priv->lookup_f = lookup_cb; 
        priv->lookup_ctx = ctx;
        return 0;
}
 
 
static void
process_get_hostbyname_ind(struct wlp_msg *msg)
{
        struct wlp_get_hostbyname_ind *ind = &msg->u.ind.u.get_hostbyname;
        struct ip_addr ip;

        ip4_addr_set_u32(&ip, ind->ip);

        owl_dbg("ip:%d.%d.%d.%d",
                ip4_addr1(&ip), ip4_addr2(&ip), ip4_addr3(&ip), ip4_addr4(&ip));
        
        if (priv->lookup_f) {
                priv->lookup_f(priv->lookup_ctx, &ip);
                priv->lookup_f = NULL;
        }
}

static void
process_connect_ind(struct wlp_msg *msg)
{
        struct wlp_connect_ind *ind = &msg->u.ind.u.connect;
        struct wlp_sock *sock;
        struct ip_addr ip;
        
        ind->xid = owl_ntohs(ind->xid);
        ind->port = owl_ntohs(ind->port);

        ip4_addr_set_u32(&ip, ind->ip);
        
        owl_dbg("id:%d ip:%s port:%d connected:%d",
                ind->xid, ip_ntoa(&ip), ind->port, ind->connected);
        
        if ((sock = sock_find(ind->xid)) == NULL) {
                owl_err("invalid socket");
                return;
        }

        if (ind->connected <= 0)
                _wlp_close(ind->xid);
        
        if (sock->conn_f) {
                sock->conn_f(sock->ctx, ind->xid,
                             &ip, ind->port, ind->connected);
        }
}

static void
process_listen_ind(struct wlp_msg *msg)
{
        struct wlp_listen_ind *ind = &msg->u.ind.u.listen;
        struct wlp_sock *lsock;
        struct ip_addr ip;
        
        ind->lxid = owl_ntohs(ind->lxid);
        ind->xid = owl_ntohs(ind->xid);
        ind->port = owl_ntohs(ind->port);

        ip4_addr_set_u32(&ip, ind->ip);
        
        owl_dbg("lsockid:%d sockid:%d ip:%s port:%d",
                ind->lxid, ind->xid, ip_ntoa(&ip), ind->port);
        
        if ((lsock = sock_find(ind->lxid)) == NULL) {
                owl_err("invalid socket");
                return;
        }
         
        if (sock_alloc(WLP_SOCK_STREAM, ind->xid) == NULL) {
                owl_err("could not alloc socket");

                /* XXX fatal */
                return;
        }

        if (lsock->listen_f) {
                lsock->listen_f(lsock->ctx, ind->lxid, ind->xid, &ip, ind->port);
        }
}

static void
process_recv_ind(struct wlp_msg *msg)
{
        struct wlp_recv_ind *ind = &msg->u.ind.u.recv;
        struct wlp_sock *sock;
        struct ip_addr ip;
        
        ind->xid = owl_ntohs(ind->xid);
        ind->len = owl_ntohs(ind->len);
        ind->port = owl_ntohs(ind->port);

        ip4_addr_set_u32(&ip, ind->ip);
        
        owl_dbg("id:%d len:%d ip:%s port:%d",
                ind->xid, ind->len, ip_ntoa(&ip), ind->port);
        
        if ((sock = sock_find(ind->xid)) == NULL) {
                owl_err("invalid socket");
                return;
        }

        if (sock->recv_f) {
                sock->recv_f(sock->recv_ctx, ind->xid, &ip, ind->port,
                             ind->data, ind->len);
        }
}

static void
process_sent_ind(struct wlp_msg *msg)
{
        struct wlp_sent_ind *ind = &msg->u.ind.u.sent;
        struct wlp_sock *sock;
        
        ind->xid = owl_ntohs(ind->xid);
        ind->len = owl_ntohs(ind->len);

        owl_dbg("id:%d len:%d", ind->xid, ind->len);
        
        if ((sock = sock_find(ind->xid)) == NULL) {
                owl_err("invalid socket");
                return;
        }

        sock->sndbuf = ind->len;
        if (sock->sent_f)
                sock->sent_f(sock->sent_ctx, ind->xid, ind->len);
}

static int
sock_nextid(void)
{
        int sockid;

        /* zero is not a valid socket id */
        
        for (sockid = 2; sockid < INT16_MAX; sockid += 2) {
                if (sock_find(sockid) == NULL)
                        return sockid;
        }

        owl_err("out of socket ids");
        return WLP_ERR_MEM;
}

static struct wlp_sock * 
sock_alloc(int type, int sockid) 
{ 
	struct wlp_sock *sock;
        
        if ((sock = owl_os_zalloc(sizeof(*sock))) == NULL) {
                owl_err("fail to alloc sock");
                return NULL;
        }
        
        if (sockid)
                sock->id = sockid;
        else
                sock->id = sock_nextid();
        sock->sndbuf = 1460;
        sock->type = type;
        list_add_tail(&sock->list, &priv->sockets);

	return sock; 
}         
 
 
static void 
sock_free(struct wlp_sock *sock) 
{ 
	owl_dbg("sock:%p sock->id:%d", sock, sock ? sock->id : 0); 
        list_del(&sock->list);
        owl_os_free(sock); 
} 

static struct wlp_sock * 
sock_find(int sockid)
{
        struct list_head *pos;

        list_for_each(pos, &priv->sockets) {
                struct wlp_sock *iter = list_entry(pos, struct wlp_sock, list);

                if (iter->id == sockid)
                        return iter;
        }
 
	return NULL; 
} 

/* wlp_service.h ---------------------------------------------------------- */

int
_wlp_set_dhcpd(int enable)
{
        struct wlp_msg *msg;
        struct wlp_set_dhcpd_req *req;

        owl_dbg("enable:%d", enable);
        
        if ((msg = msg_alloc(
                     WLP_TYPE_REQ | WLP_ID_SET_DHCPD, sizeof(*req))) == NULL)
                return WLP_ERR_MEM;

        req = &msg->u.req.u.set_dhcpd;        
        req->enable = enable;

        priv->dhcp = enable;
        
        return msg_send(msg);
}

#endif /* WLP_CONFIG_STACK_OFFLOAD */

/* FIRMWARE --------------------------------------------------------------- */

/* Process the fw upgrade procedure */
static int
archive_send(void)
{
        struct wlp_msg *msg;
        struct wlp_archive_req *req;
        int err;
        int len = 1024;
        int remain = priv->fwlen - priv->fwpos;

        if (!(len = owl_min(len, remain))) {
                owl_err("no more data");
                return WLP_ERR_SIZE;
        }

        owl_dbg("len:%d pos:%d", len, priv->fwpos);

        if ((msg = msg_alloc(
                     WLP_TYPE_REQ | WLP_ID_ARCHIVE,
                     sizeof(*req) + len)) == NULL) {
                return WLP_ERR_MEM;
        }
        
        req = &msg->u.req.u.archive;

        if ((err = archive_read(
                     req->data, WLP_FIRMWARE_DATA_OFFSET + priv->fwpos,
                     len)) < 0) {
                owl_err("fail to read firmware, pos:%d err%d",
                        priv->fwpos, err);
                msg_free(msg);
                return err;
        }

        priv->fwpos += err;
        req->len = owl_htons(err);
        remain = priv->fwlen - priv->fwpos;
        if (remain)
                owl_bitset(req->flags, HDG205_ARCHIVE_MORE);
        req->flags = owl_htons(req->flags);
        return msg_send(msg);
}

/* Read a chunk of data from the fw_read_cb */
static int
archive_read(void *buf, int offset, int len)
{
        int err;
        int pos = 0;
        int remain = len;
        while (remain > 0) {
                if ((err = priv->program_f(
                             priv->program_ctx,
                             (uint8_t *) buf + pos, offset, remain)) < 0) {
                        wlp_shutdown();
                        return err;
                }
                
                pos += err;
                remain -= err;
        }

        return len;
}

static void
archive_done(int err)
{
        priv->program_f(priv->program_ctx, NULL, 0, err);
}

/* TX PATH ---------------------------------------------------------------- */

static struct wlp_msg *
msg_alloc(int id, int len)
{
        struct wlp_msg *msg;
        int pad;
        int totlen;
        
        pad = wlp_pad(sizeof(msg->hdr) + len);
        totlen = sizeof(msg->hdr) + len + pad;
        
        if ((msg = owl_os_zalloc(totlen)) == NULL)
                return NULL;

        msg->hdr.pad = pad;
        msg->hdr.len = totlen - sizeof(msg->hdr);
        msg->hdr.id = id;
        msg->hdr.res = 0;

        return msg;
}

static void
msg_free(struct wlp_msg *msg)
{
        owl_os_free(msg);
}

static int
msg_send(struct wlp_msg *msg)
{
        int err;
        int len;

        switch (WLP_ID(msg->hdr.id)) {
        case WLP_ID_DATA:
#if WLP_CONFIG_STACK_OFFLOAD
        case WLP_ID_SEND:
        case WLP_ID_SENDTO:
#endif /* WLP_CONFIG_STACK_OFFLOAD */
                if (priv->txwin <= TXWIN_DATA) {
                        owl_info("txwin:%d need:%d", priv->txwin, TXWIN_DATA);
                        msg_free(msg);
                        return WLP_ERR_BUSY;
                }
                break;
                
        default:
                if (!priv->txwin) {
                        owl_info("txwin is 0");
                        msg_free(msg);
                        return WLP_ERR_BUSY;
                }
                break;
        };
        
        /* hdr->len does not include header size so add it now */
        len = msg->hdr.len + sizeof(msg->hdr);

        owl_dbg("msg:%p type:%d id:%d len:%d pad:%d res:%d",
                msg, WLP_TYPE(msg->hdr.id), WLP_ID(msg->hdr.id), msg->hdr.len,
                msg->hdr.pad, msg->hdr.res);

        msg->hdr.len = owl_htons(msg->hdr.len);
        msg->hdr.id = owl_htons(msg->hdr.id);
        
        owl_hexdump("TXMSG", (char *) msg, len);

        hpi_process_tx(msg, len);
        if ((err = hpi_tx(msg, len)) < 0) {
                /* This is fatal */
                owl_err("fail to send, err:%d", err);
        }
        
        priv->txwin--;
        msg_free(msg);
        
        if (err < 0) {
                notify_error(err);
                return err;
        }

        return 0;
}

/* RX PATH ----------------------------------------------------------------- */

static void
process_init_cfm(struct wlp_msg *msg)
{
        struct wlp_init_cfm *cfm = &msg->u.cfm.u.init;
        
        if (owl_os_strcmp(cfm->version, "1.1.4")) {
		/* Fix endian unless we are using the 1.1.4 firmware which
		   for some reason did not deliver these fields in the
		   proper endian */
		cfm->bufcnt = owl_ntohs(cfm->bufcnt);
		cfm->buflen = owl_ntohs(cfm->buflen);
        }

        owl_info("CFM INIT bufcnt:%d buflen:%d caps:0x%04x "
                 "hwaddr:%02x:%02x:%02x:%02x:%02x:%02x",
                 cfm->bufcnt, cfm->buflen, cfm->caps,
                 cfm->hwaddr.octet[0], cfm->hwaddr.octet[1], 
                 cfm->hwaddr.octet[2], cfm->hwaddr.octet[3], 
                 cfm->hwaddr.octet[4], cfm->hwaddr.octet[5]);
        owl_info("version:%s build:%s date:%s target:%s",
                 cfm->version, cfm->build, cfm->date, cfm->target);
        priv->flags |= FLAG_INIT;
        priv->txwin = cfm->bufcnt / ((WLP_MAX_PKT_LEN / cfm->buflen) + 1);
        owl_info("txwin:%d", priv->txwin);
        owl_os_memcpy(&priv->hwaddr, &cfm->hwaddr, sizeof(priv->hwaddr));
        owl_os_strncpy(priv->version, cfm->version, sizeof(priv->version));
}

static void
process_loopback_cfm(struct wlp_msg *msg)
{
        struct wlp_loopback_cfm *cfm = &msg->u.cfm.u.loopback;

        cfm->len = owl_ntohs(cfm->len);
        owl_dbg("CFM LOOPBACK len:%d", cfm->len);

        if (priv->loop_f)
                priv->loop_f(priv->loop_ctx, cfm->data, cfm->len);
}

static void
process_heartbeat_cfm(struct wlp_msg *msg)
{
        owl_unused(msg);

        owl_info("CFM HEARTBEAT");

        if (priv->heartbeat_f)
                priv->heartbeat_f(priv->heartbeat_ctx);
}

static void
process_archive_cfm(struct wlp_msg *msg)
{
        int err;
        owl_info("CFM ARCHIVE res:%d", msg->hdr.res);

        switch (msg->hdr.res) {
        case 0:
                /* Done */
                owl_info("firmware upgrade done");
                archive_done(0);
                break;
                
        case WLP_ERR_AGAIN:      
                /* Need more data */
                if ((err = archive_send()))
                        archive_done(err);
                break;

        default:
                owl_err("upgrade failed, err:%d", msg->hdr.res);
                archive_done(msg->hdr.res);
                break;
        };
}

static void
process_data_cfm(struct wlp_msg *msg)
{
        
	if (msg->hdr.res < 0) {
		owl_err("CFM DATA err:%d", msg->hdr.res);
        } else {
                struct wlp_data_cfm *cfm;
                cfm = &msg->u.cfm.u.data;
                
		owl_vdbg("CFM DATA rssi:%d", cfm->rssi);

                if (cfm->rssi)
                        priv->net.rssi = cfm->rssi;

                if (cfm->rate != WLP_RATE_UNKNOWN)
                        priv->txrate = cfm->rate;
        }
}

static void
process_set_channel_ind(struct wlp_msg *msg)
{
        owl_info("IND SET_CHANNEL res:%d", msg->hdr.res);

        if (priv->chan_f)
                priv->chan_f(priv->chan_ctx, msg->hdr.res);
}
                
static void
process_get_linkctx_ind(struct wlp_msg *msg)
{
        struct wlp_get_linkctx_ind *ind;

        ind = &msg->u.ind.u.get_linkctx;
        ind->len = owl_ntohs(ind->len);
        
        owl_info("IND GET_LINKCTX len:%d", ind->len);
        owl_hexdump("LINKCTX", ind->data, ind->len);

        if (priv->linkctx_f) {
                priv->linkctx_f(priv->linkctx_ctx, ind->data, ind->len);
                priv->linkctx_f = NULL;
                priv->linkctx_ctx = NULL;
        }
}

static void
process_scan_ind(struct wlp_msg *msg)
{
        struct wlp_network *net = NULL;

        /* If msg->hdr.res is non-zero, then the scan was completed and no
           network information is provided in msg */
        
        if (!msg->hdr.res) {
                /* Got network information for */
                
                struct wlp_scan_ind *ind;

                net = &ind->net;
                ind = &msg->u.ind.u.scan;

                net = &ind->net;
                net->rssi = owl_ntohs(net->rssi);
                net->snr = owl_ntohs(net->snr);
                
                if (!owl_os_memcmp(
                            &priv->net.bssid, &net->bssid,
                            sizeof(net->bssid))) {
                        
                        /* Update current network if we have new RSSI
                           and/or SNR values */
                        priv->net.rssi = net->rssi;
                        priv->net.snr = net->snr;
                }
        }

        if (priv->scan_f)
                priv->scan_f(priv->scan_ctx, net);
}

static void
process_linkup_ind(struct wlp_msg *msg)
{
        struct wlp_linkup_ind *ind = &msg->u.ind.u.linkup;
        struct wlp_network *net = &ind->net;

        net->rssi = owl_ntohs(net->rssi);
        net->snr = owl_ntohs(net->snr);
        
        owl_info("IND LINKUP ssid:%s hwaddr:%02x:%02x:%02x:%02x:%02x:%02x "
                 "auth:%d channel:%d flags:0x%02x rssi:%d snr:%d",
                 dbg_ssid(net->ssid.ssid, net->ssid.len),
                 net->bssid.octet[0], net->bssid.octet[1], 
                 net->bssid.octet[2], net->bssid.octet[3], 
                 net->bssid.octet[4], net->bssid.octet[5],
                 net->auth, net->channel, net->flags, net->rssi,
                 net->snr);

        owl_os_memcpy(&priv->net, net, sizeof(priv->net));
        owl_bitset(priv->flags, FLAG_LINK);

        if (priv->link_f)
                priv->link_f(priv->link_ctx, 1);
}

static void
process_linkdown_ind(struct wlp_msg *msg)
{
        owl_info("IND LINKDOWN res:%d", msg->hdr.res);

        owl_bitclr(priv->flags, FLAG_LINK);        

        if (priv->link_f)
                priv->link_f(priv->link_ctx, msg->hdr.res);
}

static void
process_get_rxstats_ind(struct wlp_msg *msg)
{
        struct wlp_get_rxstats_ind *ind = &msg->u.ind.u.get_rxstats;
        struct wlp_rfstats *rxstats = &ind->rxstats;

        if (priv->get_rxstats_f)
                priv->get_rxstats_f(priv->get_rxstats_ctx, rxstats);
}

/* MISC -------------------------------------------------------------------- */

static void
notify_error(int err)
{
        owl_err("err:%d", err);
        if (priv->error_f)
                priv->error_f(priv->error_ctx, err);

        /* wlp_shutdown() might have been called from error_f */
}

/* DEBUG ------------------------------------------------------------------- */

#if defined(OWL_DEBUG) && OWL_TRACE >= OWL_TRACE_INFO
static const char *
dbg_ssid(char *ssid, int ssid_len)
{
        static char ssid_str[33];
        owl_os_memcpy(ssid_str, ssid, ssid_len);
        ssid_str[ssid_len] = 0;
        return ssid_str;
}
#endif
