/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#ifndef WLP_HDG205_H
#define WLP_HDG205_H

#include <owl_cc.h>

#define WLP_FIRMWARE_VERSION_OFFSET 0
#define WLP_FIRMWARE_LENGTH_OFFSET  16
#define WLP_FIRMWARE_DATA_OFFSET    20

struct firmware {
        char version[16];
        uint32_t len;
        uint8_t data[0];
};

extern const struct firmware firmware;

#if defined(WLP_CONFIG_TRAMPOLINE)

int _wlp_init(int (*exchange_cb)(void *ctx, const void *tx, void *rx, int len),
              void (*select_cb)(void *ctx, int select),
              void (*cmd_cb)(void *ctx, uint8_t idx, uint32_t arg,
                             uint32_t *rsp, const void *data, int len),
              int (*poll_cb)(void *ctx),
              void *ctx,
              int flags);
void _wlp_set_error_cb(void (*error_cb)(void *ctx, int err), void *ctx);
int _wlp_program(int (*firmware_cb)(void *ctx, void *buf, int offset, int len),
                 void *ctx);
int _wlp_reset(void);
int _wlp_get_version(char *driver, int driver_len,
                    char *firmware, int firmware_len);
void _wlp_shutdown(void);
int _wlp_set_baudrate(uint32_t baudrate);

#else /* defined(WLP_CONFIG_TRAMPOLINE) */

#define wlp_init                _wlp_init
#define wlp_program             _wlp_program
#define wlp_reset               _wlp_reset
#define wlp_get_version         _wlp_get_version
#define wlp_shutdown            _wlp_shutdown
#define wlp_set_error_cb        _wlp_set_error_cb
#define wlp_set_baudrate        _wlp_set_baudrate
#endif /* defined(WLP_CONFIG_TRAMPOLINE) */


/* Flags for wlp_init() */
#define WLP_SDIO_HISPEED  (1 << 0)
#define WLP_SDIO_4BIT     (1 << 2)
#define WLP_HPI_POLL      (1 << 3)

#define WLP_CONFIG_HPI_AUTO 0
#define WLP_CONFIG_HPI_UART 1
#define WLP_CONFIG_HPI_SPI  2
#define WLP_CONFIG_HPI_SDIO 3

#if !defined(WLP_CONFIG_HPI)
# error "WLP_CONFIG_HPI not set"
#endif /* !defined(WLP_CONFIG_HPI) */

#if WLP_CONFIG_HPI == WLP_CONFIG_HPI_UART
# define WLP_HPI_HDR_SIZE 2
#elif WLP_CONFIG_HPI == WLP_CONFIG_HPI_SPI
# define WLP_HPI_HDR_SIZE 4
#elif WLP_CONFIG_HPI == WLP_CONFIG_HPI_SDIO
# define WLP_HPI_HDR_SIZE 4
#elif WLP_CONFIG_HPI == WLP_CONFIG_HPI_AUTO
# define WLP_HPI_HDR_SIZE 0 /* Used by server */
#endif

/* Size of header that prepends any data frames.
 *
 * sizeof(struct wlp_hdr):          6
 * sizeof(struct wlp_data_req/ind): 4
 * HPI header:                      WLP_HPI_HDR_SIZE
 *
 */
#define WLP_HEADER_SIZE   (6 + 4 + WLP_HPI_HDR_SIZE)
#define WLP_IFACES        1

/* Maximum packet size (including wl_api headers and paddings)
 *
 * Maximum packet size is obtained with the following data:
 *
 * 1500 bytes of Ethernet payload (MTU) + 14 bytes of Ethernet header +
 * WLP_HEADER_SIZE of wifi header.
 *
 * In SPI mode we will reserve space to hold the SDIO CRC.
 *
 */
#if WLP_CONFIG_HPI == WLP_CONFIG_HPI_SPI || \
    WLP_CONFIG_HPI == WLP_CONFIG_HPI_AUTO
# define WLP_MAX_PKT_LEN   (1500 + 14 + WLP_HEADER_SIZE + 4)
#else 
# define WLP_MAX_PKT_LEN   (1500 + 14 + WLP_HEADER_SIZE)
#endif

/*! \defgroup wlp_module Module specific functions.
 *
 * @{
 *
 */ 


/**
 * \brief Initialize the wifi device.
 *             
 * @param exchange_cb is a pointer to a function which reads and writes chars
 *              from the bus which the wifi device is connected. If tx is not
 *              NULL, the buffer pointed to by the tx parameter should be
 *              written to the bus. If rx is not NULL, data should be read
 *              from the bus and stored in the buffer pointed to by the rx
 *              parameter. If both tx and rx are not NULL, a write should be
 *              followed by a read. Note that exchange_cb should block until
 *              the number of bytes specified in the len parameter are
 *              read/written.
 * @param select_cb will be invoked before and after any write-read transaction.
 *              For UART interfaces, this function could e.g. disable/enable
 *              interrupts on the host to ensure the UART data flow. For SPI
 *              interfaces, this could assert/desassert the chip select signal
 *              connected to the device. If the select parameter is 1, the
 *              transaction will start when this function has returned. If the
 *              select parameter is 0, the transaction is completed.
 * @param cmd_cb is a pointer to a function which will be invoked when using
 *              SDIO as transport method to/from the device. The host sdio 
 *              driver should implement this function as follows:
 *              Set the cmd number to @param idx.
 *              Set the cmd argument to @param arg.
 *              @Param data will point to the data buf that should be 
 *              sent/received to.
 *              @Param len is the len of data to send/receive.
 *              The @param rsp should have the cmd response as pointee 
 *              after the driver has sent the cmd.
 *              
 * @param ctx is an opaque context pointer that will be passed back to
 *              \a exchange_cb and \a select_cb.
 * @param flags Initialization flags.
                WLP_SPI_HISPEED: enable up to 50MHz SPI/SDIO mode.
                WLP_SDIO_POLL: Hosts not using IRQ requires poll flag.
                WLP_SDIO_4BIT: Sets the device in SDIO 4bit mode.
 * 
 * @return
 *   - 0 on success
 *
 */
int wlp_init(int (*exchange_cb)(void *ctx, const void *tx, void *rx, int len),
             void (*select_cb)(void *ctx, int select),
             void (*cmd_cb)(void *ctx, uint8_t idx, uint32_t arg, uint32_t *rsp, 
                            const void *data, int len),
             int (*poll_cb)(void *ctx),
             void *ctx,
             int flags);


/**
 * \brief Register a function that will be invoked if an error occurs on the
 *      device.
 *
 * @param error_cb is a pointer to a function that will be called in case an
 *       error occurs on the device. The parameter err will indicate an error
 *       code.
 *
 * @param ctx is an opaque context pointer that will be passed back to
 *              \a error_cb.
 *
 *
 */
void wlp_set_error_cb(void (*error_cb)(void *ctx, int err), void *ctx);



/**
 * \brief Program device firmware.
 *
 *
 * HDG205 has en embedded flash memory to store its firmware. The driver
 * also comes with the firmware for HDG205 (wlp_firmware.c). After driver
 * intialization (\a wlp_init), the firmware can be programmed.
 *
 * Upon calling \a wlp_program the version on the HDG205 and the
 * firmware version embedded in the driver will compared. If they differ,
 * the HDG205 will be programmed.
 *
 * After programming is complete, the host must restart the HDG205 to
 * initiate the firmware switch procedure. Do not power off the HDG205 at
 * this time. The firmware switch will take about 10 seconds.
 *
 * @param firmware_cb is a pointer to a function that will be called when
 *      firmware data should be retrieved from the host.
 *
 * @param ctx is an opaque context pointer that will be passed back to
 *      \a firmware_cb.
 *
 * @return
 *      - WLP_ERR_IGNORED if firmware already matches the driver.
 *      - 0 if download is complete.
 *      - error code on failure.
 *
 */
int wlp_program(int (*firmware_cb)(void *ctx, void *buf, int offset, int len),
                void *ctx);

/**
 * \brief Reset the HDG205.
 *
 * @return
 *      - 0 on success, error code on failure.
 */
int wlp_reset(void);


/**
 * \brief Get driver and firmware versions.
 *
 * @param driver is a pointer to a buffer which will hold the driver version
 *      string upon successful return.
 * @param driver_len is the buffer size of \a driver.
 * @param firmware is a pointer to a buffer which will hold the driver version
 *      string upon successful return.
 * @param firmware_len is the buffer size of \a firmware.
 *
 * @return
 *      - 0 on success, error code on failure.
 */
int wlp_get_version(char *driver, int driver_len,
                    char *firmware, int firmware_len);

/**
 * \brief Shutdown the driver.
 * 
 * If the link is up will invoking wlp_shutdown(), the registered link callback
 * will be invoked with link 0.
 *
 */
void wlp_shutdown(void);

/**
 * \brief Change HDG205 uart baudrate
 * 
 * @param baudrate is the new baudrate that the HDG205 should use if
 * hpi is UART.
 * Instructions :
 * 1. Call wlp_set_baudrate() with desired baudrate (e.g. 230400).
 * 2. Wait for 100ms while calling wlp_poll() to get confirm from HDG205 
 * (Confirm is sent using the old baudrate).
 * 3. Switch the host:s baudrate.
 * 4. Done.
 * 
 */
int wlp_set_baudrate(uint32_t baudrate);


/*! @} */

#endif /* WLP_HDG205_H */
