/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#define OWL_TRACE OWL_TRACE_ERR
#define OWL_FILE "wlp_sdio"

#include <owl/core/owl_debug.h>
#include <owl/core/owl_os.h>
#include <hdg205/wlp_hpi.h>
#include <hdg205/wlp_proto.h>
#include <hdg205/wlp_hdg205.h>

#if WLP_CONFIG_HPI == WLP_CONFIG_HPI_SDIO

#define ARG_NO_CLOCK_SDIO_IRQ     0x8801e401
#define ARG_ENABLE_SDIO_IRQ       0x88000803
#define ARG_ENABLE_SDIO_4BIT_MODE 0x88000e02
#define ARG_IO_R_BASE             0x10000000
#define ARG_IO_W_BASE             0x90000000
#define ARG_ACKNOWLEDGE_SDIO_IRQ  0x8801e002
#define ARG_IRQ_STATUS            0x8801e000

/* Host SDIO driver min size restriction */
#define WLP_CONFIG_SDIO_MIN_WLEN  12

struct hpi_ctx {

        int (*exchange_f)(void *ctx, const void *tx, void *rx, int len);
        void (*select_f)(void *ctx, int select);
        void (*cmd_f)(void *ctx, uint8_t idx, uint32_t arg, uint32_t *rsp, 
                      const void *data, int len);
        int flags;
        void *ctx;
};

static struct hpi_ctx *priv = NULL;

int
hpi_init(int (*exchange_cb)(void *ctx, const void *tx, void *rx, int len),
         void (*select_cb)(void *ctx, int select),
         void (*cmd_cb)(void *ctx, uint8_t idx, uint32_t arg, uint32_t *rsp, 
                        const void *data, int len),
         int (*poll_cb)(void *ctx),
         void *ctx, int flags)
{
        uint32_t rsp1, rsp2;
        
        if ((priv = owl_os_zalloc(sizeof(*priv))) == NULL)
                return WLP_ERR_MEM;

        priv->flags = flags;

        priv->exchange_f = exchange_cb;
        priv->cmd_f = cmd_cb;
        priv->ctx = ctx;
        
        priv->cmd_f(priv->ctx, 3, 0, &rsp1, NULL, 0);
        priv->cmd_f(priv->ctx, 3, 0, &rsp2, NULL, 0);

        if ((rsp2 >> 16) != (rsp1 >> 16) + 1)
		return WLP_ERR_SYNC;

        priv->cmd_f(priv->ctx, 7, rsp2, NULL, NULL, 0);
        priv->cmd_f(priv->ctx, 52, ARG_ENABLE_SDIO_IRQ, NULL, NULL, 0);
        priv->cmd_f(priv->ctx, 52, ARG_NO_CLOCK_SDIO_IRQ, NULL, NULL, 0);
       
        if (priv->flags & WLP_SDIO_4BIT) 
                priv->cmd_f(priv->ctx, 52, ARG_ENABLE_SDIO_4BIT_MODE, 
                            NULL, NULL, 0);
        return 0;
}

void
hpi_shutdown(void)
{
        owl_os_free(priv);
        priv = NULL;
}

int
hpi_process_tx(void *data, int len)
{
        int totlen;
        uint16_t *ppad;
        uint16_t *plen;
        int hpi_pad = 0;

        ppad = data;
        plen = (uint16_t *) ((uint8_t *) data + sizeof(*ppad));

        /* Don't include HPI header when calculating size */
        len -= WLP_HPI_HDR_SIZE;

        if (len % 4)
                hpi_pad = 4 - len % 4;

        /* Length including hpi padding but excluding ppad field */
        *ppad = owl_cpu_to_le16(len + hpi_pad + sizeof(*plen));

        /* Payload excluding HPI header */
        *plen = owl_cpu_to_le16(len);

        totlen = len + WLP_HPI_HDR_SIZE + hpi_pad;
        
        return totlen;
}

int
hpi_tx(void *data, int len)
{

        uint32_t pos = 0;
        int writelen = 0;
        int err = 0;
        
        owl_dbg("data:%p len:%d", data, len);

        while (pos < len) {
                int remain = len - pos;

                if (remain >= 512 + WLP_CONFIG_SDIO_MIN_WLEN) {
                        /* The remaining part will be larger than the minsize */
                        writelen = 512;
                } else if (remain > 512) {
                        /* The reminaing part is larger than 512 but smaller
                           than 512 + WLP_CONFIG_SDIO_MIN_WLEN so make sure we
                           keep a part that is equal to the minsize */
                        writelen = remain - WLP_CONFIG_SDIO_MIN_WLEN;
                } else {
                        /* The reminaing part is smaller or equal to 512. It
                           can't be less than WLP_CONFIG_SDIO_MIN_WLEN since
                           we always keep a larger remainder. */
                        writelen = remain;
                }

                priv->cmd_f(priv->ctx, 53, ARG_IO_W_BASE + writelen, 
                            NULL, data + pos, writelen);
                
                if ((err = priv->exchange_f(priv->ctx, data + pos,
                                            NULL, writelen)) < 0)
                        return err;
                
                pos += writelen;
        }
        return len;
}

/* Read chunks of data from the host */
int
hpi_rx(void *data, int len)
{
        uint16_t *ppad;
        uint16_t *plen;
        int pos = 0;
        int remain = 0;
        int err;

        ppad = data;
        plen = (uint16_t *) ((uint8_t *) data + sizeof(*ppad));

        /* Ack IRQ before trying to do any reading */
        priv->cmd_f(priv->ctx, 52, ARG_ACKNOWLEDGE_SDIO_IRQ, NULL, NULL, 0);

        /* Fetch WLP_MSG_MIN_SIZE first to get the real data size. */
        priv->cmd_f(priv->ctx, 53, ARG_IO_R_BASE + WLP_MSG_MIN_SIZE, 
                    NULL, data, WLP_MSG_MIN_SIZE);
                
        if ((err = priv->exchange_f(priv->ctx, NULL, data, 
                                    WLP_MSG_MIN_SIZE)) < 0)
                return err;
        
        if (*ppad == *plen) {
                owl_err("invalid data");
                owl_hexdump("HPI RX", data, WLP_MSG_MIN_SIZE);
                return WLP_ERR_SYNC;
        }
        
        remain = owl_le16_to_cpu(*ppad) + sizeof(*ppad) - WLP_MSG_MIN_SIZE;
        
        /* Fetch more data if avail */
        while (pos < remain) {
                int readlen = owl_min(remain - pos, 512);
                      
                priv->cmd_f(priv->ctx, 53, ARG_IO_R_BASE + readlen, 
                            NULL, data + WLP_MSG_MIN_SIZE + pos, readlen);
                
                if ((err = priv->exchange_f(priv->ctx, NULL, 
                                            data + WLP_MSG_MIN_SIZE + pos, 
                                            readlen)) < 0)
                        return err;
                
                pos += readlen;
        }
        return len;
}

void
hpi_poll(void)
{
        uint32_t rsp;

        priv->cmd_f(priv->ctx, 52, 
                    ARG_IRQ_STATUS, &rsp, NULL, 0);

        if (rsp & 0x02) 
                wlp_irq();
}

#endif /* WLP_CONFIG_HPI == WLP_CONFIG_HPI_SDIO */
