/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#define OWL_TRACE OWL_TRACE_ERR
#define OWL_FILE "wlp_spi"

#include <owl/core/owl_debug.h>
#include <owl/core/owl_os.h>

#include <hdg205/wlp_hpi.h>
#include <hdg205/wlp_proto.h>
#include <hdg205/wlp_hdg205.h>

#if !defined(WLP_CONFIG_SPI_MAX_RLEN)
# define WLP_CONFIG_SPI_MAX_RLEN 0

#endif /* !defined(WLP_CONFIG_SPI_MAX_RLEN) */

#if !defined(WLP_CONFIG_SPI_MAX_WLEN)
# define WLP_CONFIG_SPI_MAX_WLEN 0
#endif /* !defined(WLP_CONFIG_SPI_MAX_WLEN) */

#if WLP_CONFIG_HPI == WLP_CONFIG_HPI_SPI

struct hpi_ctx {

        int (*exchange_f)(void *ctx, const void *tx, void *rx, int len);
        void (*select_f)(void *ctx, int select);
        void *ctx;
};

static int spi_cmd53r(void *data, int len);
static int spi_cmd53w(const void *data, int len);

static struct hpi_ctx *priv = NULL;

int
hpi_init(int (*exchange_cb)(void *ctx, const void *tx, void *rx, int len),
         void (*select_cb)(void *ctx, int select),
         void (*cmd_cb)(void *ctx, uint8_t idx, uint32_t arg, uint32_t *rsp, 
                        const void *data, int len),
         int (*poll_cb)(void *ctx),
         void *ctx, int flags)
{
        uint8_t zero[32] = { 0 };
        uint8_t cmd0[] = {
                0xff, 0xff,
                0x40, 
                0x00, 0x00, 0x00, 0x00, 
                0x01,
                0xff,
	};
	uint8_t rsp[17] = { 0 };
        
        if ((priv = owl_os_zalloc(sizeof(*priv))) == NULL)
                return WLP_ERR_MEM;
   
        priv->exchange_f = exchange_cb;
        priv->select_f = select_cb;
        priv->ctx = ctx;

        priv->select_f(priv->ctx, 1);
        priv->exchange_f(priv->ctx, zero, NULL, sizeof(zero));
        priv->exchange_f(priv->ctx, cmd0, NULL, sizeof(cmd0));
        priv->exchange_f(priv->ctx, NULL, rsp, sizeof(rsp));
	priv->select_f(priv->ctx, 0);

        owl_hexdump("RSP", rsp, sizeof(rsp));
        if(rsp[0] != 1) {
                owl_os_free(priv);
                return WLP_ERR_SYNC;
        }

        if (flags & WLP_SDIO_HISPEED) {
                uint8_t cmd52[] = {
                        0xff, 0xff, 0xff,
                        0x74, 
                        0x88, 0x01, 0xe6, 0x01, 
                        0xff,
                        0xff,
                        0xff, 0xff, 0xff 
                };
	
                priv->select_f(priv->ctx, 1);
                priv->exchange_f(priv->ctx, cmd52, NULL, sizeof(cmd52));
                priv->select_f(priv->ctx, 0);
        }

        return 0;
}

void
hpi_shutdown(void)
{
        owl_os_free(priv);
        priv = NULL;
}

int
hpi_process_tx(void *data, int len)
{
        int totlen;
        uint16_t *ppad;
        uint16_t *plen;
        int hpi_pad = 0;

        ppad = data;
        plen = (uint16_t *) ((uint8_t *) data + sizeof(*ppad));

        /* Don't include HPI header when calculating size */
        len -= WLP_HPI_HDR_SIZE;

        if (len % 4)
                hpi_pad = 4 - len % 4;

        /* Length including hpi padding but excluding ppad field */
        *ppad = owl_cpu_to_le16(len + hpi_pad + sizeof(*plen));

        /* Payload excluding HPI header */
        *plen = owl_cpu_to_le16(len);

        totlen = len + WLP_HPI_HDR_SIZE + hpi_pad;
        
        return totlen;
}

/* Send data in chunks to the host */
int
hpi_tx(void *data, int len)
{
        owl_hexdump("HPI TX", data, len);
        
        return spi_cmd53w(data, len);
}

/* Read chunks of data the host */
int
hpi_rx(void *data, int len)
{
        uint16_t *ppad;
        uint16_t *plen;
        int err;
        int remain;

        ppad = data;
        plen = (uint16_t *) ((uint8_t *) data + sizeof(*ppad));
        
        if ((err = spi_cmd53r(data, WLP_MSG_MIN_SIZE)) < 0)
                return err;

        if (*ppad == *plen) {
                owl_err("invalid data");
                owl_hexdump("HPI RX", data, WLP_MSG_MIN_SIZE);
                return WLP_ERR_SYNC;
        }

        remain = owl_le16_to_cpu(*ppad) + sizeof(*ppad) - WLP_MSG_MIN_SIZE;

        if (remain > 0) {
		if ((err = spi_cmd53r(
			     (uint8_t *) data + WLP_MSG_MIN_SIZE, remain)) < 0)
                        return err;
        }

        owl_hexdump("HPI RX", data, *ppad + sizeof(*ppad));
                
        return *plen;
}

void
hpi_poll(void)
{

}

static int
spi_cmd53w(const void *data, int len)
{
        int writelen;
        int pos = 0;
        static uint8_t cmd53[] = {
                0xff, 0xff, 0xff,  /* dummy */
                0x75, 0x90, 0x00,  /* write cmd */
                0x00, 0x00, 0x01, 0xff,  /* 0xsize01ff (2) */
                0xff, 0xff, 0xff, 0xfe,  /* status and preamble */
        };

        while (pos < len) {
#if WLP_CONFIG_SPI_MAX_WLEN
		/* Max size allowed by SPI driver on host platform */
                writelen = owl_min(len - pos, WLP_CONFIG_SPI_MAX_WLEN - 2);
#else /* WLP_CONFIG_SPI_MAX_WLEN */
		writelen = len - pos;
#endif /* WLP_CONFIG_SPI_MAX_WLEN */

		/* Max size allowed by SDIO */
		writelen = owl_min(len - pos, 512);

                /* Set 9 bits of chunk size in SDIO header */
                cmd53[6] = (uint8_t) ((writelen & 0x1ff) >> 8);
                cmd53[7] = (uint8_t) (writelen & 0xff);

                priv->select_f(priv->ctx, 1);
                priv->exchange_f(priv->ctx, cmd53, NULL, sizeof(cmd53));
                priv->exchange_f(
			priv->ctx, (uint8_t *) data + pos, NULL, writelen + 2);
                priv->select_f(priv->ctx, 0);
                
                pos += writelen;
        }
        
        return len;

}

static int
spi_cmd53r(void *data, int len)
{
        int readlen;
        int pos = 0;
        static uint8_t cmd53[] = {
                0xff, 0xff, 0xff,  /* dummy */
                0x75, 0x10, 0x00,  /* read cmd */
                0x00, 0x00, 0x01, 0xff,  /* 0xsize01ff (2) */
                0xff, 0xff, 0xff, 0xfe,  /* status and preamble */
        };

        while (pos < len) {
#if WLP_CONFIG_SPI_MAX_RLEN
		/* Max size allowed by SPI driver on host platform */
		readlen = owl_min(len - pos, WLP_CONFIG_SPI_MAX_RLEN - 4);
#else /* WLP_CONFIG_SPI_MAX_RLEN */
		readlen = len - pos;
#endif /* WLP_CONFIG_SPI_MAX_RLEN */
		
		/* Max size allowed by SDIO */
		readlen = owl_min(readlen, 512);

                cmd53[6] = (uint8_t) ((readlen & 0x1ff) >> 8);
                cmd53[7] = (uint8_t) (readlen & 0xff);

                priv->select_f(priv->ctx, 1);
                priv->exchange_f(priv->ctx, cmd53, NULL, sizeof(cmd53));
                priv->exchange_f(
                        priv->ctx, NULL, (uint8_t *) data + pos, readlen + 4);
                priv->select_f(priv->ctx, 0);

                pos += readlen;
        }

        return len;
}

#endif /* WLP_CONFIG_HPI == WLP_CONFIG_HPI_SPI */
