/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#define OWL_TRACE OWL_TRACE_NONE
#define OWL_FILE "wlp_uart"

#include <owl/core/owl_debug.h>
#include <owl/core/owl_defs.h>
#include <owl/core/owl_os.h>

#include <hdg205/wlp_hpi.h>
#include <hdg205/wlp_proto.h>

#if WLP_CONFIG_HPI == WLP_CONFIG_HPI_UART

struct hpi_ctx {
        int (*exchange_f)(void *ctx, const void *tx, void *rx, int len);
        void (*select_f)(void *ctx, int select);
        int (*poll_f)(void *ctx);
        void *ctx;
};

static struct hpi_ctx *priv = NULL;

int
hpi_init(int (*exchange_cb)(void *ctx, const void *tx, void *rx, int len),
         void (*select_cb)(void *ctx, int select),
         void (*cmd_cb)(void *ctx, uint8_t idx, uint32_t arg, uint32_t *rsp, 
                        const void *data, int len),
         int (*poll_cb)(void *ctx),
         void *ctx, int flags)
{
        int len;
        uint8_t tmp;

        if ((priv = owl_os_zalloc(sizeof(*priv))) == NULL)
                return WLP_ERR_MEM;
   
        priv->exchange_f = exchange_cb;
        priv->select_f = select_cb;
        priv->poll_f = poll_cb;
        priv->ctx = ctx;

        /* Make sure the uart buffer is empty before sending
           init command to device */
        while ((len = priv->poll_f(priv->ctx)) > 0)
                priv->exchange_f(priv->ctx, NULL, &tmp, 1);
        
        return 0;
}

void
hpi_shutdown(void)
{
        owl_os_free(priv);
        priv = NULL;
}

int
hpi_process_tx(void *data, int len)
{
        uint16_t *plen;

        plen = data;
        *plen = owl_cpu_to_le16(len - 2);

        return len;
}

/* Send data to the host */
int
hpi_tx(void *data, int len)
{
        int err;

        owl_hexdump("HPI TX", data, len);
        
        if (priv->select_f)
                priv->select_f(priv->ctx, 1);
        
        if ((err = priv->exchange_f(
                     priv->ctx, data, NULL, len)) < 0)
                return err;

        if (priv->select_f)
                priv->select_f(priv->ctx, 0);

        return len;
}

/* Read chunks of data the host */
int
hpi_rx(void *data, int len)
{
        int err;
        int totlen;
        int remain;
        struct wlp_hdr *hdr = data;
        
        if (len < WLP_MSG_MIN_SIZE)
                return WLP_ERR_PARAM;

        /* We don't get the hpi header in RX, so fake it here so we can use
           the same struct wlp_hdr in TX and RX */
        data += WLP_HPI_HDR_SIZE;
        
        if (priv->select_f)
                priv->select_f(priv->ctx, 1);
        
        if ((err = priv->exchange_f(
                     priv->ctx, NULL, data, WLP_MSG_MIN_SIZE)) < 0)
                return err;

        totlen = owl_ntohs(hdr->len) + sizeof(*hdr);
        
        /* HPI header not received in UART mode but is included in
           struct wlp_hdr */
        remain = totlen - WLP_MSG_MIN_SIZE - WLP_HPI_HDR_SIZE;

        if (remain > 0) {
                if ((err = priv->exchange_f(
                             priv->ctx, NULL,
                             data + WLP_MSG_MIN_SIZE, remain)) < 0)
                        return err;
        }

        if (priv->select_f)
                priv->select_f(priv->ctx, 0);

        owl_hexdump("HPI RX", data, totlen);

        return totlen;
}

void
hpi_poll(void)
{
        if (priv->poll_f(priv->ctx))
                wlp_irq();
}

#endif /* WLP_CONFIG_HPI == WLP_CONFIG_HPI_UART */
