/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */


/**
 * @file wlp_api.h
 *  
 * @author H&D Wireless AB
 * @brief oWL Pico API
 * 
 * \section intro Introduction
 * This is the documentation for the oWL Pico API.
 *
 * \mainpage oWL Pico API Reference Manual
 * See pico.hd-wireless.se for the lastest documentation, software and
 * tutorials.
 * 
 * oWL Pico API is a programming interface for Wi-Fi (802.11) and network
 * applications. It aims to be a complete interface for embedded
 * applications wanting to use wireless as a communications interface.
 * 
 * The oWL Pico API provides functions to support full Wi-Fi link management,
 * network address management and TCP, UDP and RAW sockets - resembling a
 * standard BSD socket interface.
 *
 * The oWL Pico software API is supported by many Wi-Fi devices from
 * H&D Wireless. Depending on the capabilities of the actual device,
 * some modules of the oWL Pico software API may not be available for
 * a specific module. E.g. the network address and socket API's are
 * only available for modules that supports an offloaded IP stack.
 *
 * For most devies, the oWL Pico implementation has a footprint of a
 * few kilobytes.
 *
 * The oWL Pico API has been designed to meet the following goals :
 * \li Simple : The API is as simple as is practicable
 * to make it easy to use. 
 * \li Minimal size : The API is usable on very resource constrained
 * platforms. 
 * \li Portable : The API is deployable on any platform with a standards
 * compliant C compiler and a UART port.
 * \li OS independent : The API is deployable on systems using a real time
 * operating system as well as with applications running on the
 * "bare metal" of a hardware platform (that is without an operating system).
 *
 *
 * \section wlp_api_princ Operation Principles
 * 
 * There are three important properties of the oWL Pico API to keep in
 * mind when programming with it.
 * 
 * The first is that oWL Pico API is \b asynchronous. For instance, when the
 * \a wlp_linkup() function is called to attempt connection with an access
 * point it will trigger a sequence of packets being exchanged with the
 * access point after which, if everything is okay, a connection has been
 * established. The \a wlp_linkup() call is asynchronous (or non-blocking)
 * which means that you don't know if the connection attempt has
 * succeeded after the call returns. You only know if the sequence was
 * successfully started or not. To find out if, and when, the connection
 * attempt was successful you must register an event handler using the
 * function \a wlp_set_link_cb(). This is true of a number of API calls
 * (which is indicated in their documentation).
 *
 * The second important property is that the oWL Pico API is \b polled. oWL
 * Pico API never executes "by itself", since it would then have to
 * support interrupts, timers, locks and other operating system dependent
 * features. Instead all asynchronous processes proceed when oWL Pico API
 * is polled by calling the \a wlp_poll() function. When \a wlp_poll() is
 * called the oWL Pico API reacts to any received packets, expires any
 * internal timers and performs any other tasks necessary for forward
 * progress. After \a wlp_poll() returns nothing will happen unless it or
 * some other oWL Pico API function is called again.
 *
 * The third important property is that oWL Pico API is \b not \b thread
 * \b safe. All oWL Pico API calls must execute in the same context since
 * the library has no knowledge of the locking mechanisms available (if
 * any).
 * 
 * \section contactinfo Contact Information
 * For further information, visit
 * <a href="http://www.hd-wireless.se/">H&D Wireless</A>.\n
 * Support and documentation: http://pico.hd-wireless.se/
 * 
 */

#ifndef WLP_API_H
#define WLP_API_H

/* Mandatory API */
#include <wlp_mgmt.h>
#include <wlp_link.h>

#define WLP_FAILURE      -1  /*!< General failure */
#define WLP_ERR_PARAM    -2  /*!< Parameter error */
#define WLP_ERR_MEM      -3  /*!< Out of memory */
#define WLP_ERR_CONN     -4  /*!< No connection */
#define WLP_ERR_STATE    -5  /*!< Invalid state */
#define WLP_ERR_PARSE    -6  /*!< Input parsing error */
#define WLP_ERR_PATH     -7  /*!< Path not found */
#define WLP_ERR_TYPE     -8  /*!< Incompatible type */
#define WLP_ERR_PERM     -9  /*!< No permission */
#define WLP_ERR_NOTSUPP  -10 /*!< Operation not supported */
#define WLP_ERR_RANGE    -11 /*!< Out of range */
#define WLP_ERR_SIZE     -12 /*!< Bad size */
#define WLP_ERR_AGAIN    -13 /*!< Retry later */
#define WLP_ERR_BUSY     -14 /*!< Busy */
#define WLP_ERR_DISABLED -15 /*!< Function is disabled */
#define WLP_ERR_PROTO    -16 /*!< Protocol error */
#define WLP_ERR_TIMEOUT  -17 /*!< Timeout */
#define WLP_ERR_IGNORED  -18 /*!< Ignored, no effect */
#define WLP_ERR_CANCEL   -19 /*!< Canceled by remote */
#define WLP_ERR_SYNC     -20 /*!< Sync error */
#define WLP_ERR_ABORT    -21 /*!< Aborted */
#define WLP_ERR_INUSE    -22 /*!< Already in use */
#define WLP_ERR_USAGE    -23 /*!< General usage error */
#define WLP_ERR_CHECKSUM -24 /*!< Checksum failure */

#endif /* WLP_API_H */
