/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#ifndef WLP_ETHERNET_H
#define WLP_ETHERNET_H

#if defined(WLP_CONFIG_TRAMPOLINE)

void _wlp_set_rx_cb(int (*recv_cb)(void *ctx, int len), void *ctx);
int _wlp_process_tx(void *ethhdr, int len, int totlen, void *wlphdr,
                    uint16_t vlanid_prio);
int _wlp_process_rx(void *data, int len, void **stripped, int *stripped_len,
                    uint16_t *vlanid_prio);
int _wlp_tx(void *data, int len);
int _wlp_rx(void *data, int len);

#else /* defined(WLP_CONFIG_TRAMPOLINE) */

#define wlp_set_rx_cb           _wlp_set_rx_cb
#define wlp_process_tx          _wlp_process_tx
#define wlp_process_rx          _wlp_process_rx
#define wlp_tx                  _wlp_tx
#define wlp_rx                  _wlp_rx

#endif /* defined(WLP_CONFIG_TRAMPOLINE) */

/*! \defgroup wlp_ethernet Ethernet layer functions
 *
 * These functions should be used for integration with IP stack.
 *
 * @{
 *
 */ 

/**
 * Register a function that will be invoked when a new packet should be
 * delivered to the network stack. It is the responsibility of the callback
 * function to free the netbuf after processing is done - it wil no longer
 * be referenced by the driver when the recv_cb returns.
 *
 */
void wlp_set_rx_cb(int (*rx_cb)(void *ctx, int len), void *ctx);

/**
 * \brief Prepare tx packet for transmission.
 *
 * This function is typically used only by the TCP/IP stack driver.
 *
 * Takes a Ethernet II frame header and generates a message passing header
 * for it.
 *
 * The caller should ensure that any frames injected into wlp_process_tx()
 * are proper Ethernet frames. The wl_api library or the device will not
 * perform any Ethernet padding if the frames are too short.
 *
 * The Ethernet header is assumed to have the following layout :
 * <dst addr:6><src addr:6><type:2>...
 * The rest of the Ethernet header buffer (if any) is ignored.
 *
 * lwIP would call this function like this (if ETH_PAD_LEN is WLP_HEADER_SIZE):
 *
 * wlp_process_tx(p->payload + WLP_HEADER_SIZE,
 *                p->len - WL_HEADER_SIZE,
 *                p->tot_len - WL_HEADER_SIZE,
 *                p->payload);
 *
 * @param ethhdr is pointer to the ethernet header
 * @param len is total length of ethdr buffer
 * @param totlen is total length of packet (not including wlphdr)
 * @param wlphdr points to a buffer of WLP_HEADER_SIZE that can hold the
 *               header. Does not have to be contigous with the rest of the
 *               packet.
 * @param vlanid_prio VLAN ID and 802.1p priority value. Ignored for legacy
 *      association (no WMM). Bits are set using following format:
 *
 *      <PRE>
 *      5|432109876543|210
 *      -+------------+---
 *      0|   VLANID   |PRI
 *      </PRE>
 *
 */
int wlp_process_tx(void *ethhdr, int len, int totlen, void *wlphdr,
                   uint16_t vlanid_prio);

/**
 *
 * Transmit an ethernet frame.
 *
 * Can be called multiple times for each fragment that should be sent, e..g
 * each pbuf in a pbuf chain. For each packet wlp_process_tx() should be called
 * first.
 *
 * @param data should be pointing to ATH header, as obtained by wl_process_tx()
 *
 */
int wlp_tx(void *data, int len);


/**
 * Must be called in recv_cb, data should point to contig buffer holding
 * num bytes passed to recv_cb().
 */
int wlp_rx(void *data, int len);


/**
 * \brief Get the number of bytes to strip.
 *
 * Processes a raw rx packet by unencrypting it (if necessary)
 * and stripping headers so as to output a 802.3 frame.
 *
 * A packet arriving from the WiFi interface can be either a data
 * packet or a message from the WiFi hardware to the WiFi driver
 * (which is implemented by the wl_api library). This means that
 * wlp_api must process every packet to decide if it is an internal
 * message or a data frame tha should be passed up to the application.
 * Data packets are prefixed with an extra header containing some
 * administrative information, and may be followed by padding bytes so
 * wlp_api also needs to strip the extra header and any padding before
 * the packet can be safely ingested by the IP stack.
 *
 * All this happens in the function \a wlp_process_rx() which \b must
 * be called on every packet received by a call to \a wlp_rx().
 *
 * If \a wlp_process_rx() decides that the packet was a command it processes
 * it and returns 0 to signal that the packet should
 * not be used by anyone else. Otherwise \a stripped is
 * pointing to the beginning of a 802.3 Ethernet frame of length
 * \a stripped_len. If the IP stack supports VLAN and QoS
 * the extra VLAN tag should be passed to the IP stack
 * together with the packet. For IP stacks without this support the VLAN tag
 * contents can safely be ignored, but it must still be filled in by
 * \a wlp_process_tx().
 *
 * The wlp_api library of the device will not perform any Ethernet padding
 * removal. The padding removal performed by wlp_process_rx() is only for
 * the padding used in the protocol shared by the host and the device.
 * This padding is mainly there to ensure that the host does not have to
 * deal with rx of odd-sized data buffers (which some DMA's have problems
 * to handle).
 *
 *
 * @param vlanid_prio VLAN ID and 802.1p priority value. Ignored for legacy
 *      association (no WMM). Bits are set using following format:
 *
 *      <PRE>
 *      5|432109876543|210
 *      -+------------+---
 *      0|   VLANID   |PRI
 *      </PRE>
 */
int wlp_process_rx(void *data, int len, void **stripped, int *stripped_len,
                   uint16_t *vlanid_prio);


/*! @} */

#endif /* WLP_ETHERNET_H */
