/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#ifndef WLP_MCAST_H
#define WLP_MCAST_H

#include <wlp_link.h>

#if defined(WLP_CONFIG_TRAMPOLINE)

void _wlp_add_mcast(struct wlp_mac_addr *hwaddr);
void _wlp_del_mcast(struct wlp_mac_addr *hwaddr);

#else /* defined(WLP_CONFIG_TRAMPOLINE) */

#define wlp_add_mcast    _wlp_add_mcast
#define wlp_del_mcast    _wlp_del_mcast

#endif /* defined(WLP_CONFIG_TRAMPOLINE) */
    
/*! \defgroup wlp_mcast Multicast filtering configuration.
 *
 * These functions manage the Wi-Fi multicast filters and are typically called
 * by the IP stack.
 *
 * IPv4 multicast packets are delivered using the Ethernet MAC address
 * range 01:00:5e:00:00:00–01:00:5e:7f:ff:ff (with an OUI owned by the IANA).
 * This range has 23 bits of available address space. The first octet (01)
 * includes the broadcast/multicast bit. The lower 23 bits of the 28-bit
 * multicast IP address are mapped into the 23 bits of available Ethernet
 * address space. This means that there is ambiguity in delivering packets.
 * If two hosts on the same subnet each subscribe to a different multicast
 * group whose address differs only in the first 5 bits, Ethernet packets
 * for both multicast groups will be delivered to both hosts, requiring
 * the network software in the hosts to discard the unrequired packets.
 *
 * @{
 *
 */


/**
 * @brief Add a multicast mac address in the adapter multicast filter list.
 *
 * By default, all multicast data will be receveid.
 *
 * Ethernet frames with a destination address that matches the address provided
 * by this function will now be received.
 *
 * IP stacks usually supports callbacks that are invoked when the multicast
 * filtering should be updated. Typically, an IP address is passed to this
 * callback. To translate the IP address into a proper multicast MAC address,
 * the following procedure can be performed.
 *
 * \code{.c}
 * struct wlp_mac_addr mcast_addr = { { 0x01, 0x00, 0x5E, 0xff, 0xff, 0xff } };
 *
 * mcast_addr.octet[0]  = 0x01;
 * mcast_addr.octet[1]  = 0x01;
 * mcast_addr.octet[2]  = 0x01;
 * mcast_addr.octet[3]  = (ipaddr >> 8) & 0xff;
 * mcast_addr.octet[3] &= ~(1 << 7);
 * mcast_addr.octet[4]  = (ipaddr >> 16) & 0xff;
 * mcast_addr.octet[5]  = (ipaddr >> 24) & 0xff;
 *
 * wlp_add_mcast(&mcast_addr);
 * \endcode
 *
 * @param hwaddr is the multicast address to add. If NULL, then all multicast
 * data will be received.
 *
 * @return 0 on success, error code otherwise.
 *
 */
int wlp_add_mcast(struct wlp_mac_addr *hwaddr);

/**
 * @brief Delete a multicast mac address in the adapter multicast filter list.
 *
 * Remove a multicast mac address that has been previously added with
 * wlp_add_mcast().
 *
 * @param hwaddr is the multicast address to add. If NULL, then no multicast
 * data will be received.
 * 
 * @return 0 on success, error code otherwise.
 *
 */
int wlp_del_mcast(struct wlp_mac_addr *hwaddr);


/*! @} */

#endif /* WLP_MCAST_H */
