/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#ifndef WLP_MGMT_H
#define WLP_MGMT_H

#if defined(WLP_CONFIG_TRAMPOLINE)

void _wlp_irq(void);
void _wlp_set_yield_cb(void (*yieldf)(void *ctx), void *ctx);
int _wlp_poll(void);
int _wlp_heartbeat(void (*heartbeat)(void *ctx), void *ctx);

#else  /* defined(WLP_CONFIG_TRAMPOLINE) */

# define wlp_irq                 _wlp_irq
# define wlp_set_yield_cb        _wlp_set_yield_cb
# define wlp_poll                _wlp_poll
# define wlp_heartbeat           _wlp_heartbeat

#endif /* defined(WLP_CONFIG_TRAMPOLINE) */


/*! \defgroup wlp_mgmt Library support functions
 *
 * These functions manage the library in general.
 *
 * @{
 *
 */ 



/**
 * \brief Handler for interrupts from device.
 *
 * This function should be called when an interrupt is received from the device.
 *
 * Since the wlp_api functions generally commuicates with the device and
 * blocks until a reponse is received, the wlp_irq() naturally needs to be
 * called from a different context (e.g. another thread or an interrupt
 * handler) to avoid infinite blocking. However, any calls to wlp_poll() should
 * not be invoked from a different context unless concurrency protection is
 * ensured by e.g. semaphores.
 *
 */
void wlp_irq(void);

/**
 * \brief Register a callback that will be invoked when wlp_api is idle.
 *
 * The yield callback can be used by the host to perform other tasks while
 * waiting for a response from the device. The yield callback will be invoked
 * when any function in the wlp_api is called that involves communication with
 * the device. The yield callback can also be used to e.g. put the host in
 * sleep mode until the response from the device is ready.
 *
 * The yield callback will only be used in case the WLP_INTERFACE_SPI was used
 * in the call to wlp_init().
 *
 * @param yieldf is a pointer to a function which will be invoked when any
 *              function in the wlp_api is waiting for a response from the
 *              device.
 * @param ctx is an opaque context pointer that will be passed back to yieldf.
 *
 */
void wlp_set_yield_cb(void (*yieldf)(void *ctx), void *ctx);


/**
 * \brief Event progress function.
 *
 * This function must be called in stand-alone environments to
 * ensure forward progress of events. The call can be made periodically from
 * the main application loop.
 *
 * Every time wlp_poll() is invoked, the callback conditions for all callbacks
 * registered through wlp_api will be evaluated and any registered callback
 * will be invoked from same context as the call to wlp_poll().
 *
 * If no callbacks are registered, wlp_poll() does not have to be invoked.
 *
 * In systems using an OS, a task could be scheduled to run this task
 * periodicallly.
 *
 * Note that this function should not be invoked from interrupt context.
 *
 * @returns
 *      - 0 if all pending activities have completed. There is no need to
 *              call wlp_poll() again unti a new irq is signalled.
 *      - WLP_ERR_AGAIN if wlp_poll() should be called again to complete
 *              pending activities.
 *
 *
 */
int wlp_poll(void);

/**
 * \brief Test if the hardware is still responding.
 *
 * This function is used to verify that the Wi-Fi chip is still
 * responsive. An unresponsive chip could be caused by errors
 * in the transport layer, firmware crashes or hardware glitches.
 *
 * A request is sent to the WiFi chip and the \a cb callback is
 * invoked once the response has been received. If the WiFi chip
 * is unresponsive then \a cb will never be invoked.
 *
 * A typical usage of this function is for watchdog purposes:
 * An application can periodically call this function and start
 * a timer that is stopped by the \a cb callback. If the timer
 * triggers the WiFi chip (or the whole platform) could be restarted.
 * The timeout should be at least a couple of seconds to avoid
 * false triggering of the restart timer in cases of high load.
 *
 * @param cb Callback that will be invoked when the hardware
 *           responds.
 * @param ctx Opaque context pointer that will be passed back to the
 *            callback.
 * @return 0 on success, error code on failure.
 */
int wlp_heartbeat(void (*cb)(void *ctx), void *ctx);

/*! @} */

#endif /* WLP_MGMT_H */
