/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */


#ifndef WLP_RFTEST_H
#define WLP_RFTEST_H

#include <owl_cc.h>

#include <wlp_link.h>

/**
 * @file wlp_rftest.h
 *  
 * @author H&D Wireless AB
 * @brief oWL Pico API
 *
 */

#include <owl_beginpack.h>


/*! \ingroup wlp_rftest
 *
 * \brief RX frame statistics.
 *
 */
struct wlp_rfstats {
        uint32_t num_frames;         /* total number of frames (crc good/bad) */
        uint32_t num_mcast;          /* good crc */  
        uint32_t num_ucast_to_us;    /* good crc */
        uint32_t num_ucast_to_other; /* good crc */
        uint32_t num_bad_crc;        /* all frames with bad crc */
        int32_t  rssi_dbm;           /* accumulated value - 
                                        divide with num_frames */
        int32_t  snr_db;             /* accumulated value - 
                                        divide with num_frames */
        int32_t  dc_re;              /* accumulated value - 
                                        divide with num_frames */
        int32_t  dc_im;              /* accumulated value - 
                                        divide with num_frames */
        int32_t  freq_error;         /* in Hz, accumulated value - 
                                        divide with num_frames */
        int32_t  delayspread;        /* accumulated value -
                                        divide with num_frames */
} OWL_PACKED;

#include <owl_endpack.h>

#if defined(WLP_CONFIG_TRAMPOLINE)

int _wlp_start_tx(uint32_t count, uint16_t interval, uint16_t size,
                  uint8_t rate);

int _wlp_stop_tx(void);
int _wlp_get_rxstats(uint8_t rate,
                     void (*rxstats_cb)(void *ctx,
                                        struct wlp_rfstats *rxstats), 
                     void *ctx);
int _wlp_clr_rxstats(void);

#else /* defined(WLP_CONFIG_TRAMPOLINE) */

# define wlp_start_tx       _wlp_start_tx
# define wlp_stop_tx        _wlp_stop_tx
# define wlp_get_rxstats    _wlp_get_rxstats
# define wlp_clr_rxstats    _wlp_clr_rxstats

#endif /* defined(WLP_CONFIG_TRAMPOLINE) */


/*! \defgroup wlp_rftest Radio test functions
 *
 * These functions can be used to test the radio.
 *
 * @{
 *
 */ 

/**
 * \brief Generates frames of programmable size and at given rate.
 *
 * The frames will be transmitted on the channel specified with
 * \a wlp_set_channel().
 *
 * @param count is the number of packets to transmit.
 * @param interval is the interval between packets in ms.
 * @param size is the size of packet payload.
 * @param rate is TX rate. Use WLP_RATE_XXX defines.
 *
 * @return 0 on success, error code on failure.
 *
 */

int wlp_start_tx(uint32_t count, uint16_t interval, uint16_t size,
                 uint8_t rate);

/**
 * \brief Stop transmitting packets.
 *
 * @return 0 on success, error code on failure.
 */
int wlp_stop_tx(void);

/**
 * \brief Get stats of received packets for the given rate.
 *
 * @param rate The rate of interest. Use WLP_RATE_XXX defines.
 * @param rxstats_cb callback invoked when rx statistics is available.
 *      Data is delivered in wlp_rfstats struct.
 * @param ctx is an opaque context pointer passed back to \a rxstats_cb.
 *
 * @return 0 on success, error code on failure.
 */
int wlp_get_rxstats(uint8_t rate,
		    void (*rxstats_cb)(void *ctx, struct wlp_rfstats *rxstats),
		    void *ctx);

/**
 * \brief Clear all rx statistics for all rates.
 *
 * @return 0 on success, error code on failure.
 */
int wlp_clr_rxstats(void);

/*! @} */

#endif /* WLP_RFTEST_H */
